/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "search_result.hpp"

#include <iostream>

#include "nmm/base/registry/ReferenceManager.hpp"


#include "../tvtimer/tvtimer.hpp"
#include "nmm/misc/dvb/channel.hpp"
#include "nmm/utils/utils.hpp"     // for string + int

#include "../dvb/dvb.hpp"

namespace NMM{

  EPGSearchResult::EPGSearchResult(string id, MMBoxApplication* _app)
    :MMBoxState(id, _app), 
     currentShortResult(0),
     sortMutex(),
     itvgserver(NULL),
     manager(NULL),
     shortResult(NULL),
     extResult(NULL),
     win_focus(SHORTRESULT),
     currentQueryID(0)
  {
  }

  EPGSearchResult::~EPGSearchResult(){}

  void EPGSearchResult::setWinProperties( int winnr, win_properties_t* prop ) {
    memcpy( &winprops[winnr], prop, sizeof( win_properties_t ) );
  }

  /******************** MMBoxState interface ********************/

  Result EPGSearchResult::doCreate(string){ //list<ProgramFilter> filter) {
    GLOBAL_DEBUG_STREAM("EPGSearchResult: create: " << stateid << endl);

    if (isBackgroundFree()) {
      if (FAILURE == app->createDefaultGraph(background)) {
        return FAILURE;
      }

      if (FAILURE == app->startDefaultGraph()) {
        return FAILURE;
      }
    }

    manager = app->getOsdManager();

    createWindow(SHORTRESULT, "Search Results");
    createWindow(EXTRESULT, "Extended Description");

    //Connect to TVGuide 
    NMMObject* o(SReferenceManager::getInstance().getService("TVGuide"));
    if(o){
      itvgserver.reset(o->getInterface<ITVGuide>());
    }else{
      GLOBAL_ERROR_STREAM("Service TVGuide not available" << endl);
    }

    //Connect to TVTimer 
    o = SReferenceManager::getInstance().getService("TVTimer");
    if(o){
      itvtimer.reset(o->getInterface<ITVTimer>());
    }else{
      GLOBAL_ERROR_STREAM("Service TVTimer not available" << endl);
    }

    showOSD();
    updateOSD();

    return SUCCESS;
  }

  Result EPGSearchResult::doKill(){
    hideOSD();
    updateOSD();
    deleteWindows();
    return SUCCESS;
  }
  Result EPGSearchResult::doInitialize(string){
    if (isBackgroundFree()) {
      if (FAILURE == app->createDefaultGraph(background)) {
        return FAILURE;
      }

      if (FAILURE == app->startDefaultGraph()) {
        return FAILURE;
      }
    }

    showOSD();
    updateOSD();
    return SUCCESS;
  }
  Result EPGSearchResult::doUninitialize(){
    hideOSD();
    updateOSD();
    return SUCCESS;
  }
  
  Result EPGSearchResult::Up(){
    if(queryResult.empty()) return FAILURE;

    int line = 0;
    switch(win_focus){
    case SHORTRESULT:{
      line = shortResult->getSelectedLine();
      shortResult->selectLine(line-1);
      if(currentShortResult-- <= 0) currentShortResult = 0;
      list<unsigned int> lineList = shortResult->getMarkedLines();
      if(lineMarked(lineList, line-1)){
        manager->setGreenColorBarText("Deselect");
      }else{
        manager->setGreenColorBarText("Select");
      }
           
      updateExtResult();
    }
      break;
    case EXTRESULT:{
      line = extResult->getSelectedLine();
      extResult->selectLine(line-1);
    }
      break;
    }
    updateOSD();
    return SUCCESS;
  }
  Result EPGSearchResult::Down(){
    if(queryResult.empty()) return FAILURE;

    int line = 0;
    int maxLine = shortResult->getTextWidget()->getNumLines();
    switch(win_focus){
    case SHORTRESULT:{
      line = shortResult->getSelectedLine();
      shortResult->selectLine(line+1);
      maxLine -= 1;
      if(currentShortResult++ >= maxLine) 
        currentShortResult = maxLine;
      list<unsigned int> lineList = shortResult->getMarkedLines();
      if(lineMarked(lineList, line+1)){
        manager->setGreenColorBarText("Deselect");
      }else{
        manager->setGreenColorBarText("Select");
      }
      
      updateExtResult();
    }
      break;
    case EXTRESULT:{
      line = extResult->getSelectedLine();
      extResult->selectLine(line+1);
    }
      break;
    }
    updateOSD();
    return SUCCESS;
  }
  Result EPGSearchResult::Left(){
    switch(win_focus){
    case SHORTRESULT:
      shortResult->deSelectWindow();
      extResult->selectWindow();
      win_focus = EXTRESULT;
      break;
    case EXTRESULT:
      shortResult->selectWindow();
      extResult->deSelectWindow();
      win_focus = SHORTRESULT;
      break;
    }
    updateOSD();
    return SUCCESS;
  }
  Result EPGSearchResult::Right(){
    switch(win_focus){
    case SHORTRESULT:
      shortResult->deSelectWindow();
      extResult->selectWindow();
      win_focus = EXTRESULT;
      break;
    case EXTRESULT:
      shortResult->selectWindow();
      extResult->deSelectWindow();
      win_focus = SHORTRESULT;
      break;
    }
    updateOSD();
    return SUCCESS;
  }
  Result EPGSearchResult::Return(){
    return SUCCESS;
  }
  Result EPGSearchResult::Back(){
    hideOSD();
    updateOSD();
    return goToLastState();
  }
  Result EPGSearchResult::Record(){
    return SUCCESS;
  }
  Result EPGSearchResult::colorKeyRed(){
    GLOBAL_DEBUG_STREAM("EPGSearchResult: Record" << endl);
    if(app->getOsdManager()->getRedColorBarText() == "") return FAILURE;

    if(queryResult.empty()) return FAILURE;

    TvTimer::TvTimer* timer = dynamic_cast<TvTimer*>(app->getState("TV-Timer"));
    try{
      bool haveTimerWithError = false;
      int lastSuccessfulTimer = -1;
      list<unsigned int> ml = shortResult->getMarkedLines();
      ml.push_back(shortResult->getSelectedLine());
      ml.sort(); ml.unique();
      for(list<unsigned int>::iterator i = ml.begin(); i != ml. end(); i++){
        int j = *i;
        list<ProgramDescription>::iterator it = queryResult.begin();
        advance(it, j-1);
        int res = timer->programTimer(*it);
        if(res < 0) {
          haveTimerWithError=true;
          shortResult->markLine(j);
        }else{
          lastSuccessfulTimer = res;
          shortResult->unmarkLine(j);
        }
      }
      
      if(!haveTimerWithError){
        char tmp[5];
        sprintf(tmp, "%i", lastSuccessfulTimer);
        changeState("TV-Timer",string(tmp));

      }else{
          manager->displayMessage("Error" , " A Timer could Not be programmed!", "OK" );
	  changeState("OSDInputState");
          manager->update();
      }
      
    }catch(const Exception &e){
      //cerr << "Something went wrong: " << e << endl;
      manager->displayMessage("ERROR" , "Timer Not programmed!", "OK" );
      changeState("OSDInputState");
      manager->update();
    }

    return SUCCESS;

  }

  Result EPGSearchResult::colorKeyGreen(){
    if(app->getOsdManager()->getGreenColorBarText() == "") return FAILURE;
    if(queryResult.empty()) return FAILURE;
    if(win_focus == EXTRESULT) return FAILURE;
    
    unsigned int sel = shortResult->getSelectedLine();
    list<unsigned int> lineList = shortResult->getMarkedLines();
  
    for(list<unsigned int>::iterator i = lineList.begin(); i != lineList.end(); i++){
      if (*i == sel){
        shortResult->unmarkLine(sel);
        manager->setGreenColorBarText("Select");
        updateOSD();
        return SUCCESS;
      }
    }
    shortResult->getTextWidget()->markLine(sel);
    manager->setGreenColorBarText("Deselect");

    updateOSD();
  
    return SUCCESS;
  }
  Result EPGSearchResult::colorKeyYellow(){
    if(app->getOsdManager()->getYellowColorBarText() == "") return FAILURE;

    if(!queryResult.empty()){
      
      MutexGuard mg(sortMutex);

      shortResult->deleteAllLines();
      extResult->deleteAllLines();
      
      queryResult.sort();
      try{
        for(list<ProgramDescription>::iterator i = queryResult.begin(); i != queryResult.end(); i++){
          if(!i->empty()){
            i->localizeTimes();
              shortResult->addLine(formatTimeString(i->getStartTime(), true) + " " 
                                   + i->getTitles().front(), i->getChannelID());
          }
        }
    
        currentShortResult = 0;
        shortResult->selectLine(1);
        updateExtResult();
      }catch(const Exception &e){
      }
      
      showOSD();
      updateOSD();

    }

    return SUCCESS;
  }

  Result EPGSearchResult::colorKeyBlue(){

    if(app->getOsdManager()->getBlueColorBarText() == "") return FAILURE;
    if(queryResult.empty()) return FAILURE;

    int line = shortResult->getSelectedLine();

    if(line != 0){
      string tvsource;
      MMBoxConfig* config = app->getMMBoxConfig();
      if (!(config->getValue("tvsource", tvsource))) {
        tvsource="dvb";
      }
           
      if(!itvgserver.get()) return FAILURE;


      try{
        string channel = itvgserver->getChannelById(shortResult->getLineURL(line));
        DvbState* dvbstate = dynamic_cast<DvbState*>(app->getState("DVB-TV"));
        int c = dvbstate->getChannelConfig()->getChannelByName(channel).getChannelNumber();
        if(c >= 0){
          if(dvbstate) dvbstate->setTvChannel(c);
          changeState("DVB-TV"); 
        }else{
          manager->displayMessage("Error" , "Channel not available", "OK" );
          manager->update();
          changeState("OSDInputState");
        }
      }catch(const Exception &e){
        cerr << "args" << endl;
      }
      return SUCCESS;
    }else{
      return FAILURE;
    }
  }

  /* ************** OSD-Handling **************** */
  Result EPGSearchResult::showOSD(){
    if(queryResult.empty()){
      manager->setColorBarText("", "", "", "");
    }else{
      manager->setColorBarText("Record", "Select", "Sort", "Switch");
    }
    manager->displayColorBar();
    shortResult->show();
    extResult->show();

    manager->update();

    return SUCCESS;
  }
  Result EPGSearchResult::hideOSD(){
    manager->hideColorBar();
    shortResult->hide();
    extResult->hide();

    manager->update();

    return SUCCESS;
  }
  Result EPGSearchResult::updateOSD(){

    if(itvgserver.get()){
      if(shortResult->getTextWidget()->getNumLines() > 0){
        string channel = itvgserver->getChannelById(shortResult->getLineURL(currentShortResult+1));
        if(itvtimer.get()){
          try{
            itvtimer->getChannelConfig().getChannelByName(channel);
            manager->setRedColorBarText("Record");
            manager->setGreenColorBarText("Select");
          }catch(const Exception &e){
            manager->setRedColorBarText("");
            manager->setGreenColorBarText("");
          }
        }
        
        try{
          DvbState* dvbstate = dynamic_cast<DvbState*>(app->getState("DVB-TV"));
          if(dvbstate){
            dvbstate->getChannelConfig()->getChannelByName(channel).getChannelNumber();
            manager->setBlueColorBarText("Switch");
          }
        }catch(const Exception &e){
          manager->setBlueColorBarText("");
        }
      }
    }

    shortResult->update();
    extResult->update();
    manager->update();
    return SUCCESS;
  }

  /* ************** private Helpers ************* */
  void EPGSearchResult::deleteWindows(){
    if(shortResult){
      delete shortResult;
      shortResult = NULL;
    }
    if(extResult){
      delete extResult;
      extResult = NULL;
    }
    
  }

  void EPGSearchResult::createWindow(wintype_e w, string h){
    IWidgetFactory& factory = app->getOSDWidgetFactory();

    if(w == SHORTRESULT){
      delete shortResult;
      shortResult = NULL;

      shortResult = new ListUnit( manager, &factory, "medium" );
      shortResult->setHeader( h );
      shortResult->setColors( winprops[w].win_active_color,
			      winprops[w].win_inactive_color,
			      winprops[w].sel_active_color,
			      winprops[w].sel_active_color2,
			      winprops[w].sel_inactive_color );
      /* set width and height of window */
      shortResult->setDimension( winprops[w].width, winprops[w].height );
      shortResult->setPosition( winprops[w].xpos, winprops[w].ypos );
      shortResult->selectWindow();
    }else if(w == EXTRESULT){
      delete extResult;
      extResult = NULL;

      extResult = new ListUnit( manager, &factory, "medium" );
      extResult->setHeader( h );
      extResult->setColors( winprops[w].win_active_color,
			    winprops[w].win_inactive_color,
			    winprops[w].sel_active_color,
			    winprops[w].sel_active_color2,
			    winprops[w].sel_inactive_color );
      
      extResult->setDimension( winprops[w].width, winprops[w].height );
      extResult->setPosition( winprops[w].xpos, winprops[w].ypos );
      extResult->deSelectWindow();
    }
  }

  void EPGSearchResult::updateExtResult(){
    
    if(!itvgserver.get()) return;
    extResult->deleteAllLines();

    try{
      list<ProgramDescription>::iterator it = queryResult.begin();
      advance(it, currentShortResult);
      if(it != queryResult.end()){
        extResult->addLine(formatTimeString(it->getStartTime(), false) + " - " + formatTimeString(it->getEndTime(), false));
        string sourceType;
        switch(it->getSourceType()){
        case ProgramDescription::XMLTV:
          sourceType = "XMLTV";
          break;
        case ProgramDescription::NXTVEPG:
          sourceType = "NxtvEPG";
          break;
        case ProgramDescription::DVB:
          sourceType = "DVB";
          break;
        default:
          sourceType = "Unknown";
          break;
        }
        extResult->addLine(it->getChannelDescription().getChannelNames().front() + "   Source: " + sourceType);
        extResult->addLine("");
        list<string> titles = it->getTitles();
        for(list<string>::iterator i = titles.begin(); i != titles.end(); i++){
          extResult->addLine(*i);
        }
        extResult->addMultiLine(it->getDescription());
      }
    }catch(const Exception &e){}
  }


  string EPGSearchResult::formatTimeString(string s, bool timeOnly){
    string ret("");
    try{
      if(!timeOnly){
        ret += s.substr(6,2) + ".";
        ret += s.substr(4,2) + ".";
        ret += s.substr(0,4) + " ";
      }
      ret += s.substr(8,2) + ":";
      ret += s.substr(10,2);
      return ret;
    }catch(const Exception &e){
      return string("");
    }

  }  
  bool EPGSearchResult::lineMarked(list<unsigned int> l, unsigned int i){
    for(list<unsigned int>::iterator j = l.begin(); j != l.end(); j++){
      if (*j == i){
        return true;
      }
    }
    return false;
  }
  
  void EPGSearchResult::startSearch(ProgramFilter filter){
    //Searching
    
    if(!itvgserver.get()) return;
    shortResult->deleteAllLines();
    extResult->deleteAllLines();
    updateOSD();

    manager->displayMessage( "Searching ..." , "Please wait!", "","","" );
    manager->update();        
    currentQueryID++;
    filter.setID(currentQueryID);

    if(itvgserver.get())queryResult = itvgserver->search(filter, this->getInterface<ITVGuidePeerObject>());
    queryResult.sort();
    queryResult.unique();
    manager->deleteMessage();
    manager->update();
    
    if(shortResult)shortResult->deleteAllLines();
    if(extResult)extResult->deleteAllLines();
    
    if(!queryResult.empty()){

      MutexGuard mg(sortMutex);

      try{
        for(list<ProgramDescription>::iterator i = queryResult.begin(); i != queryResult.end(); i++){
          if(!i->empty()){
            i->localizeTimes();
              shortResult->addLine(formatTimeString(i->getStartTime(), true) + " " 
                                   + i->getTitles().front(), i->getChannelID());
          }
        }
    
        currentShortResult = 0;
        shortResult->selectLine(1);
        updateExtResult();
      }catch(const Exception &e){
      }
    }    

    showOSD();
    updateOSD();
  }

  Result EPGSearchResult::peerResult (const ProgramDescriptionList& pdlist, 
                                      const int& id, 
                                      const list<Interface*>& /*history*/){
    cerr << "got " << pdlist.size() << " with id " << id << endl;

    MutexGuard mg(sortMutex);

    if(!hasFocus()) return FAILURE;
    if(id != currentQueryID) return FAILURE;
    if(pdlist.getList().empty()) return FAILURE;

    list<ProgramDescription> l = pdlist.getList();
    l.sort();

    bool selectLines = false; //if the queryresult from localhost was empty, we mark the first line. when we have come so far, we have at least 1 result

    if(queryResult.empty()){
      selectLines=true;
    }

    for(list<ProgramDescription>::iterator i = l.begin(); i != l.end(); i++){
      if(!i->empty()){
        i->localizeTimes();
        
        bool addPD = true;

        for(list<ProgramDescription>::iterator j = queryResult.begin(); j != queryResult.end(); j++){
          if(*i == *j){
            addPD = false;
            break;
          }
        }
        if(addPD){
          queryResult.push_back(*i);
          shortResult->addLine(formatTimeString(i->getStartTime(), true) + " " 
                               + i->getTitles().front(), i->getChannelID());
        }
      }
    }

    if(selectLines){
      currentShortResult = 0;
      shortResult->selectLine(1);
      updateExtResult();
    }

    showOSD();
    updateOSD();

    return SUCCESS;
  }

  //override operator< from ProgramDescription because we need different sorting-criterias here
  bool ProgramDescription::operator<(const ProgramDescription &pd) const{

    if(startTime.getRValue() < pd.getStartTime()){
      return true;
    }else{
      if(startTime.getRValue() == pd.getStartTime() 
         && title.getRValue().front() < pd.getTitles().front()
         ){

        return true;
      }else{
        if(startTime.getRValue() == pd.getStartTime() 
           && title.getRValue().front() == pd.getTitles().front()
           && sourceType.getRValue() < pd.getSourceType()
           ){
          return true;
        }
      }
    }

    return false;
  }

  //override operator== from ProgramDescription because we need different sorting-criterias here
  bool ProgramDescription::operator==(const ProgramDescription &pd) const{
    if(startTime.getRValue() == pd.getStartTime() 
       && sourceType.getRValue() == pd.getSourceType()
       && title.getRValue().front() == pd.getTitles().front()       
       && getChannelDescription().getChannelNames().front() == pd.getChannelDescription().getChannelNames().front()
       ){
      return true;
    }else{
      return false;
    }
    return false;
  }



}//end namespace NMM
