/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        David Maass <dama@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <stdio.h>

#include "TranscodeConfig.hpp"

#include "nmm/utils/utils.hpp"
#include "nmm/utils/gdparse/GDParse.hpp"
#include "nmm/base/registry/ClientRegistry.hpp"
#include "nmm/interfaces/file/ISinkFileHandler.hpp"

#include "nmm/base/ProxyObject.hpp"
#include "nmm/base/registry/GraphDescription.hpp"

extern list<NodeConfiguration> node_configuration;

namespace NMM {
  TranscodeConfig::TranscodeConfig(IOSDManager* _manager, IWidgetFactory* _factory, TranscodeTemplateFunctions* _state):
    gd_win(NULL),
    conf1_win(NULL),
    comment_win(NULL),
    state(_state),
    conf2_win(NULL),
    app(NULL),
    rootpath(""),
    mode(NORMAL)
  {
    manager.reset(dynamic_cast<IOSDManager*>(_manager->copy()));
    factory.reset(dynamic_cast<IWidgetFactory*>(_factory->copy()));
    
    dirreader.setNiceDir( true );
    focus = JOBS;
  
    elements.clear();
  }

  TranscodeConfig::~TranscodeConfig() {
    deleteWindows();
  }

  void TranscodeConfig::setApplication(MMBoxApplication* _app) {
    app=_app;
  }

  void TranscodeConfig::setWinProperties(int nr, win_properties_t* prop) {
    memcpy( &winprops[nr], prop, sizeof( win_properties_t ) );
  }

  void TranscodeConfig::deleteWindows() {
    if (conf1_win) {
      delete conf1_win;
      conf1_win = NULL;
    }
    if (conf2_win) {
      delete conf2_win;
      conf2_win = NULL;
    }
   
    if (gd_win) {
      delete gd_win;
      gd_win = NULL;
    }
    if (comment_win) {
      delete comment_win;
      comment_win=NULL;
    }
    
    elements.clear();
  }

  void TranscodeConfig::reset() {
    conf1_win->deleteAllLines();
    conf1_win->deSelectLine();
    conf2_win->deleteAllLines();
    conf2_win->selectLine(0);
    conf2_win->selectLine2(0);
    gd_win->deleteAllLines();
    gd_win->deSelectLine();
    comment_win->deleteAllLines();
    comment_win->deSelectLine();
    selectWindow(0);
  }

  void TranscodeConfig::createWindows() {
    focus=JOBS;
    mode=NORMAL;
    gd_win = new ListUnit(manager.get(),factory.get(),"medium");
    elements.push_back(gd_win);
    
    comment_win = new ListUnit(manager.get(),factory.get(),"medium");
    elements.push_back(comment_win);

    conf1_win = new ListUnit(manager.get(),factory.get(),"medium");
    conf2_win = new DoubleListUnit(manager.get(),factory.get(),"medium");
    conf2_win->setFirstColumnWidth(60);

    elements.push_back(conf1_win);
    elements.push_back(conf2_win);
        

    for (uint i=0;i<4;i++) {
      elements[i]->setHeader(winprops[i].name);

      /* set colors according to properties */
      elements[i]->setColors( winprops[i].win_active_color,
                              winprops[i].win_inactive_color,
                              winprops[i].sel_active_color,
                              winprops[i].sel_active_color2,
                              winprops[i].sel_inactive_color );
      
      /* set width and height of window */
      elements[i]->setDimension( winprops[i].width, winprops[i].height );
      
      /* set position of the window */
      elements[i]->setPosition( winprops[i].xpos, winprops[i].ypos );
      
    }
      
    gd_win->selectWindow();
    comment_win->deSelectWindow();
    conf1_win->deSelectWindow();
    conf2_win->deSelectWindow();
        
    gd_win->setHeader("Jobs");
    comment_win->setHeader("Job comment");
    conf1_win->setHeader("Nodes");
    conf2_win->setHeader("Options");
  }

  Result TranscodeConfig::colorKeyRed() {
    switch (focus) {
    case 1:
    case 2:
      conf1_win->deleteAllLines();
      conf2_win->deleteAllLines();
      selectWindow(0);
      return SUCCESS;
    default:
      return SUCCESS;
    }
  }

  Result TranscodeConfig::colorKeyGreen() {
    if (mode == NORMAL && gd_win->getTextWidget()->getSelectedLine() != 0) {
      int nr = atoi(gd_win->getTextWidget()->getLineURL(gd_win->getTextWidget()->getSelectedLine()).c_str());
      string url = dir[nr].path+dir[nr].name;
      manager->displayMessage( "Please Wait...", "creating Jobs" , "", "", "");
      manager->update();
      state->doSelectGDWindow( graphVector[gd_win->getSelectedLine()-1],configVector[gd_win->getSelectedLine()-1],dir[nr].name);
      manager->deleteMessage();
      manager->update();
    }
    return SUCCESS;
  }

  Result TranscodeConfig::colorKeyBlue() {
    return select();
  }

  Result TranscodeConfig::select() {
    switch (focus) {
    case JOBS:
      break;
    case COMMENT:
      break;
    case NODES:
      break;
    case CONFIGNODES:
      if (mode==CONFIG) {
        mode = NORMAL;
        if (conf2_win->getLineUrl(conf2_win->getSelectedLine()).find("edgeformat") != string::npos) {
          Format* fom = new Format();
          
          int edge;
          for (uint i=1; i<=conf2_win->getNumLines();i++) {
            string ac_line = conf2_win->getLineUrl(i);
            int pos1 = ac_line.find_first_of('=');
            int pos2 = ac_line.find_first_of('&');
            int pos3 = ac_line.find_last_of('=');
            edge = atoi(ac_line.substr(pos1+1,string::size_type(pos2-pos1-1)).c_str());
            string type = ac_line.substr(pos3+1);
            
            if (type=="int") {
              fom->addIntParamValue(conf2_win->getLineSubject(i),atoi(conf2_win->getLineContent(i).c_str()));
            } else if (type=="float") {
              fom->addFloatParamValue(conf2_win->getLineSubject(i),atof(conf2_win->getLineContent(i).c_str()));
            } else if (type=="string") {
              fom->addStringParamValue(conf2_win->getLineSubject(i),conf2_win->getLineContent(i));
            }
          }
          graphVector[gd_win->getSelectedLine()-1]->getEdge(edge).setFormat(fom);
        } else {
          int n = conf2_win->getNumLines();
          option_backup.clear();
          for (int i=1; i<=n; ++i) {
            option_backup.push_back(conf2_win->getLineContent(i));
            string val   = conf2_win->getLineUrl( i );
            string value = conf2_win->getLineContent( i );
            state->setConfiguration( val, value );
          }
        }
        int pos = conf2_win->getSelectedLine();
        conf2_win->selectLine(pos);
        conf2_win->update();
        updateColorBar();
        manager->update();
      }
      else {
        mode = CONFIG;
        int pos = conf2_win->getSelectedLine();
        conf2_win->selectLine2(pos);
        conf2_win->update();
        updateColorBar();
        manager->update();
      }
      break;
    }
    return SUCCESS;
  }

  Result TranscodeConfig::selectLeft() {
    if (mode == CONFIG) {
      conf2_win->selectLeft();
      conf2_win->update();
    } else {
      if (focus==0)
        selectWindow(3);
      else selectWindow(focus-1);
    }
    manager->update();
    return SUCCESS;
  }

  Result TranscodeConfig::selectRight() {
    if (mode == CONFIG) {
      conf2_win->selectRight();
      conf2_win->update();
    } else {
      if (focus==3)
        selectWindow(0);
      else selectWindow(focus+1);
    }
    manager->update();
    return SUCCESS;
  }

  Result TranscodeConfig::selectDown() {
    elements[focus]->selectDown();
    elements[focus]->update();
    selectupdown();
    manager->update();

    updateColorBar();
    return SUCCESS;
  }

  Result TranscodeConfig::selectUp() {
    elements[focus]->selectUp();
    elements[focus]->update();
    selectupdown();
    manager->update();
    
    updateColorBar();
    return SUCCESS;
  }

  void TranscodeConfig::selectupdown() {
    if (mode==CONFIG){
      int pos = conf2_win->getSelectedLine();
      conf2_win->selectLine2(pos);
      conf2_win->update();
    }
    switch(focus) {
    case JOBS:
      {
        comment_win->deleteAllLines();
        int nr = gd_win->getTextWidget()->getSelectedLine()-1;
        comment_win->addMultiLine(commentVector[nr]);
        comment_win->update();
        conf1_win->deleteAllLines();
        conf1_win->addLine("General Options");
        for (uint i=0;i<(nodedescriptionVector[nr]).size();i++) {
          if (nodedescriptionVector[nr][i]) {
            //just add nodes with options
            list<int> temp = graphVector[nr]->getOutEdgeList(i);
            for (list<int>::iterator iter=temp.begin();iter!=temp.end();iter++) {
              Format* fom = graphVector[gd_win->getSelectedLine()-1]->getEdge(*iter).getFormat();
              //if we have a format we have to display the values in the GUI to allow user changes!
              if (fom) {
                conf1_win->getTextWidget()->addLine(nodedescriptionVector[nr][i]->getNodeName(),itoa(i));
              }
            }
          }
        }
        conf1_win->update();
        conf2_win->deleteAllLines();
        conf2_win->update();
        break;
      }
    case NODES:
      analyzeNodeConfig();
      break;
    default:
      break;
    }
    manager->update();
  }

  void TranscodeConfig::analyzeNodeConfig() {
    if(conf1_win->getTextWidget()->getSelectedLine()-1==0) {
      globalOptions();
      conf2_win->update();
    } else {
      conf2_win->deleteAllLines();
      //get out-edges of current selected node
      list<int> temp = graphVector[gd_win->getSelectedLine()-1]->getOutEdgeList(atoi(conf1_win->getTextWidget()->getLineURL(conf1_win->getSelectedLine()).c_str()));//-2);
      //is there a format set?
      for (list<int>::iterator i=temp.begin();i!=temp.end();i++) {
        Format* fom = graphVector[gd_win->getSelectedLine()-1]->getEdge(*i).getFormat();
        //if we have a format we have to display the values in the GUI to allow user changes!
        if (fom) {
          //get a list of all set parameters, get type and values and insert them into our conf2_win doublelistunit
          parameter_list param_list = fom->getParameterList();
          for (list<Parameter*>::iterator j=param_list.begin();j!=param_list.end();j++) {
            Parameter::PType type = (*j)->getParameterType();
            DoubleListUnit::line_t line;
            string parametername = (*j)->getParameterName();
            string paramtype;
            switch(type) {
            case Parameter::INT32: {
              paramtype="int";
		list<int> vlist = fom->getIntValueList(parametername);
		for (list<int>::iterator iter=vlist.begin();iter!=vlist.end();iter++){
                char str[10];
                sprintf(str,"%i",(*iter));
                line.push_back(str);
              }
              break;
            }
            case Parameter::FLOAT: {
              paramtype="float";
              list<float> vlist = fom->getFloatValueList(parametername);
              for (list<float>::iterator iter=vlist.begin();iter!=vlist.end();iter++) {
                char str[10];
                sprintf ( str, "%f", (*iter) );
                line.push_back(str);
              }
              break;
            }
            case Parameter::STRING: {
              paramtype="string";
              list<string> vlist = fom->getStringValueList(parametername);
              for (list<string>::iterator iter=vlist.begin();iter!=vlist.end();iter++) {
                line.push_back((*iter));
              }
              break;
            }
            default:
              cerr<<"currently unsupported Format-Param"<<endl;
              return;
              break;
            }
            conf2_win->addLine(parametername,line,string("edgeformat?edge=")+(*i)+"&type="+paramtype);
          }
        }
      }
      conf2_win->update();
    }
  }

  void TranscodeConfig::selectWindow(int winid) {
    elements[focus]->deSelectWindow();
    elements[winid]->selectWindow(0);
    if (!elements[winid]->getSelectedLine())
      elements[winid]->selectLine(1);
    focus = (myfocus)winid;

    //if the new focus is in the NODES window, we have to show the global options
    if (focus==NODES)
      analyzeNodeConfig();

    update();
    updateColorBar();
    manager->update();
  }

  void TranscodeConfig::setRootpath(string path) {
    rootpath=path;
  }
  
  void TranscodeConfig::analyzeDir() {
    commentVector.clear();
    nodedescriptionVector.clear();
    dirreader.setRootDir( rootpath );
    dirreader.changeDir( "." );
    dir = dirreader.getDirEntries();
    gd_win->deleteAllLines();
    for ( uint i=0;i<graphVector.size(); i++) {
      delete graphVector[i];
      delete configVector[i];
    }
    graphVector.clear();
    configVector.clear();
    for ( unsigned int i=0; i < dir.size(); i++ ) {
      if (! dir[i].isdir ) {
        if (dirreader.getSuffix(dirreader.getFileName( dir[i].name, 30 )) == "gd") {
          //Create GraphDescription from .gd file
          gd_win->addLine( dirreader.getFileName( dir[i].name, 30 ), itoa(i) );
          string url = dir[i].path+dir[i].name;
          node_configuration.clear();
          GraphDescription* gr = new GraphDescription();
          getGraphfromGD(url,*gr);
          configVector.push_back(new list<NodeConfiguration>(node_configuration));
          commentVector.push_back(gr->getComment());
          nodedescriptionVector.push_back(gr->getNodeDescriptionList());
          graphVector.push_back(gr);
        }
      }
    }
    
    gd_win->selectLine(0);
    focus = JOBS;
    selectupdown();
    
    update();
  }
  
  void TranscodeConfig::globalOptions() {
    conf2_win->deleteAllLines();
    int listpos = -1;
    for (map<string,string>::const_iterator s = shown_options.begin();
         s != shown_options.end();
         ++s) {
      listpos++;
      const MMBoxConfig::Setting* setting;
      if (!config->getSetting(s->first, setting)) {
        cerr<<"Unknown setting: " << s->first << endl;
        GLOBAL_ERROR_STREAM("Unknown setting: " << s->first << endl);
        continue;
      }
      string val(s->second.empty() ? setting->getDescription() : s->second);
      string shortform(s->first);
      const list<string>& values(setting->getValues());
      
      DoubleListUnit::line_t line;
      
      /* add all values */
      string default_value;
      if (!option_backup.empty()) {
        default_value=option_backup[listpos];
        line.push_back(option_backup[listpos]);
      }

      int i;
      //Check which default-parameters were defined in .mmboxrc, and add them as default_value in the 
      //doublelistunit too
      //as parameters in .mmbox all start with lowercase letters, convert the first one
      string myval(val);
      while (myval.find(" ")!=string::npos)
        myval[myval.find(" ")]='_';
      for (unsigned int i=0; i<myval.size(); i++)
        myval[i] = tolower(myval[i]);

      if (config->getValue(myval,i)) {
        if (myval!="parallel") {
          if (i) line.push_back("yes"); else line.push_back("no");
        } else {
          line.push_back(itoa(i));
        }
      }

      for (list<string>::const_iterator i = values.begin();
           i != values.end();
           ++i) {
        string value = *i;
        if (!(value==default_value)) {
          //just add new values
          vector<string>::const_iterator j = find(line.begin(),line.end(),value);
          if (j==line.end())
            line.push_back( value );
        }
        if (default_value.empty()) {
          default_value = value;
        }
      }
      
      /* add human readable configuration text to list */
      conf2_win->addLine( val, line, shortform );
    }
  }

  void TranscodeConfig::updateColorBar() {
    switch(focus) {
    case CONFIGNODES:
      if (mode == NORMAL)
        manager->setColorBarText("","Start job","","Edit opt");
      else
        manager->setColorBarText("","","","Set opt");
      break;
    default:
      if (gd_win->getTextWidget()->getSelectedLine() != 0) {
        manager->setColorBarText("","Start job","","");
      } else
        manager->setColorBarText("","","","");
      break;
    }
    manager->update();
  }

  string TranscodeConfig::getScript() {
    string script = gd_win->getLineContent(gd_win->getSelectedLine());
    //cut away the ".gd"
    script = dirreader.getRootDir()+script.substr(0,script.size()-3);
    return script;
  }
}
