/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Eric Peters <epeters@graphics.cs.uni-sb.de>
 *                    Christian Gerstner <gerstner@graphics.cs.uni-sb.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <qapplication.h>
#include <qobject.h>
#include <qsocketnotifier.h>

#include <iostream>
#include <getopt.h>
#include <unistd.h>

#include "sigreader.hpp"

void usage() {
  cerr << "Usage: qvcorerecorder [-v | --verbose (false)]" << endl;
  cerr << "                      [-c | --channel (1)]" << endl;
  cerr << "                      [-s | --server/port]" << endl;
  cerr << "                      [-o | --outputfile]" << endl;
  cerr << "                      [-V | --usevisca]" <<endl;
  cerr << "                      [-h | --help]" << endl;
  cerr << "" << endl;
  exit(1);
}

void turnOffAllMessages(int verbosity)
{
  if (verbosity < 1) {
    NMM::NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbosity < 2) {
    NMM::NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbosity < 3) {
    NMM::NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbosity < 4) {
    NMM::NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  }
}

int main( int argc, char **argv )
{
  int verbosity = 0;
  // the channel for the tv card; this could be improved by setting it within the GUI
  int channel = 2;
  bool use_visca = false;
  string vncserver_port = string("alias:0");
  string outputfile = string("default.avi");
  QApplication a( argc, argv );
  
  // parse the command line parameters
  const char *optstring = "vc:s:o:hV";    

  struct option longopts[] = {
    {"verbose"     , 0, NULL, 'v'},
    {"channel"     , 1, NULL, 'c'},
    {"server/port" , 1, NULL, 's'},
    {"outputfile"  , 1, NULL, 'o'},
    {"help"        , 0, NULL, 'h'},
    {"usevisca"    , 0, NULL, 'V'},
    {NULL          , 0, NULL, 0  } // last row must be zero
  };

  int ch, optionindex;

  while(EOF != (ch=getopt_long(argc,argv,optstring,longopts,&optionindex))){ 
    switch (ch) {
    case 'v':
      verbosity++;
      break;

    case 'c':
      if (optarg) {
        channel = atoi(optarg);
      }
      break;

    case 's':
      if (optarg) {
        vncserver_port = string(optarg);
      }
      break;

    case 'o':
      if (optarg) {
        outputfile = string(optarg);
      }
      break;

    case 'h': // help
      usage();
      break;
 
    case 'V':
      use_visca = true;
      break;
      
    default:
      GLOBAL_ERROR_STREAM("Unknown command line parameter" << endl);
      usage();
    }   
  }

  // eventually turn off debug messages
  turnOffAllMessages(verbosity);
  
  /*Begin> create the NMM application */
  NMM::NMMApplication* app = ProxyApplication::getApplication(argc, argv);
  /*End< */
    
  GuiWidget w(*app);
  w.setCaption("VCORE - MainWindow");
  w.setGeometry(5,5,400,300);
  MDIWindow mdiwindow(*app);
  mdiwindow.setCaption("VCORE - Virtual Courseroom Environment");
  ContainerWidget containerwidget(mdiwindow.getWorkspace());
  
  // create our application
  SigReader* sigReader = new SigReader(&w, *app, &mdiwindow, &containerwidget);
  sigReader->setChannel(channel);
  sigReader->setUseVisca(use_visca);

  // create pipe:
  int pipefd[2] = {-1, -1};
  int vnc_pipefd[2] = {-1, -1};
  if ( pipe(pipefd) != 0 || pipe(vnc_pipefd) != 0 ) {
    GLOBAL_ERROR_STREAM("Could not create the pipe" << endl);
    throw Exception("Could not create the pipe");
  }
  sigReader->setReadFD(pipefd[0]);
  sigReader->setWriteFD(pipefd[1]);
  sigReader->setVNCReadFD(vnc_pipefd[0]);
  sigReader->setVNCWriteFD(vnc_pipefd[1]);

  QSocketNotifier* sn = new QSocketNotifier( pipefd[0], QSocketNotifier::Read, &w);
  QObject::connect( sn, SIGNAL(activated(int)), sigReader, SLOT(handleEvent()) );
  QSocketNotifier* vnc_sn = new QSocketNotifier( vnc_pipefd[0], QSocketNotifier::Read, &w);
  QObject::connect( vnc_sn, SIGNAL(activated(int)), sigReader, SLOT(handleVNCEvent()) );

  //MDIWindow
  QObject::connect(&mdiwindow, SIGNAL(mdiWindowClosed()), sigReader, SLOT(deinit()));
  QObject::connect(&mdiwindow, SIGNAL(mdiWindowClosed()), qApp, SLOT(quit()));
  QObject::connect(mdiwindow.getComboBoxMode(), SIGNAL(activated(int)) , sigReader, SLOT(switchVNCMode(int)));
  QObject::connect(mdiwindow.getButtonGroupColor(), SIGNAL(clicked(int)), sigReader, SLOT(setSVGDrawColor(int)));
  QObject::connect(mdiwindow.getButtonGroupDrawType(), SIGNAL(clicked(int)), sigReader, SLOT(setSVGDrawType(int)));
  QObject::connect(mdiwindow.getToolButtonUndo(), SIGNAL(clicked()), sigReader, SLOT(undoSVGAnnotation()));
  QObject::connect(mdiwindow.getToolButtonRedo(), SIGNAL(clicked()), sigReader, SLOT(redoSVGAnnotation()));
  QObject::connect(mdiwindow.getToolButtonClear(), SIGNAL(clicked()), sigReader, SLOT(newSVGAnnotation()));

  //Main-Interface
  QObject::connect(w.getButtonStart(), SIGNAL(clicked()), sigReader, SLOT(start()));
  QObject::connect(w.getButtonStop(), SIGNAL(clicked()), sigReader, SLOT(stop()));
  QObject::connect(w.getButtonPreview(), SIGNAL(clicked()), sigReader, SLOT(preview()));
  QObject::connect(w.getEditVNCHostAndPort(), SIGNAL(textChanged(const QString&)),
	  sigReader, SLOT(setVNCHostAndPort(const QString&)));
  QObject::connect(w.getEditVNCPassword(), SIGNAL(textChanged(const QString&)),
	  sigReader, SLOT(setVNCPassword(const QString&)));
  QObject::connect(w.getEditOutFileName(), SIGNAL(textChanged(const QString&)),
	  sigReader, SLOT(setOutputFileName(const QString&)));
  QObject::connect(w.getButtonOutFileName(), SIGNAL(clicked()), sigReader, SLOT(setOutputFileName()));

  QObject::connect(sigReader, SIGNAL(outputFileNameChanged(const QString&)),
	  w.getEditOutFileName(), SLOT(setText(const QString&)) ); 
  
  QObject::connect(w.getSliderVBR(), SIGNAL(valueChanged(int)), sigReader, SLOT(setVideoBitrate(int)) );
  QObject::connect(w.getSliderABR(), SIGNAL(valueChanged(int)), sigReader, SLOT(setAudioBitrate(int)) );
  QObject::connect(w.getCheckBoxDisplay(), SIGNAL(toggled(bool)), sigReader, SLOT(setUseDisplay(bool)) );
  if (use_visca) {
    QObject::connect(w.getCheckBoxVisca(), SIGNAL(toggled(bool)), sigReader, SLOT(setUseVisca(bool)) );
  } else {
    w.getCheckBoxVisca()->hide();
  }
  QObject::connect(w.getCheckBoxDeinterlacer(), SIGNAL(toggled(bool)), sigReader, SLOT(setUseDeinterlacer(bool)) );
  QObject::connect(w.getWidthBox(), SIGNAL(valueChanged(int)), sigReader, SLOT(setWidth(int)) );
  QObject::connect(w.getHeightBox(), SIGNAL(valueChanged(int)), sigReader, SLOT(setHeight(int)) );
  QObject::connect(w.getButtonQuit(), SIGNAL(clicked()), sigReader, SLOT(deinit()));
  QObject::connect(w.getButtonQuit(), SIGNAL(clicked()), qApp, SLOT(quit()));

  a.setMainWidget( &w );
  w.setDefaultValues(QString(outputfile.c_str()), QString(vncserver_port.c_str()));
  w.show();
  return a.exec();
}
