/* 
 * Copyright (C) 2005 Andres Oton Urbano
 *
 * Maintainer:        Andres Oton <andres.oton@gmail.com> 
 * Author:            Andres Oton <andres.oton@gmail.com>
 * From:  PlaybackNode.cpp by NMM work group to use Alsa
 *        and aplay from ALSA Project tools.
 *                     
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "ALSAPlaybackNode.hpp"
#include "nmm/utils/utils.hpp"
#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/base/registry/Plugin.hpp"

#include <sys/soundcard.h>
#include <alsa/asoundlib.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/stat.h>
#include <fcntl.h>

static NMM::ALSAPlaybackNodePlugin plugin("ALSAPlaybackNode",-1);
NMM_REGISTER_SPECIAL_PLUGIN(plugin)


namespace NMM {
	//! Formats to check in the doInit method.
	snd_pcm_format_t ALSAPlaybackNode::test_formats[] = {
		SND_PCM_FORMAT_MU_LAW,  //!< Logarithmic mu-law audio encoding
		SND_PCM_FORMAT_A_LAW,   //!< Logarithmic A-law audio encoding
		SND_PCM_FORMAT_U8,      //!< Linear encoding, unsigned 8-bit
		SND_PCM_FORMAT_S8,      //!< Linear encoding, signed 8-bit
		SND_PCM_FORMAT_U16_LE,  //!< Linear encoding, unsigned 16-bit, little endian
		SND_PCM_FORMAT_U16_BE,  //!< Linear encoding, unsigned 16-bit, big endian
		SND_PCM_FORMAT_S16_LE,  //!< Linear encoding, signed 16-bit, little endian
		SND_PCM_FORMAT_S16_BE,  //!< Linear encoding, signed 16-bit, big endian
		SND_PCM_FORMAT_UNKNOWN 
		//!< End marker. If you add or remove some formats, make sure that the last one remains 0.
	};

	//! Channel numbers to check in the doInit method.
	const int ALSAPlaybackNode::test_channels[] = {
		1,      //!< Mono
		2,      //!< Stereo
		4,      //!< 4
		5,      //!< 5
		6,      //!< 5.1
		0
		//!< End marker. If you add or remove some values, make sure that the last one remains 0.
	};

	//! Sample rates to check in the doInit method.
	const int ALSAPlaybackNode::test_sample_rates[] = { 8000, 11025, 22050, 44100, 48000, 96000, 0 };

	Format& ALSAPlaybackNode::getNMMFormat(snd_pcm_format_t alsa_format, Format& nmm_format)
	{
		switch(alsa_format) {
			case SND_PCM_FORMAT_MU_LAW:
				nmm_format.setFormatType("audio/raw");
				nmm_format.addStringParamValue(Format::encoding_param, "mu_law");
				nmm_format.addIntParamValue(Format::bitpersample_param, 8);
				break;

			case SND_PCM_FORMAT_A_LAW:
				nmm_format.setFormatType("audio/raw");
				nmm_format.addStringParamValue(Format::encoding_param, "a_law");
				break;

			case SND_PCM_FORMAT_U8:
				nmm_format.setFormatType("audio/raw");
				nmm_format.addStringParamValue(Format::encoding_param, "linear");
				nmm_format.addStringParamValue(Format::sign_param, "unsigned");
				nmm_format.addIntParamValue(Format::bitpersample_param, 8);
				break;

			case SND_PCM_FORMAT_S8:
				nmm_format.setFormatType("audio/raw");
				nmm_format.addStringParamValue(Format::encoding_param, "linear");
				nmm_format.addStringParamValue(Format::sign_param, "signed");
				nmm_format.addIntParamValue(Format::bitpersample_param, 8);
				break;

			case SND_PCM_FORMAT_U16_LE:
				nmm_format.setFormatType("audio/raw");
				nmm_format.addStringParamValue(Format::encoding_param, "linear");
				nmm_format.addStringParamValue(Format::sign_param, "unsigned");
				nmm_format.addIntParamValue(Format::bitpersample_param, 16);
				nmm_format.addStringParamValue(Format::endian_param, "little");
				break;

			case SND_PCM_FORMAT_U16_BE:
				nmm_format.setFormatType("audio/raw");
				nmm_format.addStringParamValue(Format::encoding_param, "linear");
				nmm_format.addStringParamValue(Format::sign_param, "unsigned");
				nmm_format.addIntParamValue(Format::bitpersample_param, 16);
				nmm_format.addStringParamValue(Format::endian_param, "big");
				break;

			case SND_PCM_FORMAT_S16_LE:
				nmm_format.setFormatType("audio/raw");
				nmm_format.addStringParamValue(Format::encoding_param, "linear");
				nmm_format.addStringParamValue(Format::sign_param, "signed");
				nmm_format.addIntParamValue(Format::bitpersample_param, 16);
				nmm_format.addStringParamValue(Format::endian_param, "little");
				break;

			case SND_PCM_FORMAT_S16_BE:
				nmm_format.setFormatType("audio/raw");
				nmm_format.addStringParamValue(Format::encoding_param, "linear");
				nmm_format.addStringParamValue(Format::sign_param, "signed");
				nmm_format.addIntParamValue(Format::bitpersample_param, 16);
				nmm_format.addStringParamValue(Format::endian_param, "big");
				break;
    
			default:
				ERROR_STREAM("getNMMFormat: ALSA format " << alsa_format << " not supported." << endl);
		}

		return nmm_format;
	};

	snd_pcm_format_t ALSAPlaybackNode::getALSAFormat(const Format &nmm_format)
	{
		string encoding;
		try {
			encoding = nmm_format.getStringValue(Format::encoding_param);
		} catch (FormatException e) {
			ERROR_STREAM("getALSAFormat: Incomplete format (encoding not specified)."
					<< endl);
			return SND_PCM_FORMAT_UNKNOWN;
		}

		if(encoding == "mu_law") return SND_PCM_FORMAT_MU_LAW;
		if(encoding == "a_law" ) return SND_PCM_FORMAT_A_LAW;
		if(encoding == "linear") {
			int bits;
			try {
				bits = nmm_format.getIntValue(Format::bitpersample_param);
			} catch (FormatException e) {
				ERROR_STREAM("getALSAFormat: Incomplete format (bitpersample not specified)." << endl);
				return SND_PCM_FORMAT_UNKNOWN;
			}
			string sign;
			try {
				sign = nmm_format.getStringValue(Format::sign_param);
			} catch (FormatException e) {
				ERROR_STREAM("getALSAFormat: Incomplete format (sign not specified)." << endl);
				return SND_PCM_FORMAT_UNKNOWN;
			}

			if(bits == 8) {
				if(sign == "unsigned") return SND_PCM_FORMAT_U8;
				if(sign == "signed"  ) return SND_PCM_FORMAT_S8;
				ERROR_STREAM("getALSAFormat: Incorrect format: sign is [" << sign << "]." << endl);
				ERROR_STREAM("getALSAFormat: Possible values for sign are: [unsigned|signed]." << endl);
			}

			if(bits == 16) {
				string endian;
				try {
					endian = nmm_format.getStringValue(Format::endian_param);
				} catch (FormatException e) {
					ERROR_STREAM("getALSAFormat: Incomplete format (endian not specified)." << endl);
					return SND_PCM_FORMAT_UNKNOWN;
				}

			if(sign == "unsigned") {
				if(endian == "little") return SND_PCM_FORMAT_U16_LE;
				if(endian == "big"   ) return SND_PCM_FORMAT_U16_BE;
				ERROR_STREAM("getALSAFormat: Incorrect format: endian is [" << endian << "]." << endl);
				ERROR_STREAM("getALSAFormat: Possible values for endian are [little|big]." << endl);
			}

			if(sign == "signed"  ) {
				if(endian == "little") return SND_PCM_FORMAT_S16_LE;
				if(endian == "big"   ) return SND_PCM_FORMAT_S16_BE;
				ERROR_STREAM("getALSAFormat: Incorrect format: endian is [" << endian << "]." << endl);
				ERROR_STREAM("getALSAFormat: Possible values for endian are [little|big]." << endl);
			}
			ERROR_STREAM("getALSAFormat: Incorrect format: sign is [" << sign << "]." << endl);
			ERROR_STREAM("getALSAFormat: Possible values for sign are [unsigend|signed]." << endl);
		}
		ERROR_STREAM("getALSAFormat: Only 8 or 16 bitpersample are supported by ALSAPlaybackNode."
				<< endl);
		}
		return SND_PCM_FORMAT_UNKNOWN;
	};

	ALSAPlaybackNode::ALSAPlaybackNode(const char* name,
				const StreamQueue::Mode queue_mode,
				const int size)
		: GenericBSyncSinkNode(name, queue_mode, size)
	{
		buffer_time = 0;
		buffer_frames = 0;
		period_time = 0;
		period_frames = 0;

		pcm_access = SND_PCM_ACCESS_RW_INTERLEAVED;

		registerEvent("finished", new TEDObject1<ALSAPlaybackNode,int>(this, &ALSAPlaybackNode::EventDebug));
		
		registerEvent("speed_up", new TEDObject1<ALSAPlaybackNode, int>(this, &ALSAPlaybackNode::eventSpeedUp));
		
		registerEvent("SEEK", new TEDObject3<ALSAPlaybackNode, Interval, Time, long>(this, &ALSAPlaybackNode::eventSeek));

		stream = SND_PCM_STREAM_PLAYBACK;
	}

	ALSAPlaybackNode::~ALSAPlaybackNode() 
	{
	
	}

	Result ALSAPlaybackNode::doInit() {
		DEBUG_STREAM("DoInit: testing devices"<<endl);

		// Open the device nonblocking, so the open call returns immediately.
		if (dev.empty())
			dev = "/dev/dsp";

		if(dev != "/dev/dsp")
			return FAILURE;


		if (__mixer_device.empty())
			__mixer_device = "/dev/mixer";

		if(!SSecurityCheck::getInstance().checkFilenameForWriting(dev)) {
			ERROR_STREAM("Security check not passed for file " << dev << ", check .nmmrc for options" << endl);
			return FAILURE;
		}

		if(!SSecurityCheck::getInstance().checkFilenameForWriting(__mixer_device)) {
			ERROR_STREAM("Security check not passed for file " << __mixer_device << ", check .nmmrc for options" << endl);
			return FAILURE;
		}

		int err;
		dev = "default";
		if ((err = snd_pcm_open (&pcm_handle, dev.c_str(), SND_PCM_STREAM_PLAYBACK, SND_PCM_ASYNC)) < 0) {
			ERROR_STREAM("Cannot open audio device " << dev <<" with error: "<< snd_strerror(err) << endl );
			return FAILURE;
		}
	  	DEBUG_STREAM("Audio device open."<<endl); 
		if ((err = snd_pcm_hw_params_malloc (&hwparams)) < 0) {
			ERROR_STREAM("Cannot allocate hardware parameter structure with error " << snd_strerror (err) << endl);
			return FAILURE;
		}

		/*Init hwparams with full configuration space */
		if ((err = snd_pcm_hw_params_any(pcm_handle, hwparams)) < 0) {
			ERROR_STREAM("Cannot initialize to configure this PCM device. Error: " << snd_strerror(err) << endl);
			return FAILURE;
		}	

		if ( (err = snd_pcm_hw_params_set_access(pcm_handle, hwparams, pcm_access)) < 0) {
			ERROR_STREAM("Error setting access. Error: "<< snd_strerror(err) << endl);
			return FAILURE;
		}


		for(int i=0; test_formats[i] != SND_PCM_FORMAT_UNKNOWN; i++) {
			// First we need a full configuration space
			snd_pcm_hw_params_any(pcm_handle, hwparams);
			snd_pcm_hw_params_set_access(pcm_handle, hwparams, pcm_access);

			// Test all the formats in test_formats which are supported by the hardware.
			if(snd_pcm_hw_params_set_format(pcm_handle, hwparams, test_formats[i]) < 0) {
				DEBUG_STREAM("audio device: - Format " <<test_formats[i] << "not supported."<<endl);
			} else {
				// This format is supported.
				DEBUG_STREAM("audio device: - Format " << test_formats[i] << endl);

				// Test the supported channel numbers for this format.
				// Note: If you need a channel number that is currently not supported by
				//       ALSAPlaybackNode, simply add it to the definition of the array
				//       test_channels in this file.
				for(int j=0; test_channels[j] != 0; j++) {
					int test_channel = test_channels[j];
					// First we need a full configuration space
					snd_pcm_hw_params_any(pcm_handle, hwparams);
					snd_pcm_hw_params_set_access(pcm_handle, hwparams, pcm_access);
					snd_pcm_hw_params_set_format(pcm_handle, hwparams, test_formats[i]);

					if(snd_pcm_hw_params_set_channels(pcm_handle,hwparams,test_channel)<0) {
						DEBUG_STREAM("audio device: - Format "<< test_formats[i]<< " not support "<< test_channel <<" channels."<<endl);
					} else {
						DEBUG_STREAM("audio device: - Format "<< test_formats[i]<< " support " << test_channel << " channels."<<endl);
						// Create a new format.
						Format *input_format = new Format();

						// Configure the parameters of the format according to the actual
						// hardware configuration (since this configuration works.)
						getNMMFormat(test_formats[i], *input_format);

						// Add the format to the input property.
						getOwnInputProperty() -> addNewFormat(input_format);
						if (!getOwnInputProperty()->getDefaultFormat()) {
							// workaround for missing default format
							// setting first format found to default one
							getOwnInputProperty()->setDefaultFormat(input_format);
						}

						// Set channels and sample rate (not for ac3)
						input_format->addIntParamValue(Format::channels_param, test_channels[j]);

						// Last step: Test the supported sample rates for the current channel number
						//            and format.
						// Note: If you need a sample rate that is not tested here, simpy add it to
						//       the definition of the array test_sample_rates in this file.
						for(int k=0; test_sample_rates[k] != 0; k++) {
							int test_rate = test_sample_rates[k];

							// First we need a full configuration space
							snd_pcm_hw_params_any(pcm_handle, hwparams);
							snd_pcm_hw_params_set_access(pcm_handle, hwparams, pcm_access);
							snd_pcm_hw_params_set_format(pcm_handle, hwparams, test_formats[i]);
							snd_pcm_hw_params_set_channels(pcm_handle,hwparams,test_channel);

							if( (err = snd_pcm_hw_params_test_rate(pcm_handle, hwparams, test_rate, 1)) < 0) {
								DEBUG_STREAM("audio device:     + Set sample rate to " << test_sample_rates[k]
										<< " failed. Error: "<< snd_strerror(err) << endl);
							} else {
								if(test_rate == test_sample_rates[k]) {
									// Sample rate is supported with the actual configuration.
									DEBUG_STREAM("audio device:     + Sample rate " << test_sample_rates[k] << endl);

									// Add it to the format in the input property.
								input_format->addIntParamValue(Format::sample_rate_param, test_sample_rates[k]);
								}
							}
						} // loop over the test_sample_rates
					}
				} //loop over test_channels
			}
			
		}

		return SUCCESS;
	};

	Result ALSAPlaybackNode::doDeinit()
	{
		snd_pcm_close(pcm_handle);
		return SUCCESS;
	};

	Result ALSAPlaybackNode::doInitOutput() 
	{

		DEBUG_STREAM("doInitOutput: Setting format device."<<endl);
		snd_pcm_hw_params_any(pcm_handle,hwparams);

		const Format* input_format = getInputFormat();
		if (input_format == NULL) { 
			DEBUG_STREAM("Not setting input_format."<<endl);
			return FAILURE;
		}
		Format* format;

		format = new Format( *input_format );
		getOwnWorkingInputProperty()->addNewFormat(format);
		getOwnWorkingInputProperty()->setDefaultFormat(format);

		Format tmp_format = *input_format;
		
		if(setFormatInternal(tmp_format) == FAILURE) {
			return FAILURE;
		}

		if(snd_pcm_hw_params(pcm_handle, hwparams) < 0) {
			ERROR_STREAM("audio device: Error setting HW params."<<endl);
			return FAILURE;
		}
		
		return SUCCESS;
	}

	Result ALSAPlaybackNode::doFlush() 
	{

		DEBUG_STREAM("DoFlush start"<<endl);
		snd_pcm_drop(pcm_handle);
		if(doInitOutput() == FAILURE) {
			ERROR_STREAM("doFlush(): I can set output");
			return FAILURE;
		}
		// snd_pcm_start(pcm_handle);
		DEBUG_STREAM("DoFlush end"<<endl);
		return SUCCESS;

	};

	Result ALSAPlaybackNode::doStart() 
	{
		return SUCCESS;
	};

	Result ALSAPlaybackNode::doStop()
	{   
		//This is not need and break the pause
		// snd_pcm_drain(pcm_handle);

		return SUCCESS;
	};

	Message* ALSAPlaybackNode::presentBuffer(Buffer *in_buffer)
	{
		unsigned int used = in_buffer->getUsed();
		unsigned int frames = used / (byte_per_sample*channels);
		unsigned int written = 0;
		ssize_t r;

		while(written < frames) {
			if (pcm_access == SND_PCM_ACCESS_RW_NONINTERLEAVED) {
//				written += snd_pcm_writen(pcm_handle, &(in_buffer->getData() + (written*byte_per_sample)), frames - written); 
			}
			else {

				r = snd_pcm_writei(pcm_handle, in_buffer->getData() + (written*byte_per_sample*channels), frames - written); 
				if (r == -EAGAIN || (r >= 0 && (size_t)r < frames)) {
					 snd_pcm_wait(pcm_handle, 1000);
				} else if (r == -EPIPE) {
					//COMPROBAR VALOR DE XRUN
					xrun();
				} else if (r == -ESTRPIPE) {
					//COMPROBAR VALOR DE SUSPEND
					suspend();
				} else if (r < 0) {
					ERROR_STREAM("WRITE ERROR: "<< snd_strerror(r)<<endl);
					return NULL;
				}
				if (r > 0) {
					DEBUG_STREAM("++++ Written "<<r<<" frames at buffer."<<endl);
					written += r;
				}
					
			}
		}

		// written_bytes += used;
		in_buffer->release();
      
		getOutputDelay();
		return 0;
	}

	/* I/O error handler */
	Result ALSAPlaybackNode::xrun(void)
	{
		snd_pcm_status_t *status;
		int res;

		snd_pcm_status_alloca(&status);
		if ((res = snd_pcm_status(pcm_handle, status))<0) {
			ERROR_STREAM("xrun(): status error: "<< snd_strerror(res)<<endl);
			return FAILURE;
		}
		if (snd_pcm_status_get_state(status) == SND_PCM_STATE_XRUN) {
			struct timeval now, diff, tstamp;
			gettimeofday(&now, 0);
			snd_pcm_status_get_trigger_tstamp(status, &tstamp);
			timersub(&now, &tstamp, &diff);
			DEBUG_STREAM("xrun(): underrun!!! (at least " << 
			diff.tv_sec * 1000 + diff.tv_usec / 1000.0 << " ms long."<<endl);
			if ((res = snd_pcm_prepare(pcm_handle))<0) {
				ERROR_STREAM("xrun(): prepare error: "<< snd_strerror(res)<<endl);
				return FAILURE;
			}
			return SUCCESS;         /* ok, data should be accepted again */
		} if (snd_pcm_status_get_state(status) == SND_PCM_STATE_DRAINING) {
			DEBUG_STREAM("xrun(): Status(DRAINING)"<<endl);
			// snd_pcm_status_dump(status, log);
		}
		ERROR_STREAM("xrun(): read/write error, state = "<< snd_pcm_state_name(snd_pcm_status_get_state(status)) <<endl );
		return FAILURE;
	}

	/* I/O suspend handler */
	Result ALSAPlaybackNode::suspend(void)
	{
		int res;

		DEBUG_STREAM("Suspended. Trying resume. "<<endl);
		while ((res = snd_pcm_resume(pcm_handle)) == -EAGAIN)
			sleep(1);       /* wait until suspend flag is released */
		if (res < 0) {
			DEBUG_STREAM("Failed. Restarting stream. "<<endl); 
			if ((res = snd_pcm_prepare(pcm_handle)) < 0) {
				ERROR_STREAM("suspend(): prepare error: "<< snd_strerror(res)<< endl);
				return FAILURE;
			}
		}
		DEBUG_STREAM("Suspend Done."<<endl);
		return SUCCESS;
	}

	Result ALSAPlaybackNode::oldSetDevice(string &value)
	{
		dev = value;
		DEBUG_STREAM("Set device to " << dev << endl);
		
		// check dev here?
		return SUCCESS;
	}

	Result ALSAPlaybackNode::oldGetDevice(string &value)
	{
		value = string(dev);
		return SUCCESS;
	}


	void ALSAPlaybackNode::setDevice(const string& device)
		throw (const DeviceNotFoundException&)
	{
		DEBUG_STREAM("setDevice: THIS METHOD IS IGNORED because ALSA doesn't use devices. Included by compatibility."<<endl);
		this->dev = device;
		return;
		
		/* now we can set the device */
		this->dev = device;
	}

	void ALSAPlaybackNode::setMixerDevice(const string& mixer_device)
		throw (const DeviceNotFoundException&)
	{
		DEBUG_STREAM("setMixerDevice: Still mixer device isn't supported."<<endl);
		__mixer_device = mixer_device;
   		return;

	}

	string ALSAPlaybackNode::getDevice() const {
		return dev;
	}

	string ALSAPlaybackNode::getMixerDevice() const {
		return __mixer_device;
	}


	Result ALSAPlaybackNode::eventSeek(Interval& /*i*/, Time& /*t*/, long& /*num_bytes*/) {
		ERROR_STREAM2("Seeking not implemented" << std::endl);    
		return SUCCESS;
	}
  
	Result ALSAPlaybackNode::eventSpeedUp(int& speedup) 
	{
		DEBUG_STREAM ("SpeedUp " << speedup << endl);
		return SUCCESS;
	}

	Result ALSAPlaybackNode::EventDebug(int& what) 
	{
		DEBUG_STREAM ("************************************************" << what << endl);
		return SUCCESS;
	}

	Result ALSAPlaybackNode::startTrack(const string& what)
	{
		DEBUG_STREAM ("****************************************** StartTrack " << what << endl);
		return SUCCESS;
	}

	Result ALSAPlaybackNode::endTrack()
	{
		DEBUG_STREAM ("****************************************** EndTrack " << endl);
		return SUCCESS;
	}

	Result ALSAPlaybackNode::finished()
	{
		DEBUG_STREAM ("****************************************** Finished " << endl);
		return SUCCESS;
	}

	Interval ALSAPlaybackNode::getOutputDelay()
	{
		DEBUG_STREAM("ALSAPlaybackNode::getOutputDelay" << endl);

		int buffered_bytes;
		snd_pcm_sframes_t frames;

		if(snd_pcm_status_malloc(&pcm_status) >= 0)
		{
			frames = snd_pcm_status_get_delay(pcm_status);
			buffered_bytes = frames * byte_per_sample*channels;
			
			DEBUG_STREAM("getOutputDelay: " << buffered_bytes
				<< " in internal buffer." << endl);
			Interval odelay(Rational(buffered_bytes, byte_per_second));

			DEBUG_STREAM("getOutputDelay: " << odelay << " of delay."
				<< endl);
			snd_pcm_status_free(pcm_status);

			return odelay;
		}
		else {
			ERROR_STREAM("getOutputDelay: Error while calculating delay"<<endl);
			return Interval();
		}
	};

	Result ALSAPlaybackNode::setFormatInternal(Format& input_format)
	{
		int err;
		snd_pcm_format_t format = SND_PCM_FORMAT_UNKNOWN;
		DEBUG_STREAM("Setting format to audio device."<<endl);

		// Get format parameters
		format  = getALSAFormat(input_format);
		if (format == SND_PCM_FORMAT_UNKNOWN) return FAILURE;
		channels    = input_format.getIntValue(Format::channels_param);
		sample_rate = input_format.getIntValue(Format::sample_rate_param);
		DEBUG_STREAM("Got Format: samplerate: " << sample_rate << ", channels: "
									<< channels << endl);

		try {
			bit_per_sample = input_format.getIntValue(Format::bitpersample_param);
		} catch(...) {
			bit_per_sample = 16;
		}
		if (bit_per_sample == 0) bit_per_sample = 16;

		byte_per_sample = bit_per_sample / 8;
		byte_per_second = sample_rate * channels * byte_per_sample;


		// It is important always to keep the same order in setting the parameters:
		// At first the access, format, then the channel number and at last the sample rate.
		DEBUG_STREAM("audio device: Trying to set access. " << endl);

		if ( (err = snd_pcm_hw_params_set_access(pcm_handle, hwparams, pcm_access)) < 0) {
			ERROR_STREAM("Error setting access at init output. Error: "<< snd_strerror(err) << endl);
			return FAILURE;
		}

		DEBUG_STREAM("audio device: Trying to set format " << format << endl);

		/* Set sample format */
		if (snd_pcm_hw_params_set_format(pcm_handle, hwparams, format) < 0) {
			ERROR_STREAM("Error setting format " << format << "."<<endl);
			return FAILURE;
		}

		DEBUG_STREAM("audio device: Set format to " << format << endl);

		/* Set number of channels */
		DEBUG_STREAM("audio device: Trying to set channels to " << channels << endl);
		if (snd_pcm_hw_params_set_channels(pcm_handle, hwparams, channels) < 0) {
			ERROR_STREAM("Error setting channels."<<endl);
			return FAILURE;
		}

		DEBUG_STREAM("audio device: Set channels to " << channels << endl);
		
		DEBUG_STREAM("audio device: Trying to set sample_rate to " << sample_rate << "." << endl);

		/* Set sample rate. If the exact rate is not supported */
		/* by the hardware, use nearest possible rate.         */
		unsigned int exact_rate = sample_rate;
		if (snd_pcm_hw_params_set_rate_near(pcm_handle, hwparams, &exact_rate, 0) < 0) {
			ERROR_STREAM("Error setting rate.");
			return FAILURE;
		}
		if (sample_rate != exact_rate) {
			DEBUG_STREAM("The rate " << sample_rate << " Hz is not supported by your hardware. " <<
					"==> Using " << exact_rate << " Hz instead."<<endl);
			sample_rate = exact_rate;
		}

		DEBUG_STREAM("audio device: Set sample_rate to " << sample_rate << "." << endl);

		if (buffer_time == 0 && buffer_frames == 0) {
			if( snd_pcm_hw_params_get_buffer_time_max(hwparams, &buffer_time, 0) < 0) {
				ERROR_STREAM("audio device: I can't get  buffer time."<<endl);
				return FAILURE;
			}
			if (buffer_time > 500000)
				buffer_time = 500000;
		}
		DEBUG_STREAM("audio device: buffer_time "<< buffer_time << " buffer_frames "<<buffer_frames<<endl);
		if (period_time == 0 && period_frames == 0) {
			if (buffer_time > 0)
				period_time = buffer_time / 4;
			else
				period_frames = buffer_frames / 4;
		}
		DEBUG_STREAM("audio device: period_time "<<period_time<<" period_frames "<<period_frames<<endl);
		if (period_time > 0) {
			if (snd_pcm_hw_params_set_period_time_near(pcm_handle, hwparams, &period_time, 0) < 0) {
				ERROR_STREAM("audio device: Error setting period time."<<endl);
				return FAILURE;
			}
		} else {
			if ( snd_pcm_hw_params_set_period_size_near(pcm_handle, hwparams, &period_frames, 0) < 0) {
				ERROR_STREAM("audio device: Error setting period size."<<endl);
				return FAILURE;
			}
		}
		DEBUG_STREAM("audio device: period buffer set"<<endl);
		if (buffer_time > 0) {
			if (snd_pcm_hw_params_set_buffer_time_near(pcm_handle, hwparams, &buffer_time, 0) < 0) {
				ERROR_STREAM("audio device: Error setting buffer time."<<endl);
				return FAILURE;
			}
		} else {
			if ( snd_pcm_hw_params_set_buffer_size_near(pcm_handle, hwparams, &buffer_frames) < 0) {
				ERROR_STREAM("audio device: Error setting buffer size."<<endl);
				return FAILURE;
			}
		}

		DEBUG_STREAM("audio device: Buffer set."<<endl);

		buffer_frames = buffer_time * sample_rate;
		internal_buffer_size = buffer_frames * channels * byte_per_sample;
		internal_buffer_time = Time(static_cast<long int>(static_cast<double>(internal_buffer_size) / byte_per_second));
		DEBUG_STREAM("setFormatInternal: Internal buffer size: " << internal_buffer_size << " bytes." << endl);
		DEBUG_STREAM("setFormatInternal: Internal buffer time: " << internal_buffer_time << endl);

		return SUCCESS;
	};

        void ALSAPlaybackNode::setLineInVolume( int /* vol */ ) {
		DEBUG_STREAM("setLineInVolume: mixer command isn't supported"<<endl);
	}

	int ALSAPlaybackNode::getLineInVolume() {
		DEBUG_STREAM("getLineInVolume: mixer command isn't supported"<<endl);
		return 0;
	}

	Result ALSAPlaybackNode::doSetFormat(const Format& format) 
	{
		DEBUG_STREAM("ALSAPlaybackNode::doSetFormat" << endl);
		try {
			DEBUG_STREAM("ALSAPlaybackNode::doSetFormat, samplerate: " << format.getIntValue(Format::channels_param) << ", channels: " << format.getIntValue(Format::sample_rate_param) << endl);
		}
		catch(Exception e) {
			ERROR_STREAM("Exception in doSetFormat " << e << endl);
		}
		return SUCCESS;
	}

	void ALSAPlaybackNode::increaseVolume()
	{
		system("/usr/bin/aumix -v +10");
	}

	void ALSAPlaybackNode::decreaseVolume()
	{
		system("/usr/bin/aumix -v -10");
	}

	void ALSAPlaybackNode::mute()
	{
		DEBUG_STREAM("MUTE" << endl);
		getController()->setBuffersInvalid(true);
	}

	void ALSAPlaybackNode::unmute()
	{
		DEBUG_STREAM("UN-MUTE" << endl);
		getController()->setBuffersInvalid(false);
	}

	ALSAPlaybackNodePlugin::ALSAPlaybackNodePlugin (const string& node_name, int max_instance)
		: TDevicePlugin<ALSAPlaybackNode>("/dev/dsp", true, node_name, max_instance)
	{
	}

	void ALSAPlaybackNodePlugin::checkForStandardAuxDevice(IDevice& device) throw() {
		/* we get the IAudio device to specify the node capabilities on the system */
		IAudioDevice_var i_audio(device.getParentObject()->getCheckedInterface<IAudioDevice>());

		/* then we try to setup the mixer device. This device must not available for
		a soundcard, but if we wish to use an availble mixer we must setup the device. */
		try {
			i_audio->setMixerDevice("/dev/mixer");
		} catch (const DeviceNotFoundException& de ) {
			GLOBAL_WARNING_STREAM(de << endl);
		}
	}

	void ALSAPlaybackNodePlugin::checkForAuxDevice(IDevice& device, unsigned device_num) throw() {
		/* we get the IAudio device to specify the node capabilities on the system */
		IAudioDevice_var i_audio(device.getParentObject()->getCheckedInterface<IAudioDevice>());

		string next_mixer_device = string("/dev/mixer") + device_num;
		try {
			i_audio->setMixerDevice(next_mixer_device);
		} catch (const DeviceNotFoundException& de ) {
			GLOBAL_WARNING_STREAM(de << endl);
		}
	}

	void ALSAPlaybackNode::setCLOEXEC(int fd){

		int oldflags = fcntl (fd, F_GETFD, 0);
		if (oldflags >= 0){
			oldflags |= FD_CLOEXEC;
			fcntl (fd, F_SETFD, oldflags);
			GLOBAL_DEBUG_STREAM("Setting CLOEXEC on audio_fd" << endl);
		}else{
			GLOBAL_ERROR_STREAM("Could not set CLOEXEC on audio_fd" << endl);
		}
	}

}
