/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


// general includes
#include "nmm/base/connect.hpp"
#include "nmm/base/registry/ClientRegistry.hpp"
#include "nmm/base/ProxyApplication.hpp"
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"
#include "nmm/plugins/video/osd/manager/OSDManagerNode.hpp"
#include "nmm/base/TSFactory.hpp"
#include "nmm/misc/timer/TimerEvents.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/utils/utils.hpp"

// include used interfaces
#include "nmm/interfaces/base/INode.hpp"
#include "nmm/interfaces/base/INamed.hpp"
#include "nmm/interfaces/base/IQueued.hpp"
#include "nmm/interfaces/base/sync/ISynchronizedSink.hpp"
#include "nmm/interfaces/device/audio/IAudioDevice.hpp"
#include "nmm/interfaces/video/display/IVideoDisplay.hpp"
#include "nmm/interfaces/video/display/IDisplayEvents.hpp"
#include "nmm/interfaces/video/display/IDisplayEventSender.hpp"
#include "nmm/interfaces/video/tv/ITVCard.hpp"
#include "nmm/interfaces/video/tv/IEPG.hpp"
#include "nmm/interfaces/general/ICopy.hpp"
#include "nmm/interfaces/file/ISinkFileHandler.hpp"
#include "nmm/interfaces/video/osd/manager/IWidgetFactory.hpp"
#include "nmm/interfaces/video/osd/manager/IOSDManager.hpp"

// sync
#include "nmm/base/sync/MultiAudioVideoSynchronizer.hpp"

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <getopt.h>
#include <iostream>
#include <sstream>

using namespace NMM;
/**
 * Class for receiving events.
 */
class AppEventReceiver : public IEventReceiver {
private:
  EventDispatcher* dispatcher;
  
public:
  AppEventReceiver() {
    dispatcher = new EventDispatcher();
  }
  
  virtual ~AppEventReceiver() {
    delete dispatcher;
  }
  
  EventDispatcher& getEventDispatcher() const {
    return *dispatcher;
  }
  
  virtual Result receiveEvent(const Event& event, const IEventSender* sender = 0 ) {
    return doReceiveEvent(event, sender);
  }
  
  virtual Result receiveEvent(const Event& event, const IEventSender* sender = 0 ) const {
    return doReceiveEvent(event, sender);
  }
  
  virtual Result receiveEvent(const CEvent& event, const IEventSender* sender = 0 ) {
    Result result = SUCCESS;
    
    for (unsigned int i=0; i<event.size(); ++i) {
      switch (doReceiveEvent(event[i], sender)) {
      case FAILURE:
        result = FAILURE;
        break;
        
      case DELETE:
        result = DELETE;
        break;
        
      default:
        break;
      }
    }
    
    return result;
  }

  virtual Result receiveEvent(const CEvent& event, const IEventSender* sender = 0 ) const {
    Result result = SUCCESS;
    
    for (unsigned int i=0; i<event.size(); ++i) {
      switch (doReceiveEvent(event[i], sender)) {
      case FAILURE:
        result = FAILURE;
        break;
        
      case DELETE:
        result = DELETE;
        break;
        
      default:
        break;
      }
    }
    
    return result;
  }
  
protected:
  virtual Result doReceiveEvent(const Event& event, const IEventSender* sender = 0) {
    return dispatcher->dispatch(event, sender);
  }
  
  virtual Result doReceiveEvent(const Event& event, const IEventSender* sender = 0) const {
    return dispatcher->dispatch(event, sender);
  }
};


/**
 * Class for handling events.
 * An instance of this class is registered at some nodes as an event listener.
 */
class EventCatcher {
private:
  bool exit;
  bool fullscreen;
  ThreadMutex exit_mutex;
  ThreadCondition exit_cv;
  IVideoDisplay* video_display;
  IMultiAudioVideoSynchronizer* sync;
  ITVCard* tvcard;
  IEPG* epg;
  ITextView* itv;
  IOSDManager*    manager;
  IWindow* win;
  TimerProducer* timer;
  unsigned int timer_id;
  IWindow* detailWin;
  ITextView* detailText;

  void signalExit() {
    MutexGuard em(exit_mutex);
    exit = true;
    exit_cv.notifyAll();
  }
  
public:
  EventCatcher() 
    : exit(false),
      fullscreen(false),
      exit_mutex(),
      exit_cv(exit_mutex),
      video_display(NULL),
      sync(NULL),
      tvcard(NULL),
      epg(NULL),
      itv(NULL),
      manager(NULL),
      win(NULL),
      timer(NULL)
  {
    timer_id = 0;
  }
  
  void setVideoDisplay(IVideoDisplay* value) {
    video_display = value;
    if (video_display) {
      fullscreen = video_display->getFullscreen() && video_display->hasFullscreen();
    }
  }
  
  void setSynchronizer(IMultiAudioVideoSynchronizer* value) {
    sync = value;
  }
  
  void setTVCard(ITVCard* value) {
    tvcard = value;
  }

  void setEPG(IEPG* value, ITextView* i, IOSDManager* m, IWindow* w) {
    epg = value;
    itv=i;
    manager=m;
    win=w;
  }

  void setTimer(TimerProducer* t)
  {
    timer = t;
  }

  void setDetailedWindow(IWindow* w, ITextView* i)
  {
    detailWin = w;
    detailText = i;
  }

 
  void showOSD()
  { 
    if(timer)
      {
        //cerr << "START TIMER " << timer_id << endl;
        Event* e = new Event("OSDHIDE", 1);
        e->setValue(0, new TInValue<unsigned int>(++timer_id));
        timer->registerTimer((time(0)+10), e);
      }

    itv->deleteAllLines();
    //cerr << "readInfo ... " << endl;
    itv->addLine(epg->getCurrentShow());
    itv->addLine(epg->getFollowingShow());
    //cerr << "... done" << endl;
    win->setVisible(true);
    win->update();
    manager->update();
  }

  Result hideOSD(unsigned int& id)
  {
    if(id < timer_id)
      return SUCCESS;

    //cerr << "HIDEOSD" << id << endl;
    win->setVisible(false);
    win->update();
    manager->update();

    return SUCCESS;
  }

  void showDetails()
  {

    if(detailWin->getVisible())
      {
        detailWin->setVisible(false);
        detailWin->update();
        manager->update();
      }else{
        detailText->deleteAllLines();
        detailText->addLine(epg->getCurrentShow());
        detailText->addLine(string());

        string description = epg->getDetailedDescription();

        stringstream s(description);
        int linelength = 0;
        string tmp;
        string line;

        while(s>>tmp)
          {
            linelength += tmp.length() *20; //lenth of word in pixel (approx)

            if(tmp == "/"){
              detailText->addLine(line);
              line="";
              linelength = 0;
            }else if(tmp == ":"){
              line += tmp + " ";
              detailText->addLine(line);
              line="";
              linelength = 0; 
            }else if(tmp == "..."){
              line += tmp + " ";
              detailText->addLine(line);
              detailText->addLine(string());
              line="";
              linelength = 0;
            }else{
              if(linelength < 700)
                {
                  line += tmp + " ";
                }else{
                  line += tmp + " ";
                  detailText->addLine(line);
                  line="";
                  linelength = 0;
                }
            }
          }
        detailText->addLine(line);

        detailWin->setVisible(true);
        detailWin->update();
        manager->update();
      }
  }

  Result keyReleased(string& key) {
    
    if (key == "KEY_q") { // 'q' to exit
      signalExit();
      
    } else if (key == "KEY_f") { // 'f' to toggle fullscreen
      if (!video_display) {
	cerr << "EventHandler: No VideoDisplay set!" << endl;
	return FAILURE;
      }
      if (video_display->hasFullscreen()) {
	fullscreen = !fullscreen;
	video_display->setFullscreen(fullscreen);
      } else {
	cerr << "Fullscreen not supported with current display!" << endl;
      }
      
    } else if (key == "KEY_Right") {
      detailWin->setVisible(false);
      int channel;
      string name;
      tvcard->getChannel(channel, name);
      tvcard->setChannel(channel+1);

      showOSD();
      
    } else if (key == "KEY_Left") {
      detailWin->setVisible(false);
      int channel;
      string name;
      tvcard->getChannel(channel, name);
      tvcard->setChannel(channel-1);
      
      showOSD();

    } else if (key == "KEY_o") {
      if(win->getVisible())
        {
          hideOSD(timer_id);
        }else{
          showOSD();
        }

    } else if (key == "KEY_i") {
      showDetails();
    } else {
      cerr << "Unsupported key pressed: " << key << endl;
    }
    
    return SUCCESS;
  }
  
  void handleEvents() {
    exit_mutex.lock();
    while (!exit) {
      exit_cv.wait();
    }
    exit_mutex.unlock();
  }
};


int usage( char *name ) {
  cerr << endl
       << "Usage:" << endl
       << name << " [options] [filename]" << endl
       << endl
       << "    -f --fullscreen      Fullscreen display" << endl
       << "    -c --channel         Set initial channel" << endl
       << "    -x --xaspect         X-aspect ratio" << endl
       << "    -y --yaspect         Y-aspect ratio" << endl
       << "    -R --record          record watched stream to stecified file also" << endl
       << "    -r --remote-source   Use remote registry as DVB-Source" << endl
       << "                         registry must be specified with: reghost[:regport]" << endl
       << "    -v --verbose         increase verbosity (may be given multiple times)" << endl
       << "    -h --help            Shows this help and exits." << endl;
  return 1;
};

int main(int argc, char ** argv) {
  /* Global options */
  bool fullscreen = false;
  unsigned int verbose = 0;
  int ratio_x = 4;
  int ratio_y = 3;
  unsigned int channel = 1;
  bool remote_source      = false;
  string remote_source_host = "localhost";
  int remote_source_port = 0;
  string recordfile;
  unsigned int timer_id;

  const char *optstring = "fc:x:y:R:r:vh";
  struct option longopts[] = {
    {"fullscreen",     0, NULL, 'f'},
    {"channel",        1, NULL, 'c'},
    {"xaspect",        1, NULL, 'x'},
    {"yaspect",        1, NULL, 'y'},
    {"record",         1, NULL, 'R'},
    {"remote-source",  1, NULL, 'r'},
    {"verbose",        0, NULL, 'v'},
    {"help",           0, NULL, 'h'},
    {NULL,             0, NULL, 0  } /* last row must be zero */
  };
  
  int ch, optionindex;
  while(EOF != (ch=getopt_long(argc,argv,optstring,longopts,&optionindex))){
    // which option did we get?
    switch (ch) {
    case 'f':
      fullscreen = true;
      break;
    case 'x':
      if (optarg) {
        ratio_x = atoi(optarg);
      }
      break;
    case 'c':
      channel = atoi(optarg);
      break;
    case 'y':
      if (optarg) {
        ratio_y = atoi(optarg);
      }
      break;
    case 'R':
      if(optarg) {
	recordfile = string(optarg);
      }
      break;
    case 'r':
      remote_source = true;
      if(optarg) {
        string remote_registry = string(optarg);
        string::size_type colon_pos = remote_registry.find_first_of(':');
        if (colon_pos == string::npos) {
          remote_source_host = remote_registry;
          remote_source_port = ClientRegistry::default_port;
        } else {
          remote_source_host = remote_registry.substr(0, colon_pos);
          remote_source_port = atoi(remote_registry.substr(colon_pos+1, string::npos).c_str());
        }
        if (!remote_source_port) {
          return usage(argv[0]);
        }
      }
      break;
    case 'v':
      verbose++;
      break;
    default: //includes 'h'
      return usage(argv[0]);
    }
  }
  
  /* handle *Stream output */
  if (verbose < 1) {
    NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 2) {
    NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 3) {
    NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 4) {
    NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  }
  
  NMMApplication* app = 0;
  
  /*Begin> Allocate _var variables for the INodes */
  INode_var dvb_mpeg;
  INode_var mpg2yuv;
  INode_var demux;
  INode_var mpa_dec;
  INode_var audio_play;
  INode_var display;
  INode_var copy;
  INode_var recorder;
  INode_var osd;
  /*End< */
  
  /* Allocate an event listener */
  EventCatcher ec;
  TimerProducer timer;
  AppEventReceiver receiver;
  timer.connectTo(&receiver);

  try {
    app = ProxyApplication::getApplication(argc, argv);
    
    /*Begin> Init requests for nodes */
    NodeDescription dvb_request;
    dvb_request.setNodeName("DVBReadNode");
    
    NodeDescription mpg2yuv_request;
    mpg2yuv_request.setNodeName("MPEGVideoDecodeNode");
    
    NodeDescription demux_request;
    demux_request.setNodeName("MPEGDemuxNode");
    
    NodeDescription mpa_dec_request;
    mpa_dec_request.setNodeName("MPEGAudioDecodeNode");
    
    NodeDescription audio_play_request;
    audio_play_request.setNodeName("PlaybackNode");
    
    NodeDescription display_request;
    display_request.setNodeName("XDisplayNode");

    NodeDescription copy_request;
    copy_request.setNodeName("CopyNode");

    NodeDescription recorder_request;
    recorder_request.setNodeName("GenericWriteNode");

    NodeDescription osd_request;
    osd_request.setNodeName("OSDManagerNode");

   
    /*End< */
    
    /*Begin> Get the ClientRegistry and request Nodes */
    ClientRegistry& registry = app->getRegistry();
    if (remote_source) { 
      cerr << "Adding registry " << remote_source_host << endl;
      registry.addRegistry(TCPAddress(remote_source_host, remote_source_port));
    }

    {
      RegistryLock lock(registry);
      
      list<Response> dvb_response;
      if(remote_source) {
	dvb_request.setLocation(remote_source_host);      
	dvb_response = registry.initRequest(dvb_request);
      }
      else {
	dvb_response = registry.initRequest(dvb_request);
      }
      list<Response> mpg2yuv_response = registry.initRequest(mpg2yuv_request);
      list<Response> demux_response = registry.initRequest(demux_request);
      list<Response> mpa_dec_response = registry.initRequest(mpa_dec_request);
      list<Response> audio_play_response = registry.initRequest(audio_play_request);
      list<Response> display_response = registry.initRequest(display_request);
      list<Response> osd_response = registry.initRequest(osd_request);
      list<Response> copy_response;
      list<Response> recorder_response;
      if (!recordfile.empty()) {
        copy_response = registry.initRequest(copy_request);
        recorder_response = registry.initRequest(recorder_request);
      }

      if (dvb_response.empty() ||
          mpg2yuv_response.empty() ||
          demux_response.empty() ||
          mpa_dec_response.empty() ||
          audio_play_response.empty() ||
          display_response.empty() ||
          osd_response.empty() ||
          (!recordfile.empty() && (copy_response.empty() || recorder_response.empty()))) {
        GLOBAL_ERROR_STREAM("Size of dvb_response " << dvb_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of mpg2yuv_response " << mpg2yuv_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of demux_response " << demux_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of mpa_dec_response " << mpa_dec_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of audio_play_response " << audio_play_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of display_response " << display_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of osd_response " << osd_response.size() << endl);
        if (!recordfile.empty()) {
          GLOBAL_ERROR_STREAM("Size of copy_response " << copy_response.size() << endl);
          GLOBAL_ERROR_STREAM("Size of recorder_response " << recorder_response.size() << endl);
        }
        throw Exception("A required node is not available");
      }
      
      dvb_mpeg.reset(registry.requestNode(dvb_response.front()));
      mpg2yuv.reset(registry.requestNode(mpg2yuv_response.front()));
      demux.reset(registry.requestNode(demux_response.front()));
      mpa_dec.reset(registry.requestNode(mpa_dec_response.front()));
      audio_play.reset(registry.requestNode(audio_play_response.front()));
      display.reset(registry.requestNode(display_response.front()));
      osd.reset(registry.requestNode(osd_response.front()));
      if (!recordfile.empty()) {
        copy.reset(registry.requestNode(copy_response.front()));
        recorder.reset(registry.requestNode(recorder_response.front()));
      }
    }

    /*Begin> Initialize and connect Nodes */
    INamed_var dvb_mpeg_name(dvb_mpeg->getParentObject()->getCheckedInterface<INamed>());
    dvb_mpeg_name->setName("VideoMPEG");
    
    dvb_mpeg->init();
  
    IBiDirQueuedNode_var mpa_dec_queue(mpa_dec->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
    mpa_dec_queue->setDownstreamMaxSize(10);
    mpa_dec->init();
      
    IBiDirQueuedNode_var audio_play_queue(audio_play->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
    audio_play_queue->setDownstreamMaxSize(10);
    IAudioDevice_var audio_device(audio_play->getParentObject()->getCheckedInterface<IAudioDevice>());
    audio_device->setDevice("/dev/dsp");
    audio_play->init();
    
    if (copy.get()) {
      ICopy_var icopy(copy->getParentObject()->getCheckedInterface<ICopy>());
      icopy->setNumberOfOutputs(2);
      copy->init();

      // make sure, video presentation does not influent recording
      IBiDirQueuedNode_var demux_queue(demux->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
      demux_queue->setDownstreamMaxSize(30);
      demux_queue->setDownstreamMode(IQueued::MODE_KEEP_NEWEST);
    }

    if (recorder.get()) {
      ISinkFileHandler_var file_sink(recorder->getParentObject()->getCheckedInterface<ISinkFileHandler>());
      file_sink->setFilename(recordfile);
      file_sink->setWorkingDirectory(string(getenv("PWD")));
      file_sink->setFileExtension(string("mpeg"));
      recorder->init();
    }


    // ------------ Setup OSDManager --------------------
    
    IOSDManager*    manager(osd->getParentObject()->getCheckedInterface<IOSDManager>());
    IWidgetFactory* factory(osd->getParentObject()->getCheckedInterface<IWidgetFactory>());
    
    manager->loadFont("medium", getNMMResource("font/font.xml"), "Test");

    ITextView* text = factory->createTextView();
    text->setFont( "medium" );

    IPlaneDecorator* plane = factory->createPlaneDecorator( *text );
    plane->setPlaneColor( 0x80ff0000 );

    IWindow* win = manager->createWindow();
    win->setDimension(700, 50);
    win->setPosition(10, 500);
    win->setContent( *plane);
    win->update();    

    //detailed information window
    ITextView* detailedText = factory->createTextView();
    detailedText->setFont("medium");

    IPlaneDecorator* detailedPlane = factory->createPlaneDecorator (*detailedText);
    detailedPlane->setPlaneColor( 0x80ff0000);

    IWindow* detailedWin = manager->createWindow();
    detailedWin->setVisible(false);
    detailedWin->setDimension(700, 500);
    detailedWin->setPosition(10, 10);
    detailedWin->setContent( *detailedPlane);
    detailedWin->update();

    manager->update();

    osd->init();
    // ----------------------------------------------"
    mpg2yuv->init();
    
    demux->init();
    /*End< Initialize and connect Nodes */
    
    /*Begin> Connect Graph */
    dvb_mpeg->initOutput();
    if (copy.get()) {
      connect(dvb_mpeg, copy);
      dvb_mpeg->activate();
      copy->initOutput();
      if (recorder.get()) {
        connect(copy, string("1"), recorder);
      }
      connect(copy, string("0"), demux);
      copy->activate();
    } else {
      connect(dvb_mpeg, demux);
      dvb_mpeg->activate();
    }
    
    demux->initOutput();
    connect(demux, "mpeg_video0", mpg2yuv);
    connect(demux, "mpeg_audio0", mpa_dec);
    demux->activate();
    
    mpa_dec->initOutput();
    connect(mpa_dec, audio_play);
    mpa_dec->activate();
    
    audio_play->initOutput();
    audio_play->activate();
    
    mpg2yuv->initOutput();
    connect(mpg2yuv, osd, "stream1");
    mpg2yuv->activate();
    
    osd->initOutput();
    display->init();
    connect(osd, display);
    osd->activate();

    display->initOutput();
    display->activate();
    
    if (recorder.get()) {
      recorder->initOutput();
      recorder->activate();
    }
    /*End< Connect Graph */
    
    /*Begin> Init synchronization */
    MultiAudioVideoSynchronizer av_sync;
    IMultiAudioVideoSynchronizer_var sync(av_sync.getCheckedInterface<IMultiAudioVideoSynchronizer>());
    
    ISynchronizedSink_var audio_sync(audio_play->getParentObject()->getCheckedInterface<ISynchronizedSink>());
    ISynchronizedSink_var video_sync(display->getParentObject()->getCheckedInterface<ISynchronizedSink>());
    audio_sync->setSynchronized(true);
    video_sync->setSynchronized(true);
    ISinkController_var audio_controller(audio_sync->getController());
    ISinkController_var video_controller(video_sync->getController());
    sync->addAudioSinkController(audio_controller.get());
    sync->addVideoSinkController(video_controller.get());
    /*End< */
    
    IVideoDisplay_var video_display(display->getParentObject()->getCheckedInterface<IVideoDisplay>());
    video_display->setFullscreen( fullscreen );
    
    /*Begin> Register event listeners */
    display->getParentObject()->
      registerEventListener(IDisplayEvents::keyReleased_event,
                            new TEDObject1<EventCatcher, string>(&ec, &EventCatcher::keyReleased));
    //Event* e = new Event("OSDHIDE",1);
    //s->setValue(0, TInValue<unsigned int>(timer_id++));
    receiver.getEventDispatcher().registerEventListener("OSDHIDE",
                                                        new TEDObject1<EventCatcher, unsigned int>(&ec, &EventCatcher::hideOSD));
    /*End< */
    
    /*Begin> Start Graph */
    if (copy.get()) {
      copy->start();
    }
    if (recorder.get()) {
      recorder->start();
    }
    display->start();
    audio_play->start();
    mpg2yuv->start();
    mpa_dec->start();
    demux->start();
    dvb_mpeg->start();
    osd->start();
    timer.start();

    /*End< Start Graph */
    
    ITVCard_var tvcard(dvb_mpeg->getParentObject()->getCheckedInterface<ITVCard>());
    tvcard->setChannel(channel);

    IEPG* epg(dvb_mpeg->getParentObject()->getCheckedInterface<IEPG>());
    /*Begin> Event loop */
    IDisplayEventSender_var display_event_sender(display->getParentObject()->getInterface<IDisplayEventSender>());
    if (display->isStarted() && display_event_sender.get()) {
      display_event_sender->enableDisplayEvents();

    }
    ec.setVideoDisplay(video_display.get());
    ec.setSynchronizer(sync.get());
    ec.setTVCard(tvcard.get());
    ec.setEPG(epg, text, manager, win);
    ec.setDetailedWindow(detailedWin, detailedText);
    ec.setTimer(&timer);
    ec.showOSD();
    ec.handleEvents();
    if (display->isStarted() && display_event_sender.get()) {
      display_event_sender->disableDisplayEvents();
    }
    /*End< Event loop */
    
    /*Begin> Stop nodes */
    timer.stop();
    dvb_mpeg->stop();
    demux->stop();
    mpa_dec->stop();
    mpg2yuv->stop();
    audio_play->stop();
    display->stop();
    osd->stop();
    if (copy.get()) {
      copy->stop();
    }
    if (recorder.get()) {
      recorder->stop();
    }
    /*End< Stop Nodes */
    
    /*Begin> Flush nodes */
    dvb_mpeg->flush();
    demux->flush();
    mpa_dec->flush();
    mpg2yuv->flush();
    audio_play->flush();
    display->flush();
    osd->flush();
    if (copy.get()) {
      copy->flush();
    }
    if (recorder.get()) {
      recorder->flush();
    }
    /*End< Flush Nodes */
    
    /*Begin> deactivate nodes */
    dvb_mpeg->deactivate();
    demux->deactivate();
    mpa_dec->deactivate();
    mpg2yuv->deactivate();
    audio_play->deactivate();
    display->deactivate();
    osd->deactivate();
    if (copy.get()) {
      copy->deactivate();
    }
    if (recorder.get()) {
      recorder->deactivate();
    }
    /*End< deactivate Nodes */
    
    /*Begin> Disconnect Graph */
    display->deinitOutput();
    disconnect_input(display);
    mpg2yuv->deinitOutput();
    osd->deinitOutput();
    disconnect_input(osd);

    audio_play->deinitOutput();
    disconnect_input(audio_play);
    mpa_dec->deinitOutput();
    
    disconnect_input(mpg2yuv);
    disconnect_input(mpa_dec);
    demux->deinitOutput();
    
    if (copy.get()) {
      if (recorder.get()) {
        recorder->deinitOutput();
        disconnect_input(recorder);
      }
      disconnect_input(demux);
      copy->deinitOutput();
    }
    disconnect_output(dvb_mpeg);
    dvb_mpeg->deinitOutput();
    osd->deinitOutput();
    disconnect_input(osd);
    /*End< Disconnect Graph */
    
    /*Begin> Deinit nodes */
    dvb_mpeg->deinit();
    demux->deinit();
    mpa_dec->deinit();
    mpg2yuv->deinit();
    audio_play->deinit();
    display->deinit();
    osd->deinit();
    if (copy.get()) {
      copy->deinit();
    }
    if (recorder.get()) {
      recorder->deinit();
    }
    /*End< Deinit nodes */
    
    /*Begin> Release nodes */
    if (dvb_mpeg.get()) {
        app->getRegistry().releaseNode(*dvb_mpeg);
    }
    if (mpg2yuv.get()) {
      app->getRegistry().releaseNode(*mpg2yuv);
    }
    if (demux.get()) {
        app->getRegistry().releaseNode(*demux);
    }
    if (mpa_dec.get()) {
        app->getRegistry().releaseNode(*mpa_dec);
    }
    if (audio_play.get()) {
        app->getRegistry().releaseNode(*audio_play);
    }
    if (display.get()) {
        app->getRegistry().releaseNode(*display);
    }
    if (copy.get()) {
        app->getRegistry().releaseNode(*copy);
    }
    if (recorder.get()) {
        app->getRegistry().releaseNode(*recorder);
    }
    if (osd.get()) {
        app->getRegistry().releaseNode(*osd);
    }
    /*End< Release nodes */
    
  } catch(Exception e) {
    cerr << e << endl;    
    
    if (app != NULL) {
      /*Begin> Release nodes */
      if (dvb_mpeg.get()) {
        app->getRegistry().releaseNode(*dvb_mpeg);
      }
      if (mpg2yuv.get()) {
        app->getRegistry().releaseNode(*mpg2yuv);
      }
      if (demux.get()) {
        app->getRegistry().releaseNode(*demux);
      }
      if (mpa_dec.get()) {
        app->getRegistry().releaseNode(*mpa_dec);
      }
      if (audio_play.get()) {
        app->getRegistry().releaseNode(*audio_play);
      }
      if (display.get()) {
        app->getRegistry().releaseNode(*display);
      }
      if (copy.get()) {
        app->getRegistry().releaseNode(*copy);
      }
      if (recorder.get()) {
        app->getRegistry().releaseNode(*recorder);
      }
      /*End< Release nodes */
    }
    
    return 1;
  } catch(...) {
    cerr << "Something went wrong..." << endl;
    
    if (app != NULL) {
      /*Begin> Release nodes */
      if (dvb_mpeg.get()) {
        app->getRegistry().releaseNode(*dvb_mpeg);
      }
      if (mpg2yuv.get()) {
        app->getRegistry().releaseNode(*mpg2yuv);
      }
      if (demux.get()) {
        app->getRegistry().releaseNode(*demux);
      }
      if (mpa_dec.get()) {
        app->getRegistry().releaseNode(*mpa_dec);
      }
      if (audio_play.get()) {
        app->getRegistry().releaseNode(*audio_play);
      }
      if (display.get()) {
        app->getRegistry().releaseNode(*display);
      }
      if (copy.get()) {
        app->getRegistry().releaseNode(*copy);
      }
      if (recorder.get()) {
        app->getRegistry().releaseNode(*recorder);
      }
      if (osd.get()) {
        app->getRegistry().releaseNode(*osd);
      }
      /*End< Release nodes */
    }
    
    return 1;
  }
  
  return 0;
}
