/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


// general includes
#include "nmm/base/connect.hpp"
#include "nmm/base/registry/NodeDescription.hpp"
#include "nmm/base/registry/GraphDescription.hpp"
#include "nmm/base/registry/ClientRegistry.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/base/ProxyApplication.hpp"
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"
#include "nmm/base/TSFactory.hpp"

// include used interfaces
#include "nmm/interfaces/base/INode.hpp"
#include "nmm/interfaces/base/INamed.hpp"
#include "nmm/interfaces/base/IQueued.hpp"
#include "nmm/interfaces/base/sync/ISynchronizedSink.hpp"
#include "nmm/interfaces/device/audio/IAudioDevice.hpp"
#include "nmm/interfaces/video/display/IVideoDisplay.hpp"
#include "nmm/interfaces/video/display/IDisplayEvents.hpp"
#include "nmm/interfaces/video/display/IDisplayEventSender.hpp"
#include "nmm/interfaces/video/tv/ITVCard.hpp"
#include "nmm/interfaces/file/ISinkFileHandler.hpp"
#include "nmm/interfaces/file/ITimeshifting.hpp"
#include "nmm/base/sync/MultiAudioVideoSynchronizer.hpp"


#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <getopt.h>
#include <iostream>

using namespace NMM;
/**
 * Class for receiving events.
 */
class AppEventReceiver : public IEventReceiver {
private:
  EventDispatcher* dispatcher;
  
public:
  AppEventReceiver() {
    dispatcher = new EventDispatcher();
  }
  
  virtual ~AppEventReceiver() {
    delete dispatcher;
  }
  
  EventDispatcher& getEventDispatcher() const {
    return *dispatcher;
  }
  
  virtual Result receiveEvent(const Event& event, const IEventSender* sender = 0 ) {
    return doReceiveEvent(event, sender);
  }
  
  virtual Result receiveEvent(const Event& event, const IEventSender* sender = 0 ) const {
    return doReceiveEvent(event, sender);
  }
  
  virtual Result receiveEvent(const CEvent& event, const IEventSender* sender = 0 ) {
    Result result = SUCCESS;
    
    for (unsigned int i=0; i<event.size(); ++i) {
      switch (doReceiveEvent(event[i], sender)) {
      case FAILURE:
        result = FAILURE;
        break;
        
      case DELETE:
        result = DELETE;
        break;
        
      default:
        break;
      }
    }
    
    return result;
  }

  virtual Result receiveEvent(const CEvent& event, const IEventSender* sender = 0 ) const {
    Result result = SUCCESS;
    
    for (unsigned int i=0; i<event.size(); ++i) {
      switch (doReceiveEvent(event[i], sender)) {
      case FAILURE:
        result = FAILURE;
        break;
        
      case DELETE:
        result = DELETE;
        break;
        
      default:
        break;
      }
    }
    
    return result;
  }
  
protected:
  virtual Result doReceiveEvent(const Event& event, const IEventSender* sender = 0) {
    return dispatcher->dispatch(event, sender);
  }
  
  virtual Result doReceiveEvent(const Event& event, const IEventSender* sender = 0) const {
    return dispatcher->dispatch(event, sender);
  }
};


/**
 * Class for handling events.
 * An instance of this class is registered at some nodes as an event listener.
 */
class EventCatcher {
private:
  bool exit;
  bool fullscreen;
  ThreadMutex exit_mutex;
  ThreadCondition exit_cv;
  IVideoDisplay* video_display;
  IMultiAudioVideoSynchronizer* sync;
  ITVCard* tvcard;
  ITimeshifting* ts;

  void signalExit() {
    MutexGuard em(exit_mutex);
    exit = true;
    exit_cv.notifyAll();
  }
  
public:
  EventCatcher() 
    : exit(false),
      fullscreen(false),
      exit_mutex(),
      exit_cv(exit_mutex),
      video_display(NULL),
      sync(NULL),
      tvcard(NULL)
  {
  }
  
  void setVideoDisplay(IVideoDisplay* value) {
    video_display = value;
    if (video_display) {
      fullscreen = video_display->getFullscreen() && video_display->hasFullscreen();
    }
  }

  void setTimeshifting(ITimeshifting* value){
    ts = value;
  }

  void setSynchronizer(IMultiAudioVideoSynchronizer* value) {
    sync = value;
  }
  
  void setTVCard(ITVCard* value) {
    tvcard = value;
  }

  Result keyReleased(string& key) {
    
    if (key == "KEY_q") { // 'q' to exit
      signalExit();
      
    } else if (key == "KEY_f") { // 'f' to toggle fullscreen
      if (!video_display) {
	cerr << "EventHandler: No VideoDisplay set!" << endl;
	return FAILURE;
      }
      if (video_display->hasFullscreen()) {
	fullscreen = !fullscreen;
	video_display->setFullscreen(fullscreen);
      } else {
	cerr << "Fullscreen not supported with current display!" << endl;
      }
      
    } else if (key == "KEY_p") {
      ts->setModePause();
      
    } else if (key == "KEY_Return") {
      ts->setModePlay();

    } else if (key == "KEY_w") {
      ts->setModeFastForward(2);

    } else if (key == "KEY_Right") {
      int channel;
      string name;
      tvcard->getChannel(channel, name);
      tvcard->setChannel(channel+1);

      
    } else if (key == "KEY_Left") {
      int channel;
      string name;
      tvcard->getChannel(channel, name);
      tvcard->setChannel(channel-1);
      
    } else {
      cerr << "Unsupported key pressed: " << key << endl;
    }
    
    return SUCCESS;
  }
  
  void handleEvents() {
    exit_mutex.lock();
    while (!exit) {
      exit_cv.wait();
    }
    exit_mutex.unlock();
  }
};


int usage( char *name ) {
  cerr << endl
       << "Usage:" << endl
       << name << " [options]" << endl
       << endl
       << "    -f --fullscreen      Fullscreen display" << endl
       << "    -c --channel         Set initial channel" << endl
       << "    -r --remote-source   Use remote registry as DVB-Source" << endl
       << "    -s --syncronisation  turn off synchronisation" << endl
       << "                         registry must be specified with: reghost[:regport]" << endl
       << "    -v --verbose         increase verbosity (may be given multiple times)" << endl
       << "    -h --help            Shows this help and exits." << endl;
  return 1;
};

int main(int argc, char ** argv) {
  /* Global options */
  bool fullscreen = false;
  unsigned int verbose = 0;
  unsigned int channel = 1;
  string remote_source_host = "localhost";
  string recordfile;
  bool synchronisation = true;

  GraphDescription graph;

  const char *optstring = "fc:r:vh:s";
  struct option longopts[] = {
    {"fullscreen",     0, NULL, 'f'},
    {"channel",        1, NULL, 'c'},
    {"remote-source",  1, NULL, 'r'},
    {"synchronisation",0, NULL, 's'},
    {"verbose",        0, NULL, 'v'},
    {"help",           0, NULL, 'h'},
    {NULL,             0, NULL, 0  } /* last row must be zero */
  };
  
  int ch, optionindex;
  while(EOF != (ch=getopt_long(argc,argv,optstring,longopts,&optionindex))){
    // which option did we get?
    switch (ch) {
    case 'f':
      fullscreen = true;
      break;
    case 'c':
      channel = atoi(optarg);
      break;
    case 's':
      synchronisation = false;
      break;
    case 'r':
      if(optarg) {
        remote_source_host = string(optarg);
      }
      break;
    case 'v':
      verbose++;
      break;
    default: //includes 'h'
      return usage(argv[0]);
    }
  }
  
  /* handle *Stream output */
  if (verbose < 1) {
    NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 2) {
    NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 3) {
    NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 4) {
    NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  }
  
  NMMApplication* app = 0;
  
  /* Allocate an event listener */
  EventCatcher ec;
  AppEventReceiver receiver;

  try {
    app = ProxyApplication::getApplication(argc, argv);
    
    /*Begin> Init requests for nodes */
    NodeDescription dvb;
    dvb.setNodeName("DVBReadNode");
    dvb.setSharingType(NodeDescription::EXCLUSIVE_THEN_SHARED);
    dvb.setLocation(remote_source_host);

    NodeDescription timeshift;
    timeshift.setNodeName("MPEGTimeShiftingNode");
    timeshift.setSharingType(NodeDescription::EXCLUSIVE_THEN_SHARED);
    timeshift.setLocation("sparrow");

    NodeDescription mpg2yuv;
    mpg2yuv.setNodeName("MPEGVideoDecodeNode");
    mpg2yuv.setSharingType(NodeDescription::EXCLUSIVE_THEN_SHARED);

    NodeDescription demux;
    demux.setNodeName("MPEGDemuxNode");
    demux.setSharingType(NodeDescription::EXCLUSIVE_THEN_SHARED);

    NodeDescription mpa_dec;
    mpa_dec.setNodeName("MPEGAudioDecodeNode");
    mpa_dec.setSharingType(NodeDescription::EXCLUSIVE_THEN_SHARED);

    NodeDescription audio_play;
    audio_play.setNodeName("PlaybackNode");
    audio_play.setSharingType(NodeDescription::EXCLUSIVE_THEN_SHARED);
        
    NodeDescription display;
    display.setNodeName("XDisplayNode");
    display.setSharingType(NodeDescription::EXCLUSIVE_THEN_SHARED);

   
    /*End< */
    
    /*Begin> Get the ClientRegistry and request Nodes */
    ClientRegistry& registry = app->getRegistry();

    graph.addEdge(dvb, timeshift);
    graph.addEdge(timeshift, demux);
    graph.addEdge(demux, "mpeg_video0", mpg2yuv);
    graph.addEdge(mpg2yuv, display);
    graph.addEdge(demux, "mpeg_audio0", mpa_dec);
    graph.addEdge(mpa_dec, audio_play);

    registry.requestGraph(graph);

    /*Begin> Initialize and connect Nodes */
    INode* dvb_mpeg = graph.getINode(dvb); 
    INamed_var dvb_mpeg_name(dvb_mpeg->getParentObject()->getCheckedInterface<INamed>());
    dvb_mpeg_name->setName("VideoMPEG");
    
    INode* mpa = graph.getINode(mpa_dec);
    IBiDirQueuedNode_var mpa_dec_queue(mpa->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
    mpa_dec_queue->setDownstreamMaxSize(10);
      
    INode* audio = graph.getINode(audio_play);
    IBiDirQueuedNode_var audio_play_queue(audio->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
    audio_play_queue->setDownstreamMaxSize(10);
    IAudioDevice_var audio_device(audio->getParentObject()->getCheckedInterface<IAudioDevice>());
    audio_device->setDevice("/dev/dsp");

    INode* timesh = graph.getINode(timeshift);
    ITimeshifting* itimeshift = timesh->getParentObject()->getInterface<ITimeshifting>();
    itimeshift->setFilename("tsbuffer");
    itimeshift->setFileBufferThreshhold(32*1000);
    itimeshift->enableMemBuf();

    INode* disp = graph.getINode(display);

    graph.realizeGraph();
    /*End< Initialize and connect Nodes */

    MultiAudioVideoSynchronizer*     av_sync;
    IMultiAudioVideoSynchronizer_var iav_sync;

    if(synchronisation){
      cerr << "Synchronisation enabled" << endl;
      if(graph.getISession()) {
        ISinkSynchronizer_var tmp_sync(graph.getISession()->getSinkSynchronizer());
        iav_sync.reset(tmp_sync->getParentObject()->getCheckedInterface<IMultiAudioVideoSynchronizer>());
      }else{
        av_sync = new MultiAudioVideoSynchronizer();
        iav_sync.reset(av_sync->getCheckedInterface<IMultiAudioVideoSynchronizer>());
      }

      ISynchronizedSink_var audio_sync(audio->getParentObject()->getCheckedInterface<ISynchronizedSink>());
      ISynchronizedSink_var video_sync(disp->getParentObject()->getCheckedInterface<ISynchronizedSink>());
      if (audio_sync.get() && video_sync.get()){
        ISinkController_var audio_controller(audio_sync->getController());
        ISinkController_var video_controller(video_sync->getController());
        audio_sync->setSynchronized(true);
        video_sync->setSynchronized(true);
        iav_sync->addAudioSinkController(audio_controller.get());
        iav_sync->addVideoSinkController(video_controller.get());
      }
    }

    IVideoDisplay_var video_display(disp->getParentObject()->getCheckedInterface<IVideoDisplay>());
    video_display->setFullscreen( fullscreen );

    /*Begin> Register event listeners */
    disp->getParentObject()->
      registerEventListener(IDisplayEvents::keyReleased_event,
                            new TEDObject1<EventCatcher, string>(&ec, &EventCatcher::keyReleased));
    /*End< */
    
    /*Begin> Start Graph */
    graph.activateGraph();
    graph.startGraph();
    /*End< Start Graph */
    
    ITVCard_var tvcard(dvb_mpeg->getParentObject()->getCheckedInterface<ITVCard>());
    tvcard->setChannel(channel);

    /*Begin> Event loop */
    IDisplayEventSender_var display_event_sender(disp->getParentObject()->getInterface<IDisplayEventSender>());
    if (disp->isStarted() && display_event_sender.get()) {
      display_event_sender->enableDisplayEvents();

    }
    ec.setVideoDisplay(video_display.get());
    ec.setSynchronizer(iav_sync.get());
    ec.setTVCard(tvcard.get());
    ec.setTimeshifting(itimeshift);
    ec.handleEvents();
    if (disp->isStarted() && display_event_sender.get()) {
      display_event_sender->disableDisplayEvents();
    }
    /*End< Event loop */

    
    graph.stopGraph();
    graph.unrealizeGraph();

    return 0;
  }catch(...){
    graph.stopGraph();
    graph.unrealizeGraph();
  }
}
