/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *                     
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <getopt.h>
#include <unistd.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <iostream>

#include "nmm/base/connect.hpp"
#include "nmm/base/ProxyObject.hpp"

// include used interfaces
#include "nmm/interfaces/base/INode.hpp"
#include "nmm/interfaces/device/dvd/IDVD.hpp"
#include "nmm/interfaces/file/ISinkFileHandler.hpp"
#include "nmm/interfaces/file/ITrack.hpp"
#include "nmm/interfaces/video/display/IDisplayEvents.hpp"
#include "nmm/interfaces/general/progress/IProgress.hpp"
#include "nmm/interfaces/general/progress/IProgressListener.hpp"

// include the registry
#include "nmm/base/ProxyApplication.hpp"
#include "nmm/base/registry/ClientRegistry.hpp"
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"
#include "nmm/utils/utils.hpp"

// console producer
#include "nmm/misc/console/ConsoleProducer.hpp"

// import namespace
using namespace NMM;


/**
 * Class for receiving events.
 */
class AppEventReceiver : public IEventReceiver {
private:
  EventDispatcher* dispatcher;
  
public:
  AppEventReceiver() {
    dispatcher = new EventDispatcher();
  }
  
  virtual ~AppEventReceiver() {
    delete dispatcher;
  }
  
  EventDispatcher& getEventDispatcher() const {
    return *dispatcher;
  }
  
  virtual Result receiveEvent(const Event& event, const IEventSender* sender = 0 ) {
    return doReceiveEvent(event, sender);
  }
  
  virtual Result receiveEvent(const Event& event, const IEventSender* sender = 0 ) const {
    return doReceiveEvent(event, sender);
  }
  
  virtual Result receiveEvent(const CEvent& event, const IEventSender* sender = 0 ) {
    Result result = SUCCESS;
    
    for (unsigned int i=0; i<event.size(); ++i) {
      switch (doReceiveEvent(event[i], sender)) {
      case FAILURE:
        result = FAILURE;
        break;
        
      case DELETE:
        result = DELETE;
        break;
        
      default:
        break;
      }
    }
    
    return result;
  }
  
  virtual Result receiveEvent(const CEvent& event, const IEventSender* sender = 0 ) const {
    Result result = SUCCESS;
    
    for (unsigned int i=0; i<event.size(); ++i) {
      switch (doReceiveEvent(event[i], sender)) {
      case FAILURE:
        result = FAILURE;
        break;
        
      case DELETE:
        result = DELETE;
        break;
        
      default:
        break;
      }
    }
    
    return result;
  }
  
protected:
  virtual Result doReceiveEvent(const Event& event, const IEventSender* sender = 0) {
    return dispatcher->dispatch(event, sender);
  }
  
  virtual Result doReceiveEvent(const Event& event, const IEventSender* sender = 0) const {
    return dispatcher->dispatch(event, sender);
  }
};


/**
 * Class for handling events.
 * An instance of this class is registered at some nodes as an event listener.
 */
class EventCatcher {
private:
  bool exit;
  ThreadMutex exit_mutex;
  ThreadCondition exit_cv;
  
  void signalExit() {
    MutexGuard em(exit_mutex);
    exit = true;
    exit_cv.notifyAll();
  }
  
public:
  EventCatcher() 
    : exit(false),
      exit_mutex(),
      exit_cv(exit_mutex)
  {}
  
  Result endTrack() {
    cout << "100% extracted.                                "; // main loop sends newline
    // make some spaces to overwrite actual/total count output from setProgress() method
    signalExit();
    return SUCCESS;
  }
  
  Result setProgress(u_int64_t& actual, u_int64_t& total) {
    int progress = static_cast<int>(actual * 100.0 / static_cast<float>(total));
    // print out progress like " 04% extracted, 2000/50000
    cout << " ";
    if (10 > progress) {
      cout << "0";
    }
    cout << progress << "% extracted";
    cout << ", " << actual << "/" << total;
    cout << "\r"; // no endl, we want to overwrite this again
    cout.flush(); // flush the output to make sure, its printed out.
    return SUCCESS;
  }

  Result keyReleased(string& key) {
    if (key == "KEY_q") {
      cout << endl << "Quit"; // main loop sends newline
      signalExit();
    } else {
      GLOBAL_ERROR_STREAM("Unsupported key pressed: " << key << endl);
    }
    
    return SUCCESS;
  }
  
  void handleEvents() {
    exit_mutex.lock();
    while (!exit) {
      exit_cv.wait();
    }
    exit_mutex.unlock();
  }
};



/* Global options */
bool verbose          = false;
bool show_information = false;
string device         = "/dev/cdrom";
unsigned int width    = 0;
unsigned int height   = 0;
int chapter           = 1;
int title             = 1;
string filename       = "dvd_dump.vob";
int chapterend        = 0;



void usage( char *name ) {
  cerr << endl
       << "Usage:" << endl
       << name << " [options] [filename]" << endl
       << endl
       << "    filename:             Set the filename of extracted data [" << filename << "]" << endl
       << "    -d --device:          Set the DVD device [" << device << "]" << endl
       << "    -t --title:           Select title [" << title << "]" << endl
       << "    -c --chapter:         Select chapter [" << chapter << "]" << endl
       << "    -n --chapter-end:     Select end chapter[" << chapterend << "] 0 means: read all chapters" << endl
       << "    -v --verbose:         Verbose mode (lots of messages!)" << endl
       << "    -i --information:     Shows number of title, chapters and stream informations" << endl
       << "    -h --help:            Shows this help and exits." << endl;
  exit(0);
}

void turnOffAllMessages() {
  NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
}


int fderr = -1;
int fd    = -1;

void setStderr( bool on ) {
  if ( !on ) {
    // get copy of stderr desciptor
    fderr = dup( 2 );
    // close stderr
    close( 2 );
    // open /dev/null directly, so that it has fd=2 (stderr)
    fd = open( "/dev/null", O_WRONLY );
  } else {
    // close fd(=2=/dev/null), and set it to stderr again
    dup2( fderr, fd );
  }
}

int main(int argc, char ** argv) {
  const char *optstring = "d:t:c:n:vih";
  struct option longopts[] = {
    {"device",       1, NULL, 'd'},
    {"title",        1, NULL, 't'},
    {"chapter",      1, NULL, 'c'},
    {"chapter-end",  1, NULL, 'n'},
    {"verbose",      0, NULL, 'v'},
    {"information",  0, NULL, 'i'},
    {"help",         0, NULL, 'h'},
    {NULL         ,  0, NULL, 0  } /* last row must be zero */
  };

  int ch, optionindex;

  while(EOF != (ch=getopt_long(argc,argv,optstring,longopts,&optionindex))){
    // which option did we get?
    switch (ch) {
    case 'd':
      if (optarg) {
	device = optarg;
      }
      break;
    case 't':
      if (optarg) {
	title = atoi(optarg);
      }
      break;
    case 'c':
      if (optarg) {
	chapter = atoi(optarg);
      }
      break;
    case 'n':
      if (optarg) {
	chapterend = atoi(optarg);
      }
      break;
    case 'i':
      show_information = true;
      break;
    case 'v':
      verbose = true;
      break;
    default: //includes 'h'
      usage(argv[0]);
    }
  }

  if (argc > optind + 1) { // more than one arg left
    usage(argv[0]);
  } else if (argc == optind + 1) { // hostname is specified
    filename = string(argv[optind]);
  }
  
  // maybe turn off debug messages
  if (!verbose) {
    turnOffAllMessages();
  }
  
  NMMApplication* app = 0;
  
  /*Begin> Allocate _var variables for the INodes */
  INode_var dvd;
  INode_var writer;
  /*End< */
  
  /* Allocate an event listener */
  EventCatcher ec;
  AppEventReceiver receiver;
  
  int error_level = 0;
  try {
    app = ProxyApplication::getApplication(argc, argv);

    /*Begin> Init requests for nodes */
    NodeDescription dvd_request;
    dvd_request.setNodeName("DVDReadNode");

    NodeDescription writer_request;
    writer_request.setNodeName("GenericWriteNode");
    /*End< */
    
    /*Begin> Get the ClientRegistry and request Nodes */
    ClientRegistry& registry = app->getRegistry();
    {
      RegistryLock lock(registry);
      
      list<Response> dvd_response = registry.initRequest(dvd_request);
      list<Response> writer_response = registry.initRequest(writer_request);
      
      if (dvd_response.empty() ||
          writer_response.empty()) {
        GLOBAL_ERROR_STREAM("Size of dvd_response " << dvd_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of writer_response " << writer_response.size() << endl);
        throw Exception("A required node is not available");
      }
      
      dvd.reset(registry.requestNode(dvd_response.front()));
      writer.reset(registry.requestNode(writer_response.front()));
    }
    /*End< */
    
    /*Begin> Initialize DVD */
    IDVD_var dvd_interface(dvd->getParentObject()->getCheckedInterface<IDVD>());
    dvd_interface->setDevice(device);
    
    // enable progress informations
    IProgress_var progress(dvd->getParentObject()->getInterface<IProgress>());
    if (progress.get()) {
      progress->sendProgressInformation(true);
    }

    setStderr( false );
    dvd->init();
    dvd->initOutput();
    {
      Property_var dvd_wout(dvd->getWorkingOutputProperty());
      const Format* file_format = dvd_wout->getDefaultFormat();
      if (file_format) {
        if (file_format->getFormatType() != "av/mpegps") {
          throw Exception("No PS/PES stream was found, seems to be an elementary stream!");
        }
      } else {
        throw Exception("Can't identify data stream!");
      }
    }
    
    int chapters[100];
    int titles = dvd_interface->getTitles();
    if (titles > 100) throw Exception("DVD has too many titles");
    for(int i=1; i<=titles; i++) {
      dvd_interface->select(i,1,1);
      chapters[i] = dvd_interface->getChapters(i);
    }
    if ( (title > titles) || (title < 1) ) {
      throw Exception(string("You selected title ") + title + string(", but there are only ") + titles + string(" titles on this DVD"));
    }
    if ( (chapter > chapters[title]) || (chapter < 1) ) {
      throw Exception(string("You selected chapter ") + chapter + string(", but there are only ") + chapters[title] + string(" chapters in this title"));
    }
    
    if ( show_information ) {
      cout << "DVD-Contents:" << endl;
      cout << "-------------" << endl;
      cout << "There are " << titles << " titles on this DVD." << endl;
      
      for(int i=1; i<=titles; i++) {
        cout << "Title " << i << " has " << chapters[i] << " chapters" << endl;
      }

      // fw: currently can't get these values without adding a demux AND MPEG video decoder
//       cout << endl;
//       cout << "Video Resolution: " << width << "x" << height << endl;
//       cout << "Aspect Ratio:     " << file_format->getIntValue(Format::ratio_x_param)/100 << ":" << file_format->getIntValue(Format::ratio_y_param)/100 << endl;
//       cout << "Audio Type:       " << file_format->getStringValue(Format::audio_format_string_param) << endl;
      
      dvd->deinitOutput();
      dvd->deinit();
      registry.releaseNode(*dvd);
      registry.releaseNode(*writer);
      return 0;
    }
    
    if ( (chapterend == 0) || (chapterend > chapters[title]) ) chapterend = chapters[title];
    dvd_interface->select(title, chapter, 1);
    cout << "Reading title " << title << " from chapter " << chapter << " to " << chapterend << endl;
    
    setStderr( true );
    /*End< Initialize DVD */
    
    /*Begin> Initialize writer */
    writer->init();
    
    ISinkFileHandler_var file_sink(writer->getParentObject()->getCheckedInterface<ISinkFileHandler>());
    file_sink->setFilename(filename);
    /*End< Initialize writer */
    
    /*Begin> Connect graph */
    connect(dvd, writer);
    
    dvd->activate();
    writer->initOutput();
    writer->activate();
    /*End< Connect graph */
    
    /*Begin> Start graph */
    dvd->start();
    writer->start();
    /*End< Start graph */
    
    /* Begin> Configure and start console producer */
    ConsoleProducer console;
    console.readConfig( getNMMResource("XProducer.xml") );
    console.connectTo(&receiver);
    console.start();
    /* End< Add console producer */
    
    /* Begin> Register event handlers */
    receiver.getEventDispatcher().registerEventListener(IDisplayEvents::keyReleased_event,
                                                        new TEDObject1<EventCatcher, string>(&ec, &EventCatcher::keyReleased));
    writer->getParentObject()->registerEventListener(ITrack::endTrack_event,
                                                     new TEDObject0<EventCatcher>(&ec, &EventCatcher::endTrack));
    // EDObject for progress receiving
    TEDObject2<EventCatcher, u_int64_t, u_int64_t> progress_listener(&ec, &EventCatcher::setProgress);
    writer->getParentObject()->registerEventListener(IProgressListener::setProgress_event,
                                                     &progress_listener);
    /* End< Register event handlers */
    
    cout << "Extracting stream (press 'q' to stop extraction)" << endl;
    ec.handleEvents();

    /* Begin> Deregister event handlers */
    writer->getParentObject()->removeEventListener(&progress_listener);
    /* End< Deregister event handlers */
    
    console.stop();
    
    // send a newline as progress does not!
    cout << endl;

  } catch(const Exception& e) {
    cerr << e << endl;
    error_level = 1;
  } catch(...) {
    cerr << "Something strange went wrong!" << endl;
    error_level = 1;
  }
  
  /*Begin> Clean up */
  try {
    if (writer.get() && writer->isStarted()) {
      writer->stop();
    }
    if (dvd.get() && dvd->isStarted()) {
      dvd->stop();
    }
  
    if (writer.get() && writer->isActivated()) {
      writer->flush();
    }
    if (dvd.get() && dvd->isActivated()) {
      dvd->flush();
    }
  
    if (writer.get()) {
      if (writer->isActivated()) {
        writer->deactivate();
      }
      if (writer->isOutputInitialized()) {
        writer->deinitOutput();
      }
      {
        IInputJack_var writer_in_jack(writer->getInputStream("default"));
        if (writer_in_jack.get() && writer_in_jack->isConnected()) {
          if (dvd.get() && dvd->isActivated()) {
            dvd->deactivate();
          }
        }
      }
    }
  
    if (dvd.get()) {
      if (dvd->isActivated()) {
        dvd->deactivate();
      }
      if (dvd->isOutputInitialized()) {
        dvd->deinitOutput();
      }
    }
  
    if (writer.get() && writer->isInitialized()) {
      writer->deinit();
    }
    if (dvd.get() && dvd->isInitialized()) {
      dvd->deinit();
    }
    
    ClientRegistry& registry = app->getRegistry();  
    if (writer.get()) {
      registry.releaseNode(*writer);
    }
    if (dvd.get()) {
      registry.releaseNode(*dvd);
    }
  } catch(const Exception& e) {
    cerr << "Exception during cleanup" << endl;
    cerr << e << endl;
    error_level = 1;
  } catch(...) {
    cerr << "Something strange went wrong during cleanup" << endl;
    error_level = 1;
  }
  /*End< Clean up */
  
  return error_level;
}
