/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "InputJack.hpp"
#include "Node.hpp"
#include "OutputJack.hpp"
#include "InstreamChannel.hpp"
#include "nmm/interfaces/base/INodeImpl.hpp"
#include "nmm/interfaces/base/IOutputJack.hpp"
#include "TransportStrategy.hpp"
#include "TSFactory.hpp"
#include "GenericNode.hpp"
#include <iostream>
#include "nmm/base/registry/ReferenceManager.hpp"
#include "ConnectionManager.hpp"
#include "FormatStrategyMap.hpp"

namespace NMM {
  InputJack::InputJack(Node* parent, const string& jack_tag, const string& name)
  :  Jack(parent, jack_tag, name), __transport_strategy(0)
  {
    data_jack    = this;
  }

  InputJack::~InputJack() {
    if (__transport_strategy) {
      __transport_strategy->scheduleDestruction();
    }
  }

  bool InputJack::isConnected() const 
  { 
    if((getPartner() || data_partner.get()) && getMessageReceiver()) 
      return true; 
    else 
      return false; 
  }

  void InputJack::connectTo(const OutputJack* out_jack) const
    throw (const BindingException&)
  {
    if(out_jack == 0) {
      throw BindingException(this, out_jack, "OutputJack is 0");
    }
    else {
      if (getPartner() == 0) {
        setPartner(const_cast<OutputJack*>(out_jack));
        const_cast<InputJack*>(this)->setMessageReceiver(const_cast<OutputJack*>(out_jack));
        NODE_DEBUG_STREAM2( (getParent()) , "Valid Partner connected" << endl);
      } else {
        NODE_WARNING_STREAM2( (getParent()), "Jack already connected" << endl);
        throw BindingException(this, out_jack, "InputJack already connected");
      }
    }
  }

  void InputJack::disconnect()
  {
    if(getParent()->getState() == INode::STARTED) {
    }
    setPartner(0);
    setMessageReceiver(0);
    if (__transport_strategy) 
      __transport_strategy->scheduleDestruction();

    __transport_strategy = 0;
    data_partner.reset(0);

  }

  void InputJack::setTransportStrategy(TransportStrategy* strategy) \
                 throw(const NMM::BindingException &)
    {
    /* Begin> If an old strategy is available, release it. */
    if (__transport_strategy)
      __transport_strategy->scheduleDestruction();
    /* End< */

    /* stores the new strategy */
    __transport_strategy = strategy;

    /* Begin> Handle downstream messages */
    /* If the strategy receives downstream-messages, the input_jack receives it*/
    strategy->setMessageReceiver(this);
    /* End< */

    /* Begin> Handle upstream messages. */
    /* If the the jack receives upstream-messages from its parent they are forwarded to the given strategy */
    setMessageReceiver(strategy);
    /* Signals from the stratgey, like FREE_QUEUE_SLOT are send to the jacks parent*/
    strategy->setSignalReceiver(this);
    /* End< */

    /* Begin> Configure the TransportStrategie's Buffermanager */
    /* If a TransportStrategy receives a Buffer, it needs a BufferManager to request it. */
    if (getParent()) {
      /* Request the IExternalBuffermanager */
      IExternalBufferManager* buffer_manager = dynamic_cast<IExternalBufferManager*>(getParent());
      if (buffer_manager) {
        /* and tell the strategy to use it */
        strategy->setBufferManager(buffer_manager);
      } else {
        /* Otherwise the TransportStrategy can not receive messages */
        NODE_ERROR_STREAM(getParent(), "Parent implements not the IExternalManager" << endl);
      }
    }
    /* End< */
    
    /* Begin> Configure the TransportStrategy's Format */
    if (getParent()) {
      const Format* format = getParent()->getInputFormat(getJackTag().c_str());
      if (format) {
        MESSAGE_STREAM("Configure InputJack with format " << format->getFormatType() << endl);
        strategy->configureForInputJack(*format);
      }
    }
    /* End< */
  }


  void InputJack::setTransportStrategy(const TSAddress* address) \
		  throw(const NMM::BindingException &)

  {
    /* If the address is invalid throw an Exception */
    if (!address)
      throw BindingException("Invalid Address (0-ptr)");

    /* Begin> Handle local case!
     *        For performance reason, the jack communicates with its jack partner without a
     *        TransportStrategy in a local case.
     */
    /* If the address describes a local binding, its a NMMReference describes an object in this address space*/
    if (address->getBindingType() == TSAddress::LOCAL) {
      /* Request the NMMObject with NMMReference in the given address. The type of this Object must be
       * OutputJack
       */
      OutputJack* output_jack = SReferenceManager::getInstance().getObject<OutputJack>( address->getNMMReference() );

      /* If the address points not to an OutputJack, throw an Exception */
      if (!output_jack)
        throw BindingException("No such OutputJack available");

      /* Connect this jack with Call the connectTo method for this jack */
      connectTo(output_jack);
      /* Connect the output_jack with this jack */
      output_jack->connectTo(this);
      /* The output_jack is our partner and receives Messages from out parent */
      setMessageReceiver(output_jack);
      return;
    }
    /* End< */

    /* Begin> Handle distributed case */
    /* 1. Create a TransportStrategy which can connect to the given address */
    TransportStrategy* strategy = STSFactory::getInstance().createStrategy(address->getStrategyType());

    /* If this jack can not create such a TransportStrategy throw a BindingException */
    if (!strategy)
      throw  BindingException("Can not create strategy with type "+ address->getStrategyType());

    /* 2. The strategy connects to the given address */
    strategy->setAddress(*address);

    /* 3. The connection between this object and the other object is used to send instream messagess.
     *    So configure the jack to the new strategy*/
    setTransportStrategy(strategy);

    /* 4. Tell the other jack to use it for instream message communication too.*/
    Event event(IJack::connectToJack_event);
    if ( strategy->receiveReliableEvent(event) != SUCCESS ) {
      /* Begin> Error handling */   
      if (event.getException()) {
        NODE_DEBUG_STREAM2( getParent() , *event.getException() << endl);
      } else {
        NODE_DEBUG_STREAM2( (getParent()) , "could not set TS in InputJack" << endl);
      }
      /* End< */
    }
  }

  void InputJack::connectToJack() {
    TransportStrategy* strategy = dynamic_cast<TransportStrategy*>(getCurrentSender());
    if (!strategy)
      throw BindingException();

    setTransportStrategy(strategy);
  }


  CommunicationChannel* InputJack::connectTo(const IOutputJack& partner)
      throw(const BindingException& )
  {
    // setPartner(0); // should not be needed
    data_partner.reset(dynamic_cast<IOutputJack*>(partner.copy()));
    IInputJack_var tmp_jack(getInterface<IInputJack>());
    return new InstreamChannel(*tmp_jack.get(), *data_partner.get());
  }

  ITransportStrategy* InputJack::getTransportStrategy() {
    if (__transport_strategy)
      return __transport_strategy->getInterface<ITransportStrategy>();

    return 0;
  }

  IOutputJack* InputJack::getDataPartner()
  {
    if (data_partner.get()) {
      return dynamic_cast<IOutputJack*>(data_partner.get()->copy()); 
    }
    else {
      return 0;
    }
  }
  
  TSAddress* InputJack::getDefaultAddress(const string& strategy_type) const {
    const Format* format = getParent()->getInputFormat(getJackTag().c_str());
    if (SConnectionManager::getInstance().getFormatStrategyMap().isSupported(*format, strategy_type)) {
      return NMMObject::getDefaultAddress(strategy_type);
    } else {
      return NULL;
    }
  }
}
