/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_JACK_HPP_
#define _NMM_JACK_HPP_

#include "nmm/interfaces/base/IJackImpl.hpp"
#include "StdNamedObject.hpp"
#include "ISignalReceiver.hpp"

namespace NMM {

  class Node;

  class TransportStrategy;
  /**
   * This is an (abstract) jack, superclass of InputJack and
   * OutputJack. To use a jack, one should extend InputJack and
   * OutputJack (as in SimpleInputJack and SimpleOutputJack).
   */
  class Jack : public IMessageReceiver, 
	       public IMessageSender,
               public ISignalReceiver,
	       virtual public IJackImpl, public StdNamedObject
  {
  public:

    /**
     * The main constructor. This constructor only sets the parent,
     * not the partner.
     * \param node the parent node.
     */
    Jack(Node* parent, const string& jack_tag = "default", const string& name = "Jack");

    /**
     * The main destructor.
     */
    virtual ~Jack() {};

    virtual           Node* getParent();            
    virtual     const Node* getParent() const;

    /**
     * A method to obtain a pointer to the partner of this jack.
     * The partner is not (necessarily) the jack to exchange buffers with,
     * but rather the jack this jack is logically connected to, i.e.
     * communicates with about data stream control.
     * \returns a pointer to the partner jack of this jack, or NULL if
     * this jack has no partner jack.
     */
    inline const Jack* getPartner()  const  {return __partner;}
    inline       Jack* getPartner()         {return __partner;}

    virtual  string getJackTag()       const {return __jack_tag;}

    virtual inline void receiveSignal(Signal sig) { __signal_receiver->receiveSignal(sig); }

    /**
     * Returns true, if this Jack is connected, otherwise false.
     * \return true, if this Jack is connected, otherwise false.
     * \note Still public for Node::is{Input,Output}Connected()
     */
    virtual bool isConnected() const = 0;

  protected: //interface methods
    virtual  INode* getIParent();

  protected:
    inline void  setPartner(Jack *partner)                                  const { __partner = partner; }
    inline void  setMessageReceiver(IMessageReceiver* message_receiver)           { __message_receiver = message_receiver; }
    inline       IMessageReceiver* getMessageReceiver()                           { return __message_receiver; }
    inline const IMessageReceiver* getMessageReceiver()                     const { return __message_receiver; }
    inline       ISignalReceiver*  getSignalReceiver()                            { return __signal_receiver; }
  private:
    Node*                   __parent;
    string                  __jack_tag;
    mutable Jack*           __partner;
    IMessageReceiver*       __message_receiver; //the new data partner
    ISignalReceiver*        __signal_receiver;  //the new signal partner
  };

}

#endif // _NMM_JACK_HPP_
