/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_OUTPUTJACK_HPP_
#define _NMM_OUTPUTJACK_HPP_

#include "Jack.hpp"
#include "nmm/interfaces/base/IOutputJackImpl.hpp"
#include "nmm/interfaces/base/IInputJack.hpp"

namespace NMM {

  class Node;
  class Message;
  class InputJack;
  class IInputJack;
  class SenderStrategy;
  class IMessageReceiver;
  class InputJack ;
  class CommunicationChannel;
  /**
   * This is the (abstract) superclass for all output jacks.
   */
  class OutputJack : public Jack, public IOutputJackImpl 
  {

  private:
    mutable IInputJack_var   data_partner;  ///< The jack this jack should exchange buffers with.
    OutputJack              *data_jack;     ///< The data jack of this jack.
    TransportStrategy*      __transport_strategy; //< The TransportStrategy to send messages over network
  public:

    /**
     * The main constructor. This constructor only sets the parent,
     * not the partner. 
     * \param node The parent node.
     */
    OutputJack(Node* node, const string& jack_tag = "default", const string& name = "OutputJack");

    /**
     * The main destructor.
     */
    virtual ~OutputJack() ;

    /**
     * Connect to a partner jack.
     * \param jack partner jack.
     */
    virtual void connectTo(const InputJack* partner) const
      throw(const BindingException& );

    /**
     * A method to obtain a pointer to the data jack of this jack.
     * In the case of proxy jacks, it is more efficient to exchange
     * the actual buffers between the "end" jacks instead of by a
     * chain of proxy jacks. In the case of non-proxy jacks, this
     * method will simply return a pointer to this jack.
     * \returns a pointer to the data jack of this jack, or NULL if
     * this jack has no data jack (only in special cases). If a jack
     * is its own data jack, it should return a pointer to itself, not
     * NULL.
     */
    virtual const OutputJack* getDataJack() const {return data_jack;}
    virtual       OutputJack* getDataJack()       {return data_jack;}
    /**
     * Returns true, if this Jack is connected, otherwise false.
     * \return true, if this Jack is connected, otherwise false.
     * \note Still public for MPEGDemuxNode
     */
    virtual bool isConnected() const;

  protected: //interface methods

    /**
     * A method to obtain a pointer to the data partner of this jack.
     * \returns a copy to the data partner jack of this jack, or NULL
     * if this jack has no data partner.
     */
    virtual       IInputJack* getDataPartner();

    virtual CommunicationChannel* connectTo(const IInputJack& partner)
      throw(const BindingException& );


    /**
     * Disconnect from partner jack.
     */
    virtual void disconnect();

    virtual void setTransportStrategy(const TSAddress* )throw(const BindingException&);

    virtual void connectToJack();

    virtual ITransportStrategy* getTransportStrategy();
    
    virtual TSAddress* getDefaultAddress(const string& strategy_type) const;
    
  protected:
    virtual void setTransportStrategy(TransportStrategy* strategy);
  };

}

#endif // _NMM_OUTPUTJACK_HPP_
