/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "nmm/base/Message.hpp"
#include "Node.hpp"
#include "QueuedInputJack.hpp"
#include "EventDispatcher.hpp"
#include "OutputJack.hpp"
#include "nmm/interfaces/base/INodeImpl.hpp"
#include "TransportStrategy.hpp"
#include "QueuedOutputJack.hpp"
#include "CommunicationChannel.hpp"

namespace NMM {
  QueuedInputJack::QueuedInputJack(Node* parent, 
				   const string& jack_tag /* = "default */,
				   const string& name /* = "QueuedInputJack"*/,
				   const StreamQueue::Mode mode /* = StreamQueue::DEFAULT_MODE */,
				   const unsigned int size /* = StreamQueue::DEFAULT_SIZE */)
    : InputJack(parent, jack_tag, name)
  {
    stream_queue = 
      new StreamQueue( mode, size, name.c_str(), FREE_DOWNSTR_QUEUE_SLOT, INPUT_QUEUE_NOT_EMPTY);

    stream_queue->setParentSignalReceiver(dynamic_cast<ISignalReceiver *>(parent));
  }

  QueuedInputJack::~QueuedInputJack() 
  {
    stream_queue->setPartnerSignalReceiver(0);
    delete stream_queue;
  }

  void QueuedInputJack::connectTo(const OutputJack* out_jack) const
    throw (const BindingException&)
  {
    InputJack::connectTo(out_jack);
    stream_queue->setPartnerSignalReceiver(const_cast<Jack *>(getPartner()));
  }

  void QueuedInputJack::disconnect()
  {
    InputJack::disconnect();
    stream_queue->setPartnerSignalReceiver(0);
  }

  Result QueuedInputJack::receiveMessage(Message* message, const IMessageSender*  )
  {
    DEBUG_STREAM2("Forwarding buffer to parent" << endl);

    return stream_queue->receiveMessage(message);
  }

  Result QueuedInputJack::sendMessage(Message *message)
  {
    if(getMessageReceiver() != 0) {
      DEBUG_STREAM2("Send Buffer to Jackpartner" << endl);
      DEBUG_STREAM2("sendMessage UPSTREAM" << endl);
      Result r = getMessageReceiver()->receiveMessage(message, this);
      if(r == BLOCKED) {
	DEBUG_STREAM2("Jackpartner returned BLOCKED which means its stream_queue is full now" << endl);
      }
      return r;
    }
    else {
      // no partner, must discard buffer
      DEBUG_STREAM2("QueuedInputJack> Partner-Jack is 0" << endl);
      message->release();
      return FAILURE;
    }
  }

  void QueuedInputJack::setTransportStrategy(TransportStrategy* strategy)\
			throw(const NMM::BindingException &)
  {
    InputJack::setTransportStrategy(strategy);
    stream_queue->setPartnerSignalReceiver(strategy);
    strategy->setSignalReceiver(this);
  }

  Result QueuedInputJack::flush()
  { 
    return stream_queue->flush(); 
  }

  unsigned int QueuedInputJack::getSize() const 
  { 
    return stream_queue->getSize(); 
  }

  unsigned int QueuedInputJack::getMaxSize() const 
  { 
    return stream_queue->getMaxSize(); 
  }

  Result QueuedInputJack::setMaxSize(unsigned int size) 
  { 
    return stream_queue->setMaxSize(size); 
  }

  StreamQueue::Mode QueuedInputJack::getMode() const 
  { 
    return stream_queue->getMode(); 
  }

  Result QueuedInputJack::setMode(StreamQueue::Mode mode) 
  { 
    return stream_queue->setMode(mode); 
  }

}
