/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_THREADEDNODE_HPP_
#define _NMM_THREADEDNODE_HPP_

#include "nmm/base/Node.hpp"
#include "nmm/interfaces/base/IThreadedNodeImpl.hpp"
#include "nmm/base/resourcemgr/ResourceCollector.hpp"

namespace NMM {

  /**
      This class introduces thread-functionality.
  */

  class ThreadedNode : public Node,
                       public IThreadedNodeImpl,
                       public ResourceCollector
{

  private:

    bool activated_before;

    bool thread_owner;

    Thread thread;
    /**
     * This method is merely a way to start the thread on this
     * object's run method.
     * This method will also contain a try-catch block catching
     * any exception originating from inside the run method. Since a thread
     * is a single exception domain, any exception at this point
     * would invariably call abort() without the possibility of another
     * thread (especially the main thread) catching it.
     * \param me The Node whose run method should be run, cast to void*
     *           because of pthread_create.
     * \return 0, the return value has no meaning.
     */
    static void* run_static(void *me);

    pid_t thread_pid;
    int thread_priority;

    /**
     * Set the threads process id (used in the run_static method)
     */
    void setPID(pid_t new_pid);

  protected:

    /**
     * The work method. Override this to implement the node's
     * functionality. When activated, the node will create a thread
     * which will start the run method. On deactivation, the thread
     * will be canceled (?)
     */
    virtual void run(int num_times = 0) = 0;

    /**
     * Lower priority by amount
     */
    void lowerPriority(const unsigned int& amount);


  public:

    /**
     * The main constructor.
     */
    ThreadedNode(const char* name);

   /**
     * Destructor.
     */
    virtual ~ThreadedNode();

    /**
     * This method creates the thread to run in and starts it.
     */
    virtual void initOutput()
      throw (const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);

    /**
     * This method halts the thread.
     */
    virtual void deinitOutput()
      throw (const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);

    /**
     * If set to false no thread will be created for this node.
     */
    virtual void setThreadOwner(bool owner)  {  thread_owner = owner;  }

    /**
     * Get the process id of the (running) node
     */
    pid_t getPID();

    /**
     * Set the priority level (in state INITIALIZED)
     * \param level The new priority level (valid range is 1 to 20)
     */
    Result setPriority(unsigned int level);

//------ pthread scheduling methods (only allowed for root)
    void enableCustomScheduling(Thread::Scheduling_Policy);
    void setMinPriority();
    void setMidPriority();
    void setMaxPriority();

  };

}

#endif // _NMM_THREADEDNODE_HPP_
