/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_BUFFER_HPP_
#define _NMM_BUFFER_HPP_

#include "nmm/base/Message.hpp"

namespace NMM 
{
  class BufferManager; //class Allocator;

  /**
   * This class represents a multimedia buffer. A multimedia
   * buffer is merely a continuous block of memory with some
   * header information.
   */
  class Buffer : public Message
  {
  private:
    const BufferManager *manager;  ///< The BufferManager of this buffer.
    TValue<size_t> size;           ///< The size in bytes of this buffer.
    TValue<size_t> realsize;               ///< The number of bytes actually used.
    TValue<char*> data;            ///< The actual data field.
    TValue<bool>  own_data;        ///< True if constructor allocated the data.

    void registerAttributes();

    ThreadMutex op_mutex;  ///< Mutex to protect operations like getWriteableInstance(), increaseCount() and release()
  protected:
  /** This method handles the unused buffer
    */
    virtual void erase() const;
  public:

    /**
     * The main constructor. A buffer is passed its size and
     * a pointer to the Allocator object which allocated
     * (i.e. created) it.
     * \param size The size of the buffer in bytes.
     * \param allocator The Allocator object in charge of this buffer.
     */
    Buffer (const size_t newsize, const BufferManager *master);

    /**
     * A constructor to reuse a given memory area. Same as the main
     * constructor, except that no memory is allocated for the data.
     * Instead, the char array passed is used.
     * \param size The size of the buffer in bytes.
     * \param data The actual buffer.
     * \param allocator The Allocator object in charge of this buffer.
     */
    Buffer (const size_t newsize,char* initial_data,
	    const BufferManager *master);

    Buffer(const Buffer&);

    ~Buffer ();   ///<default destructor

    /**
     * Use this method to reset the internal state of the buffer
     * This means the mandatory bit will be set to false;
     * the used size will be set to the total size of the buffer;
     * and the buffer is set to be a data buffer
     * This method should be called by the BufferManager before 
     * an already used buffer is returned.
     */
    void reset(); 

    /**
     * This method returns a writeable instance of the buffer.
     * Since Buffer-objects can be shared along the flowgraph, 
     * an operation, which modifies the data of the Buffer-object,
     * must call this method before getData().
     * Typical use: 
     * myBuffer = myBuffer->getWritableInstance(); 
     * char *data = myBuffer->getData();
     */
    Buffer *getWriteableInstance();

    /**
     * Returns a pointer to the data field. The data field is a
     * simple byte array.
     * \returns A pointer to the byte-array used.
     */
    char* getData () const;

    /**
     * Attempts to set the data field to an etxernal memory
     * area. Only use if you own the Buffer!
     */
    Result setData (const size_t size, char *data);

    /**
     * Returns the size of this buffer.
     * \returns The size of this buffer.
     */
    size_t getSize () const;

    /**
     * Returns the the used bytes of this buffer. Buffers should
     * be filled from the beginning.
     * \returns The used bytes of this buffer.
     */
    size_t getUsed () const;


    /**
     * Sets the used bytes of this buffer. The used byte count is
     * automatically clipped to the range between 0 and size inclusive.
     * \param used the new used byte count.
     */
    void setUsed (const size_t actused);

    /**
     * assignement operator 
     */
    const Buffer& operator=(const Buffer& b);

    /** This method returns a non writeabale copy of the current buffer and increases the ref-count
      */
    virtual Buffer* copy() const;
  };
}

#endif // _NMM_BUFFER_HPP_









