/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "BufferedSink.hpp"

#include "nmm/base/memorymgr/IExternalBufferManager.hpp"
#include "nmm/base/memorymgr/Buffer.hpp"

#include "nmm/base/IReceiver.hpp"

#include <sys/time.h>
#include <string.h>

#include <live/RTPSource.hh>

namespace NMM {
  
  BufferedSink::BufferedSink(UsageEnvironment& env,
                             IMessageReceiver* receiver, IMessageSender* sender,
                             IExternalBufferManager* buffer_manager,
                             unsigned buffer_size, unsigned frame_size)
    : MediaSink(env),
      NamedObject("BufferedSink"),
      __current_buffer(NULL),
      __buffer_size(buffer_size),
      __frame_size(frame_size),
      __blocked(false),
      __flush(false),
      __internal_cond(__internal_mutex),
      __buffer_manager(buffer_manager),
      __receiver(receiver),
      __sender(sender),
      __first_rtp_frame(true),
      __first_rtp_seqno(0)
  { 
    /* set time offset */
    gettimeofday(&__start_time, NULL);
  }
  
  BufferedSink::~BufferedSink() {
    /* delete pending buffer */      
    if (__current_buffer) {
      __current_buffer->release();
      __current_buffer = NULL;
    }
  }
  
  BufferedSink* BufferedSink::createNew(UsageEnvironment& env,
                                        IMessageReceiver* receiver, IMessageSender* sender,
                                        IExternalBufferManager* buffer_manager,
                                        unsigned buffer_size, unsigned frame_size) 
  {
    return new BufferedSink(env, receiver, sender, buffer_manager,
                            buffer_size, frame_size);
  }
  
  Boolean BufferedSink::continuePlaying() {
    /* check for exit condition */
    if (!fSource || __flush) {
      return False;
    }
    
    /* allocate buffer if necessary */
    if (!__current_buffer) {
      /* we must have a buffer manager */
      if(!__buffer_manager) {
        GLOBAL_ERROR_STREAM("No buffer manager" << std::endl);
        return False;
      }
      
      /* allocate buffer */
      __current_buffer = __buffer_manager->getNewBuffer(__buffer_size);
      if (!__current_buffer) {
        return False;
      }
      
      /* the buffer is initially empty */
      __current_buffer->setUsed(0);
    }
    
    /* get next frame */
    unsigned used = __current_buffer->getUsed();
    fSource->getNextFrame(reinterpret_cast<unsigned char*>(__current_buffer->getData()) + used, __buffer_size - used,
                          &afterGettingFrame, this, &onSourceClosure, this);
    
    return True;
  }
  
  void BufferedSink::afterGettingFrame(void* clientData, unsigned frameSize, unsigned /*numTruncatedBytes*/, struct timeval presentationTime, unsigned /*durationInMicroseconds*/ ) {
    /* notify the sink */
    BufferedSink* sink = static_cast<BufferedSink*>(clientData);
    sink->handleFrame(frameSize, presentationTime);
    
    /* try getting the next frame */
    sink->continuePlaying();
  }
  
  void BufferedSink::handleFrame(unsigned frameSize, struct timeval presentationTime) {
    if (__flush) {
      /* discard frame */
      return;
    }
    
    /* count bytes of frame */
    DEBUG_STREAM2("Received frame of " << frameSize << " bytes" << endl);
    unsigned used = __current_buffer->getUsed();
    used += frameSize;
    __current_buffer->setUsed(used);
    
    /* check if the source is an RTP source */
    Timestamp t;
    RTPSource* rtp_source = dynamic_cast<RTPSource*>(fSource);
    if(rtp_source) {
      /* set the valid flag based on the marker bit */
      if (rtp_source->hasBeenSynchronizedUsingRTCP()) {
	t.is_valid = true;//rtp_source->curPacketMarkerBit();
      } else {
	t.is_valid = false;
      }

      /* get sequence number to reproduce stream counter */
      uint32_t seqno = rtp_source->curPacketRTPSeqNum();
      if(__first_rtp_frame) {
        /* this is the first frame. RTP sequence numbers have random offset */
        __first_rtp_seqno = seqno;
        __first_rtp_frame = false;
      }
      __last_rtp_seqno = seqno;
      t.stream_counter = seqno - __first_rtp_seqno;
    } else {
      /* this does not happen since we always read from RTP */
      GLOBAL_ERROR_STREAM2("NO RTP SOURCE!!!" << std::endl);
      return;
    }
    
    /* subtract time offset */
    /*presentationTime.tv_sec -= __start_time.tv_sec;
    if (presentationTime.tv_usec < __start_time.tv_usec) {
      presentationTime.tv_sec--;
      presentationTime.tv_usec += 1000000;
    }
    presentationTime.tv_usec -= __start_time.tv_usec;*/
    
    /* convert to NMM timestamp */
    t.sync_time.sec = presentationTime.tv_sec;
    t.sync_time.nsec = presentationTime.tv_usec * 1000;
    __current_buffer->setTimestamp(t);
    
    /* enqueue buffer */
    if(used > 0 /*> __buffer_size - __frame_size*/) {
      MutexGuard mg(__internal_mutex);
      if (__blocked && !__flush) {
        DEBUG_STREAM2("Waiting for free queue slot" << endl);
        __internal_cond.wait();
      }
      if (__flush) {
        return;
      }
      
      DEBUG_STREAM2("Passing buffer of " << __current_buffer->getUsed() << " bytes to receiver" << endl);
      if (BLOCKED == __receiver->receiveMessage(__current_buffer, __sender)) {
        DEBUG_STREAM2("Receiver returned BLOCKED" << endl);
        __blocked = true;
      }
      __current_buffer = NULL;
    }
    /* End< */
  }
  
  void BufferedSink::receiveSignal(Signal signal) {
    switch (signal) {
    case FREE_QUEUE_SLOT:
      {
        DEBUG_STREAM2("Received signal FREE_QUEUE_SLOT from receiver" << endl);
        
        /* wake up the liveMedia thread */
        MutexGuard mg(__internal_mutex);
        __blocked = false;
        __internal_cond.notify();
      }
      break;
    default:
      break;
    }
  }
  
  void BufferedSink::flush() {
    /* wake up the thread */
    MutexGuard mg(__internal_mutex);
    __flush = true;
    __internal_cond.notify();
  }
  
} // namespace NMM
