/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "BufferedSource.hpp"

#include "nmm/base/memorymgr/Buffer.hpp"

#include <string.h>

namespace NMM {
  
  BufferedSource::BufferedSource(UsageEnvironment& env, ISignalReceiver* parent, float sec_per_frame /*= 1.0*/,
                                 IQueued::Mode queue_mode /*= IQueued::MODE_SUSPEND*/,
                                 unsigned int queue_size /*= 1*/, bool no_merge /*= false*/)
    : FramedSource(env),
      NamedObject("BufferedSource"),
      __stream_queue((StreamQueue::Mode) queue_mode, queue_size, "BufferedSource"),
      __current_buffer(NULL),
      __buffer_offset(0),
      __sec_per_frame(sec_per_frame),
      __internal_cond(__internal_mutex),
      __flush(false),
      __waiting(false),
      __no_merge(no_merge)
  {
    /* this object reads from the queue, so it is the parent of the queue */
    __stream_queue.setParentSignalReceiver(this);
    
    /* the parent writes tothe queue, so it is the partner of the queue */
    __stream_queue.setPartnerSignalReceiver(parent);

  }
  
  BufferedSource::~BufferedSource()
  {}
  
  BufferedSource* BufferedSource::createNew(UsageEnvironment& env, ISignalReceiver* parent, float sec_per_frame /*= 1.0*/,
                                            IQueued::Mode queue_mode /*= StreamQueue::MODE_SUSPEND*/,
                                            unsigned int queue_size /*= 1*/, bool no_merge /*= false*/)
  {
    return new BufferedSource(env, parent, sec_per_frame, queue_mode, queue_size, no_merge);
  }
  
  void BufferedSource::doGetNextFrame() {
    /* read data until we have a complete frame */
    unsigned char* to = fTo;
        
    fFrameSize = 0;

    /* maximum size of a frame */
    size_t max_size = fMaxSize;
        
    //static struct timezone Idunno;
    
    /* merge NMM buffers to frames */
    for(;;) {
      /* if we have no current buffer, fetch one from the queue */
      fetchData();
      
      /* check if flush was called */
      if (__flush) {
        /* if we still have a buffer, release it now */
        if(__current_buffer) {
          __current_buffer->release();
          __current_buffer = NULL;
        }
        
        /* exit */
        break;
      }
      
      /* get buffer size */
      unsigned used = __current_buffer->getUsed();
            
      /* convert NMM timestamp to RTP timestamp */
      if (!fFrameSize) {
	/* If there are too many artefacts in the video stream, try gettimeofday-timestamps */
	//gettimeofday(&fPresentationTime, &Idunno);
	fPresentationTime.tv_sec = __current_buffer->getTimestamp().sync_time.sec;
	fPresentationTime.tv_usec = __current_buffer->getTimestamp().sync_time.nsec / 1000;
      }
      
      /* set these variable to well-defined (but not necessarily meaningful) values */
      fDurationInMicroseconds = 0;
      fNumTruncatedBytes = 0;
      
      // XXX: should set the playing duration ?
      // XXX: set the marker bit for valid TS ?
      // XXX: are we truncating frames here ?
      
      /* check if maximum size has been reached */
      if ((fFrameSize + used - __buffer_offset) > max_size) {
        /* number of bytes to copy*/
        size_t size = max_size - fFrameSize;
        
        /* copy data */
        DEBUG_STREAM2("Passing frame of " << size << " bytes to RTP sink" << endl);
        memcpy(to, __current_buffer->getData() + __buffer_offset, size);
        __buffer_offset += size;
        
        /* advance pointers */
        to += size;
        fFrameSize += size;
        
        /* maximum size of a frame reached */
        break;
      } else {
        /* number of bytes to copy*/
        size_t size = used - __buffer_offset;
        
        /* copy data */        
        memcpy(to, __current_buffer->getData() + __buffer_offset, size);
        DEBUG_STREAM2("Passing frame of " << size << " bytes to RTP sink" << endl);
        __current_buffer->release();
        __current_buffer = NULL;
        
        /* advance pointers */
        to += size;
        fFrameSize += size;
        
        /* if buffers should not be merged, then this is the end of the frame */
        if(__no_merge) {
          break;
        }
      }
      
      /* stop if maximum frame size has been reached */
      if(fFrameSize == max_size) {
        break;
      }
    }
    
    /* check if frame has zero size */
    if(!fFrameSize) {
      /* we are closed */
      GLOBAL_DEBUG_STREAM("BufferedSource: source closed!" << endl);
      handleClosure(this);
      return;
    }
    
    /* switch to another task, and inform the reader that he has data */
    nextTask() = envir().taskScheduler().scheduleDelayedTask(0, reinterpret_cast<TaskFunc*>(afterGetting), this);
  }
  
  void BufferedSource::doStopGettingFrames() {
    handleClosure(this);
  }
  
  float BufferedSource::getPlayTime(unsigned numFrames) const {
    return __sec_per_frame * numFrames;
  }
  
  Result BufferedSource::writeBuffer(Buffer* buffer) {
    /* enqueue buffer */
    DEBUG_STREAM2("Placing buffer of " << buffer->getUsed() << " bytes in send queue" << endl);

    MutexGuard mg(__internal_mutex);
    Result result = __stream_queue.receiveMessage(buffer);
    if (result == BLOCKED) {
      DEBUG_STREAM2("Send queue returned BLOCKED" << endl);
    }
    return result;
  }
  
  void BufferedSource::receiveSignal(Signal signal) {
    switch (signal) {
    case QUEUE_NOT_EMPTY:
      {
        DEBUG_STREAM2("Received signal QUEUE_NOT_EMPTY from send queue" << endl);
        
        /* wake up the thread */
	//MutexGuard mg(__internal_mutex);
	__internal_cond.notify();
      }
      break;
    default:
      break;
    }
  }
  
  void BufferedSource::flush() {
    /* wake up the thread */
    __flush = true;
    
    MutexGuard mg(__internal_mutex);
    __internal_cond.notify();
    
    /* flush the send queue */
    __stream_queue.flush();
  }
  
  void BufferedSource::fetchData() {
    /* wait for a buffer or a flush */
    while(!__flush && !__current_buffer) {
      Message* message = NULL;
      {
	MutexGuard mg(__internal_mutex);
        
	/* dequeue message from queue */
	message = __stream_queue.dequeueMessage();
	
	/* check if the queue is empty */
	if(!message) {
	  __internal_cond.wait();
	}
      }
      
      /* check if we have a buffer */
      if(message) {
        Buffer* buffer = dynamic_cast<Buffer*>(message);
        if(buffer) {
          __current_buffer = buffer;
          __buffer_offset = 0;
        } else {
          /* should not happen */
          message->release();
          ERROR_STREAM2("Message is not a buffer" << std::endl);
          break;
        }
      }
    }
  }
  
}
