/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef NMM_RTP_GENERIC_HPP
#define NMM_RTP_GENERIC_HPP

#include "RTPAddress.hpp"

#include "nmm/interfaces/net/IRTPControlImpl.hpp"

/* liveMedia forwards */
class TaskScheduler;
class UsageEnvironment;
class Groupsock;

namespace NMM {
  
  /**
   *  @brief Generic RTP Connection
   *
   *  This class manages some common objects needed for an RTP session, such as the liveMedia UsageEnvironment,
   *  the TaskScheduler and the RTP and RTCP sockets. It also stores the session parameters and provides an interface
   *  for setting these parameters.
   */
  class RTPGeneric
    : public IRTPControlImpl
  {
  public:
    /**
     * @name Constructor & Destructor
     */
    //@{
    /**
     *  @brief Default constructor
     */
    RTPGeneric();
    
    /**
     *  @brief Virtual destructor
     */
    virtual ~RTPGeneric();
    //@}
    
    /**
     *  @name IRTPControl Interface
     */
    //@{
    /**
     *  @brief Set canonical name
     *  @param cname The canonical name
     *
     *  This method sets the canonical name of this RTP session. This is a free-form string
     *  which identifies an RTP session. The default is "NMM-RTP".
     */
    virtual void setCNAME(const string& cname);
    
    /**
     *  @brief Get canonical name
     *  @return The canonical name
     */
    virtual string getCNAME() const;
    
    /**
     *  @brief Set multicast TTL
     *  @param ttl The multicast TTL
     *
     *  This method sets the time to live (TTL) for multicast packets. For unicast sessions, this
     *  parameter has no effect. The integer number is the maximum number of hops until a packet
     *  is deleted.
     */
    virtual void setTTL(const unsigned int& ttl);
    
    /**
     *  @brief Get multicast TTL
     *  @return The multicast TTL
     */
    virtual unsigned int getTTL() const;
    
    /**
     *  @brief Set buffer size
     *  @param buffer_size The buffer size
     *
     *  This method sets the buffer size for receiving RTP packets.
     */
    virtual void setBufferSize(const size_t& buffer_size);
    
    /**
     *  @brief Get buffer size
     *  @return The buffer size
     */
    virtual size_t getBufferSize() const;

    /**
     *  @brief Set RTP socket receive buffer size
     *  @param recv_buffer_size Receive buffer size
     *
     *  This method sets the receive buffer size for receiving RTP sockets
     */
    virtual void setReceiveBufferSize(const size_t& recv_buffer_size);
    
    /**
     *  @brief Get RTP socket receive buffer size
     *  @return The receive buffer size
     */
    virtual size_t getReceiveBufferSize() const;
    
    /**
     *  @brief Set maximum frame size
     *  @param max_frame_size The maximum frame size (in bytes)
     *
     *  This method sets the maximum size that a received RTP frame may have. If an incoming frame
     *  is larger than this size, it may be truncated.
     */
    virtual void setMaxFrameSize(const size_t& max_frame_size);
    
    /**
     *  @brief Get maximum frame size
     *  @return The maximum frame size (in bytes)
     */
    virtual size_t getMaxFrameSize() const;
    
    /**
     *  @brief Set session bandwidth
     *  @param bandwidth The session bandwidth (in kb/s)
     *  @todo Check if this parameter is really in kb/s
     *
     *  This method sets the session bandwidth. Note that this is only an estimate which
     *  serves as a hint to the RTP library. The actual bandwidth is determined by the
     *  data which is transferred (for example the video bitrate).
     */
    virtual void setSessionBandwidth(const unsigned int& bandwidth);
    
    /**
     *  @brief Get session bandwidth
     *  @retun The session bandwidth (in kb/s)
     */
    virtual unsigned int getSessionBandwidth() const;
    //@}
    
  protected:
    /**
     *  @name Protected Access Methods
     */
    //@{
    /**
     *  @brief Get RTP socket
     *  @return The RTP socket
     *
     *  This method returns the RTP socket used by this session.
     */
    Groupsock& getRTPSocket() const;
    
    /**
     *  @brief Get RTCP socket
     *  @return The RTCP socket
     *
     *  This method returns the RTCP socket used by this session.
     */
    Groupsock& getRTCPSocket() const;
    
    /**
     *  @brief Get task scheduler
     *  @return The task scheduler
     *
     *  This method returns the liveMedia task scheduler used by this session.
     */
    TaskScheduler& getTaskScheduler() const;
    
    /**
     *  @brief Get usage environment
     *  @return The usage environment
     *
     *  This method returns the liveMedia usage environment used by this session.
     */
    UsageEnvironment& getUsageEnvironment() const;
    
    /**
     *  @brief Set RTP socket
     *  @param sock the new RTP socket
     *
     *  This method sets the RTP socket. A session does not automatically allocate sockets.
     *  This must be done by a subclass. If a socket was already set, then it is destroyed.
     *  If this object is destroyed, then the sockets are destroyed automatically.
     */
    void setRTPSocket(Groupsock* sock);
    
    /**
     *  @brief Set RTCP socket
     *  @param sock the new RTCP socket
     *
     *  This method sets the RTCP socket. A session does not automatically allocate sockets.
     *  This must be done by a subclass. If a socket was already set, then it is destroyed.
     *  If this object is destroyed, then the sockets are destroyed automatically.
     */
    void setRTCPSocket(Groupsock* sock);
    //@}
    
  private:
    /**
     *  @name Session Parameters
     */
    //@{
    /** The canonical name (CNAME).
     */
    string __cname;
    
    /** The multicast TTL.
     */
    unsigned int __ttl;
    
    /** The buffer size.
     */
    size_t __buffer_size;
    
    /** The maximum frame size.
     */
    size_t __max_frame_size;
    
    /** The receive buffer size.
     */
    size_t __recv_buffer_size;
    
    /** The session bandwidth (in kb/s).
     */
    unsigned int __bandwidth;
    //@}
    
    /**
     *  @name Session Components
     */
    //@{
    /** The liveMedia task scheduler.
     */
    TaskScheduler* __scheduler;
    
    /** The liveMedia usage environment.
     */
    UsageEnvironment* __env;
    
    /** The RTP socket.
     */
    Groupsock* __rtp_sock;
    
    /** The RTCP socket.
     */
    Groupsock* __rtcp_sock;
    //@}
  };
  
  
  inline Groupsock& RTPGeneric::getRTPSocket() const {
    return *__rtp_sock;
  }
  
  inline Groupsock& RTPGeneric::getRTCPSocket() const {
    return *__rtcp_sock;
  }
  
  inline TaskScheduler& RTPGeneric::getTaskScheduler() const {
    return *__scheduler;
  }
  
  inline UsageEnvironment& RTPGeneric::getUsageEnvironment() const {
    return *__env;
  }
  
}

#endif
