/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.timedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "RTPStrategy.hpp"

#include "nmm/base/TSFactory.hpp"
#include "nmm/base/ConnectorAcceptor.hpp"
#include "nmm/base/ConnectorAddress.hpp"
#include "nmm/base/ConnectionManager.hpp"
#include "nmm/base/memorymgr/Buffer.hpp"

#include "nmm/interfaces/base/IConnectionManager.hpp"
#include "nmm/interfaces/net/IObjectAcceptor.hpp"

#include <live/liveMedia.hh>

#include "BufferedSink.hpp"
#include "BufferedSource.hpp"
#include "LiveThread.hpp"

#include <unistd.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>

namespace NMM {
  
  RTPStrategy::RTPStrategy()
    : __local_address(NULL),
      __remote_address(NULL),
      __icm(NULL),
      __remote_ts_ref(0),
      __rtp_source(NULL),
      __buffered_sink(NULL),
      __buffered_source(NULL),
      __framer(NULL),
      __rtp_sink(NULL),
      __send_queue_size(10),
      __send_queue_mode(IQueued::MODE_SUSPEND),
      __rtcp_instance(NULL),
      __live_thread(NULL)
  {}

  RTPStrategy::~RTPStrategy() {
    /* Begin> Deactivate thread */
    if (__live_thread) {
      stop();
    }
    /* End< */
    
    /* Begin> Close media */
    if (__rtcp_instance) {
      GLOBAL_MESSAGE_STREAM("RTPSTrategy: Closing RTCP" << endl);
      Medium::close(__rtcp_instance);
      __rtcp_instance = NULL;
    }
    
    if (__buffered_sink) {
      GLOBAL_MESSAGE_STREAM("RTPSTrategy: Closing buffered sink" << endl);
      Medium::close(__buffered_sink);
      __buffered_sink = NULL;
    }
    
    if (__rtp_sink) {
      GLOBAL_MESSAGE_STREAM("RTPSTrategy: Closing RTP sink" << endl);
      Medium::close(__rtp_sink);
      __rtp_sink = NULL;
    }

    /* do not close the sources, libLive seems to do that. */
    //     if (__rtp_source) {
    //       GLOBAL_MESSAGE_STREAM("RTPSTrategy: Closing RTP source" << endl);
    //       Medium::close(__rtp_source);
    //       __rtp_source = NULL;
    //       if(__rtp_source == __framer) {
    //         __framer = NULL;
    //       }
    //     }
    
    //     if (__buffered_source) {
    //       GLOBAL_MESSAGE_STREAM("RTPSTrategy: Closing buffered source" << endl);
    //       Medium::close(__buffered_source);
    //       __buffered_source = NULL;
    //     }
    
    //     if (__framer) {
    //       GLOBAL_MESSAGE_STREAM("RTPSTrategy: Closing framer" << endl);
    //       Medium::close(__framer);
    //       __framer = NULL;
    //     }
    /* End< Close media */
  }
  
  void RTPStrategy::scheduleDestruction() {
    delete this;
  }
  
  string RTPStrategy::getStrategyType() const {
    return getStaticStrategyType();
  }
  
  void RTPStrategy::setAddress(const TSAddress& address)
    throw (const BindingException&)
  {
    /* create a local socket */
    try {
      setupReceiverSockets(dynamic_cast<const RTPAddress*>(&address));
    } catch (const Exception& e) {
      throw BindingException(e.getComment());
    }
    
    /* check if we have a fully specified RTP address */
    const RTPAddress* rtp_addr = dynamic_cast<const RTPAddress*>(&address);
    if (rtp_addr) {
      
      /* set remote address */
      delete __remote_address;
      __remote_address = rtp_addr->copy();
      
      /* set remote connection manager */
      if(rtp_addr->getConnectionManager()) {
        __icm = dynamic_cast<IConnectionManager*>(rtp_addr->getConnectionManager()->copy());
        __remote_ts_ref = rtp_addr->getTSReference();
      }
      
      return;
    }
    
    /* check if we have the address of a connection manager */
    const ConnectorAddress* connector_address = dynamic_cast<const ConnectorAddress*>(&address);
    if (connector_address) {
      /* make sure the address is valid */
      if (!connector_address->getConnectionManager()) {
        GLOBAL_ERROR_STREAM("RTPStrategy: Shared address has no IConnectionManager interface" << endl);
        throw BindingException("Shared address has no IConnectionManager interface");
      }
      
      /* get the local connection manager and request connection */
      ConnectorAddress* remote_addr = NULL;
      const ConnectionManager& local_icm = SConnectionManager::getInstance();
      local_icm.connect(__local_address, remote_addr, connector_address->getConnectionManager());
      
      /* the request was successful */
      /* check if the returned address is actually a RTP address */
      if (!remote_addr) {
        GLOBAL_ERROR_STREAM("RTPStrategy: Could not get remote RTP address" << endl);
        throw BindingException("Could not get remote RTP address");
      }
      const RTPAddress* rtp_addr = dynamic_cast<RTPAddress*>(remote_addr);
      if (!rtp_addr) {
        GLOBAL_ERROR_STREAM("RTPStrategy: Remote address is not a RTPAddress" << endl);
        throw BindingException("Remote address is not a RTPAddress");
      }
      __remote_address = rtp_addr->copy();
      
      /* check if the returned NMM reference is valid */
      if (!__remote_address->getTSReference()) {
        GLOBAL_WARNING_STREAM("RTPStrategy: Could not get NMM reference of remote transport strategy" << endl);
      }
      
      /* store the IConnectionManager interface for later use */
      __icm = dynamic_cast<IConnectionManager*>(connector_address->getConnectionManager()->copy());
      __remote_ts_ref = remote_addr->getTSReference();
      
      /* if the address points to an NMM object, connect to it */
      if (connector_address->getNMMReference()) {
        Event e(IObjectAcceptor::connectToObject_event, new TInValue<NMMReference>(connector_address->getNMMReference()));
        if (sendReliableEvent(e) == FAILURE) {
          throw BindingException("Could not connect to object");
        }
      }
      
      return;
    }
    
    throw BindingException(getStrategyType() + " can't handle address of type " + address.getTypeName());
  }
  
  void RTPStrategy::setValueStream(const IValueStream& , const OValueStream&) {
    /* Don't need value streams */
  }

  void RTPStrategy::setSendValueStream(const string&) {
    /* Don't need value streams */
  }
  
  void RTPStrategy::setReceiveValueStream(const string&) {
    /* Don't need value streams */
  }
  
  Result RTPStrategy::receiveMessage(Message* message, const IMessageSender* /*sender*/) {
    if (message->getType() == Message::BUFFER) {
      return internalSendBuffer(static_cast<Buffer*>(message));
    } else {      
      return FAILURE;
    }
  }
  
  Result RTPStrategy::sendMessage(Message* message) {
    return receiveMessage(message, 0);
  }
  
  Result RTPStrategy::sendEvent(const CEvent& /*ce*/) {
    /* Not supported */
    return FAILURE;
  }
  
  Result RTPStrategy::sendEvent(const Event& /*e*/) {
    /* Not supported */
    return FAILURE;
  }
  
  Result RTPStrategy::sendEvent(const CEvent& /*ce*/) const {
    /* Not supported */
    return FAILURE;
  }
  
  Result RTPStrategy::sendEvent(const Event& /*e*/) const {
    /* Not supported */
    return FAILURE;
  }
  
  void RTPStrategy::receiveSignal(Signal signal) {
    switch(signal) {
    case FREE_QUEUE_SLOT:
      {
        if (__buffered_sink) {
          /* Our IMessageReceiver sends a FREE_QUEUE_SLOT.
             Forward it to the BufferedSink so it can send new buffers to it. */
          __buffered_sink->receiveSignal(FREE_QUEUE_SLOT);
        }
        
        if (__buffered_source) {
          /* Our BufferedSource sends a FREE_QUEUE_SLOT.
             Forward it to out ISignalReceiver so it can send new buffers to us. */
          if(!getSignalReceiver()) {
            GLOBAL_WARNING_STREAM1("No signal receiver" << std::endl);
          } else {
	    getSignalReceiver()->receiveSignal(FREE_QUEUE_SLOT);
	  }
        }
      }
      break;
      
    default:
      break;
    }
  }
  
  Result RTPStrategy::receiveReliableEvent(Event& event, const IEventSender* sender ) const {
    /* If this TransportStrategy is the sender don't send the event again */
    if (sender == this) {
      return SUCCESS;
    }
    
    if ( find(getEventSenderList().begin(), getEventSenderList().end(), sender) != getEventSenderList().end()) {
      return NMMObject::receiveEvent(event, sender);
    }
    
    /* forward the Event to sendReliableEvent */
    return sendReliableEvent(event);
  }
  
  Result RTPStrategy::sendReliableEvent(Event& event) const {
    /* If we have no IConnectionManager, return FAILURE */
    if (!__icm) {
      GLOBAL_ERROR_STREAM("RTP sendReliableEvent failed (No IConnectionManager)" << endl);
      return FAILURE;
    }
    
    /* If we have no partner reference, return FAILURE */
    if (!__remote_ts_ref) {
      GLOBAL_ERROR_STREAM("RTP sendReliableEvent failed (No partner reference)" << endl);
      return FAILURE;
    }
    
    try {
      return __icm->sendReliableEvent(__remote_ts_ref, event);
    } catch(BindingException& e) {
      GLOBAL_ERROR_STREAM("RTP sendReliableEvent failed (" << e << ")" << endl);
      return FAILURE;
    }
  }
  
  Result RTPStrategy::sendReliableConfigEvent(Event& event) const {
    /* If we have no IConnectionManager, return FAILURE */
    if (!__icm) {
      GLOBAL_ERROR_STREAM("RTP sendReliableConfigEvent failed (No IConnectionManager)" << endl);
      return FAILURE;
    }
    
    /* If we have no partner reference, return FAILURE */
    if (!__remote_ts_ref) {
      GLOBAL_ERROR_STREAM("RTP sendReliableConfigEvent failed (No partner reference)" << endl);
      return FAILURE;
    }
    
    try {
      return __icm->sendReliableConfigEvent(__remote_ts_ref, event);
    } catch(BindingException& e) {
      GLOBAL_ERROR_STREAM("RTP sendReliableConfigEvent failed (" << e << ")" << endl);
      return FAILURE;
    }
  }
  
  void RTPStrategy::setupReceiverSockets(const RTPAddress* address)
    throw (const BindingException&)
  {
    try {
      /* delete old local address */
      delete __local_address;
      __local_address = NULL;
      
      /* delete old sockets */
      setRTPSocket(NULL);
      setRTCPSocket(NULL);
      
      /* IP address to use for creating the socket (default: INADDR_NONE) */
      const UnicastIPAddress null_ip("0.0.0.0");
      const MulticastIPAddress* mcast_ip = NULL;
      const IPAddress* ip = &null_ip;
      
      /* ports for creating the sockets (default: random) */
      unsigned rtp_port = 0;
      unsigned rtcp_port = 0;
      
      /* check what kind of address we have */
      if(address) {
        mcast_ip = dynamic_cast<const MulticastIPAddress*>(&address->getIPAddress());
        if(mcast_ip) {
          /* use multicast */
          ip = mcast_ip;      
          
          /* use ports from address */
          rtp_port = address->getRTPPort();
          rtcp_port = address->getRTCPPort();
        }
      }
      
      /* variables for getting a random port */
      int socket;
      struct sockaddr_in addr;
      socklen_t len_addr = sizeof(addr);
      addr.sin_addr = null_ip.getINetAddr();
      addr.sin_port = 0;
      
      /* Begin> Create RTP socket */
      if(rtp_port == 0) {
        /* create a socket with a random port */
        setRTPSocket(new Groupsock(getUsageEnvironment(), ip->getINetAddr(), 0, getTTL()));
        
        /* force the socket to be bound */
        socket = getRTPSocket().socketNum();
        sendto(socket, &socket, 1, 0, reinterpret_cast<struct sockaddr*>(&addr), len_addr);
        
        /* find out the local port */
        if (socket >= 0 && getsockname(socket, reinterpret_cast<struct sockaddr*>(&addr), &len_addr) == 0) {
          rtp_port = ntohs(addr.sin_port);
        }
        if (!rtp_port) {
          throw BindingException("Could not find a free port for RTP");
        }
      } else {
        /* create a socket with a fixed port */
        setRTPSocket(new Groupsock(getUsageEnvironment(), ip->getINetAddr(), rtp_port, getTTL()));
      }
      /* End< */

      /* Begin> Create RTCP socket */
      if(rtcp_port == 0) {
        /* create a socket with a random port */
        setRTCPSocket(new Groupsock(getUsageEnvironment(), null_ip.getINetAddr(), 0, getTTL()));
        
        /* force the socket to be bound */
        socket = getRTCPSocket().socketNum();
        sendto(socket, &socket, 1, 0, reinterpret_cast<struct sockaddr*>(&addr), len_addr);
        
        /* find out the local port */
        if (socket >= 0 && getsockname(socket, reinterpret_cast<struct sockaddr*>(&addr), &len_addr) == 0) {
          rtcp_port = ntohs(addr.sin_port);
        }
        if (!rtcp_port) {
          throw BindingException("Could not find a free port for RTCP");
        }
      } else {
        /* create a socket with a fixed port */
        setRTCPSocket(new Groupsock(getUsageEnvironment(), ip->getINetAddr(), rtcp_port, getTTL()));
      }
      /* End< */
      
      /* set receive buffer size */
      increaseReceiveBufferTo(getUsageEnvironment(), getRTPSocket().socketNum(), getReceiveBufferSize());
      
      /* get local IP address */
      char buffer[256];
      ::gethostname(buffer, 256);
      string hostname(buffer);
      UnicastIPAddress local_ip(hostname);
      
      /* setup local address */
      if(!mcast_ip) {
        ip = &local_ip;
      }
      __local_address = new RTPAddress(*ip, rtp_port, rtcp_port, getReference());
      GLOBAL_MESSAGE_STREAM("RTPStrategy local address is : ip_addr=" << ip->toString() <<
                            " rtp_port=" << rtp_port <<
                            " rtcp_port=" << rtcp_port << endl);
      
      /* set local connection manager interface */
      __local_address->setConnectionManager(*SConnectionManager::getInstance().getInterface<IConnectionManager>());
      __local_address->setTSReference(getReference());
 
    } catch(...) {
      /* clean up created sockets */
      setRTPSocket(NULL);
      setRTCPSocket(NULL);
      
      /* delete local address */
      delete __local_address;
      __local_address = NULL;
      
      /* rethrow exception */
      throw;
    }
  }
  
  void RTPStrategy::setupSenderSockets() throw(const BindingException&)
  {
    /* check whether we have a remote address */
    if(!__remote_address) {
      throw BindingException("No remote address");
    }
    
    /* check whether remote address is a multicast address */
    if(dynamic_cast<const MulticastIPAddress*>(&__remote_address->getIPAddress())) {
      /* nothing more to do */
      return;
    }
    
    /* Begin> set the destination address */
    const IPAddress& dest_ip(__remote_address->getIPAddress());
    unsigned int rtp_port = __remote_address->getRTPPort();
    unsigned int rtcp_port = __remote_address->getRTCPPort();
    getRTPSocket().changeDestinationParameters(dest_ip.getINetAddr(), rtp_port, getTTL());
    getRTCPSocket().changeDestinationParameters(dest_ip.getINetAddr(), rtcp_port, getTTL());
    GLOBAL_MESSAGE_STREAM("RTPStrategy remote address is: ip_addr=" << dest_ip.toString() <<
                          " rtp_port=" << rtp_port <<
                          " rtcp_port=" << rtcp_port << endl);
    /* End< */
  }
  
  void RTPStrategy::configureForInputJack(const Format& format) {
    UsageEnvironment& env(getUsageEnvironment());
    Groupsock* rtp_sock(&getRTPSocket());
    
    string type = format.getFormatType();

    /* Begin> Create RTP source depending on format*/
    if (type == Format::video_mpeg_type) {
      
      /* Create an RTP source for MPEG video */
      __rtp_source = MPEG1or2VideoRTPSource::createNew(env, rtp_sock);

    } else if (type == Format::video_h263plus_type) {
      
      /* Create an RTP source for h263plus video */
      __rtp_source = H263plusVideoRTPSource::createNew(env, rtp_sock, 34);
 
    } else if (type == Format::video_jpeg_type) {
      
      /* Create an RTP source for motion-JPEG video */
      __rtp_source = JPEGVideoRTPSource::createNew(env, rtp_sock);
 
    } else if (type == Format::audio_mpeg_type) {
      
      /* Create an RTP source for MPEG audio */
      __rtp_source = MPEG1or2AudioRTPSource::createNew(env, rtp_sock);

      // fw: audio/ac3 doesn't seem to work at the moment
      //     } else if (type == Format::audio_ac3_type) {
      
      //       /* Create an RTP source for AC3 audio */
      //       __rtp_source = AC3AudioRTPSource::createNew(env, rtp_sock);
      
    } else if (type == Format::audio_raw_type) {
    
      int channels = format.getIntValue(Format::channels_param);
      int sample_rate = format.getIntValue(Format::sample_rate_param);
      int bitpersample = format.getIntValue(Format::bitpersample_param);
      string encoding = format.getStringValue(Format::encoding_param);
      
      if (bitpersample == 16 && (sample_rate == 8000 || sample_rate == 44100 || sample_rate == 48000) && encoding == "linear") {
        if (channels == 2) {
          
          /* Create a simple RTP source */
          __rtp_source = SimpleRTPSource::createNew(env, rtp_sock, PT_AUDIO_L16_STEREO, sample_rate, "audio/raw");
          
	} else if (channels == 1) {
          
          /* Create a simple RTP source */
          __rtp_source = SimpleRTPSource::createNew(env, rtp_sock, PT_AUDIO_L16_MONO, sample_rate, "audio/raw");
          
	}
      }else if (encoding == "mu_law" && bitpersample == 8) {
        
        int sample_rate = format.getIntValue(Format::sample_rate_param);
       
        /* Create a simple RTP source */
        __rtp_source = SimpleRTPSource::createNew(env, rtp_sock, PT_AUDIO_PCMU , sample_rate, "audio/pcmu");
        
        
      } else if (encoding == "a_law" && bitpersample == 8) {
        
        int sample_rate = format.getIntValue(Format::sample_rate_param);   
        
        /* Create a simple RTP source */
        __rtp_source = SimpleRTPSource::createNew(env, rtp_sock, PT_AUDIO_PCMA , sample_rate, "audio/pcma");
      
        
      } else {
        GLOBAL_ERROR_STREAM("audio/raw only implemented with 16 bit per sample, samplerate 44100 and encoding linear or 8bit, 8000Hz and mu_law encoding" << endl);
      } 
      
    } else {
      
      GLOBAL_ERROR_STREAM("Unsupported payload format" << endl);
      throw BindingException("Unsupported payload format");
      
    }
    /* End< */
    
    /* Create buffered sink */
    __buffered_sink = BufferedSink::createNew(getUsageEnvironment(), getMessageReceiver(), this,
                                              getBufferManager(), getBufferSize(), getMaxFrameSize());

    /* Create RTCP instance */
    __rtcp_instance = RTCPInstance::createNew(getUsageEnvironment(), &getRTCPSocket(), getSessionBandwidth(),
					      reinterpret_cast<const unsigned char*>(getCNAME().c_str()), NULL, __rtp_source);
    /* Activate thread */
    start();
  }
  
  void RTPStrategy::configureForOutputJack(const Format& format) {
    setupSenderSockets();
    
    UsageEnvironment& env(getUsageEnvironment());
    Groupsock* rtp_sock(&getRTPSocket());
    
    string type = format.getFormatType();
    
    /* Begin> Create RTP sink and framer depending on format*/
    if (type == Format::video_mpeg_type) {
      
      /* Create an RTP sink for MPEG video */
      __rtp_sink = MPEG1or2VideoRTPSink::createNew(env, rtp_sock);
      
      /* Create a buffered source */
      __buffered_source = BufferedSource::createNew(env, this, 1.0f, __send_queue_mode, __send_queue_size, true);
      
      /* Create a framer for MPEG video */
      __framer = MPEG1or2VideoStreamFramer::createNew(env, __buffered_source);
      
    } else if (type == Format::video_h263plus_type) {
      
      /* Create an RTP sink for h263plus video */
      __rtp_sink = H263plusVideoRTPSink::createNew(env, rtp_sock, 34);
      
      /* Create a buffered source */
      __buffered_source = BufferedSource::createNew(env, this, 1.0f, __send_queue_mode, __send_queue_size, true);
      
      /* The source is the framer */
      __framer = __rtp_source;
      
    } else if (type == Format::video_jpeg_type) {
      
      /* Create an RTP sink for motion-JPEG video */
      __rtp_sink = JPEGVideoRTPSink::createNew(env, rtp_sock);
      
      /* Create a buffered source */
      __buffered_source = BufferedSource::createNew(env, this, 1.0f, __send_queue_mode, __send_queue_size, true);
      
      /* The source is the framer */
      __framer = __rtp_source;
    
    } else if (type == Format::video_mpeg4_type) {

      /* Create an RTP sink for MPEG4 video */
      __rtp_sink = MPEG4ESVideoRTPSink::createNew(env, rtp_sock, 96);
      
      /* Create a buffered source */
      __buffered_source = BufferedSource::createNew(env, this, 1.0f, __send_queue_mode, __send_queue_size, true);
      
      /* Create a framer for MPEG video */
      __framer = MPEG4VideoStreamFramer::createNew(env, __buffered_source);
  
    } else if (type == Format::audio_mpeg_type) {
      
      /* Create an RTP sink for MPEG audio */
      __rtp_sink = MPEG1or2AudioRTPSink::createNew(env, rtp_sock);
      
      /* Create a buffered source */
      __buffered_source = BufferedSource::createNew(env, this, 1.0f, __send_queue_mode, __send_queue_size, true);
      
      /* Create a framer for MPEG audio */
      __framer = MPEG1or2AudioStreamFramer::createNew(env, __buffered_source);

      // fw: audio/ac3 doesn't seem to work at the moment
      //     } else if (type == Format::audio_ac3_type) {
      
      //       /* Create an RTP sink for AC3 audio */
      //       __rtp_sink = AC3AudioRTPSink::createNew(env, rtp_sock);
      
      //       /* Create a buffered source */
      //       __buffered_source = BufferedSource::createNew(env, __send_queue_mode, __send_queue_size);
      
      //       /* Create a framer for AC3 audio */
      //       __framer = AC3AudioStreamFramer::createNew(env, __buffered_source);
      
    } else if (type == Format::audio_raw_type) {
      
      int channels = format.getIntValue(Format::channels_param);
      int sample_rate = format.getIntValue(Format::sample_rate_param);
      int bitpersample = format.getIntValue(Format::bitpersample_param);
      string encoding = format.getStringValue(Format::encoding_param);

      if (bitpersample == 16 && (sample_rate == 8000 || sample_rate == 44100 || sample_rate == 48000 ) && encoding == "linear") {
      if (channels == 2) {
        
          /* Create a simple RTP sink */
          __rtp_sink = SimpleRTPSink::createNew(env, rtp_sock, PT_AUDIO_L16_STEREO, sample_rate, "audio/raw", "L16");
	  
          // comment in to adapt PacketSizes in liblive
          // dynamic_cast<SimpleRTPSink*>(__rtp_sink)->setPacketSizes(250, 724);

          /* Create a buffered source */
          __buffered_source = BufferedSource::createNew(env, this, 1.0 * getBufferSize() / (2 * 2 * sample_rate),
                                                        __send_queue_mode, __send_queue_size, true);
	  
          /* The source is the framer */
          __framer = __rtp_source;
          
        } else if (channels == 1) {

          /* Create a simple RTP sink */
          __rtp_sink = SimpleRTPSink::createNew(env, rtp_sock, PT_AUDIO_L16_MONO, sample_rate, "audio/raw", "L16");
          // comment in to adapt PacketSizes in liblive
          // dynamic_cast<SimpleRTPSink*>(__rtp_sink)->setPacketSizes(250, 724);

          /* Create a buffered source */
          __buffered_source = BufferedSource::createNew(env, this, 1.0 * getBufferSize() / (2 * 1 * sample_rate),
                                                        __send_queue_mode, __send_queue_size, true);
          
          /* The source is the framer */
          __framer = __rtp_source;
          
        }
      } else if (encoding == "mu_law" && bitpersample == 8) {//pcmu
      
      // int channels = format.getIntValue(Format::channels_param);
      // int bitpersample = format.getIntValue(Format::bitpersample_param);
      int sample_rate = format.getIntValue(Format::sample_rate_param);
      
      
          
      /* Create a simple RTP sink */
      __rtp_sink = SimpleRTPSink::createNew(env, rtp_sock, PT_AUDIO_PCMU, sample_rate, "audio/pcmu", "PCMU");
      
      // comment in to adapt PacketSizes in liblive
      dynamic_cast<SimpleRTPSink*>(__rtp_sink)->setPacketSizes(128, 724);
	  
      /* Create a buffered source */
      __buffered_source = BufferedSource::createNew(env, this, 1.0 * getBufferSize()  / (channels * 1 * sample_rate),
                                                    __send_queue_mode, __send_queue_size, true);
	  
      /* The source is the framer */
      __framer = __rtp_source;
          
        
      }else if (encoding == "a_law" && bitpersample == 8) { //pcma
        
        int channels = format.getIntValue(Format::channels_param);
        // int bitpersample = format.getIntValue(Format::bitpersample_param);
        int sample_rate = format.getIntValue(Format::sample_rate_param);
        
        
        
        /* Create a simple RTP sink */
        __rtp_sink = SimpleRTPSink::createNew(env, rtp_sock, PT_AUDIO_PCMA, sample_rate, "audio/pcma", "PCMA");
        
        // comment in to adapt PacketSizes in liblive
        dynamic_cast<SimpleRTPSink*>(__rtp_sink)->setPacketSizes(128, 724);
        
        /* Create a buffered source */
        __buffered_source = BufferedSource::createNew(env, this, 1.0 * getBufferSize() / (channels * 1 * sample_rate),
                                                      __send_queue_mode, __send_queue_size);
        
        /* The source is the framer */
        __framer = __rtp_source;
        
      } else {
        GLOBAL_ERROR_STREAM("Unsupported parameter for audio/raw format" << endl);
        throw BindingException("Unsupported parameter for audio/raw format");
      }
    }else {
      GLOBAL_ERROR_STREAM("Unsupported payload format" << endl);
      throw BindingException("Unsupported payload format");
    }
      
  
    /* End< */
    
    
    /* Create RTCP instance */
    __rtcp_instance = RTCPInstance::createNew(getUsageEnvironment(), &getRTCPSocket(), getSessionBandwidth(),
                                              reinterpret_cast<const unsigned char*>(getCNAME().c_str()), __rtp_sink, NULL);
    /* Activate thread */
    start();
  }
  
  void RTPStrategy::start() {
    /* Begin> Determine source */
    MediaSource* source;
    if (__buffered_source) {
      source = __buffered_source;
    }
    if (__framer) {
      source = __framer;
    }
    if (__rtp_source) {
      source = __rtp_source;
    }
    /* End< */
    
    /* Begin> Determine sink */
    MediaSink* sink;
    if (__buffered_sink) {
      sink = __buffered_sink;
    }
    if (__rtp_sink) {
      sink = __rtp_sink;
    }
    /* End< */
    
    /* Begin> Create liveMedia thread */
    __live_thread = new LiveThread(getUsageEnvironment(), *source, *sink);
    __live_thread->start();
    GLOBAL_MESSAGE_STREAM("---- RTP THREAD STARTED ----" << endl);
    /* End< */
  }
  
  void RTPStrategy::stop() {
    if (!__live_thread) {
      return;
    }
    
    /* Cancel blocking calls and flush all queues */
    if (__buffered_sink) {
      __buffered_sink->flush();
    }
    if (__buffered_source) {
      __buffered_source->flush();
    }
    
    /* stop LIVE thread */
    __live_thread->stop();
    delete __live_thread;
    __live_thread = NULL;
    
    GLOBAL_MESSAGE_STREAM("---- RTP THREAD STOPPED ----" << endl);
  }
  
  Result RTPStrategy::internalSendBuffer(Buffer* buffer) {
    /* If we don't have a buffered source, we can't send buffers */
    if (!__buffered_source) {
      GLOBAL_ERROR_STREAM("RTP: internalSendBuffer failed because there is no BufferedSource" << endl);
      return FAILURE;
    }
    
    /* Write buffer to source. */
    return __buffered_source->writeBuffer(buffer);
  }
  
  unsigned int RTPStrategy::getSize(const string& /*= "default"*/) const {
    if(__buffered_source) {
      /* return the current size of the BufferedSource's queue */
      return __buffered_source->getStreamQueue().getSize();
    } else {
      /* there is no queue */
      return 0;
    }
  }
  
  unsigned int RTPStrategy::getMaxSize(const string& /*= "default"*/) const {
    if(__buffered_source) {
      /* return the maximum size of the BufferedSource's queue */
      return __buffered_source->getStreamQueue().getMaxSize();
    } else {
      /* return the configured queue size */
      return __send_queue_size;
    }
  }
  
  Result RTPStrategy::setMaxSize(const unsigned int& size, const string& /*= "default"*/) {
    if(__buffered_source) {
      /* set the maximum size of the BufferedSource's queue */
      __buffered_source->getStreamQueue().setMaxSize(size);
    } else {
      /* configure the queue size */
      __send_queue_size = size;
    }
    
    return SUCCESS;
  }
  
  IQueued::Mode RTPStrategy::getMode(const string& /*= "default"*/) const {
    if(__buffered_source) {
      /* set the mode of the BufferedSource's queue */
      return (IQueued::Mode) __buffered_source->getStreamQueue().getMode();
    } else {
      /* return the configured queue mode */
      return __send_queue_mode;
    }
  }
  
  Result RTPStrategy::setMode(const IQueued::Mode& mode, const string& /*= "default"*/) {
    if(__buffered_source) {
      /* set the mode of the BufferedSource's queue */
      return __buffered_source->getStreamQueue().setMode((StreamQueue::Mode) mode);
    } else {
      /* configure the queue mode */
      __send_queue_mode = mode;
      return SUCCESS;
    }
  }

  string RTPStrategy::getRTPSinkMediaType() const{
    if (__rtp_sink && __rtp_sink->sdpMediaType()) {
      return (string(__rtp_sink->sdpMediaType()));
    }
    return string("");
  }
  
  unsigned char RTPStrategy::getPayloadType() const {
    if (__rtp_source) {
      return (__rtp_source->rtpPayloadFormat());
    }
    if (__rtp_sink) {
      return (__rtp_sink->rtpPayloadType());
    }
    return 0;
  }
  
  string RTPStrategy::getRTPSinkMapLine() const{
    if (__rtp_sink && __rtp_sink->rtpmapLine()) {
      return string(__rtp_sink->rtpmapLine());
    }
    return string("");
  }
  
  string RTPStrategy::getRTPSinkAuxSDPLine() const{
    if (__rtp_sink && __rtp_sink->auxSDPLine()) {
      return string(__rtp_sink->auxSDPLine());
    }
    return string("");
  }
  
  unsigned short RTPStrategy::getRTPSeqNum() const {
    if (__rtp_sink && __rtp_sink->currentSeqNo()) {
      return __rtp_sink->currentSeqNo();
    }
    return 0;
  }

  unsigned RTPStrategy::getRTPTimestamp() const {
    if (__rtp_sink && __rtp_sink->currentTimestamp()) {
      return __rtp_sink->currentTimestamp();
    }
    return 0;
  }

  unsigned RTPStrategy::getRTPPort() const {
    if (__local_address) {
      return __local_address->getRTPPort();
    }
    return 0;
  }

  unsigned RTPStrategy::getRTCPPort() const {
    if (__local_address) {
      return __local_address->getRTCPPort();
    }
    return 0;
  }

  Format* RTPStrategy::getFormatFromPayloadType(int payload_type) {
    Format* format = 0;
    
    switch (payload_type) {
      
    case 0:
      format = new Format(Format::audio_raw_type);
      format->addIntParamValue(Format::channels_param, 1);
      format->addIntParamValue(Format::bitpersample_param, 8);
      format->addIntParamValue(Format::sample_rate_param, 8000);
      format->addStringParamValue(Format::encoding_param, "mu_law");
      format->addStringParamValue(Format::sign_param, "signed");
      break;
      
      /* Not supported yet
       * Perhaps you have to add some more params to the format
       case 3:
       format = new Format(Format::audio_gsm_type);
       format->addIntParamValue(Format::channels_param, 1);
       format->addIntParamValue(Format::sample_rate_param, 8000);
       break;
      */

    case 8:
      format = new Format(Format::audio_raw_type);
      format->addIntParamValue(Format::channels_param, 1);
      format->addIntParamValue(Format::bitpersample_param, 8);
      format->addIntParamValue(Format::sample_rate_param, 8000);
      format->addStringParamValue(Format::encoding_param, "a_law");
      format->addStringParamValue(Format::sign_param, "signed");
      break;    
    
    case 10:
      format = new Format(Format::audio_raw_type);
      format->addIntParamValue(Format::channels_param, 2);
      format->addIntParamValue(Format::bitpersample_param, 16);
      format->addIntParamValue(Format::sample_rate_param, 44100);
      format->addStringParamValue(Format::encoding_param, "linear");
      break;
      
    case 11:
      format = new Format(Format::audio_raw_type);
      format->addIntParamValue(Format::channels_param, 1);
      format->addIntParamValue(Format::bitpersample_param, 16);
      format->addIntParamValue(Format::sample_rate_param, 44100);
      format->addStringParamValue(Format::encoding_param, "linear");
      break;
      
    case 14:
      format = new Format(Format::audio_mpeg_type);
      format->addIntParamValue(Format::layer_param, 2);
      format->addIntParamValue(Format::layer_param, 3);
      format->addIntParamValue(Format::audio_codec_param, 0x55);
      //format->addWildcard(Format::sample_rate_param);
      //format->addWildcard(Format::channels_param);
      //format->addWildcard(Format::bitpersample_param);
      //format->addWildcard(Format::bitrate_param);
      break;
      
    case 26:
      format = new Format(Format::video_jpeg_type);
      format->addWildcard(Format::x_resolution_param);
      format->addWildcard(Format::y_resolution_param);
      format->addWildcard(Format::ratio_x_param);
      format->addWildcard(Format::ratio_y_param);
      format->addWildcard(Format::framerate_param);
      break;
      
    case 32:
      format = new Format(Format::video_mpeg_type);
      format->addIntParamValue(Format::version_param, 1);
      format->addIntParamValue(Format::version_param, 2);
      //format->addWildcard(Format::x_resolution_param);
      //format->addWildcard(Format::y_resolution_param);
      //format->addWildcard(Format::ratio_x_param);
      //format->addWildcard(Format::ratio_y_param);
      //format->addWildcard(Format::framerate_param);
      break;
      
    case 34:
      format = new Format(Format::video_h263plus_type);
      format->addWildcard(Format::x_resolution_param);
      format->addWildcard(Format::y_resolution_param);
      format->addWildcard(Format::ratio_x_param);
      format->addWildcard(Format::ratio_y_param);
      format->addWildcard(Format::framerate_param);
      format->addWildcard(Format::bitrate_param);
      break;

    case 96:
      format = new Format(Format::video_mpeg4_type);
      format->addWildcard(Format::x_resolution_param);
      format->addWildcard(Format::y_resolution_param);
      format->addWildcard(Format::ratio_x_param);
      format->addWildcard(Format::ratio_y_param);
      format->addWildcard(Format::framerate_param);
      format->addWildcard(Format::bitrate_param);
      format->addStringParamValue(Format::compressor_param, "DIVX");
      format->addStringParamValue(Format::compressor_param, "DX50");

      /*format = new Format(Format::video_mpeg4_type);
      format->addWildcard(Format::x_resolution_param);
      format->addWildcard(Format::y_resolution_param);
      format->addWildcard(Format::ratio_x_param);
      format->addWildcard(Format::ratio_y_param);
      format->addWildcard(Format::framerate_param);
      format->addWildcard(Format::bitrate_param);
      format->addStringParamValue(Format::compressor_param, "DIVX");*/
      break;

    default:
      format = new Format(Format::wild_wild_type);
      
    }
    
    return format;
  }

  const char RTPStrategy::__strategy_type[] = "RTPStrategy";
  
  class RTPStrategyFormatList : public TSFormatList {
  public:
    RTPStrategyFormatList() {
      /* Begin> MPEG streams (RFC 2250) */
      /* Begin> MPEG video formats used in NMM */
      Format video_mpeg(Format::video_mpeg_type);
      video_mpeg.addIntParamValue(Format::version_param, 1);
      video_mpeg.addIntParamValue(Format::version_param, 2);
      addFormat(video_mpeg);
      
      Format video_mpeg_detail(Format::video_mpeg_type);
      video_mpeg_detail.addIntParamValue(Format::version_param, 1);
      video_mpeg_detail.addIntParamValue(Format::version_param, 2);
      video_mpeg_detail.addWildcard(Format::x_resolution_param);
      video_mpeg_detail.addWildcard(Format::y_resolution_param);
      video_mpeg_detail.addWildcard(Format::ratio_x_param);
      video_mpeg_detail.addWildcard(Format::ratio_y_param);
      video_mpeg_detail.addWildcard(Format::framerate_param);
      addFormat(video_mpeg_detail);      
      /* End< MPEG video formats used in NMM */
      
      /* Begin> motion-JPEG streams (RFC 2435) */
      /* Begin> motion-JPEG video formats used in NMM */
      Format video_jpeg(Format::video_jpeg_type);
      video_jpeg.addWildcard(Format::x_resolution_param);
      video_jpeg.addWildcard(Format::y_resolution_param);
      video_jpeg.addWildcard(Format::ratio_x_param);
      video_jpeg.addWildcard(Format::ratio_y_param);
      video_jpeg.addWildcard(Format::framerate_param);
      addFormat(video_jpeg);
      /* End< motion-JPEG video formats used in NMM */

      /* Begin> h623plus streams (RFC 2429) */
      /* Begin> h623plus video formats used in NMM */
      Format video_h623plus(Format::video_h263plus_type);
      video_h623plus.addWildcard(Format::x_resolution_param);
      video_h623plus.addWildcard(Format::y_resolution_param);
      video_h623plus.addWildcard(Format::ratio_x_param);
      video_h623plus.addWildcard(Format::ratio_y_param);
      video_h623plus.addWildcard(Format::framerate_param);
      video_h623plus.addWildcard(Format::bitrate_param);
      addFormat(video_h623plus);
      /* End< motion-JPEG video formats used in NMM */

      /* Begin> mpeg4 streams */
      /* Begin> mpeg4 video format used in NMM */
      Format video_mpeg4(Format::video_mpeg4_type);
      video_mpeg4.addWildcard(Format::x_resolution_param);
      video_mpeg4.addWildcard(Format::y_resolution_param);
      video_mpeg4.addWildcard(Format::ratio_x_param);
      video_mpeg4.addWildcard(Format::ratio_y_param);
      video_mpeg4.addWildcard(Format::framerate_param);
      video_mpeg4.addWildcard(Format::bitrate_param);
      video_mpeg4.addStringParamValue(Format::compressor_param, "DIVX");
      video_mpeg4.addStringParamValue(Format::compressor_param, "DX50");
      addFormat(video_mpeg4);
      /* End< mpeg4 video format used in NMM */
      

      /* Begin> MPEG audio formats used in NMM */
      Format audio_mpeg_mime(Format::audio_mpeg_type);
      addFormat(audio_mpeg_mime);
      
      Format audio_mpeg(Format::audio_mpeg_type);
      audio_mpeg.addIntParamValue(Format::layer_param, 2);
      audio_mpeg.addIntParamValue(Format::layer_param, 3);
      audio_mpeg.addIntParamValue(Format::audio_codec_param, 0x55);
      addFormat(audio_mpeg);
      
      Format audio_mpeg_bitrate(Format::audio_mpeg_type);
      audio_mpeg_bitrate.addIntParamValue(Format::layer_param, 2);
      audio_mpeg_bitrate.addIntParamValue(Format::layer_param, 3);
      audio_mpeg_bitrate.addIntParamValue(Format::audio_codec_param, 0x55);
      audio_mpeg_bitrate.addWildcard(Format::bitrate_param);
      addFormat(audio_mpeg_bitrate);
      
      Format audio_mpeg_bitrate_detail(Format::audio_mpeg_type);
      audio_mpeg_bitrate_detail.addIntParamValue(Format::layer_param, 2);
      audio_mpeg_bitrate_detail.addIntParamValue(Format::layer_param, 3);
      audio_mpeg_bitrate_detail.addIntParamValue(Format::audio_codec_param, 0x55);
      audio_mpeg_bitrate_detail.addWildcard(Format::sample_rate_param);
      audio_mpeg_bitrate_detail.addWildcard(Format::channels_param);
      audio_mpeg_bitrate_detail.addWildcard(Format::bitpersample_param);
      audio_mpeg_bitrate_detail.addWildcard(Format::bitrate_param);
      addFormat(audio_mpeg_bitrate_detail);

      Format audio_mpeg_detai(Format::audio_mpeg_type);
      audio_mpeg_detai.addIntParamValue(Format::layer_param, 2);
      audio_mpeg_detai.addIntParamValue(Format::layer_param, 3);
      audio_mpeg_detai.addIntParamValue(Format::audio_codec_param, 0x55);
      audio_mpeg_detai.addWildcard(Format::sample_rate_param);
      audio_mpeg_detai.addWildcard(Format::channels_param);
      audio_mpeg_detai.addWildcard(Format::bitpersample_param);
      addFormat(audio_mpeg_detai);
      /* End< MPEG audio formats used in NMM */
      /* End< MPEG streams (RFC 2250) */
      
      /* Begin> raw formats (RFC 1890) */
      Format audio_raw_l16(Format::audio_raw_type);
      audio_raw_l16.addStringParamValue(Format::endian_param, "little");
      audio_raw_l16.addStringParamValue(Format::encoding_param, "linear");
      audio_raw_l16.addStringParamValue(Format::sign_param, "signed");
      audio_raw_l16.addIntParamValue(Format::sample_rate_param, 8000);
      audio_raw_l16.addIntParamValue(Format::sample_rate_param, 44100);
      audio_raw_l16.addIntParamValue(Format::sample_rate_param, 48000);
      audio_raw_l16.addIntParamValue(Format::channels_param, 1);
      audio_raw_l16.addIntParamValue(Format::channels_param, 2);
      audio_raw_l16.addIntParamValue(Format::bitpersample_param, 16);
      addFormat(audio_raw_l16);
      /* End< raw formats (RFC 1890) */

      /* Begin> pcmu formats (RFC 1890) */
      Format audio_pcmu(Format::audio_raw_type);
      audio_pcmu.addStringParamValue(Format::endian_param, "little");
      audio_pcmu.addStringParamValue(Format::encoding_param, "mu_law");
      audio_pcmu.addStringParamValue(Format::sign_param, "signed");
      audio_pcmu.addIntParamValue(Format::sample_rate_param, 8000);
      audio_pcmu.addIntParamValue(Format::sample_rate_param, 44100);
      audio_pcmu.addIntParamValue(Format::sample_rate_param, 48000);
      audio_pcmu.addIntParamValue(Format::channels_param, 1);
      audio_pcmu.addIntParamValue(Format::channels_param, 2);
      audio_pcmu.addIntParamValue(Format::bitpersample_param,8);
      addFormat(audio_pcmu);
      /* End< pcmu formats (RFC 1890) */

      /* Begin> pcma formats (RFC 1890) */
      Format audio_pcma(Format::audio_raw_type);
      audio_pcma.addStringParamValue(Format::endian_param, "little");
      audio_pcma.addStringParamValue(Format::encoding_param, "a_law");
      audio_pcma.addStringParamValue(Format::sign_param, "signed");
      audio_pcma.addIntParamValue(Format::sample_rate_param, 8000);
      audio_pcma.addIntParamValue(Format::sample_rate_param, 44100);
      audio_pcma.addIntParamValue(Format::sample_rate_param, 48000);
      audio_pcma.addIntParamValue(Format::channels_param, 1);
      audio_pcma.addIntParamValue(Format::channels_param, 2);
      audio_pcma.addIntParamValue(Format::bitpersample_param, 8);
      addFormat(audio_pcma);
      /* End< pcma formats (RFC 1890) */
    }
  };
  
  TTSFactoryRegistrator< TConnectorTSFactoryObject<RTPStrategy, ConnectorAcceptor> >
  __rtp_strategy_prototype("RTPStrategy", false, true, RTPStrategyFormatList());

  
} // namespace NMM
