/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef NMM_RTP_STRATEGY_HPP
#define NMM_RTP_STRATEGY_HPP

#include "RTPGeneric.hpp"

#include "nmm/base/TransportStrategy.hpp"
#include "nmm/base/ObjectAcceptor.hpp"

#include "nmm/interfaces/base/IQueuedImpl.hpp"

/* liveMedia forwards */
class RTPSource;
class RTPSink;
class RTCPInstance;
class FramedSource;

namespace NMM {
  
  /* NMM forwards */
  class IConnectionManager;
  class Buffer;

  /* NMM-RTP forwards */  
  class RTPAddress;  
  class BufferedSink;
  class BufferedSource;
  class LiveThread;  
  
  /**
   *  @brief RTP Transport strategy
   *
   *  A transport strategy which uses the Realtime Transport Protocol (RTP).
   */
  class RTPStrategy :
    public TransportStrategy,
    public RTPGeneric,
    public IQueuedImpl,
    public TSAddressOwner
  {
  public:
    /**
     *  @name Constructor & Destructor
     */
    //@{
    /**
     *  @brief Default constructor
     */
    RTPStrategy();

    /**
     *  @brief Destructor
     */
    ~RTPStrategy();

    /**
     *  @brief Schedule destruction
     *
     *  This method deletes this object
     */
    virtual void scheduleDestruction();
    //@}
    
    /**
     *  @name TransportStratgegy Information
     */
    //@{
    /**
     *  @brief Get transport strategy type
     *  @return The type of this transport strategy
     *
     *  The type of this transport strategy is "RTPStrategy".
     */
    virtual string getStrategyType() const;
    //@}
    
    /**
     *   @name Connect & Configure
     */
    //@{
    /**
     *  @brief Set address
     *  @param address Address to connect to
     *
     *  This method sets the address of the partner transport strategy. The address must either be a
     *  %ConnectorAddress (for indirect connection setup) or an RTPAddress (if the connection parameters
     *  are known).
     */
    virtual void setAddress(const TSAddress& address)
      throw (const BindingException&);
    
    /**
     *  @brief Set ValueStream
     *
     *  This method does nothing because this strategy doesn't use value streams.
     */
    virtual void setValueStream(const IValueStream&, const OValueStream&);
    //@}
    
    /**
     *   @name Sender and Receiver Interfaces
     */
    //@{
    /**
     *  @brief Receive an instream message
     *  @return Stream queue status
     *
     *  If the message is a buffer, then this method sends the buffer over RTP.
     *  Other message types are not supported.
     */
    virtual Result receiveMessage(Message*, const IMessageSender*);
    
    /**
     *  @brief This method sends an incoming Event reliably
     *  @return The event result, or @c FAILURE if no IConnectionManager interface is available
     *
     *  This method uses the IConnectionManager interface and the NMMReference of the
     *  partner strategy.
     */
    virtual Result receiveReliableEvent(Event&, const IEventSender* = 0)  const;
    
    /**
     *  @brief Send an instream message
     *  @return Stream queue status
     *
     *  If the message is a buffer, then this method sends the buffer over RTP.
     *  Other message types are not supported.
     */
    virtual Result sendMessage(Message*);
    
    /**
     *  @brief Send a CEvent message
     *  @return @c FAILURE
     *
     *  This transport strategy does not support out-of-band events.
     */
    virtual Result sendEvent(const CEvent&);
    
    /**
     *  @brief Send an event
     *  @return @c FAILURE
     *
     *  This transport strategy does not support out-of-band events.
     */
    virtual Result sendEvent(const Event&);
    
    /**
     *  @brief Send a CEvent message
     *  @return @c FAILURE
     *
     *  This transport strategy does not support out-of-band events.
     */
    virtual Result sendEvent(const CEvent&) const;
    
    /**
     *  @brief Send an event
     *  @return @c FAILURE
     *
     *  This transport strategy does not support out-of-band events.
     */
    virtual Result sendEvent(const Event&) const;
    
    /**
     *  @brief This method sends an Event reliably
     *  @return @c FAILURE
     *
     *  This transport strategy does not support out-of-band events.
     */
    virtual Result sendReliableEvent(Event&) const;
    
    /**
     *  @brief This method sends a configuration Event reliably
     *  @return @c FAILURE
     *
     *  This transport strategy does not support out-of-band events.
     */
    virtual Result sendReliableConfigEvent(Event&) const;
    
    /**
     *  @brief Receive a signal
     *
     *  This method handles signals from the parent of this transport strategy.
     *
     *  If a @c FREE_QUEUE_SLOT or @c FREE_UPSTR_QUEUE_SLOT event is received, then this
     *  transport strategy will continue to forward messages to its parent.
     */
    virtual void receiveSignal(Signal);
    //@}
    
    /**
     *  @name ITransportStrategyConfig interface
     */
    //@{
    /**
     *  @brief Configure for input jack
     *  @param format Stream format
     *
     *  This method configures the transport strategy for an input jack with the given format.
     *  The method fails, if the format is not supported by RTP (i.e. no payload type is available or implemented).
     */
    virtual void configureForInputJack(const Format& format);
    
    /**
     *  @brief Configure for output jack
     *  @param format Stream format
     *
     *  This method configures the transport strategy for an output jack with the given format.
     *  The method fails, if the format is not supported by RTP (i.e. no payload type is available or implemented).
     */
    virtual void configureForOutputJack(const Format& format);
    //@}
    
    /**
     *  @name IQueued interface
     */
    //@{
    /**
     *  @brief Get current send queue size
     *  @return The current size of the send queue
     */
    unsigned int getSize(const string& = "default") const;
    
    /**
     *  @brief Get maximum send queue size
     *  @return The maximum size of the send queue
     */
    unsigned int getMaxSize(const string& = "default") const;
    
    /**
     *  @brief Set maximum send queue size
     *  @param size The maximum send queue size
     */
    Result setMaxSize(const unsigned int& size, const string& = "default");
    
    /**
     *  @brief Get send queue mode
     *  @return The send queue mode
     */
    IQueued::Mode getMode(const string& = "default") const;
    
    /**
     *  @brief Set send queue mode
     *  @param mode The send queue mode
     */
    Result setMode(const IQueued::Mode& mode, const string& = "default");
    //@}
    
    /**
     *  @name Addresses and Connection Parameters
     *
     *  This is the interface used by the RTPConnector to exchange the connection parameters.
     */
    //@{
    /**
     *  @brief Get local address
     *  @return The local RTP address
     *
     *  This method returns the local RTP address. It contains all the connection parameters
     *  needed by a partner strategy to connect to this strategy.
     */
    virtual const RTPAddress& getLocalAddress() const throw();
    
    /**
     *  @brief Get remote address
     *  @return The remote RTP address
     *
     *  This method returns the remote RTP address that was set by the last RTPStrategy::setAddress call.
     */
    const RTPAddress* getRemoteAddress() const;
    //@}
    
    /** @name Static type information.
     */
    //@{
    /**
     *  @brief Get transport strategy type name
     *  @return "RTPStrategy"
     */
    static string getStaticStrategyType();
    //@}
    
    string getRTPSinkMediaType() const;

    unsigned char getPayloadType() const;

    string getRTPSinkMapLine() const;

    string getRTPSinkAuxSDPLine() const;

    unsigned short getRTPSeqNum() const;
    
    unsigned getRTPTimestamp() const;

    unsigned getRTPPort() const;
  
    unsigned getRTCPPort() const;

    Format* getFormatFromPayloadType(int payload_type);
    
  protected:
    /**
     *  @name Low Level Sending and Receiving
     */
    //@{
    /**
     *  @brief Send a buffer via RTP
     *  @param buffer Buffer to send
     *
     *  This method passes the data contained in @c buffer to the RTP stack.
     */
    virtual Result internalSendBuffer(Buffer* buffer);
    //@}
    
    /**
     *  @name Scheduling
     */
    //@{
    /**
     *  @brief Start liveMedia thread
     *
     *  This method starts the event processing by the liveMedia RTP stack.
     */
    void start();
    
    /**
     *  @brief Stop liveMedia thread
     *
     *  This method stops the event processing by the liveMedia RTP stack.
     */
    void stop();
    //@}
    
    /**
     *  @name Set up sockets
     */
    //@{
    /**
     *  @brief Setup sender sockets
     *  @exception BindingException if something goes wrong
     *
     *  This method configures the sender sockets for RTP and RTCP
     */
    void setupSenderSockets() throw (const BindingException&);
    
    /**
     *  @brief Setup receiver sockets
     *  @brief address The session address
     *  @exception BindingException if something goes wrong
     *
     *  This method configures the receiver sockets for RTP and RTCP. The sockets may later
     *  be reconfigured to be used for sending by setting a destination address.
     *
     *  The @c address parameter is interpreted as follows:
     *   - A @c NULL pointer configures the sockets for unicast with random ports
     *   - An address with a UnicastIPAddress configures the sockets for unicast with fixed ports. The IP address is ignored.
     *   - An address with a MulticastIPAddress configures the sockets for multicast with fixed ports
     *
     *  @see RTPStrategy::setupSenderSockets
     */
    void setupReceiverSockets(const RTPAddress* address) throw (const BindingException&);
    //@}
    
    /**
     *  @brief Set send value stream
     *
     *  Not used, does nothing.
     */
    virtual void setSendValueStream(const string&);
    
    /**
     *  @brief Set receive value stream
     *
     *  Not used, does nothing.
     */
    virtual void setReceiveValueStream(const string&);

  private:
    /**
     *  @brief Payload types according to RFC
     */
    enum PayloadType {
      PT_AUDIO_PCMU       = 0,
      PT_AUDIO_GSM        = 3,
      PT_AUDIO_PCMA       = 8,
      PT_AUDIO_L16_STEREO = 10,
      PT_AUDIO_L16_MONO   = 11,
      PT_AUDIO_MPEG       = 14,
      PT_VIDEO_JPEG       = 26,
      PT_VIDEO_MPEG       = 32,
    };
    
  private:
    /**
     *  @brief Addresses and Interfaces
     */
    //@{
    /** Local RTP address.
     */
    RTPAddress* __local_address;
    
    /** Remote RTP address.
     */
    RTPAddress* __remote_address;
    
    /** Remote connection manager interface.
     */
    IConnectionManager* __icm;
    
    /** NMMReference of partner strategy.
     */
    NMMReference __remote_ts_ref;
    //@}
    
    /**
     *  @brief Receiver Components
     */
    //@{
    /** RTP network source.
     */
    RTPSource* __rtp_source;
    
    /** Buffered sink to produce NMM buffers.
     */
    BufferedSink* __buffered_sink;
    //@}
    
    /**
     *  @brief Sender Components
     */
    //@{
    /** Buffered source to unpack NMM buffers.
     */
    BufferedSource* __buffered_source;
    
    /** RTP framer (may be the same as the buffered source).
     */
    FramedSource* __framer;
    
    /** RTP network sink.
     */
    RTPSink* __rtp_sink;
    
    /** Send queue size.
     */
    unsigned __send_queue_size;
    
    /** Send queue mode.
     */
    IQueued::Mode __send_queue_mode;
    //@}
    
    /**
     *  @brief Common Components
     */
    //@{
    /** RTCP instance.
     */
    RTCPInstance* __rtcp_instance;
    
    /** liveMedia event processing thread.
     */
    LiveThread* __live_thread;
    //@}
    
  private:
    /** Strategy type ("RTPStrategy").
     */
    static const char __strategy_type[];
  };
  
  
  inline const RTPAddress& RTPStrategy::getLocalAddress() const throw() {
    return *__local_address;
  }
  
  inline const RTPAddress* RTPStrategy::getRemoteAddress() const {
    return __remote_address;
  }
  
  inline string RTPStrategy::getStaticStrategyType() {
    return __strategy_type;
  }
  
} // namespace NMM

#endif // NMM_RTP_STRATEGY_HPP
