/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef NMM_TCPGENERIC_HPP
#define NMM_TCPGENERIC_HPP

#include "nmm/utils/net/SockStream.hpp"
#include "nmm/interfaces/net/ITCPControlImpl.hpp"

namespace NMM {
  class Node;



  /** This class implements serval services for a tcp conneciton. So it implements the ITCPControl
   * Interface to acces the internal parameters
   */
  class TCPGeneric : public ITCPControlImpl {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Constructor
     */
    TCPGeneric();

    /** Destructor
     */
    virtual ~TCPGeneric();
    //@}


    /** @name TCPGeneric specifc mehtods
     */
    //@{
    /** This method returns the SockStream that is used to read/write the data
     */
    virtual       SockStream& getStreamSocket()       { return __stream; }

    /** This method returns the SockStream that is used to read/write the data
     */
    virtual const SockStream& getStreamSocket() const { return __stream; }
    //@}

  protected:
    /** @name ITCPControl specific methods
     */
    //@{

    /** This method turns the Nagle algorithm on or off. This means that packets are always sent as
     * soon as possible and no unnecessary delays are introduced, at the cost of more packets in
     * the network. Expects an integer boolean flag.
     */
    virtual void enableNoDelay(const bool& value);

    /** Returns to true if the Nagle algorithm is enabled, otherwise false
     */
    virtual bool isDelayEnabled() const;

    /** Set or receive the maximum segment size for outgoing TCP packets. If this option is set before
     * connection establishment, it also changes  the  MSS value announced to the other end in the
     * initial packet. Values greater than the interface MTU are ignored and have no effect.
     */
    virtual void setMaxSegment(const int& value);

    /** This method returns the current segment size.
     */
    virtual int  getMaxSegment() const;

    /** This method enanbles/disables partial frames. If enabled  don't  send  out partial frames.
     * All queued partial frames are sent when the option is cleared again.  This is useful for
     * prepending headers before calling sendfile(2), or for throughput optimization. This option
     * cannot be combined with TCP_NODELAY.
     */
    virtual void enablePartialFrames(const bool& value);

    /** This method returns true if partial frames are enabled, otherwise false.
     */
    virtual bool isPartialFramesEnabled();

    /** This method returns the count of unread data.
     */
    virtual int  getUnreadQueueCount() const;

    /** This method returns the count of unsend data.
     */
    virtual int  getUnsendQueueCount() const;

    /** This method returns true if all data are send.
     */
    virtual bool allUrgentDataSend()   const;

    /** This method enables/disables keep connections alive
     */
    virtual void enableKeepAlive(const bool& value);

    /** This method returns true if a connection is keeped alive, otherwise false.
     */
    virtual bool getKeepAlive() const;

    /** This method set buffer size for incoming data.
     */
    virtual void setRecvBufferSize(const int& value);

    /** This method returns the buffer size for incoming data.
     */
    virtual int getRecvBufferSize() const;

    /** This method sets the buffer size for outgoing data.
     */
    virtual void setSendBufferSize(const int& value);

    /** This method returns the buffer size for outgoing data.
     */
    virtual int getSendBufferSize() const;

    /** This method sets the priority of the data
     */
    virtual void setPriority(const int& value);

    /** This method returns the used priority to send the data
     */
    virtual int getPriority() const;
    //@}

  private:
    void createException(int) const;
  private:
    SockStream      __stream;     /* Receiver's stream */
  };

}
#endif
