/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *                    Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <iostream>

#include "ClientRegistry.hpp"
#include "nmm/base/registry/PluginManager.hpp"
#include "nmm/base/NMMApplication.hpp"
#include "nmm/base/registry/Plugin.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/interfaces/net/IObjectAcceptor.hpp"

#include "nmm/base/sync/MasterSlaveSourceSynchronizer.hpp"
#include "nmm/base/sync/MultiAudioVideoSynchronizer.hpp"
#include "nmm/utils/utils.hpp"
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"

namespace NMM {

ClientRegistry::ClientRegistry()
  : NamedObject("ClientRegistry"), __is_running(true), __init_transaction(false), __condition(__condition_mutex), __time_interval(5), __member_mutex(ThreadMutex::RECURSIVE)
{
  /* Stores a pointer to the PluginManager in this address space. This is need for
   * local optimization
   */
  __internal_p_manager.reset(SPluginManager::getInstance().getInterface<IPluginManager>());

  /* start the thread to refresh ressource reservation */
  activateThread();
}

ClientRegistry::~ClientRegistry() {
  /* Begin> Deactivate the thread*/
  __condition_mutex.lock();
  __is_running = false;                 // set the running flag to false
  __condition.notify();                 // and notify the thread
  __condition_mutex.unlock();
  deactivateThread();                   // wait until thread terminates
  /* End< */

  /* Begin> delete the interfaces the ClientRegistry use*/
  for (RegistryIt it = __named_registry.begin(); it != __named_registry.end(); ++it) {
    delete it->second.server_registry;  //delete the IServerRegistry
    delete it->second.plugin_manager;   //delete the corresponding IPluginManager
  }
  /* End < */
}


IServerRegistry* ClientRegistry::getServerRegistry(const RegistryName& registry_name) {
  string my_host;
  if(registry_name.first == "localhost") {
    my_host = getHostname();
  }
  else {
    my_host = registry_name.first;
  }
  RegistryIt it = __named_registry.find(RegistryName(my_host, registry_name.second));
  /* Begin> If ServerRetgistry is unknown, try to add it */
  if (it == __named_registry.end()) {
    try {
      addRegistry(TCPAddress(registry_name.first,registry_name.second));
      it = __named_registry.find(RegistryName(my_host, registry_name.second));
    } catch (const Exception& e) {
      WARNING_STREAM( e << endl );
    }
  }
  /* End< */

  if ( it != __named_registry.end() )
    return it->second.server_registry;

  return 0;
}

const RegistryName& ClientRegistry::determineRegistryName(Interface& interface) const {
  /* Search the ReservationStruct for given interface which stores the corresponding server registry */
  CReservationIt res_it = __reservation_map.find(interface.getParentObject());
  if ( res_it == __reservation_map.end() ) {
    throw RegistryException("Given interface was not requested by this ClientRegistry or has been already released");
  }

  /* find the RegistryName for the server registry  */
  for (CRegistryIt it = __named_registry.begin(); it != __named_registry.end(); ++it) {
    if ( it->second.server_registry == res_it->second.registry ) {
      return it->first;
    }
  }

  throw RegistryException("Given interface was not requested by this ClientRegistry or has been already released");
}

IPluginManager* ClientRegistry::getPluginManager(const RegistryName& registry_name){
  string my_host;
  if(registry_name.first == "localhost") {
    my_host = getHostname();
  }
  else {
    my_host = registry_name.first;
  }
  RegistryIt it = __named_registry.find(RegistryName(registry_name.first, registry_name.second));
  if (it != __named_registry.end())
    return it->second.plugin_manager;

  return 0;
}
/**
  */
const IServerRegistry* ClientRegistry::getServerRegistry(const RegistryName& registry_name) const {
  string my_host;
  if(registry_name.first == "localhost") {
    my_host = getHostname();
  }
  else {
    my_host = registry_name.first;
  }
  CRegistryIt it = __named_registry.find(RegistryName(my_host, registry_name.second));
  if (it != __named_registry.end())
    return it->second.server_registry;

  return 0;
}


INode* ClientRegistry::requestNode(const Response& resp, NodeDescription* nd )
          throw (const RegistryException& )
{
  /* Begin> Try to request one of these nodes */
  for(list<ObjectID>::const_iterator it = resp.second.begin(); it != resp.second.end(); ++it)
    try {
      return requestNode(resp.first, *it, nd);
    } catch (const Exception& e) {
      WARNING_STREAM( e << endl);
    }
  /* End< */

  /* If no node can be requested an Exception is thrown*/
  throw RegistryException("Could not request a required node ");
}


list<Response> ClientRegistry::initRequest(const NodeDescription& nd)
{
  RegistryIt reg_it = __named_registry.find(RegistryName(nd.getLocation(),nd.getPort()));
  if (reg_it == __named_registry.end()) {
    try {
      addRegistry(TCPAddress(nd.getLocation(), nd.getPort()));  
    } catch (const Exception& e) {
      ERROR_STREAM ( e << endl);
      throw RegistryException(string("Can not connect to registry which is running on host '") + nd.getLocation()+string("' to request node '") + nd.getObjectName() + string("'"));
    }
  }

  list<Response> response_list;
  /* If no specific location was set, forward the request to all registries */
  if (!nd.isLocationEnabled()) {
    /* 1. Begin> search the local registry */
    RegistryIt it = __named_registry.find(RegistryName(getHostname(),ClientRegistry::default_port));
    if (it != __named_registry.end())
      response_list = it->second.server_registry->initRequest(nd);
    else {
      ERROR_STREAM ("No local server registry available on host " << getHostname() << endl);
    }


    /* 2. searches the other registries */
    for (RegistryIt it = __named_registry.begin(); it != __named_registry.end(); ++it) {
      if (it->first != RegistryName(getHostname(),ClientRegistry::default_port)) {
        try {
          list<Response> tmp_list = it->second.server_registry->initRequest(nd);
          if (!tmp_list.empty())
            response_list.insert(response_list.end(), tmp_list.begin(), tmp_list.end());
        } catch(const Exception& e) {
          WARNING_STREAM(e << endl);
        }
      }
    }
  }
  /* Otherwise search the specific Server-Registry */
  else {
    RegistryIt it = __named_registry.find(RegistryName(nd.getLocation(), nd.getPort()));
    if (it != __named_registry.end())
      response_list = it->second.server_registry->initRequest(nd);
  }

  /* Return the response list */
  return response_list;
}

list< Response >  ClientRegistry::initDistributedRequest(const NodeDescription& nd, string& hostname) {
  RegistryIt reg_it = __named_registry.find(RegistryName(nd.getLocation(),nd.getPort()));
  if (reg_it == __named_registry.end()) {
    try {
      addRegistry(TCPAddress(nd.getLocation(), nd.getPort()));
    } catch (const Exception& e) {
      ERROR_STREAM ( e << endl);
      throw RegistryException(string("Can not connect to registry which is running on host '") + nd.getLocation()+string("' to request node '") + nd.getObjectName() + string("'"));
    }
  }

  list<Response> response_list;

  // sort server_registrys by speed left
  __registry_speeds.clear();
  IResourceManager_var dist_res_mgr;
  // build a map with cpu speed left for all server registrys

  for (RegistryIt it = __named_registry.begin(); it != __named_registry.end(); ++it) {
    if (it->first != RegistryName(getHostname(),ClientRegistry::default_port)) {
      dist_res_mgr.reset(it->second.server_registry->getResourceManager());
      uint speed;
      float free;
      dist_res_mgr->getCPUSpeed(speed,free);
      __registry_speeds.insert(pair<uint,RegistryName>(speed,RegistryName(it->first)));
    }
  }

  // request the node
  for (map<uint,RegistryName>::reverse_iterator it = __registry_speeds.rbegin(); it != __registry_speeds.rend(); ++it) {
    if (it->second != RegistryName(getHostname(),ClientRegistry::default_port)) { // dont return a local node
      try {
        list<Response> tmp_list = __named_registry[it->second].server_registry->initRequest(nd);
        if (!tmp_list.empty())
          response_list.insert(response_list.end(), tmp_list.begin(), tmp_list.end());
          dist_res_mgr.reset(__named_registry[it->second].server_registry->getResourceManager());
          dist_res_mgr->setReservationPending();
          cout << " requested node from " << it->second.first << endl;
          hostname = it->second.first;
          break;
      } catch(const Exception& e) {
        WARNING_STREAM(e << endl);
      }
    }
  }

  /* Return the response list */
  return response_list;

}


void ClientRegistry::initTransaction()
  throw (const RegistryException&)
{
  MutexGuard guard(__member_mutex);
  __init_transaction = true;
  for (RegistryIt it = __named_registry.begin(); it != __named_registry.end(); ++it)  {
    initTransaction( it->second.server_registry );
  }
}

void ClientRegistry::initTransaction( IServerRegistry* server_registry ) {
  server_registry->initTransaction(UserInfo());
  __locked_registry.insert(pair<IServerRegistry*, long>(server_registry, server_registry->getTransactionTimeout()));
}

void ClientRegistry::releaseTransaction() {
  MutexGuard guard(__member_mutex);
  for (RegistryIt it = __named_registry.begin(); it != __named_registry.end(); ++it) {
    it->second.server_registry->closeTransaction(UserInfo());
  }
  __locked_registry.clear();
  __init_transaction = false;
}


void ClientRegistry::releaseNode(INode& i_node) {
  MutexGuard m(__member_mutex);

  /* search the proxy object this interface */
  ReservationIt it = __reservation_map.find(i_node.getParentObject());
  if (it == __reservation_map.end()) {
    ERROR_STREAM("Could not releases Node: Unknown node: " << i_node.getParentObject()->getCheckedInterface<INamed>()->getName() << endl);
    return;
  }

  /* If the ClientRegistry allocates the Node we call the releaseLocalNode method. In this case the
   * ClientRegistry has a pointer to the Node
   */
  if (it->second.node) {
    releaseLocalNode(it);
  }
  /* Otherwise we forward the method call to the corresponding IRegistry */
  else {
    it->second.registry->releaseNode(it->second.registry_id, it->second.reference, it->second.node_id, UserInfo());
  }
  /* remove it from the map */
  __reservation_map.erase(it);
}

void ClientRegistry::releaseLocalNode(ReservationIt& it) {
  /* 1. We destroy the node */
  it->second.node_factory->releaseNode(*it->second.node);

  /* 2. We release the reservation and remove it from the internal struct*/
  it->second.registry->releaseReservation(it->second.registry_id, it->second.node_id, UserInfo());
}

NodeDescription ClientRegistry::getNodeDescription(const RegistryID& reg_id, const ObjectID& node_id)
{
  /* Search the Server-Registry that manages the Registry with given RegistryID */
  RegistryIt registry = findRegistryStruct(reg_id);
  if (registry != __named_registry.end())
    /* Return ths NodeDescription */
    return registry->second.server_registry->getNodeDescription(reg_id, node_id);

  throw RegistryException("getNodeDescription: Unknown Registry");
}


INode* ClientRegistry::requestNode(const list<Response>& request_list, NodeDescription* nd)
  throw (const RegistryException& )
{
  /* Try to request one of the Nodes in the response-list */
  for (list<Response>::const_iterator it = request_list.begin(); it != request_list.end(); ++it) {
    for(list<ObjectID>::const_iterator it_2 = it->second.begin(); it_2 != it->second.end(); ++it_2)
      try {
        /* Try to request the node */
        return requestNode(it->first, *it_2, nd);
      } catch (const Exception& e) {
        WARNING_STREAM(e.getComment() << endl);
      }
  }
  throw RegistryException("No such node available");
}


INode* ClientRegistry::requestNode(const RegistryID& reg_id, const ObjectID& node_id, NodeDescription* nd)
  throw (const RegistryException& )
{
  /* search the ServerRegistry that manages the RegistryID */
  RegistryIt tmp_registry = findRegistryStruct(reg_id);

  /* If no such registry availble throw an exception */
  if (tmp_registry == __named_registry.end())  {
    throw RegistryException("RequestNode: Unknown Registry");
  }

  /* Lock the internal structs against the run-method */
  MutexGuard guard(__member_mutex);

  /** If the Server-Registry runs on our host and optimization is allowed, allocate the Node in our addresspace */
  if (
       (
         SPluginManager::getInstance().getInterfaceList().front()->isSystemTo(*tmp_registry->second.server_registry) ||
         SPluginManager::getInstance().getInterfaceList().front()->isLocalTo(*tmp_registry->second.server_registry)
       ) && tmp_registry->second.is_optimization_allowed
     )
  {
    DEBUG_STREAM2("Create Local Node " << endl);
    /* create a local node */
    return createLocalNode(tmp_registry, reg_id, node_id, nd);
  } else {
    DEBUG_STREAM2("Create Distributed Node " << endl );
    /* create a distributed node */
    return createDistributedNode(tmp_registry, reg_id, node_id, nd);
  }
}

INode* ClientRegistry::createDistributedNode(RegistryIt registry_it, const RegistryID& reg_id, const ObjectID& node_id, NodeDescription* nd)
  throw (const RegistryException& )
{
  /* request the node by the sever-registry */
  INode* tmp_node = 0;
  if (nd) {
    tmp_node = registry_it->second.server_registry->requestNode(reg_id, node_id, UserInfo(), *nd);
  }
  else {
    NodeDescription tmp_nd;
    tmp_node = registry_it->second.server_registry->requestNode(reg_id, node_id, UserInfo(), tmp_nd);
  }

  if (!tmp_node)
    throw RegistryException("Could not Request INode");

  /*Begin> We create the reservation struct */
  ReservationStruct tmp_struct;
  tmp_struct.registry_id      = reg_id;
  tmp_struct.reference        = tmp_node->getParentReference();
  tmp_struct.registry         = registry_it->second.server_registry;
  tmp_struct.node_id          = node_id;
  /* Store, how long we can use this node*/
  tmp_struct.reservation_time = registry_it->second.server_registry->getNodeLeasingTime();
  /*End< */

  /* Store the created reservation struct in our reservation map*/
  __reservation_map.insert(pair<NMMObject*, ReservationStruct>(tmp_node->getParentObject(), tmp_struct));

  /* return the interface */
  return tmp_node;
}


INode* ClientRegistry::createLocalNode(RegistryIt registry_it, const RegistryID& reg_id, const ObjectID& node_id, NodeDescription* nd)
  throw (const RegistryException& )
{
  /* store the IServer-Registry in a temporary pointer */
  IServerRegistry* registry = registry_it->second.server_registry;

  /* 1. reserve the node */
  FactoryTicket ticket = registry->reserveNode(reg_id, node_id, UserInfo());

  /* 2. request the library to load */
  string plugin = registry->getPluginInfo(reg_id, node_id);

  try {
    /*Begin> Create the node and set the ID's*/
    IPluginManager* pm = registry_it->second.plugin_manager;
    if (!pm) {
      throw RegistryException("No PluginManager available on this system");
    }
    /* load the library and set the same ObjectID's as in the system's NodeFactory*/
    loadLocalPlugin(pm, plugin);

    /* search the node-factory that can create the node */
    NodeFactory* factory      = SLocalNodeFactory::getInstance().findNodeFactory(ticket.getNodeFactoryID());

    /* if no such NodeFactory is available throw an exception */
    if ( !factory ) {
      registry->releaseReservation(reg_id, node_id, UserInfo());
      throw RegistryException("No NodeFactory available to create the node");
    }

    /* Create the node */
    Node* tmp_node = &factory->createNode(ticket);
    /*End< */

    /* get the interface */
    INode* i_tmp_node = tmp_node->getInterface<INode>();

    /*Begin> We create the reservation struct */
    ReservationStruct tmp_struct;
    tmp_struct.registry_id  = reg_id;
    tmp_struct.reference    = tmp_node->getReference();
    tmp_struct.registry     = registry;
    tmp_struct.node_factory = factory;
    tmp_struct.node_id      = node_id;
    tmp_struct.node         = tmp_node;
    /* Store, how long we can use this node*/
    tmp_struct.reservation_time = registry->getNodeLeasingTime();
    /*End< */

    /* store the reservation struct */
    __reservation_map.insert(pair<NMMObject*, ReservationStruct>(i_tmp_node->getParentObject(), tmp_struct));

    /* update the NodeDescription if required */
    if (nd) {
      Registry::updateNodeDescription(*nd, registry->getNodeDescription(reg_id, node_id) );
    }

    /* return the INode */
    return i_tmp_node;
  }
  /* Begin> Error handling */
  catch (const FactoryException& fe) {
    /* release the reservation if an error occured */
    registry->releaseReservation(reg_id, node_id, UserInfo());
    throw RegistryException(fe.getComment());
  }
  /* End< */
}



void ClientRegistry::loadLocalPlugin(IPluginManager* pm, const string& plugin)
  throw (const RegistryException& )
{
  try {
    if (SPluginManager::getInstance().isPluginUsed(plugin))
      return;

    /* 1. load the library in our address space*/
    __internal_p_manager->loadPlugin(plugin);

    /* 2. Begin> Syncing node id's instead of activating registry-part of the Plugin !*/
    /* Get the plugin-array of the library */
    Plugin** plugins = SPluginManager::getInstance().requestPluginArray(plugin);
    if (!plugins) {
      GLOBAL_ERROR_STREAM( "No Plugins available" << endl);
      return;
    }

    int plugin_size = SPluginManager::getInstance().getPluginSize(plugin);
    if (!plugin_size) {
      GLOBAL_ERROR_STREAM( "No Plugins available" << endl);
      return;
    }

    vector<list<ObjectID> > node_id_vector;
    for (int i = 0; i < plugin_size; ++i ) {
      /* If the ObjectID's are not set */
      if (plugins[i]->getSize() == 0) {
        /* get the ObjectID form the global PluginManager of the Server-Registry */
        list<ObjectID> node_id = pm->getPluginID(plugin, i);
        /* Set the ObjectID's */
        node_id_vector.push_back( node_id );
      } else {
        node_id_vector.push_back( list<ObjectID>() );
      }
    }
    SPluginManager::getInstance().defineObjectIDs( plugin, node_id_vector );
    /** End< */

    /* 3. If the plugin is not registered by a NodeFactory, register it */
    if (!SPluginManager::getInstance().isFactoryPartActive(plugin))
      __internal_p_manager->activateFactoryPart(plugin);
  } catch (const Exception& ce){
    throw RegistryException( ce.getComment() ) ;
  }
}

ClientRegistry::RegistryIt ClientRegistry::findRegistryStruct(const RegistryID& reg_id ) {
  for (RegistryIt it = __named_registry.begin(); it != __named_registry.end(); ++it) {
    if (it->second.server_registry->isSubRegistry(reg_id))
      return it;
  }

  return __named_registry.end();
}

void ClientRegistry::addRegistry(const TSAddress& address, bool is_optimization_allowed)
  throw (const BindingException&)
{
  RegistryName reg_name = RegistryName(address.getHostname(), address.getPort());

  /* Begin> If a ServerRegistry with given name already exist, don't add it again */
  RegistryIt it = __named_registry.find(reg_name);
  if (it != __named_registry.end()) {
    /* Try to contact the ServerRegistry */
    try {
      it->second.server_registry->getTransactionTimeout();
      DEBUG_STREAM("A ServerRegistry with name " << reg_name.first << " already exist and is not added again! " << endl);
      return;
    } catch (const Exception& e){
      ERROR_STREAM( "Can not connect already existing ServerRegistry" << endl);
      ERROR_STREAM( "Error message : " << e << endl);
      delete it->second.plugin_manager;
      delete it->second.server_registry;
      __named_registry.erase(it);
    }
  }
  /* End< */

  ProxyObject*        new_registry = 0;     ///< ProxyObject for the new registry
  RegistryStruct      registry_struct;      ///< Struct for the internal map
  IObjectAcceptor_var i_connector;          ///< var_pointer to the temoorary ObjectAcceptor
  try {
    /* Create a ProxyObject for the ServerRegistry with automatic memory management */
    new_registry = new ProxyObject(true);

    /* and connect to the given address */
    new_registry->connectToObject(address);

    /* Get the IObjectAcceptor to access the services available on this Application.
     * This interface can be 0 if the address stores a valid NMMReference.
     */
    i_connector.reset(new_registry->getInterface<IObjectAcceptor>());

    /* Begin> Connect to the ServerRegistry */
    if (i_connector.get()) {
      i_connector->connectToService("Registry");
    }
    /* End< */

    /* Can the ClientRegistry allocate Nodes in local address space (if possible) */
    registry_struct.is_optimization_allowed = is_optimization_allowed;
    /* store an IServerRegistry pointer */
    registry_struct.server_registry         = new_registry->getCheckedInterface<IServerRegistry>();

    /* Begin> if its a local or a system Registry add the local plugin manager and if optimization is allowed */
    if ( SPluginManager::getInstance().isLocalTo(*registry_struct.server_registry) )
    {
      registry_struct.plugin_manager = SPluginManager::getInstance().getCheckedInterface<IPluginManager>();
    } else {
      /* Create a ProxyObject for the PluginManager with automatic memory management */
      ProxyObject* new_plugin_manager = new ProxyObject(true);

      /* and connect to the given address */
      new_plugin_manager->connectToObject(address);

      /* Get the IObjectAcceptor to access the PluginManager */
      IObjectAcceptor_var p_connector(new_plugin_manager->getInterface<IObjectAcceptor>());

      /* Begin> Connect to the PluginManager */
      if (p_connector.get()) {
        p_connector->connectToService("PluginManager");
      }
      /* End< */

      /* add the PluginManager */
      registry_struct.plugin_manager = new_plugin_manager->getCheckedInterface<IPluginManager>();
    }

    __named_registry.insert(pair<RegistryName, RegistryStruct>(reg_name, registry_struct));
    /* End< */

    /* If the registry is added during a transaction, we call initTransaction on the new registry as well */
    MutexGuard mg( __member_mutex );
    if ( __init_transaction ) {
      initTransaction( registry_struct.server_registry );
    }
  /* Begin> Error handling */
  } catch (const Exception& e) {
    ERROR_STREAM( e << endl);
    /* Begin> memory cleanup */
    if (!registry_struct.server_registry && ! i_connector.get())
      delete new_registry;
    delete registry_struct.server_registry;
    delete registry_struct.plugin_manager;
    /* End< */
    throw BindingException(e.getComment());   ///< convert the Exception to a BindingException to rethrow it
  }
  /* End< */
}

map<RegistryName, list<Response> > ClientRegistry::initRequest(const NodeDescription& nd, const list<RegistryName>& registry_list)
{
  /*Begin> Check if all registries are known, otherwise add them */
  for (list<RegistryName>::const_iterator it = registry_list.begin(); it != registry_list.end(); ++it) {
    RegistryIt reg_it = __named_registry.find(RegistryName(it->first,it->second));
    if (reg_it == __named_registry.end()) {
      try {
        addRegistry(TCPAddress(it->first,it->second));
      } catch (const Exception& e) {
        ERROR_STREAM ( e << endl);
        throw RegistryException(string("Can not connect to registry which is running on host '") + nd.getLocation()+string("' to request node '") + nd.getObjectName() + string("'"));      
      }
    }
  }

  /* allocate the reponse_map */
  map<RegistryName, list<Response> > response_list;

  /* If the user specifies a location, ...*/
  if (nd.isLocationEnabled()) {
    /* ... find the IServerRegistry */
    RegistryIt reg_it = __named_registry.find(RegistryName(nd.getLocation(), nd.getPort()));
    if (reg_it != __named_registry.end()) {
      /* and store the response */
      response_list.insert(pair<RegistryName, list<Response> >(RegistryName(nd.getLocation(), nd.getPort()), reg_it->second.server_registry->initRequest(nd)));
    }
  } else {
    NodeDescription tmp_nd(nd);
    tmp_nd.enableLocation();
    /* forward the request to all registries in the list */
    for (list<RegistryName>::const_iterator it = registry_list.begin(); it != registry_list.end(); ++it)
    {
      /* find the IServerRegistry */
      RegistryIt reg_it = __named_registry.find(*it);
      if (reg_it != __named_registry.end()) {
        tmp_nd.setLocation(it->first);
        tmp_nd.setPort(it->second);
        /* and store the response */
        response_list.insert(pair<RegistryName, list<Response> >(*it, reg_it->second.server_registry->initRequest(nd)));
      }
      else {
        ERROR_STREAM("Unknwon registry " << it->first << endl);
      }
    }
  }
  /* return the response */
  return response_list;
}


void ClientRegistry::refreshNodeReservation() {
  MutexGuard guard(__member_mutex);

  /* iterate over all reservation entry */
  for (ReservationIt it = __reservation_map.begin(); it != __reservation_map.end(); ++it) {
    it->second.reservation_time -= __time_interval;
    /*If the time left is less then the time until we check again, the ClientRegistry must refresh the reserfation in this cylce*/
    if (it->second.reservation_time <= __time_interval) {
      try {
        it->second.reservation_time = it->second.registry->refreshNodeLeasing(it->second.registry_id,it->second.node_id, UserInfo());
      } catch (const Exception& e) {
        ERROR_STREAM(e << endl);
      }
    }
  }
}

void ClientRegistry::refreshTransactionLock() {
  MutexGuard guard(__member_mutex);
  /* Begin> For each locked ServerRegistry */
  for (map<IServerRegistry*, long>::iterator it = __locked_registry.begin(); it != __locked_registry.end(); ++it)
  {
    try {
      /* remove the time wait on the condition */
      it->second -= __time_interval;
      /* and if the left time is less than the time interval */
      if (it->second <= __time_interval)
        /* refresh the transaction */
        it->second = it->first->refreshTransaction(UserInfo());
    } catch ( Exception& e) {
      ERROR_STREAM(e << endl);
    }
  }
}

void ClientRegistry::run() {
  /* While the ClientRegistry is running */
  while(__is_running) {
    __condition_mutex.lock();
    __condition.relativeTimedWait(__time_interval,0);
    __condition_mutex.unlock();
    try {
      /* refresh the reservation of reserved nodes */
      refreshNodeReservation();

      /* refresh the reserved access-control */
      refreshTransactionLock();
    } catch (const Exception& e) {
      ERROR_STREAM("Error in run method " << e << endl);
    }
  }
}

void ClientRegistry::resetInternalReservation() {
  /* forward the method call to all Server-Registries */
  for (RegistryIt it = __named_registry.begin(); it != __named_registry.end(); ++it) {
    it->second.server_registry->resetInternalReservation();
  }
}

void ClientRegistry::resetSpecificInternalReservation(const UserInfo& user_info)
{
  /* forward the method call to all Server-Registries */
  for (RegistryIt it = __named_registry.begin(); it != __named_registry.end(); ++it) {
    it->second.server_registry->resetSpecificInternalReservation(user_info);
  }
}


void ClientRegistry::requestExclusiveGraph(GraphDescription& gd, const RegistryName& registry_name)
  throw (const RegistryException& )
{
  list<RegistryName> registry_list;
  IServerRegistry* registry = 0;

  /* Begin> If no session_location is set we try to find the best place to init a session */
  if (registry_name.first.empty()) {
    registry_list = getBestRegistry(gd);
    /* Some Error handling: If the desired registry is not available we throw an exception */
    if ( registry_list.empty() ) {
      throw RegistryException ("No ServerRegistry available that can be searched for sessions" );
    }
  } else {
    /* Otherwise we only search the given session_location */
    if (registry_name.first == "localhost")
      registry_list.push_back( RegistryName(getHostname(), ClientRegistry::default_port) );
    else
      registry_list.push_back( registry_name );
  }
  /* End< */

  /* init the session and return the session interface */
  ISession* session = 0;

  /* Try to request the graph from one of the registries */
  for (list<RegistryName>::iterator it = registry_list.begin(); it != registry_list.end(); ++it) {
    /* Get the corresponding server registry */
    registry = getServerRegistry(*it);

    /* Begin> init the session */
    if ( !registry ) {
      throw RegistryException(string("No ServerRegistry with name ") + it->first + string(" available."));
    }
    GraphDescription tmp_gd(gd);
    /* init the session for the graphdescription */
    session = registry->initSession(tmp_gd, UserInfo(), 0);
    /* If a session is available, leave the loop */
    if (session) {
      /* Begin> Store the session */
      ISessionIt it = __session_map.insert(pair<ProxyObject*, ISessionStruct>(session->getParentObject(), ISessionStruct())).first;
      it->second.session = session;
      it->second.server_registry = registry;
      /* End< */
    }
    /* <End */

    try {
      /* Begin> If the best registry is our localhost, we initialize only the session and allocate
       *        the remaining nodes on our local address space.
       */
      if ( it->first == "localhost" || it->first == getHostname() ) {
        /* allocate the reaming nodes */
        cout << "Request remaining nodes : Session runs on localhost " << endl;
        requestRemainingNodes(tmp_gd);
      } else {
        /* Begin> Set the flag for nodes that should be requested */
        for (unsigned int i = 0; i < tmp_gd.getNodeDescriptionList().size(); ++i) {
          if ( tmp_gd.getNodeDescriptionList()[i]->getLocation().empty() ||
               ( tmp_gd.getNodeDescriptionList()[i]->getLocation() == it->first && tmp_gd.getNodeDescriptionList()[i]->getPort() == it->second )
             )
          {
            tmp_gd.setRequestFlag(*gd.getNodeDescriptionList()[i], true);
            cout << "Request Node " << tmp_gd.getNodeDescriptionList()[i]->getNodeName() << " from host '" << it->first << "'" << it->second << endl;
          }
          else
          {
            tmp_gd.setRequestFlag(*gd.getNodeDescriptionList()[i], false);
            cout << "Request Node " << tmp_gd.getNodeDescriptionList()[i]->getNodeName() << " from host '" << tmp_gd.getNodeDescriptionList()[i]->getLocation() << "'" << tmp_gd.getNodeDescriptionList()[i]->getPort() << endl;
          }
        }
        /* End< */
        DEBUG_STREAM( "Request graph : Session runs on host " << it->first << endl);
        /* request the Graph */
        registry->requestGraph(tmp_gd, UserInfo());
        requestRemainingNodes(tmp_gd);
      }

      gd = tmp_gd;
      /* Update the Graphdescription, to tell the Session the new INodes. */
      gd.getISession()->setGraphDescription(gd);
      return;    
    } catch (const Exception& e) {
      WARNING_STREAM( "Error Requesting session from host " << it->first << endl);
      WARNING_STREAM( e << endl );
      /* if something went wrong, release the graph */
      try {
        releaseGraph(tmp_gd);
      } catch (const Exception& e) {
        ERROR_STREAM(e << endl);
      }
    }
    /* End< */
  }

  /* If no session returns, throw an exception */
  throw RegistryException("Could not request an exclusive graph");
}


void ClientRegistry::requestGraph(GraphDescription& gd, const RegistryName& registry_name)
  throw (const RegistryException&)
{
  /* Begin> Add all required registries */
  for ( unsigned int i = 0; i < gd.getNodeDescriptionList().size(); ++i) {
    if ( gd.getNodeDescriptionList()[i] ) {
      if ( gd.getNodeDescriptionList()[i]->isLocationEnabled() ) {
        try {
          addRegistry(TCPAddress(gd.getNodeDescriptionList()[i]->getLocation(), gd.getNodeDescriptionList()[i]->getPort()));
        } catch (const Exception& e) {
          WARNING_STREAM ( e << endl );
          throw RegistryException(string("Can not connect to registry which is running on host '") +gd.getNodeDescriptionList()[i]->getLocation()+string("' to request node '") + gd.getNodeDescriptionList()[i]->getObjectName() + string("'"));
        }
      }
    }
  }
  /* End< */  

  /* Lock the Registries */
  RegistryLock lock(*this);
  
  /* If ALL NodeDescriptions are exclusive use the old requestGraph  */
  if (gd.isSharingType(NodeDescription::EXCLUSIVE) ) {
    try {
      /* Use the 'standard' request graph method */
      requestExclusiveGraph(gd, registry_name);
      /* and configure the out-of-band bindings */
      gd.configureOutOfBandBindings();

      gd.getISession()->setGraphDescription(gd);
      return;
    } catch (const Exception& e) {
      throw RegistryException( e.getComment() );
    }
  }

  /* Begin> If ALL NodeDescriptions are exclusive_or_shared */
  if (gd.isSharingType(NodeDescription::EXCLUSIVE_OR_SHARED)) {
    try {
      /* First try to request a exclusive Graph */
      requestExclusiveGraph(gd, registry_name);
      /* and configure the out-of-band bindings */
      gd.configureOutOfBandBindings();
      return;
    } catch ( const Exception& e) {
      ERROR_STREAM( e << endl);
      ERROR_STREAM("Could not request exclusive graph try to get a shared graph" << endl);
      /* If exclusive is not possible try to request a shared one */
      try {
        /* Otherwise try to request a shared Graph */
        requestSharedGraph(gd, registry_name);
        /* and configure the out-of-band bindings */
        gd.configureOutOfBandBindings();
        /* return the sessions */
        return;
      } catch (const Exception& e2) {
        /* Convert exception to a RegistryException */
        throw RegistryException(e.getComment() + ".\n" + e2.getComment());
      }
    }
  }
  /*End< */

  /* Begin> If ALL NodeDescriptions are shared_or_exclusive */
  if (gd.isSharingType(NodeDescription::SHARED_OR_EXCLUSIVE)) {
    try {
      /* First try to request a shared Graph */
      requestSharedGraph(gd, registry_name);
      /* then request the remaining nodes */
      requestRemainingNodes(gd);
      /* and configure the out-of-band bindings */
      gd.configureOutOfBandBindings();
      /* return the session */
      return;
    } catch ( const Exception& e) {
      WARNING_STREAM(e << endl);
      WARNING_STREAM("Could not request exclusive graph try top get shared graph" << endl);
      try {
        /* If exclusive is not possible try to request a shared one */
        requestExclusiveGraph(gd, registry_name);
        gd.configureOutOfBandBindings();
        return;
      } catch (const Exception& e2) {
        /* Convert exception to a RegistryException */
        throw RegistryException(e.getComment() + ".\n" + e2.getComment());
      }
    }
  }
  /*End< */

  try {
    /* Begin> handle remaining cases */
    requestSharedGraph(gd, registry_name);
    /* and configure the out-of-band bindings */
    requestRemainingNodes(gd);
    /* congigiure the out-of-band bindings */
    gd.configureOutOfBandBindings();
    /* return the session */
    return;
    /* End< */
  } catch (const Exception& e) {
    /* Convert exception to a RegistryException */
    throw RegistryException(e.getComment());
  }
}


void ClientRegistry::requestSharedGraph(GraphDescription& gd, const RegistryName& registry_name)
  throw (const RegistryException& )
{
  list<RegistryName> registry_list;
  IServerRegistry* registry = 0;

  /* Begin> If a session is specified for a shared graph, we try to request it from this host */
  if ( gd.getISession() ) {
    ISessionIt it = __session_map.find(gd.getISession()->getParentObject());
    if (it != __session_map.end()) {
      try {
        /* request Graph */
        return requestSharedGraph( gd, it->second.server_registry,  getRegistryName(*it->second.server_registry) );
      } catch (const Exception& e) {
        WARNING_STREAM( "Error Requesting session from predefined Session " << endl);
        WARNING_STREAM( e << endl );
      }
    } else {
      ERROR_STREAM( "Specified ISession for a shared graph is unknown to the ClientRegistry-> Ignore this session " << endl);
    }
  }
  /* End< */

  /* Begin> If no session_location is set we try to find the best place to search for a session */
  if (registry_name.first.empty() ) {
    /* create a list with 'best' server registries at the beginning */
    registry_list = getBestRegistry(gd);

    /* Some Error handling */
    if ( registry_list.empty() )  {
      throw RegistryException ("No ServerRegistry available that can be searched for sessions" );
    }
    /* End< */
  } else {
    /* Begin> Otherwise we add the given session_location into the list*/
    if (registry_name.first == "localhost")
      registry_list.push_back( RegistryName(getHostname(),ClientRegistry::default_port) );
    else
      registry_list.push_back( registry_name );
  }

  /* Begin> Try to request the session from one of the registries. The first match will be returned */
  for (list<RegistryName>::iterator it = registry_list.begin(); it != registry_list.end(); ++it) {
    /* Get the corresponding server registry */
    registry = getServerRegistry( *it );
    if ( !registry ) {
      WARNING_STREAM ("No ServerRegistry with name "<< it->first << " available.");
    }

    try {
      /* Try to request the graph */
      return requestSharedGraph(gd, registry, *it);
    } catch (const Exception& e) {
      WARNING_STREAM( "Error Requesting session from host " << it->first << endl);
      WARNING_STREAM( e << endl );
    }
  }
  /* End< */

  /* If no session returns, throw an exception */
  throw RegistryException("No Session for sharing available ");
}

void ClientRegistry::requestSharedGraph(GraphDescription& gd, IServerRegistry* registry, const RegistryName& reg_name)
    throw (const RegistryException&)
{
  /* Create a copy of the graph description */
  GraphDescription tmp_gd(gd);

  /* Begin> Enable the request flag for nodes that should be requested */
  for (unsigned int i = 0; i < tmp_gd.getNodeDescriptionList().size(); ++i) {
    if ( tmp_gd.getNodeDescriptionList()[i]->getLocation().empty() ||
       ( tmp_gd.getNodeDescriptionList()[i]->getLocation() == reg_name.first && tmp_gd.getNodeDescriptionList()[i]->getPort() == reg_name.second )
      )
    {
      tmp_gd.setRequestFlag(*tmp_gd.getNodeDescriptionList()[i], true);
      cout << "Request Node " << tmp_gd.getNodeDescriptionList()[i]->getNodeName() << " from host '" << reg_name.first << "' " << reg_name.second << endl;
    }
    else
    {
      tmp_gd.setRequestFlag(*tmp_gd.getNodeDescriptionList()[i], false);
      cout << "Request Node " << tmp_gd.getNodeDescriptionList()[i]->getNodeName() << " from host '" << tmp_gd.getNodeDescriptionList()[i]->getLocation() << "'" << tmp_gd.getNodeDescriptionList()[i]->getPort() << endl;
    }
  }
  /* End< */

  /* Try to request the session */
  registry->requestSharedGraph(tmp_gd, UserInfo());

  /* If a session is available, leave the loop */
  if (tmp_gd.getISession()) {
    /* Begin> Store the session */
    ISession* session = dynamic_cast<ISession*>(tmp_gd.getISession()->copy());	 
    ISessionIt s_it = __session_map.insert(pair<ProxyObject*, ISessionStruct>(session->getParentObject(), ISessionStruct())).first;
    /* We need a copy for the internal session management */
    s_it->second.session = session;
    s_it->second.server_registry = registry;
    /* End< */

    /* assign the tmp_gd to the gd again */
    gd = tmp_gd;

    /* Update the Graphdescription, to tell the Session the new INodes. */
    gd.getISession()->setGraphDescription(gd);

    /* return a copy of the session interface for the application */
    return;
  }

  /* If no session returns, throw an exception */
  throw RegistryException("No Session for sharing available ");
}

void ClientRegistry::registerGraphDescription(GraphDescription& gd)
  throw ( const RegistryException& ) 
{
  list<RegistryName> registry_list;
  IServerRegistry* registry = 0;

  registry_list = getBestRegistry(gd);
  /* Some Error handling: If the desired registry is not available we throw an exception */
  if ( registry_list.empty() ) {
    throw RegistryException ("No ServerRegistry available that can be searched for sessions" );
  }
  
  /* init the session and return the session interface */
  ISession* session = 0;

  /* Try to request the graph from one of the registries */
  list<RegistryName>::iterator it = registry_list.begin(); 
  
  /* Get the corresponding server registry */
  registry = getServerRegistry(*it);

  /* Begin> init the session */
  if ( !registry ) {
    throw RegistryException(string("No ServerRegistry with name ") + it->first + string(" available."));
  }
  
  GraphDescription tmp_gd(gd);
  /* init the session for the graphdescription */
  session = registry->initSession(tmp_gd, UserInfo(), 0);
  
  if (session) {
    delete session;
    gd = tmp_gd;
    return;
  }
  
  throw RegistryException("Error while init session for GraphDescription");
}

void ClientRegistry::requestRemainingNodes(GraphDescription& gd)
  throw (const RegistryException& )
{
  if ( gd.getISession() ) {
    ISessionIt it = __session_map.find(gd.getISession()->getParentObject());
    if ( it == __session_map.end() )
      throw RegistryException("Internal error in ClientRegistry: Graphdescription does not maintain a Session!");
    try{
    /* Begin> Request the remaining nodes */
    for (unsigned int i = 0; i < gd.getNodeDescriptionList().size(); ++i) {
      if (! gd.getINode( *gd.getNodeDescriptionList()[i] ) ) {
        cout << "Must request " << gd.getNodeDescriptionList()[i]->getNodeName() << endl;

        /* Check wheter the node is declared as SHARED ... */
        if ( gd.getNodeDescriptionList()[i]->getSharingType() == NodeDescription::SHARED ) {
          /* Then we must not allocate it. */
          throw RegistryException(string ("Node " ) + gd.getNodeDescriptionList()[i]->getNodeName() + string(" is requested as SHARED, but can not be requestes as SHARED" ));
        }

        list<Response> resp;
        if ( gd.getNodeDescriptionList()[i]->isLocationEnabled() ){
          cout << "Try to request Node from Host " << gd.getNodeDescriptionList()[i]->getLocation() << endl;
        }

        /* init the request */
        resp = initRequest(*gd.getNodeDescriptionList()[i]);
        if(resp.empty()) {
          /* release the already requested nodes */
          string node_name = gd.getNodeDescriptionList()[i]->getNodeName();
          releaseGraph(gd);
          throw RegistryException(string("Node ") + node_name + string(" is not available"));
        }

        INode* inode = requestNode(resp.front(), gd.getNodeDescriptionList()[i]);

        /* store the node inside the graphdescription */
        gd.setINode(*gd.getNodeDescriptionList()[i], *inode);

        if ( gd.getNodeDescriptionList()[i]->getSharingType() == NodeDescription::EXCLUSIVE
          || gd.getNodeDescriptionList()[i]->getSharingType() == NodeDescription::EXCLUSIVE_OR_SHARED
          || gd.getNodeDescriptionList()[i]->getSharingType() == NodeDescription::SHARED_OR_EXCLUSIVE
           )
        {
          /* We request the Node Exclusive */
          gd.getNodeDescriptionList()[i]->setSharingType(NodeDescription::EXCLUSIVE);
        } else {
          /* Otherwise we request it EXCLUSIVE_THEN_SHARED */
          gd.getNodeDescriptionList()[i]->setSharingType(NodeDescription::EXCLUSIVE_THEN_SHARED);
        }

        /* store the node inside the session list, because we must release it.*/
        it->second.node_list.push_back(inode);
      }
    }
    } catch (const Exception& e) {
      ERROR_STREAM(e << endl);
      throw RegistryException(e.getComment());
    }
    /* End< */
  } else {
      throw RegistryException("Internal error in ClientRegistry: No Session object found");
  }
}

list<RegistryName> ClientRegistry::getBestRegistry(GraphDescription& gd)
    throw (const RegistryException&)
{
  map<RegistryName, int> count_map;
  /* Initialize the map */
  for (unsigned int i = 0; i < gd.getNodeDescriptionList().size(); ++i) {
    if ( ! gd.getNodeDescriptionList()[i]->getLocation().empty() ) {
      count_map.insert(pair<RegistryName, int>(RegistryName(gd.getNodeDescriptionList()[i]->getLocation(), gd.getNodeDescriptionList()[i]->getPort()), 0));
    } else {
      count_map.insert(pair<RegistryName, int>(RegistryName(getHostname(), ClientRegistry::default_port), 0));
    }
  }

  /* Begin> Create a map with amount of nodes for each ServerRegistry.
            Include only nodes with sharing type != EXCLUSIVE */
  for (unsigned int i = 0; i < gd.getNodeDescriptionList().size(); ++i)
  {
    if ( ! gd.getNodeDescriptionList()[i]->getLocation().empty()
         && ( gd.getNodeDescriptionList()[i]->getSharingType() != NodeDescription::EXCLUSIVE )
       )
    {
      count_map[RegistryName(gd.getNodeDescriptionList()[i]->getLocation(), gd.getNodeDescriptionList()[i]->getPort())] += 1;
    } else {
      if ( gd.getNodeDescriptionList()[i]->getSharingType() != NodeDescription::EXCLUSIVE ) {
        count_map[RegistryName(getHostname(), ClientRegistry::default_port)] += 1;
      }
    }
  }
  /* End< */

  /* Begin> ServerRegistry with most nodes is the best */
  list<RegistryName> return_list;
  while ( ! count_map.empty() ) {
    int count = 0;
    map<RegistryName, int>::iterator value = count_map.end();

    for (map<RegistryName, int>::iterator it = count_map.begin(); it != count_map.end(); ++it){
      if (count <= it->second ) {
        count = it->second;
        value = it;
      }
    }
    /* Insert the registry in the list */
    if ( value != count_map.end() ) {
      return_list.push_back(value->first);
      /* erase it from count_map*/
      count_map.erase(value);
    }
  }
  /* End< */

  /* return the server registry */
  return return_list;
}



void ClientRegistry::releaseGraph(GraphDescription& gd) {
  /* Begin> Check whether the given graph description is used with a session */
  if (!gd.getISession()) {
    throw RegistryException("Could not release a Graph, that does not maintain a session");
  }

  ISessionIt it = __session_map.find(gd.getISession()->getParentObject());
  if (it == __session_map.end()) {
    throw RegistryException("Could not release a Graph, that does not maintain a session" );
  }

  DEBUG_STREAM( "Must delete a session " << endl );

  /* Begin> Create a list of INodes that were locally allocated by the ClientRegistry because it
   *        must release the nodes as well. If the Node has no IServerRegistry interface which
   *        refers to its maintaining ServerRegistry the ClientRegistry allocates the Node.
   */
  list<INode*> release_list;
  for (list<INode*>::iterator i = it->second.node_list.begin(); i != it->second.node_list.end(); ++i) {
    IServerRegistry_var tmp_reg( (*i)->getParentObject()->getRegistry() ) ;
    /* If no IServerRegistry was set, the ClientRegistry allocates the nodes */
    if ( !tmp_reg.get() )
      release_list.push_back(*i);
  }

  releaseTransaction();

  /* Tell the corresponding server registry thate the Session should be released */
  it->second.server_registry->releaseSession(*it->second.session, UserInfo());

  /* */
  for (list<INode*>::iterator i = release_list.begin(); i != release_list.end(); ++i) {
    releaseNode(**i);
  }

  /* remove the entry from session_map*/
  __session_map.erase(it);

  /* Clear the Graphdescription because the INodes can no longer be used */
  gd.clear();
  /* End< */
}

INode* ClientRegistry::requestCompositeNode(GraphDescription&)
        throw (const RegistryException&)
{
//  return __named_registry["localhost"].server_registry->requestCompositeNode(gd);
  return 0;
}


Interface* ClientRegistry::requestNMMObject(const ObjectDescription& desc)
{
  /* Begin> Search the ServerRegistry that should be used */
  IServerRegistry* registry = 0;
  if ( desc.getLocation().empty() )
    registry = getServerRegistry( RegistryName( getHostname(), ClientRegistry::default_port) );
  else
    registry = getServerRegistry( RegistryName( desc.getLocation(), desc.getPort()) );
  /* End< */

  /* If no one available, throw an exception */
  if ( !registry )
    throw RegistryException("Could not request SourceSynchronizer from unknown Registry " + desc.getLocation() );

  /* Allocate the object and return its generic Interface */
  return allocateNMMObject( registry, desc );
}

void ClientRegistry::releaseNMMObject(const Interface& ) {

}

Interface* ClientRegistry::allocateNMMObject(IServerRegistry* registry, const ObjectDescription& desc) {
  NMMObject* object = 0;
  Interface* interface = 0;

  /* Begin> if local optimization is allowed, allocate the object in our address space */
  if (isOptimizationAllowed(registry)) {
    DEBUG_STREAM( "Allocate object " << desc.getObjectName() << " in local address space" << endl);
    if ( desc.getObjectName() == "MultiAudioVideoSynchronizer") {
      object = new MultiAudioVideoSynchronizer() ;
    } else if ( desc.getObjectName() == "MasterSlaveSourceSynchronizer") {
      object = new MasterSlaveSourceSynchronizer();
    }

    interface = object->getInterface<Interface>();
  }
  /* End< */
  /* Begin> Otherwise, the server registry should allocate the object */
  else {
    MutexGuard guard(__member_mutex);
    interface = registry->requestNMMObject(desc);
    object = interface->getParentObject();
  }
  /* End */

  /* store the allocated pointer for memory management */
  __object_map.insert(pair<NMMObject*, IServerRegistry*>(object, registry));

  /* return the interface pointing to the object */
  return interface;
}


bool ClientRegistry::isOptimizationAllowed(const IServerRegistry* registry) const {
  for (CRegistryIt it = __named_registry.begin(); it != __named_registry.end(); ++it)
    if (it->second.server_registry == registry)
      if ((
         SPluginManager::getInstance().getInterfaceList().front()->isSystemTo(*it->second.server_registry) ||
         SPluginManager::getInstance().getInterfaceList().front()->isLocalTo(*it->second.server_registry)
       ) && it->second.is_optimization_allowed
      )
        return it->second.is_optimization_allowed;

  return false;
}

void ClientRegistry::setOptimization(bool b, const RegistryName& registry_name ) {
  RegistryIt reg_it = __named_registry.find( registry_name );
  if (reg_it == __named_registry.end() && !registry_name.first.empty()) {
    try {
      addRegistry(TCPAddress(registry_name.first, registry_name.second));
    } catch (const Exception& e) {
      ERROR_STREAM ( e << endl);
    }
  }

  /* Begin> If no specific server registry is set, disbale the optimization for all ServerRegistries */
  if (registry_name.first.empty()) {
    for (RegistryIt it = __named_registry.begin(); it != __named_registry.end(); ++it)
      it->second.is_optimization_allowed = b;
  } else {
    RegistryIt it = __named_registry.find( registry_name );
    if (it != __named_registry.end() )
      it->second.is_optimization_allowed = b;
    else {
      ERROR_STREAM("Can not disable optimization on unknown ServerRegistry " << registry_name.first << " : " << registry_name.second << endl);
    }
  }
}

RegistryName ClientRegistry::getRegistryName( const IServerRegistry& registry ) const {
  for (map<RegistryName, RegistryStruct>::const_iterator it = __named_registry.begin(); it != __named_registry.end(); ++it)
  {
    if (it->second.server_registry->getParentObject() == registry.getParentObject())
      return it->first;
  }
  return RegistryName();
}

ClientRegistry::ISessionStruct::~ISessionStruct() {
  delete session;
  //server_registry must not be deleted, because we use a pointer from the __named_registry map
}



const int ClientRegistry::default_port = 22801;

}//end namespace


