/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef NMM_FACOTRYOBJECT_HPP
#define NMM_FACTORYOBJECT_HPP

#include "nmm/base/Node.hpp"

namespace NMM {


/** \class FactoryObject
  * A FactoryObject register themself by the local Registry object and provides
  * an interface to create and destroy a specific node object.
  */
class FactoryObject {
public: 
/** @name Constructor & Destructor
  */
//@{
/** constructor
  * \param plugin_name name unique node name which is normaly the class name
  * \param node_node Name of the node which is the class-name 
  */
  FactoryObject(const string& plugin_name, const string& node_name)
    : __plugin_name(plugin_name), __node_name(node_name) {}

/** destructor
  */
  virtual ~FactoryObject() {}
//@}

/** @name Factory methods
  */
//@{
/** abstract method which creates an instance of a node
  */
  virtual  Node* createNode()         = 0;

/** abstract method which releases a specific node
  */
  virtual void  releaseNode(Node*)    = 0;

/** This virtual copy constructor creates a copy of the FactoryObject.
  *
  * The caller of this method is responsible to delete the copy
  * \return Pointer to a copy of this FactoryObject.
  */
  virtual FactoryObject* copy()        const = 0;
//@}

/** @name Node information
  */
//@{
/** Returns the node type which can be instantiated by this class
  */
  inline const string& getNodeName()   const {return __node_name;}

/** Returns the unique node name which can be instantiated by this class.
  *
  * Normally this is the class name
  * \return string which contains the unique node name
  */
  inline const string& getPluginName() const {return __plugin_name;}
//@}
private:
  string __plugin_name;   // stores the plugin name
  string __node_name;     // stores the node name
};


/** \class TFactoryObject
  * A TFactoryObject (TemplateFactoryObject) provides a generic implementation of class
  * NodeFactory.
  *
  * This class can be used for all standard (software-based) Nodes which uses a default constructor.
  */
template <class T>
class TFactoryObject : public FactoryObject {
public:
/** @name Constructor & Destructor
  */
//@{
/** constructor
  * \param plugin_name name unique node name which is normaly the class name
  * \param node_node Name of the node which is the class-name
  */
  TFactoryObject(const string& plugin_name, const string& node_name) : FactoryObject(plugin_name, node_name) {};

/** destructor
  */
  virtual ~TFactoryObject(){};
//@}

/** @name Factory methods
  */
//@{
/** This method creates an instance of a node by calling new T();
  */
  virtual Node* createNode()                  { T* tmp_object = new T(); object_list.push_back(tmp_object); return tmp_object;}

/** This method destroyes a node
  */
  virtual void  releaseNode(Node* node)       { object_list.remove(node); delete node; }

/** This virtual copy constructor creates a copy of the FactoryObject.
  *
  * The caller of this method is responsible to delete the copy
  * \return Pointer to a copy of this FactoryObject.
  */
  virtual FactoryObject* copy()         const { return new TFactoryObject<T>(*this); }
//@}
private:
  list<Node*> object_list;
};


} //end namespace
#endif
