/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *                    Marco Lohse        <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "NodeDescription.hpp"

namespace NMM {
NodeDescription::NodeDescription()
  : __internal_id(-1), __sharing_type(NodeDescription::EXCLUSIVE_OR_SHARED), __config_events(5)
{
  __config_events.useTValue(false);

  for (unsigned int i = 0; i< __config_events.size(); ++i)
    __config_events[i] = 0;

  setTypeName("NodeDescription");
  setNamespace("NMM");
  registerAttributes();
  setNodeName("");
  setNodeType(INode::UNKNOWN);
  setURL("");
}

NodeDescription::NodeDescription(const string& node_name)
  : ObjectDescription(node_name), __internal_id(-1), __sharing_type(NodeDescription::EXCLUSIVE_OR_SHARED), __config_events(5)
{
  __config_events.useTValue(false);

  for (unsigned int i = 0; i< __config_events.size(); ++i)
    __config_events[i] = 0;

  setTypeName("NodeDescription");
  setNamespace("NMM");
  registerAttributes();
  setNodeType(INode::UNKNOWN);
  setURL("");
}



NodeDescription::~NodeDescription() 
{
}

NodeDescription::NodeDescription(const NodeDescription& nd)
  : ObjectDescription(nd), __internal_id(nd.getInternalID()), __config_events(5)
{
  __config_events.useTValue(false);

  for (unsigned int i = 0; i< __config_events.size(); ++i)
    __config_events[i] = 0;

  registerAttributes();
  *this = nd;
}

bool NodeDescription::allowsSharing() const
{
  if(__sharing_type.getValue() == NodeDescription::EXCLUSIVE_OR_SHARED || 
     __sharing_type.getValue() == NodeDescription::SHARED_OR_EXCLUSIVE || 
     __sharing_type.getValue() == NodeDescription::SHARED ||
     __sharing_type.getValue() == NodeDescription::EXCLUSIVE_THEN_SHARED ) {
    return true;
  }
  else {
    return false;
  }
}

bool NodeDescription::isSubset(const NodeDescription& nd) const 
{
  /*Begin> Check isSubset in super class */
  if(ObjectDescription::isSubset(nd) == false) {
    return false;
  }
  /*End< */

  /*Begin> Compare the NodeType if it is set*/
  if (nd.getNodeType() != INode::UNKNOWN) {
    if (getNodeType() != nd.getNodeType()) {
      return false;
    }
  }
  /*End< */

  /*Begin> Check the input properties*/
  if (!isPropertySubset(getInputProperty(), nd.getInputProperty())) {
    return false;
  }
  /*End< */

  /*Begin> Check the output properties*/
  if (!isPropertySubset(getOutputProperty(), nd.getOutputProperty())) {
    return false;
  }
  /*End< */

  /* If all parameters match return true */
  return true;
}

bool NodeDescription::isPropertySubset(const Property& own_property, const Property& property) const 
{
  for (list<Format*>::const_iterator it = property.getFormatList().begin(); it != property.getFormatList().end(); ++it)
  {
    if (!own_property.isFormatIncluded(**it, (*it)->getJackTag()))
      return false;
  }
  return true;
}

void NodeDescription::registerAttributes() {
  insertArgument(&__url);
  insertArgument(&__node_type);
  insertArgument(&__input_property);
  insertArgument(&__output_property);
  insertArgument(&__internal_id);
  insertArgument(&__sharing_type);
  insertArgument(&__config_events);
}

void NodeDescription::addNodeEvent(const Event& event, INode::State state) {
  unsigned int pos = 0; //stores the position inside the cevent
  switch (state) {
    case INode::CONSTRUCTED:
      pos = 0;
      break;
    case INode::INITIALIZED:
      pos = 1;
      break;
    case INode::OUTPUT_INITIALIZED:
      pos = 2;
      break;
    case INode::ACTIVATED:
      pos = 3;
      break;
    case INode::STARTED:
      pos = 4;
      break;
  }

  if (__config_events[pos] == 0)
    __config_events[pos] = new CEvent();

  __config_events[pos]->insertEvent(__config_events[pos]->size(), event);
}

CEvent* NodeDescription::getNodeEvent(INode::State state) {
  switch (state) {
    case INode::CONSTRUCTED:
      return __config_events[0];
    case INode::INITIALIZED:
      return __config_events[1];
    case INode::OUTPUT_INITIALIZED:
      return __config_events[2];
    case INode::ACTIVATED:
      return __config_events[3];
    case INode::STARTED:
      return __config_events[4];
  }

  return 0; //to avoid warnings
}

const CEvent* NodeDescription::getNodeEvent(INode::State state) const{
  switch (state) {
    case INode::CONSTRUCTED:
      return __config_events[0];
    case INode::INITIALIZED:
      return __config_events[1];
    case INode::OUTPUT_INITIALIZED:
      return __config_events[2];
    case INode::ACTIVATED:
      return __config_events[3];
    case INode::STARTED:
      return __config_events[4];
  }

  return 0; //to avoid warnings
}

TConstructor<NodeDescription>    __nmm__nodedescription_info("NodeDescription", "NMM");
TValueConstructor< NodeDescription::SharingType > __NodeDEscription_SharingType(TypeInfo("NodeDescription::SharingType", "NMM"));



NMM::OValueStream& operator<<(NMM::OValueStream& ov, const NMM::NodeDescription::SharingType& desc) {
  ov.beginType(NMM::TypeInfo("NodeDescription::SharingType", "NMM"));
  ov << reinterpret_cast<const int&>(desc);
  ov.endType();
  return ov;
}

NMM::IValueStream& operator>>(NMM::IValueStream& iv, NMM::NodeDescription::SharingType& desc){
  if(iv.getCurrentTypeInfo() != NMM::TypeInfo("NodeDescription::SharingType", "NMM")) {
    throw InvalidTypeException(NMM::TypeInfo("NodeDescription::SharingType", "NMM"), iv.getCurrentTypeInfo());
  }
  iv.goDown();
  iv >> reinterpret_cast<int&>(desc);
  iv.goUp();
  return iv;
}

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const list<NMM::NodeDescription>& desc) {
  return writeList<NMM::NodeDescription>(ov, desc, "NodeDescription");
}

NMM::IValueStream& operator>>(NMM::IValueStream& ov, list<NMM::NodeDescription>& string_list) {
  return readList<NMM::NodeDescription>(ov, string_list, "NodeDescription");
}

} //end namespace


NMM::OValueStream& operator<<(NMM::OValueStream& ov, const vector<NMM::NodeDescription>& desc) {
  return writeVector<NMM::NodeDescription>(ov, desc, "NodeDescription");
}

NMM::IValueStream& operator>>(NMM::IValueStream& ov, vector<NMM::NodeDescription>& string_list) {
  return readVector<NMM::NodeDescription>(ov, string_list, "NodeDescription");
}

