/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef NMM_NODEFACTORY_HPP
#define NMM_NODEFACTORY_HPP

#include "NodeFactoryStructs.hpp"

namespace NMM {
class FactoryException : public Exception {
public:
  FactoryException(const string& s) : Exception(s) {};
};

class Node;
class FactoryObject;
class Dll;

/** This abstract class is a blass class for all NodeFactories used in NMM.
  * A NodeFactory is used to create and deletes Nodes which are reserved by a Registry.
  * This base class defines abstract method to create and destroy nodes. It also implements
  * methods that can be used to manage several NodeFactories. NodeFactories are identified
  * by a NodeFactoryID that stores their class name.
  */
class NodeFactory : public NamedObject{
public:
/** @name Constructor & Destructor
  */
//@{
/** Constructor
  * \param name of the NodeFactory, normally the class name
  * \param parent Pointer of the parent NodeFactory
  */
  NodeFactory(const char* name, NodeFactory* parent = 0);

/** Destructor
  */
  virtual ~NodeFactory();
//@}

/** @name factory methods
  * These methods can be used to create and destroy Nodes
  */
//@{
/** This method creates a Node from a given FactoryTicket. If the node can not be created for some reason
  * an exception is thrown. The FactoryTicket is creates by a Registry, if the node is reserved. After
  * this method the FactoryTicket is invalid and can not be used again.
  * \param ticket The ticket is created by a registry and allows the maintainer to create the node.
  * \return A refernce to the new Node
  */
  virtual  Node& createNode (const FactoryTicket& ticket)
    throw (const FactoryException& ) = 0;

/** This method releases a creates node. The NodeFactory can delete this node or store it for a later use.
  * \param node The node that should be released
  */
  virtual  void  releaseNode (const Node& node)
    throw (const FactoryException& ) = 0;
//@}

/** @name NodeFactory management
  * These methods are for internal use 
  */
//@{
/** This method  returns the NodeFactoryID of the NodeFactory. This ID contains the name of the NodeFactory
  * which is used to identify the Factory
  */
  virtual const NodeFactoryID& getFactoryID() const {return __factory_id;}

/** This method adds a new child NodeFactory
  */
  void addNodeFactory(NodeFactory& factory);

/** This method removes a child NodeFactroy
  */
  void removeNodeFactory(NodeFactory& factory);

/** This mehtod searches a NodeFactory for a given NodeFactoryID
  * \param factory_id The NodeFacotryID of the requested NodeFactory
  */
  NodeFactory* findNodeFactory(const NodeFactoryID& factory_id);

/** This method returns all children which are registered by this NodeFactory
  */
  const list<NodeFactory*>& getChildList()    const {return __child_list;}

/** This method returns a pointer to the parent NodeFactory
  */
  const NodeFactory* getParent()              const {return __parent;}

/** This method returns a pointer to the parent NodeFactory
  */
        NodeFactory* getParent()                    {return __parent;}

/** This method sets the parent NodeFactory.
  */
  void  setParent(NodeFactory* factory);
//@}
protected:
/** This method sets the valid flag inside a FactoryTicket to true
  */
  inline void setTicketInvalid(const FactoryTicket& fp) const  {fp.setValid(false); }
private:
  NodeFactory*        __parent;                 ///< pointer to the parent NodeFactory
  list<NodeFactory*>  __child_list;             ///< list of children NodeFacotries
  NodeFactoryID       __factory_id;             ///< The NodeFactoryID with the name of the NodeFactoryID
};

} //end namespace
#endif
