/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef NMM_PLUGINMANAGER_HPP
#define NMM_PLUGINMANAGER_HPP

#include "LTDll.hpp"

#include "nmm/interfaces/base/registry/IPluginManagerImpl.hpp"
#include "nmm/base/StdNamedObject.hpp"

#include <map>
#include <list>
#include <string>

namespace NMM {
class Registry;
class NodeFactory;
class Exception;
class Plugin;

/** \class PluginManager
  * The PluginManager loads libraries which contain NMM-Nodes that can be use in an application.
  * To find the plugins in a library, each library export a global variable of name "nmm_plugins"
  * that is an array of plugins. This array is NULL-terminated. A plugin search and register nodes
  * that can be used on the system.
  */
class PluginManager : public IPluginManagerImpl, public StdNamedObject {
friend class CreateUsingNew<PluginManager>;
public:

/** @name Local methods
  * This methods can not be accessed through an interface, because they can only be used from local objects.
  */
//@{
/** This method returns the array of plugins from a specific library. If the library is not load the
  * return value is NULL.
  * \param library The name of the library
  */
  Plugin** requestPluginArray(const string& library);

/** This method returns the array of plugins from a specific library. If the library is not load the
  * return value is NULL.
  * \param library The name of the library
  */
  Plugin** const  requestPluginArray(const string& library) const;

/** This method returns true if the plugin is load, otherwise false.
  * \param library The name of the library
  * \so loadLibrary
  */
  bool isPluginUsed(const string& library )          const;

/** This method returns true if the factory part is active, so the nodes in a library can be
  * created by a NodeFactory.
  * \param library The name of the library
  */
  bool isFactoryPartActive(const string& library )   const;

/** This method returns true if the registry part is active, so the nodes in a library are managed
  * by a Registry.
  * \param library The name of the library
  */
  bool isRegistryPartActive(const string& library)   const;

  void defineObjectIDs(const string& plugin, const vector<list<ObjectID> >& id_list);

  int getPluginSize(const string& plugin)            const;
//@}
protected:
/** @name Get and set parameters.
  * This methods can be accessed through an interface.
  */
//@{
/** This method returns the list od ObjectID's that are set in the i'th plugin in a library.
  * The list is only valid if the plugin was registered to a Registry
  * \param library the name of the library
  * \param index the i'th plugin
  */
  virtual list<ObjectID> getPluginID(const string& library, const unsigned int& index );

/** This method returns a list of all liraries that can be load.
  */
  virtual list<string> getPlugins()          const;

/** This method returns a list of libraries that are currently in use.
  */
  virtual list<string> getUsedPlugins()      const;

/** This method returns a list of pathes that are searched for libraries.
  */
  virtual list<string> getPluginPath()       const;

/** This method adds library path which is searched if a library is load. The path must be a
  * valid Unix path on the current system.
  */
  virtual void addPluginPath(const string& path);
//@}


/** @name Load/remove libraries
  * This methods can be accessed through an interface.
  */
//@{
/** This method loads a library, register it with a Registry by calling activateRegisryPart()
  * and with a NodeFactory by calling activateFactoryPart().
  * 
  * \param library The name of the library that should be load
  * \sa activateRegistryPart()
  * \sa activateFactoryPart()
  */
  virtual void loadLibrary(const string& library)
    throw (const Exception& );

/** This method loads a library.
  * \param library The name of the library that should be load
  * \sa activateRegistryPart()
  * \sa activateFactoryPart()
  */
  virtual void loadPlugin(const string& library)
    throw (const Exception& );

/** This method loads all libraries in a given directory. 
  */        
  virtual void loadExternalPlugins(const string& path)
    throw (const Exception& );

/** This method removes all libraries in a given directory. 
  */        
  virtual void removeExternalPlugins(const string& path)
    throw (const Exception& );
    
/** This method removes a library with given name. This method calls the method
  * deactivateRegistryPart() and deactivateFactoryPart().
  * \param library The name of the library that should be removed
  * \sa deactivateRegistryPart()
  * \sa deactivateFactoryPart()
  */
  virtual void removeLibrary(const string& library)
    throw (const Exception& );

/** This method removes the plugins od a library from the registry.
  * \param library The name of the library that should be removed from a registry
  */
  virtual void deactivateRegistryPart(const string& library)
    throw (const Exception&);

/** This method removes the plugins of a library from the NodeFactory.
  * \param library The name of the library that should be removed from a NodeFactory
  */
  virtual void deactivateFactoryPart(const string& library)
    throw (const Exception&);

/** This method registers the plugins of library with a Registry.
  * \param library The name of the library that should be registered with a Registry
  */
  virtual void activateRegistryPart(const string& library)
    throw (const Exception&);

/** This method registers the plugins of library with a Registry.
  * \param library The name of the library that should be registered with a NodeFactory
  */
  virtual void activateFactoryPart(const string& plugin)
    throw (const Exception&);
//@}

private: //methods
/** initialize the Registry part. The plugin must be loaded
  */
  void activateRegistryPart(const string& plugin, Registry& registry)
    throw (const Exception& );

/** initialize the NodeFactory part
  */
  void activateFactoryPart(const string& plugin, NodeFactory& node_factory)
    throw (const Exception& );

/** reads the content of a directory
  */
  list<string> readFolder(const string&  folder)  const
    throw (const Exception& );

private:
/** @name Constructor & Destructor
  */
//@{
/** Constructor
  */
  PluginManager();

/** Destructor
  */
  virtual ~PluginManager();

/** Disabled Copy-Constuctor to avoid the instantiation of a second PluginManager
  */
  PluginManager(const PluginManager& );

  void applyDefinedObjectIDs(const string& plugin);

//@}

private:
/** private struct which stores the Dll object, the corresponding Registry and
  * NodeFactroy
  */
  struct PluginStruct {
    PluginStruct() : dll_object(0), reg_active(0), fac_active(0), plugins(0) {}
    LTDll*       dll_object;                    ///< The LTDll object that loads the library
    Registry*    reg_active;                    ///< Stores the Registry which is used to register the plugins
    NodeFactory* fac_active;                    ///< Stores the NodeFactory which is used to create the NodeFactory
    Plugin**     plugins;                       ///< The plugins insde a library, the last element is NULL
    list<string> alias;                         ///< stores alias names of the plugin
  };

  typedef map<string, PluginStruct>::iterator InternalIt;
  typedef map<string, PluginStruct>::const_iterator CInternalIt;

  list<string>              __plugin_path;      ///< Stores a list of pathes wich are searched for libraries
  map<string, PluginStruct> __plugin_map;       ///< Stores the loaded libraries
  map<string, vector<list<ObjectID> > > __defined_id_map;
  
  static const string       __plugin_symbol;    ///< Stores the symbol "nmm_plugins", wich is used to search for plugins in a library
};

/* Typedef to access the PluginManager with the method call SPluginManager::getInstance() */
typedef Singleton<PluginManager, CreateUsingNew> SPluginManager;

}//end namespace
#endif
