/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <iostream>
#include "RMIndexVertex.hpp"

 namespace NMM {
   RMIndexVertex::RMIndexVertex(string dim_name,bool isleaf) : StdNamedObject("RMIndex") {
     d_name = dim_name;
     leaf = isleaf;
     index = map<QualityParameter, RMIndexVertex*>();
     vertices = map<QualityParameter, RMVertex*>();
     DEBUG_STREAM2("RMIndexVertex: " << dim_name << endl);
   }

   RMVertex* RMIndexVertex::getVertex(QualityParameter value) {
     if ( ! leaf ) {  // the vertex is not a leaf
       ERROR_STREAM2("RMIndexVertex: not at leaf level" << endl);
       return 0;
     }
     if ( vertices.find(value) == vertices.end()) { // there is no vertex with the specified value
       DEBUG_STREAM2("RMIndexVertex: no vertex with value "<< value << " at dimension " << d_name << endl);
       return 0;
     }

     return vertices[value];
   }

   RMVertex* RMIndexVertex::getLowerVertex(QualityParameter value) {
     if ( ! leaf ) {  // the vertex is not a leaf
       return 0;
     }
     bool dummy = false;
     if ( vertices.find(value) == vertices.end()) { // there is no vertex with the specified value
       vertices[value] = 0;                         // so we add a dummy
       dummy = true;
     }
     map<QualityParameter,RMVertex*>::iterator map_it = vertices.find(value);
     if (map_it == vertices.begin()) { // first value
       if (dummy) {  // delete dummy
         vertices.erase(value);
       }
       return 0;
     }
     map_it--;
     
     if (dummy) {  // delete dummy
       vertices.erase(value);
     }

     return map_it->second;
   }

   RMVertex* RMIndexVertex::getHigherVertex(QualityParameter value) {
     if ( ! leaf ) {  // the vertex is not a leaf
       return 0;
     }
     bool dummy = false;
     if ( vertices.find(value) == vertices.end()) { // there is no vertex with the specified value
       vertices[value] = 0;                         // so we add a dummy
       dummy = true;
     }
     map<QualityParameter,RMVertex*>::iterator map_it = vertices.find(value);
     map_it++;
     if (dummy) {  // delete dummy
       vertices.erase(value);
     }
     if (map_it == vertices.end()) // first value
       return 0;

     return map_it->second;
   }


   RMIndexVertex* RMIndexVertex::getNextIndex(const QualityParameter value) {
   if ( leaf ) {  // the vertex is a leaf
       ERROR_STREAM2("RMIndexVertex: no next index in leaf" << endl);
       return 0;
     }
     if ( index.find(value) == index.end()) { // there is no index-vertex with the specified value
       DEBUG_STREAM2("RMIndexVertex: value "<< value << " not in index " << d_name << endl);
       return 0;
     }

     return index[value];
   }
   
   RMIndexVertex* RMIndexVertex::getLowerIndex(QualityParameter value) {
     QualityParameter dummy;
     return getLowerIndex( value, dummy);
   }

   RMIndexVertex* RMIndexVertex::getLowerIndex(QualityParameter value, QualityParameter&  lower_value) {
     if ( leaf ) {  // the vertex is a leaf
       return 0;
     }
     bool dummy = false;
     map<QualityParameter,RMIndexVertex*>::iterator map_it = index.find(value);
     if ( map_it == index.end()) { // there is no index-vertex with the specified value
       index[value] = 0;           // so we add a dummy
       dummy = true;
     }
     map_it = index.find(value);
     if (map_it == index.begin()) { // first value
       if (dummy) { // delete dummy
         index.erase(value);
       }
       lower_value = 0;
       return 0;
     }  
     map_it--;

     if (dummy) { // delete dummy
       index.erase(value);
     }
     
     lower_value = map_it->first;
     return map_it->second;
   }

   RMIndexVertex* RMIndexVertex::getHigherIndex(QualityParameter value) {
     QualityParameter dummy;
     return getHigherIndex( value, dummy);
   }
   
   RMIndexVertex* RMIndexVertex::getHigherIndex(QualityParameter value, QualityParameter&  higher_value) {
     if ( leaf ) {  // the vertex is a leaf
       return 0;
     }
     map<QualityParameter,RMIndexVertex*>::iterator map_it = index.find(value);
     bool dummy = false;
     if ( map_it == index.end()) { // there is no index-vertex with the specified value
       index[value] = 0;           // so we add a dummy
       dummy = true;
     }
     map_it = index.find(value);
     map_it++;
     if (dummy) { // delete dummy
       index.erase(value);
     }

     if (map_it == index.end()) { // no higher value
       higher_value = 0;
       return 0;
     }  

     higher_value = map_it->first;  
     return map_it->second;

   }

   Result RMIndexVertex::addLeaf(QualityParameter value,RMVertex* vertex) {
     DEBUG_STREAM2("RMIndexVertex: add leaf" << endl);
     if ( ! leaf) { // the vertex is not a leaf
       ERROR_STREAM2("RMIndexVertex: vertex is not a leaf!" << endl);
       return FAILURE;
     }

     if (vertices.find(value) != vertices.end()) {  // vertex with specified value is already existing
       return FAILURE;
     }

     vertices[value] = vertex;
     DEBUG_STREAM2("RMIndexVertex: vertex added" << endl);
     return SUCCESS;
   }

   Result RMIndexVertex::addIndex(QualityParameter value,RMIndexVertex* vertex) {
     DEBUG_STREAM2("RMIndexVertex: add index" << endl);
     if ( leaf) { // the vertex is a leaf
       ERROR_STREAM2("RMIndexVertex: vertex is a leaf!" << endl);
       return FAILURE;
     }
     if (index.find(value) != index.end()) {  // index with specified value is already existing
       return FAILURE;
     }

     index[value] = vertex;

     return SUCCESS;
   }
   
   Result RMIndexVertex::eraseIndex(QualityParameter value) {
     if ( leaf ) {
       vertices.erase(value);
     } else {  
       index.erase(value);
     }  
     return SUCCESS;
   }

   void RMIndexVertex::deleteVertices() {
     if ( isLeaf() ) { // on leaf-level we delete all vertices, on upper levels we first delete all indices
       // the vertices are stored in a list (which is member of RMMesh)
       // and are deleted on destruction of the RMMesh
       vertices.clear();
       DEBUG_STREAM2("all leaves deleted" << endl);
     } else {
       for (map<QualityParameter,RMIndexVertex*>::iterator i = index.begin(); i != index.end(); ++i) {
         DEBUG_STREAM2("deleting index " << d_name << " " << i->first << endl);
         i->second->deleteVertices();
       }
       index.clear();
       DEBUG_STREAM2("all indices deleted" << endl);
     }
   }

  void RMIndexVertex::printmap() {
    cerr << d_name << ":" << endl;
    if (leaf) {
      for (map<QualityParameter, RMVertex*>::iterator i = vertices.begin(); i != vertices.end(); ++i) {
        cerr << i->first << endl;
      }
    } else {
      for (map<QualityParameter, RMIndexVertex*>::iterator i = index.begin(); i != index.end(); ++i) {
        cerr << i->first << endl;
      }
    }  
  }
	
 }
	
