/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "RMVertex.hpp"
 
namespace NMM {

unsigned int RMVertex::id_counter = 1;

   RMVertex::RMVertex() : NMM::SerializedValue("RMVertex", "NMM")  {
     registerValues();
     dimensions = map<string,VEdge>();
   }

   RMVertex::RMVertex(list<string> &dim, bool dummy) : NMM::SerializedValue("RMVertex", "NMM") {
     registerValues();
     for (LI i = dim.begin(); i != dim.end(); ++i) {
       dimensions.getPValue()->insert(pair<string,VEdge>(*i, VEdge()));
     }
     if (!dummy) {
       id = id_counter++;
     } else {
       id = 0; // id 0 indicates a dummy node
     }
   }

   RMVertex::RMVertex(Format &description) : NMM::SerializedValue("RMVertex", "NMM") {
      registerValues();
      parameter_list desc_dim = description.getParameterList();

      typedef list<Parameter*>::iterator PLI;

      // the format-parameters are vertex-dimensions
      for (PLI i = desc_dim.begin(); i != desc_dim.end(); ++i) {
        Parameter::PType param_type = (*i)->getParameterType();
        string param_name = (*i)->getParameterName();
        switch (param_type) {
          case Parameter::INT32:
            dimensions.getRValue()[param_name] = VEdge(QualityParameter(static_cast<int>(description.getIntValue( param_name ))));
            break;
          case Parameter::FLOAT:
            dimensions.getRValue()[param_name] = VEdge(QualityParameter(description.getFloatValue( param_name )));
            break;
          case Parameter::STRING:
            dimensions.getRValue()[param_name] = VEdge(QualityParameter(description.getStringValue( param_name )));
          default:
            break;
        }
      }
      id = id_counter++;
    }


   RMVertex::RMVertex(const RMVertex& rmv)  : NMM::SerializedValue(rmv)  {
     registerValues();
     *this = rmv;
   }

   Result RMVertex::setProcessingTime(float val) {
     // FIXME recompute slopes
     resources.getRValue().setProcessingTime(val);
     return SUCCESS;
   }

   Result RMVertex::setBandwidth(float val) {
     resources.getRValue().setBandwidth(val);
     return SUCCESS;
   }

   Result RMVertex::setValue(string dim_name, QualityParameter val) {
     dimensions.getRValue()[dim_name].value = val;
     return SUCCESS;
   }

   Result RMVertex::setLower(string dim_name, RMVertex* vertex) {
     dimensions.getRValue()[dim_name].lower = vertex;
     return SUCCESS;
   }

   Result RMVertex::setHigher(string dim_name, RMVertex* vertex) {
     dimensions.getRValue()[dim_name].higher = vertex;
     return SUCCESS;
   }

   const list<string> RMVertex::getDimensions() {
     list<string> ret_val;

     for (MI i = dimensions.getRValue().begin(); i != dimensions.getRValue().end(); ++i) {
       ret_val.push_back(i->first);
     }
     return ret_val;
   }


   void RMVertex::registerValues() {
     insertArgument(&dimensions);   //register dimensions
     insertArgument(&resources);    //register cpu_usage
     insertArgument(&id);           //register id
   }

   void RMVertex::update(vector<RMVertex*>& vertex_index) {
     for (MI i = dimensions.getRValue().begin(); i != dimensions.getRValue().end(); ++i) {
       i->second.id2ptr(vertex_index);
     }
   }

   void RMVertex::prepare() {
     for (MI i = dimensions.getRValue().begin(); i != dimensions.getRValue().end(); ++i) {
       i->second.ptr2id();
     }
   }

   NMM::OValueStream& operator<<(NMM::OValueStream& ov, const map<string, RMVertex::VEdge>& string_map) {
    return writeMap<string, RMVertex::VEdge>(ov, string_map, string_type_info, TypeInfo("VEdge", "NMM"));
   }

   NMM::IValueStream& operator>>(NMM::IValueStream& ov, map<string, RMVertex::VEdge>& string_map) {
    return readMap<string, RMVertex::VEdge>(ov, string_map, string_type_info, TypeInfo("VEdge", "NMM"));
   }

   NMM::IValueStream& operator>>(NMM::IValueStream& iv, QualityParameter::QType& t)
   {
     iv >> reinterpret_cast<int&>(t);
     return iv;
   }

   NMM::OValueStream& operator<<(NMM::OValueStream& ov, const QualityParameter::QType& t)
   {
     ov << reinterpret_cast<const int&>(t);
     return ov;
   }

   NMM::OValueStream& operator<<(NMM::OValueStream& ov, const list<RMVertex>& vertex_list) {
     return writeList<RMVertex>(ov, vertex_list, "Vertex_Vector");
   }

   NMM::IValueStream& operator>>(NMM::IValueStream& iv, list<RMVertex>& vertex_list) {
     return readList<RMVertex>(iv, vertex_list, "Vertex_Vector");
   }

   TValueConstructor<map<string, RMVertex::VEdge> > __nmm_map_string_vertex ("map_string_VEdge", "std");
   TValueConstructor<list<RMVertex> > __nmm_list_vertex ("RMVertex_list", "std");
   TConstructor<RMVertex::VEdge > __nmm_vertex_vedge ("VEdge", "NMM");
   TConstructor<RMVertex > __nmm_vertex ("RMVertex", "NMM");

}
