/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef _NMM_RMVERTEX_HPP_
#define _NMM_RMVERTEX_HPP_

#include "nmm/base/format/Format.hpp"
#include "QualityParameter.hpp"
#include "Resources.hpp"
#include <map>
#include <string>
#include <list>
#include <vector>

namespace NMM {

/**
 * store cpu-usage for a specific format of a node
 */
class RMVertex : public NMM::SerializedValue {
  public:
    struct VEdge : public NMM::SerializedValue  {

      VEdge() :  NMM::SerializedValue("RMVEdge", "NMM")  {
        registerValues();
        lower = 0;
        lower_id = 0;
        higher = 0;
        higher_id = 0;
      }

      VEdge(const VEdge& ve)  : NMM::SerializedValue("RMVEdge", "NMM")  {
        registerValues();
        *this = ve;
      }

      VEdge(const QualityParameter val) : NMM::SerializedValue("RMVEdge", "NMM")  {
        registerValues();
        value = val;
        lower = 0;
        lower_id = 0;
        higher = 0;
        higher_id = 0;
      }

      VEdge* copy() const { return new VEdge(*this); } //virtual CopyConstructor

      VEdge& operator= (const VEdge& ve) {
        value = ve.value.getRValue();
        lower = ve.lower;
        lower_id = ve.lower_id.getRValue();
        higher = ve.higher;
        higher_id = ve.higher_id.getRValue();
        return *this;
      }

      NMM::TValue<QualityParameter>  value;
      RMVertex*                      lower;
      NMM::TValue<unsigned int>      lower_id;
      RMVertex*                      higher;
      NMM::TValue<unsigned int>      higher_id;

      void registerValues() {
        insertArgument(&value);      //register value
        insertArgument(&lower_id);   //register lower
        insertArgument(&higher_id);  //register higher
      }

      void id2ptr(vector<RMVertex*>& vertex_index) {
        lower = vertex_index[lower_id.getRValue()];
        higher = vertex_index[higher_id.getRValue()];
      }

      void ptr2id() {
        lower_id = 0;
        higher_id = 0;
        if (lower) {
          lower_id = lower->getID();
        }
        if (higher) {
          higher_id = higher->getID();
        }
      }

    }; // end VEdge

  private:
  
    friend OValueStream& operator<<(NMM::OValueStream& ov, const map<string, RMVertex::VEdge>& string_map);
    friend IValueStream& operator>>(NMM::IValueStream& iv, map<string, RMVertex::VEdge>& string_map);

    typedef list<string>::iterator LI;
    typedef map<string, VEdge>::iterator MI;
    typedef map<string, VEdge>::const_iterator CMI;

    static unsigned int id_counter;

  public:

    RMVertex();

    /*!
     * default constructor
     * \params dim a list of the quality dimensions for which all values are initialized with 0
     * \params dummy a dummy vertex has no id and no lower/higher pointers will point to it
     */
    RMVertex(list<string> &dim, bool dummy = false);

    /*!
     *
     */
    RMVertex(Format &description);

    /*!
     * copy constructor
     */
    RMVertex(const RMVertex&);

    /*!
     * \returns the dimensionality of the vertex
     */
    const unsigned int dimensionality() const {
      return dimensions.getRValue().size();
    }

    /*!
     * set the cpu-usage-value
     */
    Result setProcessingTime(float val);

    /*!
     * set the cpu-usage-value
     */
    Result setBandwidth(float val);

    /*!
     * set the value of a dimension
     */
    Result setValue(string dim_name, QualityParameter value);

    /*!
     * setup a pointer to a vertex that differs only by a lower value in the specified dimension
     * this is done while adding a vertex to a mesh
     */
    Result setLower(string dim_name, RMVertex* vertex);

    /*!
     * setup a pointer to a vertex that differs only by a lower value in the specified dimension
     * this is done while adding a vertex to a mesh
     */
    Result setHigher(string dim_name, RMVertex* vertex);

    /*!
     * \returns the value of the specified dimension
     */
    QualityParameter value(string dim) { return dimensions.getRValue()[dim].value.getRValue();}

    /*!
     * \returns a pointer to a vertex that differs only by a lower value in the specified dimension or 0 if there if no such vertex
     */
    RMVertex* lower(string dim) { return dimensions.getRValue()[dim].lower;}

    /*!
     * \returns a pointer to a vertex that differs only by a higher value in the specified dimension or 0 if there if no such vertex
     */
    RMVertex* higher(string dim) { return dimensions.getRValue()[dim].higher;}

    /*!
     * \returns the cpu-usage value
     */
//    const float getUsageValue() const { return cpu_usage.getRValue(); }
    const float getProcessingTime() const { return resources.getRValue().getAverageProcessingTime(); }

    /*!
     * \returns the cpu-usage value
     */
    const float getBandwidth() const { return resources.getRValue().getAverageBandwidth(); }

    /*!
     * \returns an alphabetically sorted list
     */
    const list<string> getDimensions();

    /*!
     * \returns the unique id of the vertex or 0 if its a dummy vertex
     */
    const unsigned int getID() const {
      return id.getRValue();
    }

    friend ostream& operator<< (ostream& stream, const RMVertex& rmv) {

      stream << " VERTEX: ";
      for (CMI i = rmv.dimensions.getRValue().begin(); i != rmv.dimensions.getRValue().end(); ++i) {
        stream << "<" << i->first << ": "
        << i->second.value.getRValue() << ">";
        const RMVertex* vpl = i->second.lower;
        if ( vpl != 0 ) {
          stream << "<-id:" << vpl->id.getRValue() << ">";
        }
        const RMVertex* vph = i->second.higher;
        if ( vph ) {
          stream << "<+id:" << vph->id.getRValue() << ">";
        }
      }
      stream <<  "<cpu_usage: "<< rmv.getProcessingTime() << "><bandwidth: " << rmv.getBandwidth() << "><id: " << rmv.id.getRValue() << ">";
      return stream;
    }

    // Serialization
    RMVertex*    copy()         const { return new RMVertex(*this); } //virtual CopyConsturctor

    RMVertex&    operator= (const RMVertex& v) {
      id = v.id.getRValue();
//      cpu_usage = v.cpu_usage.getRValue();
      resources = v.resources.getRValue();

      for (CMI i = v.dimensions.getRValue().begin(); i != v.dimensions.getRValue().end(); ++i) {
        dimensions.getRValue().insert(*i);
      }
      return *this;
    }

    void update(vector<RMVertex*>& vertex_index);
    void prepare();

  private:
    void registerValues();

   //The attributes
    NMM::TValue<map<string,VEdge> >         dimensions;
    NMM::TValue<unsigned int>               id;
//    NMM::TValue<float>                      cpu_usage;
    NMM::TValue<Resources>                  resources;

};

OValueStream& operator<<(NMM::OValueStream& ov, const map<string, RMVertex::VEdge>& string_map);
IValueStream& operator>>(NMM::IValueStream& iv, map<string, RMVertex::VEdge>& string_map);

OValueStream& operator<<(NMM::OValueStream& ov, const list<RMVertex>& vertex_list);
IValueStream& operator>>(NMM::IValueStream& iv, list<RMVertex>& vertex_list);
}

#endif
