/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <iostream>
#include <sys/time.h>
#include <sstream>
#include "ResourceCollector.hpp"
#include "nmm/base/Node.hpp"

//serialisation
#include "nmm/base/serialize/MagicNumberValueStream.hpp"
#include "nmm/base/serialize/StdStream.hpp"

namespace NMM {

  ResourceCollector::ResourceCollector() {
    // initialize members
    resource_probing_enabled = false;
    last_access_time.tv_sec = last_access_time.tv_usec     = 0;
    last_used_time.tv_sec   = last_used_time.tv_usec       = 0;
    last_system_time.tv_sec = last_system_time.tv_usec     = 0;
    percent_cpu_used        = percent_cpu_used_since_start = 0;
    bytes_received          = bytes_sent                   = 0;
    res_container                                          = 0;
    incoming_size           = outgoing_size                = 0;
    max_incoming_size       =                                0;
    processing_time         = cpu_processing_time          = 0;
    user_cpu_restriction                                   = 0.0;
    buffer_probe_counter    = buffer_counter               = 0;
    buffer_probe_interval   = 2;
    start_time   = Time(0);
    stop_time    = Time(0);
    running_time = Interval (0);
    node_ptr = 0;
    error_stream = debug_stream = warning_stream = 0;

    rm_mesh = 0;

    cpu_info = "";
    actual_format = 0;

  }

  ResourceCollector::~ResourceCollector() {
    if (res_container)
        delete res_container;
    if (actual_format)
        delete actual_format;
//    if (rm_mesh) {
//      delete rm_mesh;
//    }
  }

  Result ResourceCollector::enableResourceProbing() {

    resource_probing_enabled = true;

    // get the message streams
    if ( (node_ptr = dynamic_cast<Node*>(this))) {
      debug_stream   = node_ptr->debugStream2();
      error_stream   = node_ptr->errorStream2();
      warning_stream = node_ptr->warningStream2();

      if (debug_stream) {
        *debug_stream << "Resource Collector: pointer to Node initialized" << endl;
      }
    } else {
      node_ptr = 0;
      error_stream = debug_stream = warning_stream = 0;
    }

    return SUCCESS;
  }

  Result ResourceCollector::setStartTime() {
    gettimeofday(&last_access_time, NULL); // get actual time
    start_time = Time(last_access_time);         // and save it
    stop_time  = Time(0); // set stop time to 0;
    buffer_counter = 0;

    return SUCCESS;
  }

  Result ResourceCollector::setStopTime() {
    timeval temp_time;
    gettimeofday(&temp_time, NULL);
    stop_time = Time(temp_time);
    start_time = Time(0);
    running_time += stop_time - start_time;
    return SUCCESS;
  }

  Result ResourceCollector::disableResourceProbing() {
    resource_probing_enabled = false;

    return SUCCESS;
  }

  Result ResourceCollector::assignDefaultResourceContainer() {
    if (!res_container) {
        res_container =  new ResourceContainer;
        res_container->addThread(getpid());
    }

    return SUCCESS;
  }

  unsigned int ResourceCollector::getCPUUsage(const bool& since_start) {
    if (since_start) {
      Interval cput= Interval(getUsedCPUTime_i());
      Interval wct = Interval(getUsedTime_i());
      float wct_s = wct.nsec;
      wct_s /= 1000000000L;
      wct_s += wct.sec;
      float cput_s = cput.nsec;
      cput_s /= 1000000000L;
      cput_s += cput.sec;
      return (unsigned int) (100 * ( cput_s / wct_s));
    } else {
      return percent_cpu_used;
    }
  }

/*

  Result ResourceCollector::setSentBufferSize(size_t size) {
      bytes_sent += size;
    return SUCCESS;
  }

  unsigned long ResourceCollector::getSentBufferSize() {
    return bytes_sent;
  }

  Result ResourceCollector::setReceivedBufferSize(size_t size) {
      bytes_received += size;
    return SUCCESS;
  }
*/
  unsigned long ResourceCollector::getMaxReceivedBufferSize() {
    return max_incoming_size;
  }

  const timeval ResourceCollector::getUsedTime_i() {
    timeval temp_time,ret_val;


    if ((stop_time == Time(0)) && (start_time == Time(0))) { // the node is not running
      ret_val.tv_sec = running_time.sec;
      ret_val.tv_usec = (running_time.nsec * 1000);

      return ret_val;
    }

    if (stop_time == Time(0)) {
      // if the node is started and has not been stopped (=> it is running)
      // get actual time to calculate the time the is running
      gettimeofday(&temp_time, NULL);
    }

    // compute time passed since start
    Interval ret_time = Time(temp_time) - start_time;
    ret_time += running_time;
    ret_val.tv_sec = ret_time.sec;
    ret_val.tv_usec = (ret_time.nsec * 1000);

    return ret_val;
  }

  const timeval ResourceCollector::getUsedCPUTime_i() {
    timeval ret_val;

    ret_val.tv_sec = (res_container->getCPUUsage() / 100);
    ret_val.tv_usec = (res_container->getCPUUsage() - (100 * ret_val.tv_sec)) * 10;
    return ret_val;
  }

  void ResourceCollector::restrictCPUUsage(const float& level) {
    if ((level > 0.0) && (level < 1.0))
        user_cpu_restriction = level;
  }

  void ResourceCollector::manageCPULoad() {
//    float percent = ((float) cpu_processing_time / (float) processing_time);

//    if (percent < user_cpu_restriction) // limit reached?
//      return; // no

    Interval sleeped_interval = processing_time - cpu_processing_time;

    if (sleeped_interval < Interval(0)) {
      if (warning_stream) {
        *warning_stream << "Resource Collector: more cpu time used than wall clock time for a single thread !" << endl;
        *warning_stream << "Resource Collector: increasing buffer_probe_intervall !" << endl;
      }
      buffer_probe_interval++;
      return;
    }

    Interval sleep_interval = cpu_processing_time;
    sleep_interval *= ((1 / user_cpu_restriction) - 1);

    sleep_interval -= sleeped_interval;

    if ( (float) sleep_interval > 0) {
      timespec sleep_time = {sleep_interval.sec, sleep_interval.nsec};
      timespec rem_sleep_time; // requested and remaining sleep time

      const timespec* req_sleep_time = &sleep_time;
      if (debug_stream) {
        *debug_stream << "Resource Collector: sleeping to lower cpu load (" << (float) sleep_interval << " sec), total usage is set to " << ( 100 *user_cpu_restriction) << " %" << endl;
      }
      nanosleep( req_sleep_time, &rem_sleep_time);

    }
  }

  const timeval ResourceCollector::getSystemTime() {
    timeval retval;
    gettimeofday(&retval,NULL);
    return retval;
  }

  void ResourceCollector::setBufferProbeInterval(const unsigned int& n) {
    buffer_probe_interval = n;
  }

  Result ResourceCollector::setCPU(const string& ci) {
    cpu_info = ci;

    return SUCCESS;
  }

  Result ResourceCollector::loadMesh(const string& s_mesh) {
    if (s_mesh.size() == 0) {
      return FAILURE;
    }
    if (rm_mesh == 0) { // initialize an empty mesh;
      list<string> empty;
      empty.clear();
      rm_mesh = new RMMesh(empty);
    }

    stringstream w_stream(s_mesh);
//    w_stream << s_mesh;

    MagicNumberIValueStream miv;
    StdIStream i_s(&w_stream);
    miv.setIPartnerStream(&i_s);
    miv.readData();

    miv >> *rm_mesh;

    if (actual_format) {
      delete actual_format;
    }
    actual_format = new Format();
    rm_mesh->getFormat(*actual_format);

    return SUCCESS;
  }
  
  Result ResourceCollector::loadMeshes(const string& format_id, const string& s_mesh) {
    s_meshes.insert(pair<string,string>(format_id,s_mesh));
    return SUCCESS;
  }

  Result ResourceCollector::getMesh(string& s_mesh) {
    if (rm_mesh == 0) {
      return FAILURE;
    }

    stringstream w_stream;
    StdOStream stream(&w_stream);

    MagicNumberOValueStream ov;
    ov.setOPartnerStream(&stream);

    ov << *rm_mesh;
    ov.writeData();

    s_mesh = w_stream.str();

    return SUCCESS;
  }

  float ResourceCollector::getUsage(const Format& description) {
      if (debug_stream) {
        *debug_stream << "Resource Collector: getUsage for " << description <<  endl;
      }
    if ((rm_mesh == 0) || (&description == 0)) {
      return 0.0F;
    }
    Format rm_format;
    parseFormat(description, rm_format);

    return rm_mesh->getUsage(const_cast<Format*>(&rm_format));
  }

  float ResourceCollector::getBandwidth(const Format& description) {
      if (debug_stream) {
        *debug_stream << "Resource Collector: getBandwidth for " << description <<  endl;
      }
    if ((rm_mesh == 0) || (&description == 0)) {
      return 0.0F;
    }
    Format rm_format;
    parseFormat(description, rm_format);

    return rm_mesh->getBandwidth(const_cast<Format*>(&rm_format));
  }

  float ResourceCollector::getCPUperBuffer() {
    if (buffer_counter > 0) {
      return (getUsedCPUTime() / buffer_counter);
    }
    return 0;
  }

  Result ResourceCollector::getVertex(const Format& description, RMVertex& vertex) {
    if (rm_mesh == 0) {
      return FAILURE;
    }

    Format rm_format;
    parseFormat(description, rm_format);

    RMVertex* ret_vert = rm_mesh->getVertex(&rm_format);
    if (ret_vert) {
      vertex = *ret_vert;
      return SUCCESS;
    } else {
      return FAILURE;
    }
  }

  Result ResourceCollector::updateVertex(const Format& description, const float& cpu_usage, const float& bandwidth) {
    if (rm_mesh == 0) {
      if (error_stream) {
        *error_stream << "Resource Collector: updateVertex without loaded mesh" << endl;
      }
      return FAILURE;
    }

    Format rm_format;
    parseFormat(description, rm_format);

    RMVertex* vertex = rm_mesh->getVertex(&rm_format);
    if (vertex) {
      vertex->setBandwidth(bandwidth);
      vertex->setProcessingTime(cpu_usage);
    } else {
      return (rm_mesh->addVertex(&rm_format, cpu_usage, bandwidth));
    }
    return SUCCESS;
  }

  Result ResourceCollector::addVertex(const Format& description, const float& cpu_usage, const float& bandwidth) {
    Format rm_format;
    parseFormat(description, rm_format);

    if (rm_mesh != 0) {
      return (rm_mesh->addVertex(&rm_format, cpu_usage, bandwidth));
    } else {
      return FAILURE;
    }
  }

  Result ResourceCollector::setupMesh(const Format& dimensions) {
    if (rm_mesh != 0) {
      return FAILURE;
    }
    Format* new_format = new Format();
    *new_format = dimensions;

    setActualFormat(new_format);

    Format rm_format;
    parseFormat(dimensions, rm_format);

    // parse format to list<string>

    list<string> dims;
    const parameter_list desc_dim = rm_format.getParameterList();

    // the format-parameters are vertex-dimensions
    for (list<Parameter*>::const_iterator i = desc_dim.begin(); i != desc_dim.end(); ++i) {
      dims.push_back((*i)->getParameterName());
    }

    rm_mesh = new RMMesh(dims);
    rm_mesh->setFormat(*new_format);

    return SUCCESS;
  }

  Result ResourceCollector::savePoints(const string& filename) {
    return rm_mesh->savePoints(filename);
  }

  Result ResourceCollector::parseFormat(const Format& in_format, Format& out_format) {

    // make a copy of the input-format
    Format new_format = in_format;

    int aspect_x;
    int aspect_y;
    bool included = true;

    try {
      aspect_x = in_format.getIntValue("aspect_x");
      aspect_y = in_format.getIntValue("aspect_y");
    }
    catch (ParameterExceptionNotExist e) {
      included = false;
    }

    if ( included) {
      new_format.deleteParameter("aspect_x");
      new_format.deleteParameter("aspect_y");
      new_format.addFloatParamValue("aspect",aspect_x/aspect_y);
    }

    out_format = new_format;

    return SUCCESS;
  }

  float ResourceCollector::getActualUsage() {
    if (actual_format == 0) {
      if (error_stream) {
        *error_stream << "no actual format !!" << endl;
      }
      return 0.0F;
    }
    return getUsage(*actual_format);
  }

  float ResourceCollector::getActualBandwidth() {
    if (actual_format == 0) {
      if (error_stream) {
        *error_stream << "no actual format !!" << endl;
      }
      return 0.0F;
    }

    // first try to take saved measures
    if (getBuffersPerSecond() == 0) {
      return getBandwidth(*actual_format);
    }

    // calcuulate the bw
    float bw;
    bw = getActualBufferSize();
		float bps = getBuffersPerSecond();
		bw = (bw * bps) / 1024;
    return bw;
  }

  float ResourceCollector::getActualBufferSize() {
    if (actual_format == 0) {
      if (error_stream) {
        *error_stream << "no actual format !!" << endl;
      }
      return 0.0F;
    }
    float bw = getMaxReceivedBufferSize();
    if ( bw == 0.0F ) {
      bw = getBandwidth(*actual_format);
    }
    return bw;
  }

  void ResourceCollector::setActualFormat(Format* new_format) {
    if (actual_format) {
      delete actual_format;
    }

    actual_format = new_format;
  }

  bool ResourceCollector::getActualFormat(Format& ret_format) {
    ret_format = *actual_format;
    return true;
  }

  float ResourceCollector::getBuffersPerSecond() {
    Interval rt = getUsedCPUTime();
		if (rt != Interval(0) ) {
    	return (buffer_counter / rt);
		}
		return 0.0F;

  }


/********************* Interface only *****************************/

  Interval ResourceCollector::getUsedTime() {
    return Interval(getUsedTime_i());
  }
  Interval ResourceCollector::getUsedCPUTime() {
    return Interval(getUsedCPUTime_i());
  }

}
