/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef _NMM_RESOURCE_COLLECTOR_HPP_
#define _NMM_RESOURCE_COLLECTOR_HPP_

#include "nmm/interfaces/base/resourcemgr/IResourceCollectorImpl.hpp"
#include "ResourceContainer.hpp"
#include "RMMesh.hpp"

namespace NMM {

  /**
   * \class ResourceCollector
   * \brief This class implements resource-probing
   */
  class ResourceCollector : public IResourceCollectorImpl {

    bool resource_probing_enabled;

    ResourceContainer* res_container;

    Node* node_ptr;                     //!> pointer to the underlying node, used for debug/error output
    ostream* debug_stream;
    ostream* error_stream;
    ostream* warning_stream;

    Time start_time;                    //!> The last time the node was started
    Time stop_time;                     //!> The last time the node was stopped
    Interval running_time;              //!> The time the node was running before the last stop
    struct timeval last_access_time;    //!> The time of the last probeCPUUsage-call
    struct timeval last_used_time;      //!> CPU time used at the time of the last probeCPUUsage-call
    struct timeval last_system_time;    //!> System time used at the time of the last probeCPUUsage-call

    unsigned int percent_cpu_used;      //!> Percent of CPU-time used between last 2 calls to probeCPUUsage
    unsigned int percent_cpu_used_since_start; //!> Percent of CPU-time used between the start of the node and the last call to probeCPUUsage
    unsigned long bytes_received,bytes_sent;

    string cpu_info;

    Format* actual_format;
    
  protected:
    Result setStartTime();
    Result setStopTime();

    unsigned long incoming_size,max_incoming_size,outgoing_size;
    Time arrival_cpu_time,finishing_cpu_time;
    Time arrival_time,finishing_time;
    Interval processing_time,cpu_processing_time; //!> Amount of time the last incoming buffer needed to be processed
    float user_cpu_restriction;        //!> Manually set maximum cpu-time the node is allowed to use (0.0 = no limit)

    void manageCPULoad();              //!> Sleeps depending on the user_cpu_restriction
    const timeval getSystemTime();
    int buffer_probe_counter;
    int buffer_probe_interval;         //!> cpu usage is probed (and managed) after each buffer_probe_interval buffers

    //! /returns true if resource-probing is enabled
    inline const bool isResourceProbing() const {return resource_probing_enabled;}

    //! returns Amount of time the node used the CPU since it was started
    const struct timeval getUsedCPUTime_i();

    //! returns Amount of wall-clock time since the node was started
    const struct timeval getUsedTime_i();

    unsigned int buffer_counter;

    RMMesh* rm_mesh;

    void setActualFormat(Format*);

    map <string, string> s_meshes;

  public:
    //! The main constructor
    ResourceCollector();
    ~ResourceCollector();

    /*!
     * assign a default resource-container to the Node
     * ( called after thread_creation )
     */
    Result assignDefaultResourceContainer();

//    Result setSentBufferSize(size_t size);
//    unsigned long getSentBufferSize();
//    Result setReceivedBufferSize(size_t size);

    ResourceContainer* const getResourceContainer() const {
      return res_container;
    }

    
/*    Result deleteResourceContainer(ResourceContainer&);
    Result addToContainer(ResourceContainer&,pid_t);
    Result removeFromContainer(ResourceContainer&,pid_t);
    Resources probeContainer(ResourceContainer&);
*/

/******************* Interface ********************/

    Interval getUsedTime();
    Interval getUsedCPUTime();
    unsigned long getMaxReceivedBufferSize();

    void setBufferProbeInterval(const unsigned int&);

    //! enable resource-probing in generic layer (output in DEBUG_STREAM2)
    Result enableResourceProbing();

    //! disable resource-probing in generic layer
    Result disableResourceProbing();

    //! restrict the nodes cpu-usage to level (range 0 - 1)
    void restrictCPUUsage(const float& level);

    //! returns how much of the cpu was used (in percent)
    /*!
     * \param since_start : if true cpu-usage since starting the node is returned
     *                      if false result of last probeCPUUsage is returned
     */
    unsigned int getCPUUsage(const bool& since_start = false);

    float getCPUperBuffer();

    Result setCPU(const string& ci);

    /***** mesh */

    /**
     * load a previously serialized mesh
     */
    Result loadMesh(const string& s_mesh);
    Result loadMeshes(const string& format_id,const string& s_mesh);

    /**
     * output's the mesh as string
     */
    Result getMesh(string& s_mesh);


    float getUsage(const Format& description);
    float getBandwidth(const Format& description);

    Result getVertex(const Format& description, RMVertex& vertex);
    Result updateVertex(const Format& description, const float& cpu_usage, const float& bandwidth);

    Result setupMesh(const Format& dimensions);
    Result addVertex(const Format& description, const float& cpu_usage, const float& bandwidth);

    /**
     * save the measured points to a file readable by HnSRTree-2.0
     */
    Result savePoints(const string& filename);

    float getActualUsage();
    float getActualBandwidth();
    float getActualBufferSize();

    bool getActualFormat(Format& ret_format);

  private:
    /**
     * the connection format of a node needs to parsed before it can be used by the ResourceCollector
     * (the aspect-ratio gets changed)
     */
    Result parseFormat(const Format& in_format, Format& out_format);

    float getBuffersPerSecond(); // returns how many buffers are processed within 1 second of cpu-time


 }; //end class


}

#endif
