/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef _NMM_RESOURCE_MANAGER_HPP_
#define _NMM_RESOURCE_MANAGER_HPP_

#include "nmm/base/registry/ServerRegistry.hpp"
#include "nmm/interfaces/base/resourcemgr/IResourceManagerImpl.hpp"
#include "nmm/utils/NamedObject.hpp"
#include "nmm/utils/proc/CPUInfo.hpp"
#include "nmm/utils/iperf/IPerf.hpp"
#include "nmm/base/resourcemgr/RMConfigManager.hpp"
#include <vector>
#include <list>
#include <string>

namespace NMM {

//class ServerRegistry;

class ResourceManager : public NamedObject , public IResourceManagerImpl {

  public:

    ResourceManager();
    ~ResourceManager();

    /**
     * setup a pointer to the registry
     */
    void setRegistry(IRegistry* ptr);

    /**
     * set the number of usable network interfaces
     * this clears all previous reservations
     */
    void setNetworkInterfaces(const unsigned int&);

    /**
     * add a connected Node to reserve the needed cpu resources
     * \param interface the node will use this network interface
     * \param id the node has this id (needed to release the node)
     */
    Result addNode(const INode*, uint& id , const uint& interface = 0);

    /**
     * release a previously added node (and free its reserved resources)
     */
    Result releaseNode(const INode*, const uint& id);

    // cpu info
    void getCPUInfo(string& name, uint& bogo_mips, uint& num, float& free); ///< return infos about the cpu
    void getCPUDescription(string& description);                            ///< return a description about the cpu consisting of name and speed (used by the configuration manager)
    void getCPUFree(float& free);                                           ///< return how much of the cpu is free (by the system)
    void setCPUUpdateIntervall(const unsigned int& seconds);
    void getCPULeft(float& left);                                           ///< return how much of the cpu if free restpecting reservations

    /**
     * start an iperf-server on port 22802
     */
    void startServer(/*const unsigned int& port*/);

    /**
     * stop the iperf-server
     */
    void stopServer();

    /**
     * test the speed of a single network connection
     */
    float testConnection(const string& hostname);

    /**
     * test the speed of simultaneous network connections
     * \param hosts a list of hosts to test
     * \param speeds an empty list that will be filled with the measured speeds
     */
    void testConnections(const list<string>& hosts, list<float>& speeds/*, const uint& sport*/);

    /**
     * set which hosts are connected to our client-registry
     */
    Result setHosts(const list<string>& ihosts);

   /**
    *
    */
    void setReservationPending();

    /**
     *
     */
    void commitReservation();

    void getCPUSpeed(uint& bogo_mips, float& free);
    
    string getMesh(string node_name, string format_id);
    list<pair<string,string> > getMeshes(string node_name);
    void setMesh(string node_name, string format_id, string mesh);
    Result updateNodeInfo(const INode* node);

  private:
    static uint node_counter;

    unsigned int cpu_num;
    vector<float> free_cpu;
    vector<float> used_cpu;

    unsigned int interface_num;
    vector<float> free_net;
    vector<float> used_net;

    bool reservation_pending;

    bool iperf_server; ///< true if iperf-server is running
    uint iperf_port;

    IRegistry_var server_registry;  ///< pointer to the server-registry
    bool has_server_registry;       ///< indicates that the pointer to the server-registry was set

    RMConfigManager rmcm;
    list<string> hosts;           ///< map containing all hosts that should be conntacted (and their interface)

    struct RMNodeInfo {
      string name;
      uint on_cpu;
      float usage;
      uint interface;
      float bandwidth;
      uint id;
    };

    list<RMNodeInfo> running_nodes;

    CPUInfo  cpu_info;
    IPerf    i_perf;
};

}

#endif
