/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */



#include "ComplexValue.hpp"

namespace NMM {

/** SerializedValue implementation  **/

SerializedValue::SerializedValue(bool reconstruction_mode, Value::MemoryMgm memory_mgm)
  : Value(Value::COMPOUND,memory_mgm), __reconstruction_mode(reconstruction_mode)
{

}

SerializedValue::SerializedValue(const string& type_name, const string& ns, bool reconstruction_mode,  Value::MemoryMgm memory_mgm)
    : Value(Value::COMPOUND, memory_mgm), __type_info(type_name, ns), __reconstruction_mode(reconstruction_mode)
{

}


SerializedValue::SerializedValue(const TypeInfo& ti,  bool reconstruction_mode, Value::MemoryMgm memory_mgm)
    : Value(Value::COMPOUND, memory_mgm), __type_info(ti), __reconstruction_mode(reconstruction_mode)
{

}

SerializedValue::SerializedValue(const SerializedValue& sv)
  : Value(sv), __type_info(sv.getTypeInfo()), __reconstruction_mode(sv.__reconstruction_mode)
{
  //don't copy the vector from sv!
}


const SerializedValue& SerializedValue::operator=(const SerializedValue& sv) {
  Value::operator=(sv);
  return assign(sv);
}

const SerializedValue& SerializedValue::assign(const SerializedValue& sv) {
  __type_info = sv.getTypeInfo();
  return *this;
}

void  SerializedValue::assignFromStream(IValueStream& ivs) {
  if ( ivs.getCurrentTypeInfo() != getTypeInfo() ) {
    GLOBAL_ERROR_STREAM (" SerializedValue > Invalid type or namespace:" << endl);
    GLOBAL_ERROR_STREAM (" SerializedValue > Current type : " << ivs.getCurrentType() << "                Expected type: " << getTypeName() << endl);
    GLOBAL_ERROR_STREAM (" SerializedValue > Current namespace : " << ivs.getCurrentNamespace() << "      Expected namespace: " << getNamespace() << endl);
    ivs.nextElement();
    return;
  }

  if (!ivs.goDown()) {
    GLOBAL_ERROR_STREAM ("SerializedValue seems to be empty");
    ivs.nextElement();
    return;
  }

  unsigned int i = 0;
  while(!ivs.isEmpty()) {
    /** If the value already exist, use this one.
      */
    if (i < __arg_array.size()) {
      if (getArgument(i)) {
        getArgument(i)->assignFromStream(ivs);
      }
    } else {
      if (__reconstruction_mode) {
        Value* value = SValueFactory::getInstance().createValue(ivs.getCurrentTypeInfo());
        value->createFromStream(ivs);
        __arg_array.push_back(value);
      } else {
        GLOBAL_ERROR_STREAM("Runtime error> Stream has unhandled Elements with type name " << ivs.getCurrentType() << endl);
        ivs.nextElement();
      }
    }
    ++i;
  }

  ivs.goUp();

  updateAttributes();
}

void  SerializedValue::createFromStream(IValueStream& ivs) {
  if ( ivs.getCurrentTypeInfo() != getTypeInfo() ) {
    GLOBAL_ERROR_STREAM ("SerializedValue > Invalid type or namespace:" << ivs.getCurrentType() << endl);
    GLOBAL_ERROR_STREAM ("SerializedValue > Current type : " << ivs.getCurrentType() << "      Expected type: " << getTypeName() << endl);
    GLOBAL_ERROR_STREAM ("SerializedValue > Current namespace : " << ivs.getCurrentNamespace() << "      Expected namespace: " << getNamespace() << endl);
    ivs.nextElement();
    return;
  }

  if (!ivs.goDown()) {
    GLOBAL_ERROR_STREAM ("SerializedValue seems to be empty");
    ivs.nextElement();
    return;
  }

  unsigned int i = 0;
  while(!ivs.isEmpty()) {
    /** If the value already exist, use this one.
      */
    if (i < __arg_array.size()) {
      if (getArgument(i)) {
        getArgument(i)->createFromStream(ivs);
      }
    } else {
      if (__reconstruction_mode) {
        Value* value = SValueFactory::getInstance().createValue(ivs.getCurrentTypeInfo());
        value->createFromStream(ivs);
        __arg_array.push_back(value);
      } else {
        GLOBAL_ERROR_STREAM("Runtime error> Stream has unhandled Elements with type name " << ivs.getCurrentType() << endl);
        ivs.nextElement();
      }
    }
    ++i;
  }
  ivs.goUp();

  updateAttributes(); 
}


void  SerializedValue::toStream(OValueStream& ovs) const {
  /* Call the prepare method */
  const_cast<SerializedValue*>(this)->prepareAttributes();

  /* Start a new type */
  ovs.beginType(getTypeName(), getNamespace());

  /* Begin> Write the registered values to the stream */
  for (size_t i = 0; i< getSize(); ++i)
    if (getArgument(i))
      getArgument(i)->toStream(ovs);
    else {
      GLOBAL_ERROR_STREAM2("Could not serialze 0ptr in SerializedValue " << getTypeName() << endl);
    }
  /* End< */

  /* End the new type */
  ovs.endType();
}


void  SerializedValue::releaseValue()           const {
  if (getMemoryStrategy() == INVALUE)
    delete this;
}

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const list<string>& string_list) {
  return writeList<string>(ov, string_list, "string");
}
NMM::IValueStream& operator>>(NMM::IValueStream& ov, list<string>& string_list) {
  return readList<string>(ov, string_list, "string");
}

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const list<float>& string_list){
  return writeList<float>(ov, string_list, "float");
}

NMM::IValueStream& operator>>(NMM::IValueStream& ov,list<float>& string_list) {
  return readList<float>(ov, string_list, "float");
}


NMM::OValueStream& operator<<(NMM::OValueStream& ov, const list<double>& string_list) {
  return writeList<double>(ov, string_list, "double");
}

NMM::IValueStream& operator>>(NMM::IValueStream& ov,list<double>& string_list)  {
  return readList<double>(ov, string_list, "double");
}

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const list<u_int32_t>& string_list) {
  return writeList<u_int32_t>(ov, string_list, "unsigned_int");
}

NMM::IValueStream& operator>>(NMM::IValueStream& ov,list<u_int32_t>& string_list) {
  return readList<u_int32_t>(ov, string_list, "unsigned_int");
}

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const list<u_int16_t>& string_list) {
  return writeList<u_int16_t>(ov, string_list, "short_unsigned_int");
}

NMM::IValueStream& operator>>(NMM::IValueStream& ov,list<u_int16_t>& string_list){
  return readList<u_int16_t>(ov, string_list, "short_unsigned_int");
}

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const list<int32_t>& string_list){
  return writeList<int32_t>(ov, string_list, "int");
}

NMM::IValueStream& operator>>(NMM::IValueStream& ov,list<int32_t>& string_list) {
  return readList<int32_t>(ov, string_list, "int");
}

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const list<int16_t>& string_list) {
  return writeList<int16_t>(ov, string_list, "short_int");
}

NMM::IValueStream& operator>>(NMM::IValueStream& ov,list<int16_t>& string_list) {
  return readList<int16_t>(ov, string_list, "short_int");
}

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const list<long>& string_list){
  return writeList<long>(ov, string_list, "long");
}


NMM::IValueStream& operator>>(NMM::IValueStream& ov,list<long>& string_list) {
  return readList<long>(ov, string_list, "long");
}

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const list<unsigned long>& string_list) {
  return writeList<unsigned long>(ov, string_list, "unsigned_long");
}

NMM::IValueStream& operator>>(NMM::IValueStream& ov,list<unsigned long>& string_list) {
  return readList<unsigned long>(ov, string_list, "unsigned_long");
}

/* Pair implementations */

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const pair<int32_t, int32_t>& string_list) {
  return writePair<int32_t, int32_t>(ov, string_list, "int", "int");
}

NMM::IValueStream& operator>>(NMM::IValueStream& ov, pair<int32_t, int32_t>& string_list) {
  return readPair<int32_t, int32_t>(ov, string_list, "int", "int");
}

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const pair<u_int64_t, u_int64_t>& string_list) {
  return writePair<u_int64_t, u_int64_t>(ov, string_list, "unsigned_long_long", "unsigned_long_long");
}

NMM::IValueStream& operator>>(NMM::IValueStream& ov, pair<u_int64_t, u_int64_t>& string_list) {
  return readPair<u_int64_t, u_int64_t>(ov, string_list, "unsigned_long_long", "unsigned_long_long");
}


NMM::OValueStream& operator<<(NMM::OValueStream& ov, const map<string, string>& string_map) {
  return writeMap<string, string>(ov, string_map, string_type_info, string_type_info);
}

NMM::IValueStream& operator>>(NMM::IValueStream& ov, map<string, string>& string_map) {
  return readMap<string, string>(ov, string_map, string_type_info, string_type_info);
}

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const map<string, int>& string_int_map) {
  return writeMap<string, int>(ov, string_int_map, string_type_info, int_type_info);
}

NMM::IValueStream& operator>>(NMM::IValueStream& ov, map<string, int>& string_int_map) {
  return readMap<string, int>(ov, string_int_map, string_type_info, int_type_info);
}

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const vector<bool>& bool_vector) {
  return writeVector<bool>(ov, bool_vector, bool_type_info.t_name);
}

NMM::IValueStream& operator>>(NMM::IValueStream& ov, vector<bool>& bool_vector) {
  return readBoolVector(ov, bool_vector, bool_type_info.t_name);
}


TValueConstructor<list<string> >                  __nmm_list_string_value("list_string", "std");
TValueConstructor<list<float> >                   __nmm_list_float_value ("list_float", "std");
TValueConstructor<list<double> >                  __nmm_list_double_value("list_double", "std");
TValueConstructor<list<u_int32_t> >               __nmm_list_u_int_value ("list_unsigned_int", "std");
TValueConstructor<list<u_int16_t> >               __nmm_list_su_int_value("list_short_unsigned_int", "std");
TValueConstructor<list<int32_t> >                 __nmm_list_int_value   ("list_int", "std");
TValueConstructor<list<int16_t> >                 __nmm_list_s_int_value ("list_short_int", "std");
TValueConstructor<list<long> >                    __nmm_list_long_value  ("list_long", "std");
TValueConstructor<list<unsigned long> >           __nmm_list_u_long_value ("list_unsigned_long", "std");

TValueConstructor<pair<int32_t, int32_t> >        __nmm_pair_int_int_value ("pair_int_int", "std");
TValueConstructor<pair<u_int64_t, u_int64_t> >    __nmm_pair_unigned_long_long_unsigned_long_long ("pair_unsigned_long_long_unsigned_long_long", "std");

TValueConstructor<map<string, string> >           __nmm_map_string_string ("map_string_string", "std");
TValueConstructor<map<string, int> >              __nmm_map_string_int ("map_string_int", "std");

TValueConstructor<vector<bool> >                  __nmm_vector_bool ("vector_bool", "std");
}




NMM::IValueStream& readBoolVector(NMM::IValueStream& iv, vector<bool>& v_vector, const string& type_name)
{
  /* Begin> Type checking */
  if (iv.getCurrentType() != "vector_"+type_name) {
    GLOBAL_ERROR_STREAM("List> Invalid type assignment" << endl);
    GLOBAL_ERROR_STREAM ("List> Current type : " << iv.getCurrentType() << "                Expected type: " << "vector_"+type_name << endl);
    GLOBAL_ERROR_STREAM ("List> Current namespace : " << iv.getCurrentNamespace() << "      Expected namespace: " << "std" << endl);

    iv.nextElement();
    return iv;
  }
  /* End< */

  /* Begin> Access internal attributes */
  if (!iv.goDown()) {
    GLOBAL_DEBUG_STREAM("List seems to be empty" << endl);
    iv.nextElement();
    return iv;
  }
  /* End< */

  size_t size;

  /* read the size of the vector */
  iv >> size;

  v_vector.resize(size);

  /* Begin> While the Stream contains elements, read them */
  bool tmp;
  for (size_t i = 0; i < size; ++i) {
    iv >> tmp;
    v_vector[i] = tmp;
  }
  /* End< */

  if (!iv.isEmpty()) {
    GLOBAL_ERROR_STREAM("Unhandled elements from vector_" << type_name << " in IValueStream" << endl);
  }

  iv.goUp();

  return iv;
}
