/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_MAGICNUMBERVALUESTREAM_HPP_
#define _NMM_MAGICNUMBERVALUESTREAM_HPP_


#include <string>
#include <list>
#include <vector>

#include "nmm/nmm_hash_map.hpp"

#include "ValueStream.hpp"
#include "nmm/utils/Singleton.hpp"

namespace NMM {

/** The MagicNumberManager manages the MagicNumber for specific type information. A magic number
  * represents a TypeInfo as a simple integer instead of two strings. The MagicNumberStream
  * uses this class to map a given TypeInfo to a magic number and vice versa. The developer can
  * register a MagicNumber for new structures by this class. He must ensure that the MagicNumber
  * are unique in the network.
  *
  * This class is designed as Singleton.
  */
class MagicNumberManager {
  friend class CreateUsingNew<MagicNumberManager>;
public:
/** @name Access MagicNumbers
  */
//@{
/** This method retruns the TypeInfo of a magic number. If the TypeInfo is unknown, a NULL-ptr is returned.
  * The returned pointer MUST not be deleted.
  * \param magic_number The magic number that should be converted in a TypeInfo
  */
  const TypeInfo* getTypeInfo(int magic_number);

/** This method returns a magic number to a given TypeInfo.
  * \param t_info The TypeInfo that should be converted in a magic number.
  */
  inline int getMagicNumber(const TypeInfo& t_info) { return getMagicNumber(t_info.t_namespace, t_info.t_name); }

/** This method returns a magic number to a given TypeInfo.
  * \param type_name The type name of the struct that should be converted in a magic number.
  * \param t_namespace The namespace of the struct that should be converted in a magic number.
  */
  int getMagicNumber(const string& type_name, const string& t_namespace);

/** This method registers a TypeInfo with a specific magic number. The caller of this method
  * is responsible to ensure that the magic number is unique.
  */
  void addMagicNumber(const TypeInfo& t_info, int magic_number);
//@}

private:
/** @name Constructor & Destructor
  */
//@{
/** Constructor
  */
  MagicNumberManager(){};

/** Destructor
  */
  ~MagicNumberManager(){};

/** Disabled Copy Constructor
  */
  MagicNumberManager(const MagicNumberManager&);
//@}
private:
  /* hash method for a string */
  struct string_hash
  {
    size_t operator()(const string& __s) const { return HASH_MAP_STL::__stl_hash_string(__s.c_str()); }
  };

  HASH_MAP_STL::hash_map<string, HASH_MAP_STL::hash_map<string, int, string_hash >, string_hash> __type_info_map;     ///< This hash_map assigns a TypeInfo a MagicNumber. The first string is the namespace, the second one the type_name
  HASH_MAP_STL::hash_map<int, TypeInfo*>                                           __magic_number_map;  ///< This hash_map assigns a MagicNumber a TypeInfo.
};

/* Typedef to access the LocalRegistry with the method call SLocalRegistry::getInstance() */
typedef Singleton<MagicNumberManager, CreateUsingNew> SMagicNumberManager;

/** This class can be used to automatically register a new struct by the MagicNumberManager.
  * If an instance of this object is created, it register the given TypeInformation and MagicNumber.
  * The developer can create a global instance in its .cpp file of the new struct. If the
  * struct is load as object file or the application links against the .o file, the struct
  * is automatically registered
  */
class MagicNumberRegister {
public:
/** @name Constructor & Destructor
  */
//@{
/** Constructor
  * \param t_info The type info of the struct.
  * \param magic_number The corresponding magic number
  */
  MagicNumberRegister(const TypeInfo& t_info, int magic_number) { SMagicNumberManager::getInstance().addMagicNumber(t_info, magic_number); }
//@}
};


/** This class is used as base class for the MagicNumberIValueStream and MagicNumberOValueStream.
  * It defines the Magigc number for the C++ base types, like int, float, etc..
  */
class GenericMagicNumber{
public:
  enum MagicNumber{
    COMPLEX_END = 0, COMPLEX_BEGIN, 
    U_INT16, INT16, U_INT32, INT32, U_INT64, INT64, BOOL,
    UNSIGNED_LONG, LONG, DOUBLE, FLOAT, CHAR, UNSIGNED_CHAR, CHAR_PTR, UNSIGNED_CHAR_PTR,
    STRING, FINFLAG
  };

  static const TypeInfo* magic_number_map[];        ///< Map with TypeInfo for the magic number

  static const int magic_number[];                  ///< Map with the integer values for the magic number
};

/** This class implements an IValueStream that uses magic numbers instead of the TypeInfo
  * struct to save the type information.
  */
class MagicNumberIValueStream : public IValueStream, public GenericMagicNumber {
public:
/** @name Constructor & Destructor
  */
//@{
/** Constructor
  */
  MagicNumberIValueStream();

/** Destructor
  */
  virtual ~MagicNumberIValueStream();

/** virtual copy constructor
  */
  virtual MagicNumberIValueStream* copy() const { return new MagicNumberIValueStream(); }
//@}


/** @name The operators
  */
//@{
/** This method reads an u_int16_t i from the stream and returns a reference to the stream.
  */   
  virtual MagicNumberIValueStream& operator>>(u_int16_t& i);

/** This method reads an int16_t i from the stream and returns a reference to the stream.
  */
  virtual MagicNumberIValueStream& operator>>(int16_t& i);

/** This method reads an int32_t i from the stream and returns a reference to the stream.
  */
  virtual MagicNumberIValueStream& operator>>(int32_t&   i);

/** This method reads an u_int32_t i from the stream and returns a reference to the stream.
  */
  virtual MagicNumberIValueStream& operator>>(u_int32_t& i);

/** This method reads an int64_t i from the stream and returns a reference to the stream.
  */
  virtual MagicNumberIValueStream& operator>>(int64_t&   i);

/** This method reads an u_int64_t i from the stream and returns a reference to the stream.
  */
  virtual MagicNumberIValueStream& operator>>(u_int64_t& i);

/** This method reads a bool i from the stream and returns a reference to the stream.
  */
  virtual MagicNumberIValueStream& operator>>(bool&      i);

/** This method reads an unsigned long i from the stream and returns a reference to the stream.
  */
  virtual MagicNumberIValueStream& operator>>(unsigned long& i);

/** This method reads a long i from the stream and returns a reference to the stream.
  */
  virtual MagicNumberIValueStream& operator>>(long&   i);

/** This method reads a double i from the stream and returns a reference to the stream.
  */
  virtual MagicNumberIValueStream& operator>>(double& i);

/** This method reads a float i from the stream and returns a reference to the stream.
  */
  virtual MagicNumberIValueStream& operator>>(float&  i);

/** This method reads a char i from the stream and returns a reference to the stream.
  */
  virtual MagicNumberIValueStream& operator>>(char&   i);

/** This method reads an unsigned char i from the stream and returns a reference to the stream.
  */
  virtual MagicNumberIValueStream& operator>>(unsigned char& i);

/** This method reads a stl-string i from the stream and returns a reference to the stream.
  */
  virtual MagicNumberIValueStream& operator>>(string& i);
//@}

/** @name Read char-array
  */
//@{
/** This method reads size chars and store they in the given buffer.
  * \param buffer The buffer to store the read char's. The minimum size of the array must be size
  * \param size The size of char's that should be read
  */
  virtual void readBuffer(char* buffer, unsigned int size);

/** This method reads size unsigned chars and store they in the given buffer.
  * \param buffer The buffer to store the read unsigned char's. The minimum size of the array must be size
  * \param size The size of char's that should be read
  */
  virtual void readBuffer(unsigned char* buffer, unsigned int size);

/** This method reads all data need to use this MagicNumber ValueStream
  */
  virtual void readData();
//@}

/** @name Navigate through the attributes
  */
//@{
/** This method accesses the internal attributes of complex structure.
  */
  virtual bool goDown();

/** This method leaves the internal attributes of complex structure.
  */
  virtual void goUp();

/** This method returns true if no values are in the ValueStream.
  */
  virtual bool isEmpty()          const;

/** This method reads all information from the next element
  */
  virtual bool nextElement();
//@}

/** @name Navigate through the attributes
  */
//@{
/** This method returns the current type information in a TypeInfo struct
  */
  virtual const TypeInfo& getCurrentTypeInfo() const;

/** This method returns the type name of the current element. The Stream must not be empty!
  * \sa isEmpty();
  */
  virtual const string& getCurrentType() const;

/** This method returns the namespace of the current element. The Stream must not be empty!
  * \sa isEmpty();
  */
  virtual const string& getCurrentNamespace() const;

/** This method returns the type name of this IValueStream.
  */
  virtual const string& getIValueStreamType() const { return __i_value_stream_type; }
//@}
private:
  const TypeInfo*         __current_type_info;              ///< stores a pointer to the current TypeInfo
  TypeInfo                __complex_type_info;              ///< intenal TypeInfo, if no magic number is used
  int                     __next_element;                   ///< magic number to the next element.         
  vector<const TypeInfo*> __open_stack;                     ///< stack of TypeInfo to read complex structure
  bool                    __is_empty;                       ///< is true if no attributes are in the stream
  bool                    __stream_empty;                   ///< is true if the end of the stream is reached
  MagicNumberManager*     __magic_number_manager;           ///< Pointer to the MagicNumberManager to avoid an method call
public:
  static const string     __i_value_stream_type;            ///< stores the TypeInfo of this ValueStream
  static const string     __empty_string;                   ///< empty string for return value
};


/** This OValueStream reads values, that are written with the MagicNumberOValueStream.
  */
class MagicNumberOValueStream : public OValueStream, public GenericMagicNumber {
public:
/** @name Constructor & Destructor
  */
//@{
/** Constructor
  */
  MagicNumberOValueStream();

/** Destructor
  */
  virtual ~MagicNumberOValueStream();

/** virtual copy constructor
  */
  virtual MagicNumberOValueStream* copy() const { return new MagicNumberOValueStream(); }
//@}

/** @name The operators
  */
//@{
/** This method writes a string i to the stream and returns a reference to the stream.
  */  
  virtual MagicNumberOValueStream& operator<<(const string& i)         ;

/** This method writes an u_int16_t i to the stream and returns a reference to the stream.
  */
  virtual MagicNumberOValueStream& operator<<(const u_int16_t& )      ;

/** This method writes an int16_t i to the stream and returns a reference to the stream.
  */
  virtual MagicNumberOValueStream& operator<<(const int16_t& i)      ;

/** This method writes an u_int32_t i to the stream and returns a reference to the stream.
  */
  virtual MagicNumberOValueStream& operator<<(const u_int32_t& i)    ;

/** This method writes an int16_t i to the stream and returns a reference to the stream.
  */
  virtual MagicNumberOValueStream& operator<<(const int32_t& i)        ;

/** This method writes an u_int64_t i to the stream and returns a reference to the stream.
  */
  virtual MagicNumberOValueStream& operator<<(const u_int64_t& i)     ;

/** This method writes an int64_t i to the stream and returns a reference to the stream.
  */
  virtual MagicNumberOValueStream& operator<<(const int64_t& i)       ;

/** This method writes an bool i to the stream and returns a reference to the stream.
  */
  virtual MagicNumberOValueStream& operator<<(const bool& i)          ;

/** This method writes an unsigned long i to the stream and returns a reference to the stream.
  */
  virtual MagicNumberOValueStream& operator<<(const unsigned long& i)  ;

/** This method writes an long i to the stream and returns a reference to the stream.
  */
  virtual MagicNumberOValueStream& operator<<(const long& i)          ;

/** This method writes an double i to the stream and returns a reference to the stream.
  */
  virtual MagicNumberOValueStream& operator<<(const double& i)        ;

/** This method writes an float i to the stream and returns a reference to the stream.
  */
  virtual MagicNumberOValueStream& operator<<(const float& i)         ;

/** This method writes an char i to the stream and returns a reference to the stream.
  */
  virtual MagicNumberOValueStream& operator<<(const char& i)          ;

/** This method writes an unsigend char i to the stream and returns a reference to the stream.
  */
  virtual MagicNumberOValueStream& operator<<(const unsigned char& i)  ;

/** This method must be called if no other values are written.
  */
  virtual void writeData() ;
//@}

/** @name Write char-array
  */
//@{
/** This method writes an char buffer of given size to the stream.
  * \param buffer Poitner to the char array.
  * \param size Size of the buffer.
  */
  virtual void writeBuffer(const char* buffer, unsigned int size);

/** This method writes an unsigned char buffer of given size to the stream.
  * \param buffer Poitner to the unsigned char array.
  * \param size Size of the buffer.
  */
  virtual void writeBuffer(const unsigned char* buffer, unsigned int size);
//@}

/** @name Write TypeInfo
  */
//@{
/** This method starts a new compound type and tries to map given information to a magic number.
  * If no such magic number exist, the type information are written.
  * \param type The type name of the struct
  * \param ns namespace of the struct
  */
  virtual void beginType(const string& type, const string& ns = "std");
  virtual void beginType(int m_number);


/** This method set the end flag of a current type and must be called after all attributes of a
  * complex struct are written.
  */
  virtual void endType() ;

/** This method returns the type name of this ValueStream
  */
  virtual const string& getOValueStreamType() const { return __o_value_stream_type; }
//@}
private:
  MagicNumberManager*     __magic_number_manager;          ///< Pointer to the MagicNumberManager
public:
  static const string     __o_value_stream_type;           ///< static string with the class name
};



}

#endif // _NMM_ValueStream_HPP_


