/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_XMLVALUESTREAM2_HPP_
#define _NMM_XMLVALUESTREAM2_HPP_


#include <string>
#include <list>
#include <vector>
#include <iostream>
#include "ValueStream.hpp"
#include "nmm/utils/xml/xml++.hpp"

namespace NMM {
/** This class implements an IValueStream, which converts the data from a string that is stored
  * in a XML structure.
  */
class XMLIValueStream : public IValueStream {
public:
/** @name Constructor & Destructor
  */
//@{
/** Constructor
  */
  XMLIValueStream(XMLNode* = 0);

/** Constructor that reads the data from given string.
  */
  XMLIValueStream(const string&);

/** Destructor
  */
  virtual ~XMLIValueStream();

/** Virtual copy constructor
  */
  XMLIValueStream* copy() const { return new XMLIValueStream(); }
//@}

/** @name The operators
  */
//@{
/** This method reads an u_int16_t i from the stream and returns a reference to the stream.
  */
  virtual XMLIValueStream& operator>>(u_int16_t& i);

/** This method reads an int16_t i from the stream and returns a reference to the stream.
  */
  virtual XMLIValueStream& operator>>(int16_t& i);

/** This method reads an int32_t i from the stream and returns a reference to the stream.
  */
  virtual XMLIValueStream& operator>>(int32_t&   i);

/** This method reads an u_int32_t i from the stream and returns a reference to the stream.
  */
  virtual XMLIValueStream& operator>>(u_int32_t& i);

/** This method reads an int64_t i from the stream and returns a reference to the stream.
  */
  virtual XMLIValueStream& operator>>(int64_t&   i);

/** This method reads an u_int64_t i from the stream and returns a reference to the stream.
  */
  virtual XMLIValueStream& operator>>(u_int64_t& i);

/** This method reads a bool i from the stream and returns a reference to the stream.
  */
  virtual XMLIValueStream& operator>>(bool&      i);

/** This method reads an unsigned long i from the stream and returns a reference to the stream.
  */
  virtual XMLIValueStream& operator>>(unsigned long& i);

/** This method reads a long i from the stream and returns a reference to the stream.
  */
  virtual XMLIValueStream& operator>>(long&   i);

/** This method reads a double i from the stream and returns a reference to the stream.
  */
  virtual XMLIValueStream& operator>>(double& i);

/** This method reads a float i from the stream and returns a reference to the stream.
  */
  virtual XMLIValueStream& operator>>(float&  i);

/** This method reads a char i from the stream and returns a reference to the stream.
  */
  virtual XMLIValueStream& operator>>(char&   i);

/** This method reads an unsigned char i from the stream and returns a reference to the stream.
  */
  virtual XMLIValueStream& operator>>(unsigned char& i);

/** This method reads a stl-string i from the stream and returns a reference to the stream.
  */
  virtual XMLIValueStream& operator>>(string& i);

/** This method reads all data need to use this MagicNumber ValueStream
  */
  virtual void readData();
//@}

/** @name Read char-array
  */
//@{
/** This method reads size chars and store they in the given buffer.
  * \param buffer The buffer to store the read char's. The minimum size of the array must be size
  * \param size The size of char's that should be read
  */
  virtual void readBuffer(char* buffer, unsigned int size);

/** This method reads size unsigned chars and store they in the given buffer.
  * \param buffer The buffer to store the read unsigned char's. The minimum size of the array must be size
  * \param size The size of char's that should be read
  */
  virtual void readBuffer(unsigned char* buffer, unsigned int size);
//@}


/** @name Navigate through the attributes
  */
//@{
/** This method accesses the internal attributes of complex structure.
  */
  virtual bool goDown();

/** This method leaves the internal attributes of complex structure.
  */
  virtual void goUp();

/** This method returns true if no values are in the ValueStream.
  */
  virtual bool isEmpty()          const;

/** This method reads all information from the next element
  */
  virtual bool nextElement();
//@}

/** @name Navigate through the attributes
  */
//@{
/** This method returns the current type information in a TypeInfo struct
  */
  virtual  const TypeInfo& getCurrentTypeInfo() const;

/** This method returns the type name of the current element. The Stream must not be empty!
  * \sa isEmpty();
  */
  virtual const string& getCurrentType() const;

/** This method returns the namespace of the current element. The Stream must not be empty!
  * \sa isEmpty();
  */
  virtual const string& getCurrentNamespace() const;

/** This method returns the type name of this IValueStream.
  */
  virtual const string& getIValueStreamType() const { return __i_value_stream_type; }
//@}

/** @name XML specific methods
  */
//@{
/** This method returns a pointer to the XML-Root node
  */
  const XMLNode* getRootNode()  const {return tree.root();}

/** This method returns the XMLTree
  */
  inline const XMLTree& getXMLTree()   const {return tree;}

/** This method returns the XMLTree
  */
  inline       XMLTree& getXMLTree()         {return tree;}

/** This method reads the data from a given string
  */
  virtual void fromString(const string& data);
//@}

protected:
  void setRootNode(XMLNode* node) {tree.set_root(node);}
private:
/* Type information of the IValueStream*/
  list<XMLNode*> parent_stack;                        ///< list of nodes that are currently used
  list<XMLNodeIterator> pos_stack;                    ///< current position in the XMLTree

  XMLTree tree;                                       ///< The XML-tree structure

  template <typename D>
  void readType (D& value, const string& type_id);
  mutable TypeInfo __current_type;
public:
  static const string __i_value_stream_type;          ///< The type name of this class
  static const string __empty_string;
};


/** This class implements an OValueStream, which converts the data to string and stores them
  * in a XML structure.
  */
class XMLOValueStream : public OValueStream {
public:
/** @name Constructor & Destructor
  */
//@{
/** Constructor
  */
  XMLOValueStream();

/** Destructor
  */
  virtual ~XMLOValueStream();

/** Virtual copy constructor
  */
  XMLOValueStream* copy() const { return new XMLOValueStream(); }
//@}

/** @name The operators
  */
//@{
/** This method writes a string i to the stream and returns a reference to the stream.
  */
  virtual XMLOValueStream& operator<<(const string& i)         ;

/** This method writes an u_int16_t i to the stream and returns a reference to the stream.
  */
  virtual XMLOValueStream& operator<<(const u_int16_t& )      ;

/** This method writes an int16_t i to the stream and returns a reference to the stream.
  */
  virtual XMLOValueStream& operator<<(const int16_t& i)      ;

/** This method writes an u_int32_t i to the stream and returns a reference to the stream.
  */
  virtual XMLOValueStream& operator<<(const u_int32_t& i)    ;

/** This method writes an int32_t i to the stream and returns a reference to the stream.
  */
  virtual XMLOValueStream& operator<<(const int32_t& i)        ;

/** This method writes an u_int64_t i to the stream and returns a reference to the stream.
  */
  virtual XMLOValueStream& operator<<(const u_int64_t& i)     ;

/** This method writes an int64_t i to the stream and returns a reference to the stream.
  */
  virtual XMLOValueStream& operator<<(const int64_t& i)       ;

/** This method writes an bool i to the stream and returns a reference to the stream.
  */
  virtual XMLOValueStream& operator<<(const bool& i)          ;

/** This method writes an unsigned long i to the stream and returns a reference to the stream.
  */
  virtual XMLOValueStream& operator<<(const unsigned long& i)  ;

/** This method writes an long i to the stream and returns a reference to the stream.
  */
  virtual XMLOValueStream& operator<<(const long& i)          ;

/** This method writes an double i to the stream and returns a reference to the stream.
  */
  virtual XMLOValueStream& operator<<(const double& i)        ;

/** This method writes an float i to the stream and returns a reference to the stream.
  */
  virtual XMLOValueStream& operator<<(const float& i)         ;

/** This method writes an char i to the stream and returns a reference to the stream.
  */
  virtual XMLOValueStream& operator<<(const char& i)          ;

/** This method writes an unsigend char i to the stream and returns a reference to the stream.
  */
  virtual XMLOValueStream& operator<<(const unsigned char& i)  ;

/** This method must be called if no other values are written.
  */
  virtual void writeData() ;
//@}

/** @name Write char-array
  */
//@{
/** This method writes an char buffer of given size to the stream.
  * \param buffer Poitner to the char array.
  * \param size Size of the buffer.
  */
  virtual void writeBuffer(const char* buffer, unsigned int size);

/** This method writes an unsigned char buffer of given size to the stream.
  * \param buffer Poitner to the unsigned char array.
  * \param size Size of the buffer.
  */
  virtual void writeBuffer(const unsigned char* buffer, unsigned int size);
//@}

/** @name Write TypeInfo
  */
//@{
/** This method starts a new compound type and tries to map given information to a magic number.
  * If no such magic number exist, the type information are written.
  * \param type The type name of the struct
  * \param ns namespace of the struct
  */
  virtual void beginType(const string& type, const string& ns = "std");

/** This method set the end flag of a current type and must be called after all attributes of a
  * complex struct are written.
  */
  virtual void endType() ;

/** This method returns the type name of this ValueStream
  */
  virtual const string& getOValueStreamType() const { return __o_value_stream_type; }
//@}

/** @name XML specific methods
  */
//@{
/** This method returns the Root Node
  */
  inline  XMLNode* getRootNode()            {return tree.root();}

/** This method returns the XML tree
  */
  inline  const XMLTree& getXMLTree() const {return tree;}

/** This method returns the XML tree
  */
  inline        XMLTree& getXMLTree()       {return tree;}

/** This method converts the XML tree to a string.
  */
  virtual string toString() const;
//@}

private:
  list<XMLNode*> parent_stack;                ///< stack of used xml nodes

  strstream stream;                           ///< stream to convert the data
  XMLTree tree;                               ///< The XML tree

  XMLNode* latest_node;                       ///< The current node
  XMLNodeIterator current_pos;                ///< The current position in the XML-Tree

  template<class T>
  void writeType(const T& value, const string& type);

  template<class T>
  void writeType(const T& value, const string& type, const string& ns);
public:
  static const string __o_value_stream_type; ///< The type name of this class
};



template<class T>
void XMLOValueStream::writeType(const T& value, const string& type) {
  XMLNode* tmp_node = latest_node->add_child("simple");
  tmp_node->add_property("type", type);
  tmp_node->add_property("Value", convertToString(value));
}

template<class T>
void XMLOValueStream::writeType(const T& value, const string& type, const string& ns) {
  XMLNode* tmp_node = latest_node->add_child("simple");
  tmp_node->add_property("type", type);
  tmp_node->add_property("Value", convertToString(value));
  tmp_node->add_property("c_ns", ns);
}

template <typename D>
void XMLIValueStream::readType (D& value, const string& type_id) {
  /* Begin> Check if stream is empty */
  if (isEmpty()) {
    cerr << "Stream is Empty for type id : " << type_id << endl;
    return;
  }
  /* End< */

  /* Compate the types */
  if (getCurrentType() != type_id) {
    cerr << "Try type conversion from type " << getCurrentType() << " to " << type_id << endl;
  }

  /* stream to convert the string to a specific type */
  strstream stream;
  XMLProperty* prop =  (**pos_stack.begin())->property("Value");
  if (prop) {
    stream << prop->value();
    /* read the value */
    stream >> value;
  }
  /* read the type information of the next element */
  nextElement();
}

}

#endif // _NMM_ValueStream_HPP_


