/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "LocalTime.hpp"

namespace NMM {
  
  LocalTime::LocalTime()
    : SerializedValue(getStaticTypeInfo()),
      __second(0),
      __minute(0),
      __hour(0),
      __day_of_month(0),
      __day_of_week(0),
      __day_of_year(0),
      __month(0),
      __gmt_offset(0),
      __year(0),
      __is_dst(0),
      __time_zone()
  {
    memset(&__tm_struct, 0, sizeof(struct ::tm));
    registerAttributes();
  }
  
  LocalTime::LocalTime(const Time& time, bool local /*=true*/)
    : SerializedValue(getStaticTypeInfo()),
      __second(0),
      __minute(0),
      __hour(0),
      __day_of_month(0),
      __day_of_week(0),
      __day_of_year(0),
      __month(0),
      __gmt_offset(0),
      __year(0),
      __is_dst(0),
      __time_zone()
  {
    registerAttributes();
    reset(time, local);
  }
  
  LocalTime::LocalTime(const struct ::tm& tm_struct)
    : SerializedValue(getStaticTypeInfo()),
      __second(0),
      __minute(0),
      __hour(0),
      __day_of_month(0),
      __day_of_week(0),
      __day_of_year(0),
      __month(0),
      __gmt_offset(0),
      __year(0),
      __is_dst(0),
      __time_zone()
  {
    registerAttributes();
    reset(tm_struct);
  }
  
  LocalTime::LocalTime(const LocalTime& other)
    : SerializedValue(other),
      __second(other.__second.getRValue()),
      __minute(other.__minute.getRValue()),
      __hour(other.__hour.getRValue()),
      __day_of_month(other.__day_of_month.getRValue()),
      __day_of_week(other.__day_of_week.getRValue()),
      __day_of_year(other.__day_of_year.getRValue()),
      __month(other.__month.getRValue()),
      __gmt_offset(other.__gmt_offset.getRValue()),
      __year(other.__year.getRValue()),
      __is_dst(other.__is_dst.getRValue()),
      __time_zone(other.__time_zone.getRValue())
  {
    registerAttributes();
  }
  
  void LocalTime::reset(const struct ::tm& tm_struct) {
    __second.setValue(tm_struct.tm_sec);
    __minute.setValue(tm_struct.tm_min);
    __hour.setValue(tm_struct.tm_hour);
    __day_of_month.setValue(tm_struct.tm_mday);
    __day_of_week.setValue(tm_struct.tm_wday);
    __day_of_year.setValue(tm_struct.tm_yday);
    __month.setValue(tm_struct.tm_mon);
    __year.setValue(tm_struct.tm_year);
    
    // XXX: if tm_isdst is negative, the information is not available!
    __is_dst.setValue(tm_struct.tm_isdst == 1);
    
    // XXX: the following fields are GNU extensions!
    __gmt_offset.setValue(tm_struct.tm_gmtoff);
    __time_zone.setValue(string(tm_struct.tm_zone));
    memcpy(&__tm_struct, &tm_struct, sizeof(struct ::tm));
  }
  
  void LocalTime::reset(const Time& time, bool local /*=true*/) {
    struct ::tm tm_value;
    
    // XXX: there seems to be no conversion function, but the manual says that the VALUE of time_t
    // XXX: is always the number of seconds since the epoch (like timeval.tv_sec???).
    // XXX: only the TYPE of time_t is not known, but the manual also says it is either a floating point
    // XXX: or integer type, so this conversion should be possible.
    time_t ct = static_cast<time_t>(time.sec);
    
    if (local) {
      ::localtime_r(&ct, &tm_value);
    } else {
      ::gmtime_r(&ct, &tm_value);
    }
    reset(tm_value);
  }
  
  string LocalTime::toString(const string& format) const {
    char buffer[4096];
    size_t size = sizeof(buffer);
    ::strftime(buffer, size, format.c_str(), &__tm_struct);
    return string(buffer);
  }
  
  TypeInfo LocalTime::getStaticTypeInfo() {
    return TypeInfo("LocalTime", "NMM");
  }
  
  void LocalTime::registerAttributes() {
    registerAttribute(&__second);
    registerAttribute(&__minute);
    registerAttribute(&__hour);
    registerAttribute(&__day_of_month);
    registerAttribute(&__day_of_week);
    registerAttribute(&__day_of_year);
    registerAttribute(&__month);
    registerAttribute(&__year);
    registerAttribute(&__is_dst);
    registerAttribute(&__time_zone);
  }
  
  void LocalTime::updateAttributes() {
    __tm_struct.tm_sec = __second.getRValue();
    __tm_struct.tm_min = __minute.getRValue();
    __tm_struct.tm_hour = __hour.getRValue();
    __tm_struct.tm_mday = __day_of_month.getRValue();
    __tm_struct.tm_mon = __month.getRValue();
    __tm_struct.tm_year = __year.getRValue();
    __tm_struct.tm_wday = __day_of_week.getRValue();
    __tm_struct.tm_isdst = __is_dst.getRValue();
    
    // XXX: the following are GNU extensions
    __tm_struct.tm_gmtoff = __gmt_offset.getRValue();
    
    // XXX: does this really work?
    __tm_struct.tm_zone = __time_zone.getRValue().c_str();
  }
  
  static TConstructor<LocalTime> __local_time_ctor(LocalTime::getStaticTypeInfo());
  
}
