/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_RATIONAL_HPP_
#define _NMM_RATIONAL_HPP_

#include "nmm/utils/NamedObject.hpp"
#include "nmm/base/serialize/ValueStream.hpp"

#include <iosfwd>

namespace NMM {

  //! RationalDivisionByZeroException
  /*!
   *  This exception is thrown if a division by zero occurs with rational numbers.
   *  That means perhaps the denominator is zero (in the constructor or if you try
   *  to build the reciprocal value) or you try to divide a rational number by
   *  zero.
   */
  class RationalDivisionByZeroException {
  public:
    RationalDivisionByZeroException();
  };

  //! This class provides basic support for rational numbers.
  class Rational {
  private:
    int64_t numerator;    //!< The numerator. 
                          //!< It also carries the sign information.
    int64_t denominator;  //!< The denominator.
                          //!< The denominator is always greater than 0.

  public:  
    //! Computes the greatest common divisors of two given int64_tegers.
    static int64_t greatestCommonDivisor(int64_t a, int64_t b);

    //! Computes the leastCommonMultiple of two given int64_tegers.
    static int64_t leastCommonMultiple(int64_t a, int64_t b);

    //! The main constructor.
    /*! 
     *  \param numerator   The numerator can be an arbitrary integer.
     *  \param denominator The denominator must be non-zero.
     *  Throws RationalDivisionByZeroException if the denominator is zero.
     */
    Rational(const int64_t& numerator = 0, const int64_t& denominator = 1)
      throw (const RationalDivisionByZeroException&);

    //! Returns the numerator.
    inline int64_t getNumerator() const
    {
      return numerator;
    };

    //! Returns the denominator.
    inline int64_t getDenominator() const
    {
      return denominator;
    };

    //! Completely reduces the fraction.
    void reduce();

    //! Inverts the value of the fraction.
    /*!
     *  Throws RationalDivisionByZeroException if the rational number was zero.
     */
    void invert()
      throw (const RationalDivisionByZeroException&);

    //! Returns a new Rational number with the reciprocal value.
    /*!
     *  Throws RationalDivisionByZeroException if the rational number was zero.
     */
    Rational getReciprocal()
      throw (const RationalDivisionByZeroException&);

    //! Returns the negative value of a rational number.
    Rational& operator-();

    //! Adds a reational number to another.
    Rational& operator+=(const Rational& b);
  
    //! Subtracts a rational number from another.
    Rational& operator-=(const Rational& b);

    //! Multiplies a rational number with another.
    Rational& operator*=(const Rational& b);

    //! Divides a rational number by another.
    /*!
     *  If the divisor is zero, RationalDivisionByZeroException is thrown.
     */
    Rational& operator/=(const Rational& b)
      throw (const RationalDivisionByZeroException&);

    //! Returns true iff a and b are equal.
    friend bool operator== (const Rational& a, const Rational& b);

    //! Returns true iff a and b are not equal.
    friend bool operator!= (const Rational& a, const Rational& b);

    //! Returns true iff a is greater than b.
    friend bool operator>  (const Rational& a, const Rational& b);

    //! Returns true iff a is greater than or equal to b.
    friend bool operator>= (const Rational& a, const Rational& b);

    //! Returns true iff a is less than b.
    friend bool operator<  (const Rational& a, const Rational& b);

    //! Returns true iff a is less than or equal to b.
    friend bool operator<= (const Rational& a, const Rational& b);

    //! The output operator for Rational.
    /*! 
     * Represents the rational number in the form "numerator / denominator".
     */
    friend ostream& operator<< (ostream& s, const Rational& a);
    friend IValueStream& operator>> (IValueStream& s, Rational& a);
  };

  //! Adds two rational numbers and returns the sum as a new object.
  Rational operator+ (const Rational& a, const Rational& b);

  //! Subtracts two rational numbers and returns the difference as a new object.
  Rational operator- (const Rational& a, const Rational& b);

  //! Multiplies two rational numbers and returns the product as a new object.
  Rational operator* (const Rational& a, const Rational& b);

  //! Divides two rational numbers and returns the quotient as a new object.
  Rational operator/ (const Rational& a, const Rational& b);

  ostream& operator<< (ostream& s, const Rational& a);

  /*  This method writes a Rational into a OValueStream. This is required if Events with a
   *  Rational parameter should be transported over a net connection.
   */
  OValueStream& operator<<(OValueStream& s, const Rational&);

  /*  This method reads a Rational from a IValueStream. This is required if Events with a
   *  Rational parameter should be transported over a net connection.
   */
  IValueStream& operator>>(IValueStream& s, Rational&);

  inline bool operator== (const Rational& a, const Rational& b)
  {
    return a.numerator == b.numerator && a.denominator == b.denominator;
  };

  inline bool operator!= (const Rational& a, const Rational& b)
  {
    return a.numerator != b.numerator || a.denominator != b.denominator;
  };

  inline bool operator>  (const Rational& a, const Rational& b)
  {
    return a.numerator * b.denominator >  b.numerator * a.denominator;
  };

  inline bool operator>= (const Rational& a, const Rational& b)
  {
    return a.numerator * b.denominator >= b.numerator * a.denominator;
  };

  inline bool operator<  (const Rational& a, const Rational& b)
  {
    return a.numerator * b.denominator <  b.numerator * a.denominator;
  };

  inline bool operator<= (const Rational& a, const Rational& b)
  {
    return a.numerator * b.denominator <= b.numerator * a.denominator;
  };

}

#endif // _NMM_RATIONAL_HPP_
