/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */



#ifndef _NMM_STREAM_TIMER_HPP_
#define _NMM_STREAM_TIMER_HPP_

#include "nmm/base/Message.hpp"
#include "nmm/utils/NamedObject.hpp"
#include "nmm/base/Node.hpp"

#include "TimedElement.hpp"

namespace NMM {

  //! This class represents a Timer for a multimedia stream.
  /*!
   *  Nodes that want to change messages' timestamps should have a 
   *  streamtimer. It can be used to set the sync_time and the 
   *  stream_counter.
   *  So each stream must have its own timer.
   *  An error correction will be added, too. (The finite number precision 
   *  may cause some problems with long media streams.) 
   */
  class StreamTimer : public TimedElement, 
                      public NamedObject {

  public:
    enum MODE { REAL_TIME,   /*! In this mode, the streamtimer uses a
			       clock to generate the creation times.
			       This is useful for sources like CameraNode
			       or RecordNode.                              */
                CONST_RATE   /*! This mode is used by file reader nodes
			       or other nodes that can create the messages
			       independend of the real creation time of 
			       their data content. A buffer rate is taken
			       to compute the actual creation timestamp of
			       the message.                                */
    };

    static const MODE DEFAULT_MODE = REAL_TIME;
    /*! The default mode is REAL_TIME, because in this mode the stream
        timer can work without knowledge of a buffer rate.                 */

  private:
    Node* parent;            //!< The node which the timer belongs to.
    int rate;                //!< The buffer rate.
    Interval interval;       //!< The interval between two buffers.
    Timestamp timestamp;     //!< The timestamp.
    MODE mode;               //!< The counter's mode.

    bool interval_set;       //!< Internal flag: true iff the interval
                             //!< between two buffers has been set.
                             //!< (This can be done by setting the rate, too.)
    bool delay_set;          //!< Interval flag: true iff the 
                             //!< presentation_delay has been set.
 
  public:
    //! Constructor in which only the mode is set, no parent node.
    StreamTimer(const MODE mode);

    //! The main constructor.
    /*!
     *  The name of the timer is set to the name of the node + " Timer".
     *  The timer is set to the specified mode.
     */
    StreamTimer(const Node* parent = 0,
		const MODE mode = DEFAULT_MODE);

    //! The main destructor.
    virtual ~StreamTimer();

    //! Resets the StreamTimer.
    /*!
     *  After a reset, interval or rate has to be set again in CONST_RATE mode.
     */
    void reset();

    //! Sets the mode of the StreamTimer.
    /*!
     *  This method should not be called while playing!
     *  The mode should be known already at compile time (it is thought to  
     *  be called by the constructor of the owner node).
     *  \param mode The mode of the StreamTimer.
     *  \return SUCCESS if the mode was valid.
     */
    Result setMode(const MODE mode);
    
    //! Sets the StreamTimer's buffer rate.
    /*!
     *  The interval between two buffers is computed by this rate.
     *  \param rate - the buffer rate (buffers per second)
     *  \return SUCCESS if the rate is not zero and non-negative.
     */
    Result setRate(const float rate);
    
    //! Sets the time interval between two buffers directly.
    /*!
     *  This method can be used if the buffer rate is not an integer
     *  (for example the mpeg - video - stream).
     *  \param interval The time between two buffers in nanoseconds.
     *  \return SUCCESS if the interval is not zero and non-negative.
     */
    Result setInterval(const Interval interval);

    //! Sets the sync_time and stream_counter of the message.
    /*!
     *  The sync_time is created as discribed above. The stream_counter
     *  is automatically increased with each call of this method.
     *  \param message The message whose timestamp should be set.
     *  \return FAILURE if the interval or rate has not been set (in 
     *                  CONST_RATE - Mode)
     *         SUCCESS otherwise
     */
    Result setTimestamp(Message* message);
  };
}

#endif // _NMM_STREAM_TIMER_HPP_
