/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <sstream>
#include <math.h>
#include "Types.hpp"
#include "Rational.hpp"
#include "nmm/base/serialize/Value.hpp"

namespace NMM {

  //-------------------------------- Time -----------------------------------

  Time::Time() 
  {
    nsec = 0;
    sec  = 0;
  };

  Time::Time(const long int& nanoseconds) 
  {
    nsec = nanoseconds % nano_per_sec;
    sec  = nanoseconds / nano_per_sec;

    if(nsec < 0) {
      nsec += nano_per_sec;
      sec--;
    }
  };

  Time::Time(const long int& seconds, const long int& nanoseconds)
  {
    nsec = nanoseconds;
    sec  = seconds;

    while(nsec < 0) {
      nsec += nano_per_sec;
      sec--;
    }
    while(nsec > nano_per_sec) {
      nsec -= nano_per_sec;
      sec++;
    }
  };

  Time::Time(const struct timeval& time) 
  {
    nsec = time.tv_usec * 1000L;
    sec  = time.tv_sec;

    while(nsec < 0) {
      nsec += nano_per_sec;
      sec--;
    }
    while(nsec > nano_per_sec) {
      nsec -= nano_per_sec;
      sec++;
    }
  };

  Time::Time(const Rational& time)
  {
    sec = time.getNumerator() / time.getDenominator();
    double nanosec = static_cast<double>(time.getNumerator()) / time.getDenominator();
    nanosec -= sec;          // so the absolute value of nanosec is less than 1.
    nanosec *= nano_per_sec;

    nsec = static_cast<long int>(nanosec);

    if(nsec < 0) {
      nsec += nano_per_sec;
      sec--;
    }
  };

  Time& Time::operator+=(const Interval& interval) 
  {
    nsec += interval.nsec;
    sec  += interval.sec;
    if(nsec >= nano_per_sec) {
      sec++;
      nsec -= nano_per_sec;
    }

    return *this;
  };

  Time& Time::operator-=(const Interval& interval) 
  {
    nsec -= interval.nsec;
    sec  -= interval.sec;
    if(nsec < 0) {
      sec--;
      nsec += nano_per_sec;
    }

    return *this;
  };

  Time operator+(const Time& a, const Interval& b) {
    Time result = a;
    result += b;

    return result;
  };

  Time operator-(const Time& a, const Interval& b) {
    Time result = a;
    result -= b;
    
    return result;
  };

  string Time::getPrettyString() const {
    ostringstream strstr;
    strstr.width(4);
    strstr.fill('0');
    strstr << sec << ".";
    strstr.width(9);
    strstr.fill('0');
    strstr << nsec;
    return strstr.str();
  }


  ostream& operator<<(ostream& s, const Time& time)
  {
    if (time.sec < 0) {
      if (time.sec == -1) {
	s << "-";
      }
      return s << time.sec + 1 << " sec, " << nano_per_sec - time.nsec << " nanosec";
    }
    return s << time.sec << " sec, " << time.nsec << " nanosec";
  };


  IntervalDivisionByZeroException::IntervalDivisionByZeroException() 
  {
    GLOBAL_ERROR_STREAM2("Interval - Division by zero exception." << endl);
  };

  //------------------------------- Interval --------------------------------

  Interval::Interval() {
    nsec = 0;
    sec  = 0;
  };

  Interval::Interval(const long int& nanoseconds) {
    nsec = nanoseconds % nano_per_sec;
    sec  = nanoseconds / nano_per_sec;

    if(nsec < 0) {
      nsec += nano_per_sec;
      sec--;
    }
  };

  Interval::Interval(const long int& seconds, const long int& nanoseconds)
  {
    nsec = nanoseconds;
    sec  = seconds;

    while(nsec < 0) {
      nsec += nano_per_sec;
      sec--;
    }
    while(nsec > nano_per_sec) {
      nsec -= nano_per_sec;
      sec++;
    }
  };

  Interval::Interval(const struct timeval& time) {
    nsec = time.tv_usec * 1000L;
    sec  = time.tv_sec;

    while(nsec < 0) {
      nsec += nano_per_sec;
      sec--;
    }
    while(nsec > nano_per_sec) {
      nsec -= nano_per_sec;
      sec++;
    }
  };

  Interval::Interval(const Rational& time)
  {
    sec = time.getNumerator() / time.getDenominator();
    double nanosec = static_cast<double>(time.getNumerator()) / time.getDenominator();
    nanosec -= sec;          // so the absolute value of nanosec is less than 1.
    nanosec *= nano_per_sec;

    nsec = static_cast<long int>(nanosec);

    if(nsec < 0) {
      nsec += nano_per_sec;
      sec--;
    }
  };

  Interval& Interval::operator+=(const Interval& interval) {
    nsec += interval.nsec;
    sec  += interval.sec;
    if(nsec >= nano_per_sec) {
      sec++;
      nsec -= nano_per_sec;
    }
    return *this;
  };

  Interval& Interval::operator-=(const Interval& interval) {
    nsec -= interval.nsec;
    sec  -= interval.sec;
    if(nsec < 0) {
      sec--;
      nsec += nano_per_sec;
    }
    return *this;
  };

  Interval& Interval::operator*=(const float& a)
  {
    if(a == 0) {
        nsec = sec = 0;
        return *this;
    }

    int64_t new_nsec = static_cast<int64_t>(a * nsec);
    sec  = static_cast<long int>(a * ((float) sec));

    while (new_nsec >= nano_per_sec) {
      new_nsec -= nano_per_sec;
      sec++;
    };

    while (new_nsec <= (nano_per_sec * -1)) {
      new_nsec += nano_per_sec;
      sec--;
    };

    nsec = static_cast<long int>(new_nsec);

    return *this;
  };

  Interval& Interval::operator/=(const int& a)
    throw (const IntervalDivisionByZeroException&)
  {
    if(a == 0) throw IntervalDivisionByZeroException();

    double b;
    nsec /= a;
    nsec += static_cast<long int>(modf(static_cast<double>(sec) / a, &b) * nano_per_sec);
    sec  /= a;

    if(nsec >= nano_per_sec) {
      nsec -= nano_per_sec;
      sec++;
    }

    return *this;
  };

  Interval operator-(const Interval& interval) {
    Interval result;
    result -= interval;
    
    return result;
  };

  Interval operator+(const Interval& a, const Interval& b) {
    Interval result = a;
    result += b;

    return result;
  };

  Interval operator-(const Interval& a, const Interval& b) {
    Interval result = a;
    result -= b;

    return result;
  };

  Interval operator- (const Time& a, const Time& b) {
    Interval result;
    result.nsec = a.nsec - b.nsec;
    result.sec  = a.sec  - b.sec;
    if(result.nsec < 0) {
      result.nsec += nano_per_sec;
      result.sec--;
    }
    return result;
  };

  void suspendThread(const Interval& interval) {
    if(interval > Interval(0)) {
      struct timeval tv;
      tv.tv_usec = interval.nsec / 1000;
      tv.tv_sec  = interval.sec;
      select(0, 0, 0, 0, &tv);
    }
  };


  ostream& operator<<(ostream& s, const Interval& interval)
  {
    if (interval.sec < 0) {
      if (interval.sec == -1) {
	s << "-";
      }
      return s << interval.sec + 1 << " sec, " << nano_per_sec - interval.nsec << " nanosec";
    }
    return s << interval.sec << " sec, " << interval.nsec << " nanosec";
  };

  OValueStream& operator<<(OValueStream& s, const Time& time) {
    s.beginType("Time", "NMM");
    s << time.sec;
    s << time.nsec;
    s.endType();

    return s;
  }

  IValueStream& operator>>(IValueStream& s, Time& time) {
    if (s.getCurrentType() != "Time" && s.getCurrentNamespace() != "NMM") {
      GLOBAL_ERROR_STREAM("Invalid Type! Can not assign " << s.getCurrentType()<< " to an NMM::Time. "<< endl);
      s.nextElement();
      return s;
    }

    if (!s.goDown()) {
      GLOBAL_ERROR_STREAM("Could not access the time values " << endl);
      s.nextElement();
      return s;
    }

    s >> time.sec;
    s >> time.nsec;

    s.goUp();

    return s;
  }


  OValueStream& operator<<(OValueStream& s, const Interval& interval)
  {
    s.beginType("Interval", "NMM");
    s << interval.sec;
    s << interval.nsec;
    s.endType();

    return s;
  };


  IValueStream& operator>>(IValueStream& s,Interval& interval)  {
    if (s.getCurrentType() != "Interval" && s.getCurrentNamespace() != "NMM") {
      GLOBAL_ERROR_STREAM("Invalid Type! Can not assign " << s.getCurrentType()<< " to an NMM::Interval. "<< endl);
      s.nextElement();
      return s;
    }

    if (!s.goDown()) {
      GLOBAL_ERROR_STREAM("Could not access the interval values " << endl);
      s.nextElement();
      return s;
    }

    s >> interval.sec;
    s >> interval.nsec;

    s.goUp();

    return s;
  };

  TValueConstructor<Time> nmm_time_prototype("Time", "NMM");
  TValueConstructor<Interval> interval_prototype("Interval", "NMM");

  string Interval::getPrettyString() const {
    ostringstream strstr;
    strstr.width(4);
    strstr.fill('0');
    strstr << sec << ".";
    strstr.width(9);
    strstr.fill('0');
    strstr << nsec;
    return strstr.str();
  }
}
