
#ifndef NMM_IRegistryImpl_HPP
#define NMM_IRegistryImpl_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/NMMObject.hpp"
#include "IRegistry.hpp"

#include "nmm/interfaces/base/INodeImpl.hpp"

namespace NMM
{
/** \class IRegistry
  * Base class for all Registry Interfaces.
  */

  class IRegistryImpl  : public virtual NMMObject { 
  public:
    IRegistryImpl() { addInterface(new IRegistry(*this)); registerEvents(); };

    virtual ~IRegistryImpl() {};

  protected:
/** This method requests a Node and returns its INode interface. If the Node can not be requested
  * an exception is thrown. This method calls the method reserveNode() to reserve the Node
  * and instantiateNode() to create it.
  * \param reg The RegistryID of the Registry which manages the node.
  * \param node_id The ObjectID of the Node that is requested.
  * \param info The UserInfo of the application that requests the node.
  * \sa reserveNode()
  * \sa instantiateNode()
  */

    virtual INode* requestNode(const RegistryID& reg, const ObjectID& node_id, const UserInfo& info, NodeDescription& nd) 
      throw(const RegistryException&) = 0;
/** This method requests a Object and returns its interface. If the objecte can not be requested
  * an exception is thrown. 
  * \param reg The RegistryID of the Registry which manages the node.
  * \param node_id The ObjectID of the object that is requested.
  * \param info The UserInfo of the application that requests the object.
  */

    virtual Interface* requestObject(const RegistryID& reg, const ObjectID& node_id, const UserInfo& info) 
      throw(const RegistryException&) = 0;
/** This method releases a requested Node. This method calls the method releasReservation() to
  * to release the reservation and destroyNode() to delete it.
  * \param reg The RegistryID of the Registry which manages the node.
  * \param node_id The ObjectID of the Node that is released.
  * \param info The UserInfo of the application that requested the node.
  * \sa releaseReservation()
  * \sa destroyNode()
  */

    virtual void releaseNode(const RegistryID& reg_id, const NMMReference& ref, const ObjectID& node_id, const UserInfo& user_info) = 0;
/** This method releases a requested object. 
  * \param reg The RegistryID of the Registry which manages the object.
  * \param node_id The ObjectID of the object that is released.
  * \param info The UserInfo of the application that requested the object.
  */

    virtual void releaseObject(const RegistryID& reg_id, const NMMReference& ref, const ObjectID& node_id, const UserInfo& user_info) = 0;
/** This method searches for Nodes with certain capabilities, which are described in the
  * NodeDescription.
  * \param node_desc The NodeDescription with the requirements of a searched node.
  */

    virtual list<Response >  initRequest(const NodeDescription& node_desc) const = 0;
/** This method searches for objects with certain capabilities, which are described in the
  * ObjectDescription.
  * \param node_desc The ObjectDescription with the requirements of a searched object.
  */

    virtual list<Response >  initObjectRequest(const ObjectDescription& node_desc) const = 0;
/** This method searches for a specific node that matches the given NodeDescription. The NodeDescription
  * is created by an application and stores the information, like formats and interfaces, that must be
  * supported by the node.
  * \param registry_id The RegistryID of the Registry that should be searched.
  * \param node_desc The NodeDescription that must be supported by the nodes
  * \return A list of ObjectID's from nodes that support the requirements in the NodeDescription
  */

    virtual list<ObjectID >  searchNode(const RegistryID& registry_id, const NodeDescription& node_desc) const = 0;
/** This method reserves a node which is managed by a Registry. If the reservation is succesfull,
  * the Registry returns a FactoryTickt that allows the caller to instantiate a Node by a NodeFactory.
  * \param reg_id The RegistryID of the Registry that should be searched.
  * \param node_id The ObjectID of the Node which is used to identify the Node inside the Registry
  * \param user_info The UserInfo identifies the user that requests the Node. If the node is released
  *        this parameter must be specified again, whereby no other user can release the reserved node.
  */

    virtual FactoryTicket reserveNode(const RegistryID& reg_id, const ObjectID& node_id, const UserInfo& info) 
      throw(const RegistryException&) = 0;
/** This method creates an instance of a reserved node. If the node can be created its INode interface is
  * returned. Otherwise an exception is thrown.
  * \param reg_id The RegistryID of the Registry that should be searched.
  * \param ticket The FactoryTicket must be send to allow the Registry to instantiate the node
  * \param user_info The UserInfo identifies the user that requests the Node. If the node is released
  *        this parameter must be specified again, whereby no other user can release the reserved node.
  * \sa destroyNode
  */

    virtual INode* instantiateNode(const RegistryID& reg_id, const FactoryTicket& ticket, const UserInfo& info) 
      throw(const RegistryException&) = 0;
/** This method releases a reserved node. The node is not destroyed, only th reservation is released.
  * \param reg_id The RegistryID of the Registry that should be searched.
  * \param node_id The ObjectID of the reserved node
  * \param user_info The UserInfo identifies the user that requests the Node. If the node is released
  *        this parameter must be specified again, whereby no other user can release the reserved node.
  * \sa reserveNode
  */

    virtual void releaseReservation(const RegistryID& reg_id, const ObjectID& node_id, const UserInfo& info) = 0;
/** This method destroys a created node
  * \param reg_id The RegistryID of the Registry that should be searched.
  * \param reference The NMMReference of the instantiated node.
  * \param user_info The UserInfo identifies the user that requests the Node. If the node is released
  *        this parameter must be specified again, whereby no other user can release the reserved node.
  * \sa instantiateNode
  */

    virtual void destroyNode(const RegistryID& reg_id, const NMMReference& reference, const UserInfo& info) = 0;
/** This method returns the maximum instance count for a specific node.
  * \param reg_id The RegistryID of the Registry that should be searched.
  * \param node_id The ObjectID of the node.
  */

    virtual int getMaxInstanceCount(const RegistryID& reg_id, const ObjectID& node_id) const = 0;
/** This method returns the current instance count for a specific node.
  * \param reg_id The RegistryID of the Registry that should be searched.
  * \param node_id The ObjectID of the node.
  */

    virtual int getInstanceCount(const RegistryID& reg_id, const ObjectID& node_id) const = 0;
/** This method returns the library that must be load for a specific node.
  * \param reg_id The RegistryID of the Registry that should be searched.
  * \param node_id The ObjectID of the node.
  */

    virtual string getPluginInfo(const RegistryID& reg_id, const ObjectID& node_id) const = 0;
/** This method returns all ObjectIDs of nodes that are available.
  * \param reg_id The RegistryID of the Registry that should be searched.
  * \param node_id The ObjectID of the node.
  */

    virtual list<ObjectID >  getNodeList(const RegistryID& reg_id) const = 0;
/** This method returns all ObjectIDs of nodes that are currently reserved and not available.
  * \param reg_id The RegistryID of the Registry that should be searched.
  */

    virtual list<ObjectID >  getReservedNodeList(const RegistryID& reg_id) const = 0;
/** This method returns a short NodeInfo for all nodes that are managed in a specific registry.
  * \param reg_id The RegistryID of the Registry that should be searched.
  */

    virtual list<NodeInfo >  getNodeInfoList(const RegistryID& reg_id) const = 0;
/** This method returns the complete NodeDescription of aspecifc node.
  * \param reg_id The RegistryID of the Registry that should be searched.
  * \param node_id The ObjectID of the node.
  */

    virtual NodeDescription getNodeDescription(const RegistryID& reg_id, const ObjectID& node_id) const = 0;
/** This method returns all NodeDescriptions stored by a Registry.
  * \param reg_id The RegistryID of the Registry that should be searched.
  */

    virtual list<NodeDescription >  getNodeDescriptionList(const RegistryID& reg_id) = 0;
/** This method returns the NodeFactoryID of the NodeFactory that can instantiate a specific node.
  * \param reg_id The RegistryID of the Registry that should be searched.
  * \param node_id The ObjectID of the node.
  */

    virtual NodeFactoryID getNodeFactoryID(const RegistryID& reg_id, const ObjectID& node_id) const = 0;
/** This method releases ALL Nodes.
 */

    virtual void resetInternalReservation() = 0;
/** This method releases ALL Nodes reserved by specifc user.
 */

    virtual void resetSpecificInternalReservation(const UserInfo& user_info) = 0;
/** This method enables the access control. If the access control is enabled, no other
  * application can access the Registry.
  * \param info The UserInfo of the application that tries to enable the access control.
  */

    virtual void initTransaction(const UserInfo& info) 
      throw(const RegistryException&) = 0;
/** This method disables the access control, so other applications can access the Registry again.
  * \param info The UserInfo of the application that tries to disable the access control.
  */

    virtual void closeTransaction(const UserInfo& info) = 0;
/** This method returns the RegistryID of this Registry.
  */

    virtual RegistryID getRegistryID() const = 0;
/** This method returns the RegistryID's of all child-Registries.
  */

    virtual list<RegistryID >  getChildList(const RegistryID& registry_id) const = 0;
/** This method returns true, if the Registry with given RegistryID is managed by this
  * Registry or by one the children of the Registry.
  * \param reg_id The RegistryID of the searched Registry.
  */

    virtual bool isSubRegistry(const RegistryID& reg_id) const = 0;

    virtual RegistryMemento getMemento() = 0;

    virtual void setMemento(const RegistryMemento& rm) = 0;


  private:
    void registerEvents();
  };

}

#endif
