
#ifndef NMM_IResourceCollector_HPP
#define NMM_IResourceCollector_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/Interface.hpp"
#include "nmm/base/Event.hpp"
#include <memory>

#include "nmm/base/format/Format.hpp"
#include "nmm/base/resourcemgr/RMVertex.hpp"

namespace NMM
{

  class IResourceCollector : virtual public Interface { 
  public:
    IResourceCollector() {  setTypeInfo(TypeInfo(IResourceCollector::IResourceCollector_type, "NMM")); };

    IResourceCollector(NMMObject& parent) : Interface(parent) {  setTypeInfo(TypeInfo(IResourceCollector::IResourceCollector_type, "NMM")); setEventReceiver(parent); };

    virtual ~IResourceCollector() {};

    static const char* IResourceCollector_type;

    virtual  string getInterfaceType() const { return IResourceCollector_type; }

    Interface* copy() const { return new IResourceCollector(*this); }
    static const char* enableResourceProbing_event;
    static const char* disableResourceProbing_event;
    static const char* setBufferProbeInterval_event;
    static const char* restrictCPUUsage_event;
    static const char* getUsedCPUTime_event;
    static const char* getUsedTime_event;
    static const char* getMaxReceivedBufferSize_event;
    static const char* getCPUUsage_event;
    static const char* setCPU_event;
    static const char* loadMesh_event;
    static const char* loadMeshes_event;
    static const char* getMesh_event;
    static const char* savePoints_event;
    static const char* getUsage_event;
    static const char* getActualUsage_event;
    static const char* getActualBandwidth_event;
    static const char* getActualBufferSize_event;
    static const char* getCPUperBuffer_event;
    static const char* getVertex_event;
    static const char* updateVertex_event;
    static const char* addVertex_event;
    static const char* setupMesh_event;
    static const char* getActualFormat_event;
    /**
     * activate resource-management
     */

    virtual Result enableResourceProbing()
      throw(const Exception&);
    /**
     * deactivate resource-management
     */

    virtual Result disableResourceProbing()
      throw(const Exception&);
    /**
     * set the number of buffers to be processed between
     * the resource probes
     */

    virtual void setBufferProbeInterval(const unsigned int& value)
      throw(const Exception&);
    /**
     * restrict the cpu usage of a node
     */

    virtual void restrictCPUUsage(const float& level)
      throw(const Exception&);
    /**
     * returns how long the node used the cpu
     */

    virtual Interval getUsedCPUTime()
      throw(const Exception&);
    /**
     * returns how long the node was started
     */

    virtual Interval getUsedTime()
      throw(const Exception&);
    /**
     * returns the size of largest incoming buffer
     */

    virtual unsigned long getMaxReceivedBufferSize()
      throw(const Exception&);
    /**
     * returns how many percent of the cputime the nodes thread used
     */

    virtual unsigned int getCPUUsage(const bool& since_start = false)
      throw(const Exception&);
    /**
     * tell the node on what cpu it is running
     */

    virtual Result setCPU(const string& ci)
      throw(const Exception&);
    /**
     * load the serialized mesh
     */

    virtual Result loadMesh(const string& s_mesh)
      throw(const Exception&);

    virtual Result loadMeshes(const string& format_id, const string& s_mesh)
      throw(const Exception&);
    /**
     * get the serialized mesh
     */

    virtual Result getMesh(string& s_mesh)
      throw(const Exception&);
    /**
     * save the measured points to a file readable by HnSRTree-2.0
     */

    virtual Result savePoints(const string& filename)
      throw(const Exception&);
    /**
     * compute estimated cpu usage of a certain format
     */

    virtual float getUsage(const Format& description)
      throw(const Exception&);
    /**
     * get the estimated cpu usage for the actual connection format
     */

    virtual float getActualUsage()
      throw(const Exception&);
    /**
     * get the estimated bandwidth for the actual connection format
     */

    virtual float getActualBandwidth()
      throw(const Exception&);
   /**
     * get the size of an arriving buffer
     */

    virtual float getActualBufferSize()
      throw(const Exception&);
    /**
     * return (average) measured cpu usage per processed buffer
     */

    virtual float getCPUperBuffer()
      throw(const Exception&);
    /**
     * return the vertex that corresponds to the given format
     */

    virtual Result getVertex(const Format& description, RMVertex& vertex)
      throw(const Exception&);
    /**
     * update resource-usage of an already existing vertex
     */

    virtual Result updateVertex(const Format& description, const float& cpu_usage, const float& bandwidth)
      throw(const Exception&);
    /**
     * add a new vertex
     */

    virtual Result addVertex(const Format& description, const float& cpu_usage, const float& bandwidth)
      throw(const Exception&);
    /**
     * setup a new mesh
     */

    virtual Result setupMesh(const Format& dimensions)
      throw(const Exception&);
    /**
     * get the actually used format
     */

    virtual bool getActualFormat(Format& ret_format)
      throw(const Exception&);


  };

  typedef auto_ptr<IResourceCollector> IResourceCollector_var;




}

#endif
