
#ifndef NMM_ITimeshifting_HPP
#define NMM_ITimeshifting_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/Interface.hpp"
#include "nmm/base/Event.hpp"
#include <memory>

#include "nmm/interfaces/base/INode.hpp"

namespace NMM
{
  /** Timeshifting interface.
   * The interface for the timeshifting node.
   */

  class ITimeshifting : virtual public Interface { 
  public:    /** Main working mode.
     * The main working mode controls what happens to incoming data.
     */

  enum Mode {       /** Live mode.
       * Every arriving buffer is forwarded and, if enabled, backed up in memory for instant replay.
       */

    LIVE ,       /** File buffer mode.
       * Buffers and events are buffered in a file buffer.
       * This mode is divided up in different play modes.
       */

    FILEBUF  };
    /** Play mode.
     * The play mode controls what data is sent to the output if the main mode is FILEBUF.
     */

  enum PlayMode {       /** Playback mode.
       * Buffers are sent in the order of arrival, but possibly delayed.
       */

    PLAY ,       /** Pause mode.
       * No buffers are sent.
       */

    PAUSE ,       /** Fast forward mode.
       * Only I-Frames are sent, and a certain number of buffers are skipped, depending on a speed parameter.
       * If there are no more buffers, the node automatically switches to FB_LIVE mode.
       */

    FAST_FORWARD ,       /** Rewind mode.
       * Like fast-forward mode, except that buffers are sent in reverse order and
       * if there are no more buffers, the node automatically switches to PLAY mode.
       */

    REWIND ,       /** Buffered live mode.
       * Arriving buffers are sent immediately.
       */

    FB_LIVE  };

    ITimeshifting() {  setTypeInfo(TypeInfo(ITimeshifting::ITimeshifting_type, "NMM")); };

    ITimeshifting(NMMObject& parent) : Interface(parent) {  setTypeInfo(TypeInfo(ITimeshifting::ITimeshifting_type, "NMM")); setEventReceiver(parent); };

    virtual ~ITimeshifting() {};

    static const char* ITimeshifting_type;

    virtual  string getInterfaceType() const { return ITimeshifting_type; }

    Interface* copy() const { return new ITimeshifting(*this); }
    static const char* enableMemBuf_event;
    static const char* disableMemBuf_event;
    static const char* isMemBuf_event;
    static const char* keepData_event;
    static const char* getMode_event;
    static const char* getPlayMode_event;
    static const char* setModeLive_event;
    static const char* setModeFBLive_event;
    static const char* setModePlay_event;
    static const char* setModePause_event;
    static const char* setModeFastForward_event;
    static const char* setModeRewind_event;
    static const char* setFilename_event;
    static const char* setFileBufferThreshhold_event;
    static const char* setBufferSize_event;
    static const char* startRecord_event;
    static const char* stopRecord_event;
    static const char* isRecording_event;
    static const char* saveToFile_event;
    static const char* getReadPosition_event;
    static const char* getStartTime_event;
    static const char* getLivePosition_event;
    static const char* getLiveOffset_event;
    static const char* getFileLength_event;
    /** @name Memory Buffer
     */
//@{
    /** Enable memory buffer.
     * Enable a memory buffer for instant replay that holds up to @a size buffers.
     * @param size Maximum number of buffers in the memory buffer
     */

    virtual void enableMemBuf(const unsigned int& size = 60)
      throw(const Exception&);
    /** Disable memory buffer.
     * Disable and free the memory buffer. Set the @a play flag for instant replay.
     * @param play If true, start playback at the beginning of the buffer
     */

    virtual void disableMemBuf(const bool& play = false)
      throw(const Exception&);
    /** Is memory buffer enabled.
     * @return true if the memorybuffer is enabled
     */

    virtual bool isMemBuf()
      throw(const Exception&);
    /** Keep incoming data.
     * Control whether to keep incoming data (and forward no buffers) or not.
     * @param keep if true, keep data
     */

    virtual void keepData(const bool& keep)
      throw(const Exception&);
//@}
    /** @name Modes and Play Modes
     */
//@{
    /** Get current main mode.
     * @return Current main mode
     */

    virtual ITimeshifting::Mode getMode() const
      throw(const Exception&);
    /** Get current play mode.
     * @return Current play mode, if the main mode is FILEBUF
     */

    virtual ITimeshifting::PlayMode getPlayMode() const
      throw(const Exception&);
    /** Switch to LIVE mode.
     * Switch the node into the LIVE main mode. This completely disables timeshifting and file buffering.
     * @param enable_memory_buffer Enable memory buffering
     * @param send_sync_reset Send a ISyncReset::syncReset event after the mode change
     */

    virtual void setModeLive(const bool& enable_memory_buffer = true, const bool& send_sync_event = true)
      throw(const Exception&);
    /** Set play mode to FB_LIVE.
     * Switch the node into FILEBUF mode and set the play mode to FB_LIVE.
     * @param send_sync_reset Send a ISyncReset::syncReset event after the mode change
     */

    virtual void setModeFBLive(const bool& send_sync_event = true)
      throw(const Exception&);
    /** Set play mode to PLAY.
     * Set the play mode to PLAY. The main mode must be FILEBUF!
     * @param send_sync_reset Send a ISyncReset::syncReset event after the mode change
     */

    virtual void setModePlay(const bool& send_sync_reset = true)
      throw(const Exception&);
    /** Set play mode to PAUSE.
     * Switch the node into FILEBUF mode and set the play mode to PAUSE.
     */

    virtual void setModePause()
      throw(const Exception&);
    /** Set play mode to FAST_FORWARD.
     * Set the play mode to FAST_FORWARD. The main mode must be FILEBUF!
     * Also set the playback speed, which is the number of I-Frames to skip, to @a speed.
     * If there are no more buffers to send, the play mode is imediately set to FB_LIVE.
     * @param speed Playback speed (number of I-Frames to skip for every I-frame sent)
     */

    virtual void setModeFastForward(const int& speed = 5)
      throw(const Exception&);
    /**
     * Set play mode to REWIND.
     * Set the play mode to REWIND. The main mode must be FILEBUF!
     * Also set the playback speed, which is the number of I-Frames to skip, to @a speed.
     * If there are no more buffers to send, the play mode is imediately set to PLAY.
     * @param speed Playback speed (number of I-Frames to skip for every I-frame sent)
     */

    virtual void setModeRewind(const int& speed = -5)
      throw(const Exception&);
//@}
    /** @name File Buffer
     */
//@{
    /** Set file buffer filename.
     * Use another file for buffering if current file is empty.
     * @param filename Buffer filename
     */

    virtual void setFilename(const string& filename)
      throw(const Exception&);
    /** Set file buffer threshold.
     * Sets the file buffer threshold, which is the maximum number of GOPs to store in the file buffer.
     * If the buffer is full, the oldest recorded GOPs is overwritten.
     * @param threshold File buffer threshold
     */

    virtual void setFileBufferThreshhold(const unsigned int& threshold) /* limited(INode::CONSTRUCTED) */
      throw(const Exception&, const IllegalStateException&);
    /** Set maximum buffer size.
     * Sets the maximum size of arriving buffers to @a size. This should be the maximum size of a GOP.
     * @param size Maximum buffer size
     * @return SUCCESS if the method succeeded, otherwise FAILURE
     */

    virtual Result setBufferSize(const unsigned int& size)
      throw(const Exception&);
    /** Enter recording mode.
     * Switch the file buffer into recording mode. In this mode, no blocks are overwritten and the file buffer
     * is not deleted after the node is destroyed.
     */

    virtual void startRecord()
      throw(const Exception&);
    /** Leave recording mode.
     * Switch the file buffer back from recording mode into to mode.
     */

    virtual void stopRecord()
      throw(const Exception&);
    /** Is file buffer in recording mode.
     * @return True if the file buffer is in recording mode
     */

    virtual bool isRecording()
      throw(const Exception&);
    /** Save file buffer to file.
     * Saves the file buffer (without events) to a file.
     * @param filename Save filename
     */

    virtual Result saveToFile(const string& filename)
      throw(const Exception&);
//@}
    /** @name Timing
     */
//@{
    /** Get read position.
     * This method returns the current read position in the timeshifting buffer.
     * In FILEBUF mode, this value is between the values returned by getStartTime and getLivePosition.
     * In LIVE mode, this returns the same value as getLivePosition.
     * @return Current read position
     */

    virtual Time getReadPosition() const /* limited(INode::STARTED) */
      throw(const Exception&, const IllegalStateException&);
    /** Get starting time.
     * This method returns the (absolute) time when the tiomeshifting buffer begins.
     * In LIVE mode, this is the same value as returned by getLivePosition.
     * @return The start time of the timeshifting buffer
     */

    virtual Time getStartTime() const /* limited(INode::STARTED) */
      throw(const Exception&, const IllegalStateException&);
    /** Get live position.
     * This method returns the (absolute) time when the timeshifting buffer ends.
     * In LIVE mode, this is the same value as returned by getStartTime.
     * @return The end time of the timeshifting buffer
     */

    virtual Time getLivePosition() const /* limited(INode::STARTED) */
      throw(const Exception&, const IllegalStateException&);
    /** Get live offset.
     * This method computes the "live" offset when timeshifting. This is the offset between the
     * live position and the current read position.
     * In LIVE mode, the live offset is always zero
     * @return The live offset
     */

    virtual Interval getLiveOffset() const /* limited(INode::STARTED) */
      throw(const Exception&, const IllegalStateException&);
    /** Get file size.
     * This method computes the length of the timeshifting buffer as a time interval.
     * In LIVE mode, the file length is always zero
     * @return The length of the timeshifting buffer
     */

    virtual Interval getFileLength() const /* limited(INode::STARTED) */
      throw(const Exception&, const IllegalStateException&);

//@}

  };

  typedef auto_ptr<ITimeshifting> ITimeshifting_var;

  OValueStream& operator<<(OValueStream& ov, const ITimeshifting::Mode& t);
  IValueStream& operator>>(IValueStream& iv, ITimeshifting::Mode& t);
  OValueStream& operator<<(OValueStream& ov, const ITimeshifting::PlayMode& t);
  IValueStream& operator>>(IValueStream& iv, ITimeshifting::PlayMode& t);

  /** Time shifting listener.
   * Provides methods for Events sent by TimeshiftingNodes.
   */

  class ITimeshiftingListener : virtual public Interface { 
  public:
    ITimeshiftingListener() {  setTypeInfo(TypeInfo(ITimeshiftingListener::ITimeshiftingListener_type, "NMM")); };

    ITimeshiftingListener(NMMObject& parent) : Interface(parent) {  setTypeInfo(TypeInfo(ITimeshiftingListener::ITimeshiftingListener_type, "NMM")); setEventReceiver(parent); };

    virtual ~ITimeshiftingListener() {};

    static const char* ITimeshiftingListener_type;

    virtual  string getInterfaceType() const { return ITimeshiftingListener_type; }

    Interface* copy() const { return new ITimeshiftingListener(*this); }
    static const char* timeshiftingBegin_event;
    static Event* create_timeshiftingBegin();
    static const char* live_event;
    static Event* create_live();
    static const char* setProgress_event;
    static Event* create_setProgress(const Time& ts_time);
  /* instream (protected) and non-instream (public) */        /** Time shifting begins.
     * This method signals, that rewinding the timeshifting buffer has
     * reached its beginning and so, the TimeshiftingNode has changed
     * its internal state from ITimeshifting::REWIND to
     * ITimeshifting::PLAY.
     */

    virtual Result timeshiftingBegin() /* in-stream */ 
      throw(const Exception&);
        /** Live mode.
     * this method signals, that fast-forwarding the timeshifting
     * buffer has reached its end and so, the TimeshiftingNode has
     * changed its internal state from ITimeshfting::FAST_FORWARD to
     * ITimeshifting::FB_LIVE.
     */

    virtual Result live() /* in-stream */ 
      throw(const Exception&);
        /** Set progress.
     * TimeshiftingNodes send progress as a absolute Time, which is
     * the Time of the stream currently displayed.
     * @param ts_time The current Time of the (timeshifted) stream.
     */

    virtual Result setProgress(const Time& ts_time) /* in-stream */ 
      throw(const Exception&);


  };

  typedef auto_ptr<ITimeshiftingListener> ITimeshiftingListener_var;




}

#endif
