
#ifndef NMM_ITVGuideImpl_HPP
#define NMM_ITVGuideImpl_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/NMMObject.hpp"
#include "ITVGuide.hpp"

#include "nmm/interfaces/misc/tvguide/ITVGuidePluginImpl.hpp"
#include "nmm/interfaces/misc/tvguide/ITVGuidePeerObjectImpl.hpp"

namespace NMM
{

  class ITVGuideImpl  : public virtual NMMObject { 
  public:
    ITVGuideImpl() { addInterface(new ITVGuide(*this)); registerEvents(); };

    virtual ~ITVGuideImpl() {};

  protected:
    /**
     * Adds a channel to the channel-list
     *\param id ID of the channel wich is used to find show of this channel
     *\param name Name of the channel
     */

    virtual Result insertChannel(const string& id = "", const string& name = "") = 0;
    /**
     * Adds a channel to the channel-list
     *\param cd ChannelDescription to be added
     */

    virtual Result insertChannelDescription(const ChannelDescription& cd) = 0;
   /**
     * Adds an Event to the database
     *\param pd PrgoramDescription, containing the inrformation about the event
     */

    virtual Result insertEvent(const ProgramDescription& pd) = 0;
    /**
     * Sets the directory, where the XML-Files are located/stored
     *\param path path to XML-Files
     */

    virtual void setDirectory(const string& path = "") = 0;
    /**
     * returns the path, where the XML-Files are stored
     *\return Path to XML-Files
     */

    virtual string getDirectory() = 0;
    /**
     * Returns the Channels, for wich informaion is available
     *\return Channel-list
     */

    virtual list<ChannelDescription >  getChannelsList() = 0;
    /**
     * Returns the channel with the given ID
     *\param id ID of the channel
     *\return name of the channel, if available, empty string otherwise
     */

    virtual string getChannelById(const string& id) = 0;
    /**
     *Return the ID of the Channel with the given name
     *\param name name o the channel
     *\return ID of the requested channel, empty string if not found
     */

    virtual string getChannelByName(const string& name) = 0;
    /**
     *Returns the Event with the given starttime, endtime and channel-id
     *\param start StartTime of the Event
     *\param end EndTime of the Event
     *\param id channel-id of the event
     *\return ProgramDescription containg the information about the event
     */

    virtual ProgramDescription getEvent(const string& start, const string& end, const string& id) = 0;
    /**
     *Returns all show on the specified day and channel
     *\param day date of the day
     *\param chan channel
     *\return List of ProgramDescripytions containing the information of the shows
     */

    virtual list<ProgramDescription >  getEventsOnDayOnChannel(const string& day, const string& chan) = 0;
    /**
     * Returns a List with Events containing the given search-phrase
     *\param phrase Phrase to search for
     *\param peer Interface to the application the it is notified of new events
     *\return Event-Hitlist
     */

    virtual list<ProgramDescription >  search(const ProgramFilter& pf, const ITVGuidePeerObject* peer = NULL) = 0;
    /**
     * adds a ExtendedDescription to a event (needed, when source is DVB)
     */

    virtual void addExtDescription(const string& start, const string& end, const string& id, const string& desc) = 0;
    /**
     * returns the current event on the specified channel
     *\param channel channel_ID
     *@return current event in the specified channel
     */

    virtual ProgramDescription getCurrentEventOnChannel(const string& channel) = 0;
    /**
     * returns the next to current event on the specified channel
     *\param channel channel_ID
     *@return next event in the specified channel
     */

    virtual ProgramDescription getFollowingEventOnChannel(const string& channel) = 0;
    /** 
     * returns a list with all categories, stored in the DB
     *@return strin-list with categories
     */

    virtual list<string >  getCategories() = 0;
    /**
     * this method starts the object and starts all relateed plugins
     */

    virtual Result start() = 0;
    /**
     * stops all related plugins
     */

    virtual Result stop() = 0;
    /**
     * initialized the requested plugins. Plugins must be rewuestetd before using 'getGrabberInterface'
     */

    virtual Result init() = 0;
    /**
     * deinitializes all requested plugins
     */

    virtual Result deinit() = 0;
    /**
     * requests the TVGuide-Plugin-Interface specified in type
     *@return ITVGuideUpdatePlugin* interface of the requested grabber
     *@param type Type of grabber, wich is its plugin-name, eg IXMLTVPlugin
     */

    virtual ITVGuidePlugin* getGrabberInterface(const string& type) = 0;
    /**
       add a Host to query, when database-Information is incomplete
     */

    virtual void addPeerHost(const string& host, const int& port) = 0;
    /**
       clear the list of peer-hosts
     */

    virtual void clearPeerHosts() = 0;
    /**
       clean up the database. specify the amount of days from today wich data should be kept. older data will be erased
    */

    virtual void cleanDatabase(const int& days) = 0;


  private:
    void registerEvents();
  };

}

#endif
