/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        chwellner <chwellner@graphics.cs.uni-sb.de>
 *                     
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "ProgramFilter.hpp"
#include <stdio.h>

namespace NMM{

  ProgramFilter::ProgramFilter()
    : NMM::SerializedValue("ProgramFilter","NMM"),
      starttime(""),
      endtime(""),
      phrase(""),
      starttimeEdited(false),
      id(0)
  {
    registerValues();
  }

  ProgramFilter::ProgramFilter(const ProgramFilter& pd)
    : NMM::SerializedValue("ProgramFilter","NMM"),
      starttime(""),
      endtime(""),
      phrase(""),
      starttimeEdited(false),
      id(0)
  {
    registerValues();
    *this=pd;
  }

  ProgramFilter::~ProgramFilter(){}
  
  void ProgramFilter::setStartTime(string st){
    starttime = st;
  }

  string ProgramFilter::getStartTime() const {
    return starttime.getRValue();
  }
  
  void ProgramFilter::setEndTime(string et) {
    endtime = et;
  }
  string ProgramFilter::getEndTime() const {
    return endtime.getRValue();
  }

  void ProgramFilter::setSearchPhrase(string p){
    phrase = p;
  }
  string ProgramFilter::getSearchPhrase() const{
    return phrase.getRValue();
  }
    

  void ProgramFilter::addCategory(string c){
    (categories.getRValue()).push_back(c);
  }
  list<string> ProgramFilter::getCategories() const {
    return categories.getRValue();
  }
    
  void ProgramFilter::addChannel(string c){
    (channels.getRValue()).push_back(c);
    (channels.getRValue()).sort();
    (channels.getRValue()).unique();
    
  }
  list<string> ProgramFilter::getChannels() const {
    return channels.getRValue();
  }

  ProgramFilter* ProgramFilter::copy(){
    return new ProgramFilter(*this);
  }

  void ProgramFilter::registerValues(){
    insertArgument(&starttime);
    insertArgument(&endtime);
    insertArgument(&phrase);
    insertArgument(&categories);
    insertArgument(&channels);
    insertArgument(&starttimeEdited);
    insertArgument(&visitedPeerHosts);
    insertArgument(&id);
  }

  void ProgramFilter::localizeTimes(){
    starttime = transformTime(starttime.getRValue());
    endtime   = transformTime(endtime.getRValue());
  }
  
  string ProgramFilter::transformTime(string s){
    try{
      
      time_t localTime = time(NULL);
      localtime(&localTime); //Just to be sure, that external time-variables of time.h are initialized
      
      time_t xmlTime = time(NULL);
      struct tm *abs_time = localtime(&xmlTime);
      abs_time->tm_sec   = 0;
      abs_time->tm_hour  = atoi( (s.substr(8,2)).c_str() );
      abs_time->tm_min   = atoi( (s.substr(10,2)).c_str() );
      abs_time->tm_mday  = atoi( (s.substr(6,2)).c_str() );
      abs_time->tm_mon   = atoi( (s.substr(4,2)).c_str() )-1;
      abs_time->tm_year  = atoi( (s.substr(0,4)).c_str() )-1900;
      xmlTime = mktime( abs_time );


      unsigned int pos = s.find_first_of(' ');
      int xmlDifference= 0;
      if(pos != string::npos){
        //if no timezone-info is available, assume UTC
        xmlDifference = atoi(s.substr(pos+1,5).c_str())/100;
      }
      char timeString[25];
      //Compute the difference between the xml-Timezone and our timezone and update the outputtime accordingly
      xmlTime -= ((-1*timezone) - (xmlDifference*3600));
      double timeDifference = -1*timezone/60/60 * 100; //in hours
      strftime(timeString, sizeof(timeString), "%Y%m%d%H%M", localtime(&xmlTime));
      sprintf(timeString, "%s %0+5.0f", timeString, timeDifference);
      return string(timeString);
      
    }catch (...){
      return s;
    }    

  }



  void ProgramFilter::addPeerHost(string name){
    visitedPeerHosts.getRValue().push_back(name);
    visitedPeerHosts.getRValue().sort();
    visitedPeerHosts.getRValue().unique();
  }
  
  bool ProgramFilter::hasPeerHost(string name){
    for(list<string>::iterator i = visitedPeerHosts.getRValue().begin(); i != visitedPeerHosts.getRValue().end(); i++){
      if (*i == name) return true;
    }
    return false;
  }
 
  void ProgramFilter::clearPeerHosts(){
    visitedPeerHosts.getRValue().clear();
  }



  bool ProgramFilter::operator==(const ProgramFilter &pf) const{
    if(starttime.getRValue() == pf.getStartTime()
       && endtime.getRValue() == pf.getEndTime() 
       && phrase.getRValue() == pf.getSearchPhrase()
       && categories.getRValue() == pf.getCategories()
       && channels.getRValue() == pf.getChannels()){
      return true;
    }else{
      return false;
    }
    
  }


  OValueStream& operator<<(OValueStream& os, const list<ProgramFilter>& pf){
    return writeList<ProgramFilter>(os, pf, "ProgramFilter");
  }
  IValueStream& operator>>(IValueStream& is, list<ProgramFilter>& pf){
    return readList<ProgramFilter>(is, pf, "ProgramFilter");
  }
  
  TConstructor<ProgramFilter> nmm_pf("ProgramFilter", "NMM");
  TValueConstructor<list<ProgramFilter> > nmm_pf_list("list_ProgramFilter","NMM");

}
