
#ifndef NMM_IDisplayEvents_HPP
#define NMM_IDisplayEvents_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/Interface.hpp"
#include "nmm/base/Event.hpp"
#include <memory>


namespace NMM
{
  /**
   * Display events.
   * Defines events that can be sent by the display nodes and caught by the application.
   * The most common use is for capturing keyboard and mouse events by the application.
   *
   * The application should only register itself for the events of this interface it is
   * actually interested in. Some implementations might check for which events listeners
   * have been registered to optimize event handling.
   */

  class IDisplayEvents : virtual public Interface { 
  public:    /**
     * Mouse buttons.
     * These are identifiers for the three mouse buttons and bitmasks for the mouse
     * button state.
     */

  enum MouseButtons { 
    MOUSE_BUTTON_LEFT = 1 , 
    MOUSE_BUTTON_RIGHT = 2 , 
    MOUSE_BUTTON_MIDDLE = 4  };

    IDisplayEvents() {  setTypeInfo(TypeInfo(IDisplayEvents::IDisplayEvents_type, "NMM")); };

    IDisplayEvents(NMMObject& parent) : Interface(parent) {  setTypeInfo(TypeInfo(IDisplayEvents::IDisplayEvents_type, "NMM")); setEventReceiver(parent); };

    virtual ~IDisplayEvents() {};

    static const char* IDisplayEvents_type;

    virtual  string getInterfaceType() const { return IDisplayEvents_type; }

    Interface* copy() const { return new IDisplayEvents(*this); }
    static const char* keyPressed_event;
    static Event* create_keyPressed(const string& key);
    static const char* keyReleased_event;
    static Event* create_keyReleased(const string& key);
    static const char* keysym_keyPressed_event;
    static Event* create_keysym_keyPressed(const unsigned long& keysym);
    static const char* keysym_keyReleased_event;
    static Event* create_keysym_keyReleased(const unsigned long& keysym);
    static const char* mouseButtonPressed_event;
    static Event* create_mouseButtonPressed(const int& button, const int& x, const int& y, const int& buttons, const bool& shift);
    static const char* mouseButtonReleased_event;
    static Event* create_mouseButtonReleased(const int& button, const int& x, const int& y, const int& buttons, const bool& shift);
    static const char* mouseMoved_event;
    static Event* create_mouseMoved(const int& x, const int& y, const int& buttons, const bool& shift);
    static const char* windowExposed_event;
    static Event* create_windowExposed();
    static const char* windowClosed_event;
    static Event* create_windowClosed();
    static const char* windowResized_event;
    static Event* create_windowResized(const int& xsize, const int& ysize);
  /* instream (protected) and non-instream (public) */        /**
     * Key was pressed.
     * Sent by the display to notify the application that a key was pressed in the context of
     * the display window.
     * @param key the identifier of the pressed key
     */

    virtual Result keyPressed(const string& key) /* in-stream */ 
      throw(const Exception&);
        /**
     * Key was released.
     * Sent by the display to notify the application that a key was released in the context of
     * the display window.
     * @param key the identifier of the pressed key
     */

    virtual Result keyReleased(const string& key) /* in-stream */ 
      throw(const Exception&);
        /**
     * Key was pressed.
     * Sent by the display to notify the application that a key was pressed in the context of
     * the display window.
     * @param key the KeySym of the pressed key
     */

    virtual Result keysym_keyPressed(const unsigned long& keysym) /* in-stream */ 
      throw(const Exception&);
        /**
     * Key was released.
     * Sent by the display to notify the application that a key was released in the context of
     * the display window.
     * @param key the KeySym of the pressed key
     */

    virtual Result keysym_keyReleased(const unsigned long& keysym) /* in-stream */ 
      throw(const Exception&);
        /**
     * Mouse button was pressed.
     * Sent by the display to notify the application that a mouse button was pressed in the context of
     * the display window.
     * @param button identifies the mouse button
     * @param x x coordinate of mouse pointer
     * @param y y coordinate of mouse pointer
     * @param buttons mouse button state bitmask
     * @param shift shift key is held down
     */

    virtual Result mouseButtonPressed(const int& button, const int& x, const int& y, const int& buttons, const bool& shift) /* in-stream */ 
      throw(const Exception&);
        /**
     * Mouse button was released.
     * Sent by the display to notify the application that a mouse button was released in the context of
     * the display window.
     * @param button identifies the mouse button
     * @param x x coordinate of mouse pointer
     * @param y y coordinate of mouse pointer
     * @param buttons mouse button state bitmask
     * @param shift shift key is held down
     */

    virtual Result mouseButtonReleased(const int& button, const int& x, const int& y, const int& buttons, const bool& shift) /* in-stream */ 
      throw(const Exception&);
        /**
     * Mouse was moved
     * Sent by the display to notify the application that th mouse was moved over
     * the display window.
     * @param x x coordinate of mouse pointer
     * @param y y coordinate of mouse pointer
     * @param buttons mouse button state bitmask
     * @param shift shift key is held down
     */

    virtual Result mouseMoved(const int& x, const int& y, const int& buttons, const bool& shift) /* in-stream */ 
      throw(const Exception&);
        /**
     * The display window was exposed.
     * Sent by the display to notify the application that the user exposed
     * the display window.
     */

    virtual Result windowExposed() /* in-stream */ 
      throw(const Exception&);
        /**
     * The display window was closed.
     * Sent by the display to notify the application that the user closed
     * the display window.
     */

    virtual Result windowClosed() /* in-stream */ 
      throw(const Exception&);
        /**
     * The display window was resized.
     * Sent by the display to notify the application that the user resized
     * the display window.
     * @param xsize new x size of window
     * @param ysize new y size of window
     */

    virtual Result windowResized(const int& xsize, const int& ysize) /* in-stream */ 
      throw(const Exception&);


  };

  typedef auto_ptr<IDisplayEvents> IDisplayEvents_var;

  OValueStream& operator<<(OValueStream& ov, const IDisplayEvents::MouseButtons& t);
  IValueStream& operator>>(IValueStream& iv, IDisplayEvents::MouseButtons& t);



}

#endif
