
#ifndef NMM_IDisplayEventsImpl_HPP
#define NMM_IDisplayEventsImpl_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/NMMObject.hpp"
#include "IDisplayEvents.hpp"


namespace NMM
{
  /**
   * Display events.
   * Defines events that can be sent by the display nodes and caught by the application.
   * The most common use is for capturing keyboard and mouse events by the application.
   *
   * The application should only register itself for the events of this interface it is
   * actually interested in. Some implementations might check for which events listeners
   * have been registered to optimize event handling.
   */

  class IDisplayEventsImpl  : public virtual NMMObject { 
  public:
    IDisplayEventsImpl() { addInterface(new IDisplayEvents(*this)); registerEvents(); };

    virtual ~IDisplayEventsImpl() {};

  protected:
    /**
     * Key was pressed.
     * Sent by the display to notify the application that a key was pressed in the context of
     * the display window.
     * @param key the identifier of the pressed key
     */

    virtual Result keyPressed(const string& key) /* in-stream */  = 0;
    /**
     * Key was released.
     * Sent by the display to notify the application that a key was released in the context of
     * the display window.
     * @param key the identifier of the pressed key
     */

    virtual Result keyReleased(const string& key) /* in-stream */  = 0;
    /**
     * Key was pressed.
     * Sent by the display to notify the application that a key was pressed in the context of
     * the display window.
     * @param key the KeySym of the pressed key
     */

    virtual Result keysym_keyPressed(const unsigned long& keysym) /* in-stream */  = 0;
    /**
     * Key was released.
     * Sent by the display to notify the application that a key was released in the context of
     * the display window.
     * @param key the KeySym of the pressed key
     */

    virtual Result keysym_keyReleased(const unsigned long& keysym) /* in-stream */  = 0;
    /**
     * Mouse button was pressed.
     * Sent by the display to notify the application that a mouse button was pressed in the context of
     * the display window.
     * @param button identifies the mouse button
     * @param x x coordinate of mouse pointer
     * @param y y coordinate of mouse pointer
     * @param buttons mouse button state bitmask
     * @param shift shift key is held down
     */

    virtual Result mouseButtonPressed(const int& button, const int& x, const int& y, const int& buttons, const bool& shift) /* in-stream */  = 0;
    /**
     * Mouse button was released.
     * Sent by the display to notify the application that a mouse button was released in the context of
     * the display window.
     * @param button identifies the mouse button
     * @param x x coordinate of mouse pointer
     * @param y y coordinate of mouse pointer
     * @param buttons mouse button state bitmask
     * @param shift shift key is held down
     */

    virtual Result mouseButtonReleased(const int& button, const int& x, const int& y, const int& buttons, const bool& shift) /* in-stream */  = 0;
    /**
     * Mouse was moved
     * Sent by the display to notify the application that th mouse was moved over
     * the display window.
     * @param x x coordinate of mouse pointer
     * @param y y coordinate of mouse pointer
     * @param buttons mouse button state bitmask
     * @param shift shift key is held down
     */

    virtual Result mouseMoved(const int& x, const int& y, const int& buttons, const bool& shift) /* in-stream */  = 0;
    /**
     * The display window was exposed.
     * Sent by the display to notify the application that the user exposed
     * the display window.
     */

    virtual Result windowExposed() /* in-stream */  = 0;
    /**
     * The display window was closed.
     * Sent by the display to notify the application that the user closed
     * the display window.
     */

    virtual Result windowClosed() /* in-stream */  = 0;
    /**
     * The display window was resized.
     * Sent by the display to notify the application that the user resized
     * the display window.
     * @param xsize new x size of window
     * @param ysize new y size of window
     */

    virtual Result windowResized(const int& xsize, const int& ysize) /* in-stream */  = 0;


  private:
    void registerEvents();
  };

}

#endif
