
#ifndef NMM_IOSDManager_HPP
#define NMM_IOSDManager_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/Interface.hpp"
#include "nmm/base/Event.hpp"
#include <memory>

#include "IWidget.hpp"
#include "IWindow.hpp"
#include "IBitmapReader.hpp"

namespace NMM
{
  /**
   * OSD Manager.
   * Interface for a OSD manager.
   */

  class IOSDManager : virtual public Interface { 
  public:
    IOSDManager() {  setTypeInfo(TypeInfo(IOSDManager::IOSDManager_type, "NMM")); };

    IOSDManager(NMMObject& parent) : Interface(parent) {  setTypeInfo(TypeInfo(IOSDManager::IOSDManager_type, "NMM")); setEventReceiver(parent); };

    virtual ~IOSDManager() {};

    static const char* IOSDManager_type;

    virtual  string getInterfaceType() const { return IOSDManager_type; }

    Interface* copy() const { return new IOSDManager(*this); }
    static const char* loadFont_event;
    static const char* getFontList_event;
    static const char* update_event;
    static const char* getWidget_event;
    static const char* getWindow_event;
    static const char* createWindow_event;
    static const char* deleteWindow_event;
    static const char* deleteWidget_event;
    static const char* displayMessage_event;
    static const char* deleteMessage_event;
    static const char* displayTime_event;
    static const char* createTime_event;
    static const char* resetTime_event;
    static const char* hideTime_event;
    static const char* setTime_event;
    static const char* getTime_event;
    static const char* displaySymbol_event;
    static const char* hideSymbol_event;
    static const char* displayProgressBar_event;
    static const char* hideProgressBar_event;
    static const char* setMaxProgress_event;
    static const char* setProgress_event;
    static const char* updateProgress_event;
    static const char* displayColorBar_event;
    static const char* hideColorBar_event;
    static const char* setColorBarText_event;
    static const char* setRedColorBarText_event;
    static const char* setGreenColorBarText_event;
    static const char* setYellowColorBarText_event;
    static const char* setBlueColorBarText_event;
    static const char* getRedColorBarText_event;
    static const char* getGreenColorBarText_event;
    static const char* getYellowColorBarText_event;
    static const char* getBlueColorBarText_event;
    static const char* messageLeft_event;
    static const char* messageRight_event;
    static const char* messageID_event;
    static const char* setPause_event;
    static const char* sendPauseEvent_event;
    static const char* getBitmapReader_event;
    static const char* checkCharDimension_event;
    static const char* getMaxWidth_event;
    static const char* getMaxHeight_event;
    static const char* setLineSpace_event;
    static const char* getLineSpace_event;
    /**
     * add font to list
     * @param id font identifier
     * @param xmfilename filename of the XML file where the font is described
     * @param fontname as a XML font file can describe several fonts, this parameter describes
     *        the desired fontname, if this name is empty, the first font is loaded
     */

    virtual Result loadFont(const string& id, const string& xmlfilename, const string& fontname)
      throw(const Exception&);
    /**
     * Get list of all fonts.
     */

    virtual list<string >  getFontList() const
      throw(const Exception&);
    /**
     * Update the OSD.
     * This will update all widgets.
     */

    virtual Result update()
      throw(const Exception&);
    /**
     * Get widget.
     * Returns an interface to access a widget.
     * @param id widget id
     * @return an interface to access the widget
     */

    virtual IWidget* getWidget(const int& id)
      throw(const Exception&);
    /**
     * Get window.
     * Returns an interface to access the window of a widget.
     * @param id window id
     * @return an interface to access the window
     */

    virtual IWindow* getWindow(const int& id)
      throw(const Exception&);
    /**
     * Create a window.
     * Creates a new window and returns an interface for it.
     * @return interface for window
     */

    virtual IWindow* createWindow()
      throw(const Exception&);
    /**
     * Deletes a window.
     * Removes the window from the manager's window list and destroys it.
     */

    virtual void deleteWindow(const IWindow& window)
      throw(const Exception&);
    /** Delete a widget. */

    virtual void deleteWidget(const IWidget& widget)
      throw(const Exception&);
    /**
     * Display a message.
     * @param header message header
     * @param text message text
     * @param label1 button label 1 text
     * @param label2 button label 2 text
     * @param label3 button label 3 text
     * @param sel initial selection
     */

    virtual Result displayMessage(const string& header, const string& text, const string& label1, const string& label2 = "", const string& label3 = "", const unsigned int& sel = 0)
      throw(const Exception&);
    /** Delete current message. */

    virtual Result deleteMessage()
      throw(const Exception&);
    /**
     * Display time.
     * @param header header text
     * @param x x position
     * @param y y position
     */

    virtual Result displayTime(const string& header = "", const unsigned int& x = 500, const unsigned int& y = 410)
      throw(const Exception&);

    virtual Result createTime(const bool& ct)
      throw(const Exception&);

    virtual Result resetTime()
      throw(const Exception&);

    virtual Result hideTime()
      throw(const Exception&);

    virtual Result setTime(const u_int64_t& time)
      throw(const Exception&);

    virtual u_int64_t getTime()
      throw(const Exception&);
    /**
     * Display symbol.
     * @param symbol symbol id
     * @param x x position
     * @param y y position
     */

    virtual Result displaySymbol(const string& symbol, const unsigned int& x = 600, const unsigned int& y = 50)
      throw(const Exception&);

    virtual Result hideSymbol()
      throw(const Exception&);

    virtual Result displayProgressBar(const string& header = "", const unsigned int& x = 50, const unsigned int& y = 470, const unsigned int& width = 620)
      throw(const Exception&);

    virtual Result hideProgressBar()
      throw(const Exception&);

    virtual void setMaxProgress(const u_int64_t& max)
      throw(const Exception&);

    virtual void setProgress(const u_int64_t& progress)
      throw(const Exception&);

    virtual void updateProgress()
      throw(const Exception&);
    /**
     * Displays the color bar which can be used to show the functionality of the four color buttons of a remote control.
     * @param x x position
     * @param y y position
     */

    virtual Result displayColorBar(const unsigned int& x = 80, const unsigned int& y = 480)
      throw(const Exception&);
    /**
     * Hides the color bar. 
     */

    virtual Result hideColorBar()
      throw(const Exception&);
    /**
     * Sets the text of all four color fields in the color bar.
     */

    virtual void setColorBarText(const string& red, const string& green, const string& yellow, const string& blue)
      throw(const Exception&);
    /**
     * Sets the text of the red color field in the color bar.
     */

    virtual void setRedColorBarText(const string& text)
      throw(const Exception&);
    /**
     * Sets the text of the green color field in the color bar.
     */

    virtual void setGreenColorBarText(const string& text)
      throw(const Exception&);
    /**
     * Sets the text of the yellow color field in the color bar.
     */

    virtual void setYellowColorBarText(const string& text)
      throw(const Exception&);
    /**
     * Sets the text of the blue color field in the color bar.
     */

    virtual void setBlueColorBarText(const string& text)
      throw(const Exception&);
    /**
     * Returns the text of the red color field in the color bar.
     */

    virtual string getRedColorBarText()
      throw(const Exception&);
    /**
     * Returns the text of the green color field in the color bar.
     */

    virtual string getGreenColorBarText()
      throw(const Exception&);
    /**
     * Returns the text of the yellow color field in the color bar.
     */

    virtual string getYellowColorBarText()
      throw(const Exception&);
    /**
     * Returns the text of the blue color field in the color bar.
     */

    virtual string getBlueColorBarText()
      throw(const Exception&);

    virtual Result messageLeft()
      throw(const Exception&);

    virtual Result messageRight()
      throw(const Exception&);

    virtual string messageID()
      throw(const Exception&);
    /**
     * Set pause mode.
     * the pause flag is a hint to the OSD manager that it receives no buffers and must set its
     * working flag explicitely to redraw the OSD whenever it has been updated.
     */

    virtual void setPause(const bool& value)
      throw(const Exception&);
    /**
     * Send pause event.
     * Make the OSD manager send a pause event.
     */

    virtual void sendPauseEvent()
      throw(const Exception&);

    virtual IBitmapReader* getBitmapReader()
      throw(const Exception&);
    /**
     * Gets width and height of a character.
     * @param id font id
     * @param c character that dimension is desired
     * @param x here the width is saved
     * @param y here the height is saved
     */

    virtual void checkCharDimension(const string& id, const unsigned char& c, unsigned int& x, unsigned int& y) const
      throw(const Exception&);
    /**
     * Get maximum width of font.
     * @param id font id
     * @return the maximum width of a character
     */

    virtual unsigned int getMaxWidth(const string& id) const
      throw(const Exception&);
    /**
     * Get maximum height of font.
     * @param id font id
     * @return the maximum height of a character
     */

    virtual unsigned int getMaxHeight(const string& id) const
      throw(const Exception&);
    /**
     * Set space (pixels) between 2 lines.
     * @param id font id
     */

    virtual void setLineSpace(const string& id, const unsigned int& space)
      throw(const Exception&);
    /**
     * Get line space.
     * @param id font id
     */

    virtual unsigned int getLineSpace(const string& id) const
      throw(const Exception&);


  };

  typedef auto_ptr<IOSDManager> IOSDManager_var;




}

#endif
