/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "TVChannelList.hpp"

#include <fstream>

#include "nmm/utils/xml/xml++.hpp"

#include "nmm/base/serialize/MagicNumberValueStream.hpp"
#include "nmm/base/serialize/StdStream.hpp"

namespace NMM{

  /* ****************************************
   *           ChannelNotFoundException     *
   **************************************** */
  ChannelNotFoundException::ChannelNotFoundException()
    : Exception("ChannelNotFound")
  {
    registerAttributes();
  }

  ChannelNotFoundException::ChannelNotFoundException(string channel) 
    : Exception("Channel " + channel + "not found")
  {
    channelName = channel;
    registerAttributes();
  }

  ChannelNotFoundException::ChannelNotFoundException(unsigned int channel) 
    : Exception("Channel not found")
  {
    channelNumber = channel;
    registerAttributes();
  }
  
  void ChannelNotFoundException::registerAttributes(){
    registerAttribute(&channelName);
    registerAttribute(&channelNumber);
  }
 
  TConstructor<ChannelNotFoundException> channel_not_found_exception("ChannelNotFoundException", "NMM");

  /* ****************************************
   *           TVChannel         *
   **************************************** */  
  TVChannel::TVChannel()
    :NMM::SerializedValue("TVChannel","NMM"),
          channelName(""),
          channelNumber(0)
  {
    registerAttributes();
  }

  TVChannel::TVChannel(string name)
    :NMM::SerializedValue("TVChannel","NMM"),
          channelName(name),
          channelNumber(0)
  {
    registerAttributes();
  }

  TVChannel::TVChannel(const TVChannel& t)
    :SerializedValue("TVChannelDecription","NMM"),
          channelName(""),
          channelNumber(0)
  {
    registerAttributes();
    *this=t;
  }

  TVChannel::~TVChannel(){}

  void TVChannel::registerAttributes(){
    registerAttribute(&channelName);
    registerAttribute(&channelNumber);
    registerAttribute(&nameAliases);
    registerAttribute(&cards);
  }

  bool TVChannel::operator<(const TVChannel &cd) const{
    if(channelNumber.getRValue() < cd.getChannelNumber()){
      return true;
    }else{
      return false;
    }
  }

  void TVChannel::addCard(string URL, unsigned int channelNumber){
    cards.getRValue()[URL] = channelNumber;
  }

  TVChannel::cardList_t TVChannel::getCards(){ 
    return cards.getRValue();
  }
  
  string TVChannel::getName(){
    return channelName.getRValue();
  }

  list<string> TVChannel::getAliases(){
    return nameAliases.getRValue();
  }

  void TVChannel::addAlias(string a){
    nameAliases.getRValue().push_back(a);
    nameAliases.getRValue().sort(); 
    nameAliases.getRValue().unique();
  }

  OValueStream& operator<<(OValueStream& os, const list<TVChannel>& tvcd){
    return writeList<TVChannel>(os, tvcd, "TVChannel");
  }
  IValueStream& operator>>(IValueStream& is, list<TVChannel>& tvcd){
    return readList<TVChannel>(is, tvcd, "TVChannel");
  }

  TConstructor<TVChannel> nmm_tvchanneldescription("TVChannel", "NMM");
  TValueConstructor<list<TVChannel> > nmm_tvchanneldescription_list("list_TVChannel","NMM");



  /* ****************************************
   *           TVChannelList                *
   **************************************** */

  TVChannelList::TVChannelList()
    :SerializedValue("TVChannelList", "NMM"),
     channelNumber(0)
  {
    registerAttributes();
  }

  TVChannelList::TVChannelList(string filename)
    :SerializedValue("TVChannelList", "NMM"),
     channelNumber(0)
  {
    load(filename);
    registerAttributes();
  }

  TVChannelList::TVChannelList(const TVChannelList& t)
    :SerializedValue("TVChannelList", "NMM"),
     channelNumber(0)
  {
    registerAttributes();
    *this=t;
  }

  TVChannelList::~TVChannelList(){}

  void TVChannelList::registerAttributes(){
    registerAttribute(&channels);
    registerAttribute(&channelNumber);
  }

  void TVChannelList::addChannel(TVChannel cd){
    channels.getRValue().push_back(cd);
  }
  
  TVChannel TVChannelList::getChannelByNumber(unsigned int number){

    if(number >= channels.getRValue().size())     throw ChannelNotFoundException(number);

    list<TVChannel>::iterator i = channels.getRValue().begin();
    advance(i,number);
    return *i;
    
  }

  TVChannel TVChannelList::getChannelByName(string name){
    for(list<TVChannel>::iterator i = channels.getRValue().begin(); i != channels.getRValue().end(); i++){
      string s1(i->getChannelName()); transform(s1);
      string s2(name); transform(s2);
      if(s1==s2) return *i;
      list<string> aliases = i->getChannelAliases();
      for(list<string>::iterator j = aliases.begin(); j != aliases.end(); j++){
        string s1(*j); transform(s1);
        string s2(name); transform(s2);
        if(s1==s2) return *i;
      }
    }
    throw ChannelNotFoundException(name);
  }

  list<TVChannel> TVChannelList::getChannels(){
    return channels.getRValue();
  }

  Result TVChannelList::load(string filename){

    ifstream stdstream;
    stdstream.open(filename.c_str());

    if(stdstream){
      try{
        
        StdIStream s(&stdstream);
        MagicNumberIValueStream vs;
        vs.setIPartnerStream(&s);
        vs.readData();
        
        TVChannelList tmp;
        while(!vs.isEmpty()){
          vs >> tmp;
        }
        stdstream.close();

        *this = tmp;
        return SUCCESS;

      }catch(const Exception &e){
        GLOBAL_ERROR_STREAM("TVChannelList: Error reading channelList: e" << endl);
      }
      stdstream.close();
    }else{
      GLOBAL_ERROR_STREAM("TVChannelList: Error reading channelList: Failed to open " << filename << endl);
    }
    return SUCCESS;

  }

  Result TVChannelList::write(string filename){
    try{
      ofstream stdstream(filename.c_str());
      if(!stdstream){
        GLOBAL_ERROR_STREAM("TVChannelList: Error writing channelList: Failed to open " << filename << endl);
        return FAILURE;
      }

      StdOStream s(&stdstream);
      MagicNumberOValueStream xmls;
      xmls.setOPartnerStream(&s);
      xmls << *this;
      xmls.writeData();
      stdstream.close();
    }catch(const Exception &e){
      GLOBAL_ERROR_STREAM("TVChannelList: Error writing channelList: e" << endl);
      return FAILURE;
    }

    return SUCCESS;
  }
  OValueStream& operator<<(OValueStream& os, const list<TVChannelList>& nmm_tvchannellist){
    return writeList<TVChannelList>(os, nmm_tvchannellist, "TVChannelList");
  }
  IValueStream& operator>>(IValueStream& is, list<TVChannelList>& nmm_tvchannellist){
    return readList<TVChannelList>(is, nmm_tvchannellist, "TVChannelList");
  }

  TConstructor<TVChannelList> nmm_tvchannellist("TVChannelList", "NMM");
  TValueConstructor<list<TVChannelList> > nmm_tvchannellist_list("list_TVChannelList","NMM");
}
