
#ifndef NMM_IVISCACameraImpl_HPP
#define NMM_IVISCACameraImpl_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/NMMObject.hpp"
#include "IVISCACamera.hpp"


namespace NMM
{

  class IVISCACameraImpl  : public virtual NMMObject { 
  public:
    IVISCACameraImpl() { addInterface(new IVISCACamera(*this)); registerEvents(); };

    virtual ~IVISCACameraImpl() {};

  protected:
    /**
     * IF_Clear 
     * 
     * Clears the command buffers in the camera and 
     * cancels the command currently beeing executed.
     *
     * this function will be called automaticly by the init
     * function 
     *
     * input  : the number of the camera [1;7,8]
     *          camera_id 8 means broadcast, all 
     *          cameras will clear their command
     *          buffers
     * output : SUCCESS when success, else FAILURE
     */

    virtual Result IF_Clear(const unsigned char& value) = 0;
    /**
     * AddressSet
     *
     * Scan all cameras connected to the VISCA-bus 
     * and dispatch one VISCA-address for each camera found.
     *
     * This function will be called automaticly by the 
     * init function
     *
     * input  : none
     * output : SUCCESS when success, else FAILURE
     */

    virtual Result AddressSet() = 0;
    /**
     * getNumberOfCameras
     *
     * Returns the number of cameras connected to the VISCA-bus
     *
     * input  : none
     * output : number of cameras found
     */

    virtual Result getNumberOfCameras(unsigned char& value) = 0;
    /**
     * CommandCancel
     * 
     * Cancel the command currently executed by the camera
     * with the address <camera identificator> and the socket
     * number <socket id>.
     *
     * input  : camera identificator and socket id
     * output : SUCCESS, if the command is canceled
     *          FAILURE, if not
     */

    virtual Result CommandCancel(const unsigned char& value1, const unsigned char& value2) = 0;
    /**
     * Power functions
     *
     * The following functions controlls the powerstate 
     * of the cameras. Each camera can be controlled seperatly,
     * selectecd by the <camera id>. The AutoPowerOff-function sets 
     * the timer after the camera goes in standby-mode, when no 
     * VISCA-cammand arrived 
     *
     * input  : camera id
     * output : SUCCESS if success, else FAILURE
     */

    virtual Result PowerOn(const unsigned char& value) = 0;

    virtual Result PowerOff(const unsigned char& value) = 0;

    virtual Result AutoPowerOff(const unsigned char& value1, const int& value2) = 0;
    /**
     * Zoom functions
     *
     * The following functions controls the zoom of the camera. 
     * All that functions expect the <camera_id> as parameter.
     * Some of them allow additional speed settings 
     *
     * - The Range for the Zoom:
     *   optical zoom : [    0;16384]
     *   digital zoom : [16384;28672]
     * - ZoomDZoom(On|Off) turn the DigitalZoom On or Off.
     */

    virtual Result ZoomStop(const unsigned char& value) = 0;

    virtual Result ZoomTeleStd(const unsigned char& value) = 0;

    virtual Result ZoomWideStd(const unsigned char& value) = 0;

    virtual Result ZoomTeleVar(const unsigned char& value1, const unsigned char& value2) = 0;

    virtual Result ZoomWideVar(const unsigned char& value1, const unsigned char& value2) = 0;

    virtual Result ZoomDirect(const unsigned char& value1, const int& value2) = 0;

    virtual Result ZoomDigitalZoomOnOff(const unsigned char& value1, const bool& value2) = 0;
    /**
     * Focus functions
     * 
     * the Focus* functions are controling the settings of te
     * camera's focus. All functions expect the <camera_id> 
     * as first parameter, some an additional parameter
     */

    virtual Result FocusStop(const unsigned char& value) = 0;

    virtual Result FocusFarStd(const unsigned char& value) = 0;

    virtual Result FocusNearStd(const unsigned char& value) = 0;

    virtual Result FocusFarVar(const unsigned char& value1, const unsigned char& value2) = 0;

    virtual Result FocusNearVar(const unsigned char& value1, const unsigned char& value2) = 0;

    virtual Result FocusDirect(const unsigned char& value1, const unsigned short& value2) = 0;

    virtual Result FocusAutoFocus(const unsigned char& value) = 0;

    virtual Result FocusManualFocus(const unsigned char& value) = 0;

    virtual Result FocusAutoManual(const unsigned char& value) = 0;

    virtual Result FocusOnePushTrigger(const unsigned char& value) = 0;

    virtual Result FocusInfinity(const unsigned char& value) = 0;

    virtual Result FocusAFSensHigh(const unsigned char& value) = 0;

    virtual Result FocusAFSensLow(const unsigned char& value) = 0;

    virtual Result FocusNearLimit(const unsigned char& value1, const unsigned short& value2) = 0;
    /**
     * WhiteBalance
     */

    virtual Result WhiteBalance(const unsigned char& value1, const unsigned char& value2, const unsigned char& value3) = 0;

    virtual Result WhiteBalanceAuto(const unsigned char& value) = 0;

    virtual Result WhiteBalanceIndoor(const unsigned char& value) = 0;

    virtual Result WhiteBalanceOutdoor(const unsigned char& value) = 0;

    virtual Result WhiteBalanceOnePush(const unsigned char& value) = 0;

    virtual Result WhiteBalanceAT(const unsigned char& value) = 0;

    virtual Result WhiteBalanceManual(const unsigned char& value) = 0;

    virtual Result WhiteBalanceOnePushTrigger(const unsigned char& value) = 0;
    /**
     * RGain
     */

    virtual Result RGainReset(const unsigned char& value) = 0;

    virtual Result RGainUp(const unsigned char& value) = 0;

    virtual Result RGainDown(const unsigned char& value) = 0;

    virtual Result RGainDirect(const unsigned char& value1, const unsigned short& value2) = 0;
    /**
     * BGain
     */

    virtual Result BGainReset(const unsigned char& value) = 0;

    virtual Result BGainUp(const unsigned char& value) = 0;

    virtual Result BGainDown(const unsigned char& value) = 0;

    virtual Result BGainDirect(const unsigned char& value1, const unsigned short& value2) = 0;
    /**
     * Automatic Exposure 
     */

    virtual Result AutomaticExposureFull(const unsigned char& value) = 0;

    virtual Result AutomaticExposureManual(const unsigned char& value) = 0;

    virtual Result AutomaticExposureShutterPriority(const unsigned char& value) = 0;

    virtual Result AutomaticExposureIrisPriority(const unsigned char& value) = 0;

    virtual Result AutomaticExposureGainPriority(const unsigned char& value) = 0;

    virtual Result AutomaticExposureBright(const unsigned char& value) = 0;

    virtual Result AutomaticExposureShutterAuto(const unsigned char& value) = 0;

    virtual Result AutomaticExposureIrisAuto(const unsigned char& value) = 0;

    virtual Result AutomaticExposureGainAuto(const unsigned char& value) = 0;
    /**
     * Slow Shutter
     */

    virtual Result SlowShutter(const unsigned char& value1, const bool& value2) = 0;
    /**
     * Shutter
     */

    virtual Result ShutterReset(const unsigned char& value) = 0;

    virtual Result ShutterUp(const unsigned char& value) = 0;

    virtual Result ShutterDown(const unsigned char& value) = 0;

    virtual Result ShutterDirect(const unsigned char& value1, const unsigned short& value2) = 0;
    /**
     * Iris
     */

    virtual Result IrisReset(const unsigned char& value) = 0;

    virtual Result IrisUp(const unsigned char& value) = 0;

    virtual Result IrisDown(const unsigned char& value) = 0;

    virtual Result IrisDirect(const unsigned char& value1, const unsigned short& value2) = 0;
    /**
     * Gain
     */

    virtual Result GainReset(const unsigned char& value) = 0;

    virtual Result GainUp(const unsigned char& value) = 0;

    virtual Result GainDown(const unsigned char& value) = 0;

    virtual Result GainDirect(const unsigned char& value1, const unsigned short& value2) = 0;
    /**
     * Bright
     */

    virtual Result BrightReset(const unsigned char& value) = 0;

    virtual Result BrightUp(const unsigned char& value) = 0;

    virtual Result BrightDown(const unsigned char& value) = 0;

    virtual Result BrightDirect(const unsigned char& value1, const unsigned short& value2) = 0;
    /**
     * Exposure Compensation
     */

    virtual Result ExposureCompensationOnOff(const unsigned char& value1, const bool& value2) = 0;

    virtual Result ExposureCompensationReset(const unsigned char& value) = 0;

    virtual Result ExposureCompensationUp(const unsigned char& value) = 0;

    virtual Result ExposureCompensationDown(const unsigned char& value) = 0;

    virtual Result ExposureCompensationDirect(const unsigned char& value1, const unsigned short& value2) = 0;
    /** 
     * BackLight
     */

    virtual Result BackLightOnOff(const unsigned char& value1, const bool& value2) = 0;
    /**
     * Aperture
     */

    virtual Result ApertureReset(const unsigned char& value) = 0;

    virtual Result ApertureUp(const unsigned char& value) = 0;

    virtual Result ApertureDown(const unsigned char& value) = 0;

    virtual Result ApertureDirect(const unsigned char& value1, const unsigned short& value2) = 0;
    /**
     * Wide Mode
     */

    virtual Result WideModeOff(const unsigned char& value) = 0;

    virtual Result WideModeCinema(const unsigned char& value) = 0;

    virtual Result WideMode16to9Full(const unsigned char& value) = 0;
    /**
     *  Mirror image on or off
     */

    virtual Result LR_Reverse(const unsigned char& value1, const bool& value2) = 0;
    /**
     *  Freeze image on or off
     */

    virtual Result Freeze(const unsigned char& value1, const bool& value2) = 0;
    /**
     * Picture Effects
     *
     * These functions allows to turn internal effects on or off. 
     * They expect as parameter the <camera_id>.
     */

    virtual Result PictureEffectOff(const unsigned char& value) = 0;

    virtual Result PictureEffectPastel(const unsigned char& value) = 0;

    virtual Result PictureEffectNegArt(const unsigned char& value) = 0;

    virtual Result PictureEffectSepia(const unsigned char& value) = 0;

    virtual Result PictureEffectBW(const unsigned char& value) = 0;

    virtual Result PictureEffectSolarize(const unsigned char& value) = 0;

    virtual Result PictureEffectMosaic(const unsigned char& value) = 0;

    virtual Result PictureEffectSlim(const unsigned char& value) = 0;

    virtual Result PictureEffectStretch(const unsigned char& value) = 0;
    /**
     * Digital Effects
     *
     * These functions allows to turn internal digital effects 
     * on or off. They expect as parameter the <camera_id>.
     *
     * Please don't call DigitalEffect( in unsigned char value1, in unsigned char ), its a parent
     * functions for all the other effect functions  
     */

    virtual Result DigitalEffectOff(const unsigned char& value) = 0;

    virtual Result DigitalEffectStill(const unsigned char& value) = 0;

    virtual Result DigitalEffectFlash(const unsigned char& value) = 0;

    virtual Result DigitalEffectLumi(const unsigned char& value) = 0;

    virtual Result DigitalEffectTrail(const unsigned char& value) = 0;

    virtual Result DigitalEffectLevel(const unsigned char& value1, const unsigned char& value2, const unsigned char& value3) = 0;
    /**
     * Memory (user defined presets)
     */

    virtual Result Memory(const unsigned char& value1, const unsigned char& value2, const unsigned char& value3) = 0;

    virtual Result MemoryReset(const unsigned char& value1, const unsigned char& value2) = 0;

    virtual Result MemorySet(const unsigned char& value1, const unsigned char& value2) = 0;

    virtual Result MemoryRecall(const unsigned char& value1, const unsigned char& value2) = 0;
    /**
     * DataScreen (OSD)
     */

    virtual Result DataScreenOnOff(const unsigned char& value1, const bool& value2) = 0;
    /**
     * IR (remote controller)
     */

    virtual Result IRReceiveOnOff(const unsigned char& value1, const bool& value2) = 0;

    virtual Result IRReceiveReturn(const unsigned char& value1, const bool& value2) = 0;
    /** 
     * PanTiltDriveMotion
     *
     * input  : camera-identifier value1,
     *          motion x-direction, motion y-direction, 
     *          speed x-direction, speed y-direction
     * output : SUCCESS, if command send with success
     *          FAILURE, if an error occured
     * 
     * used for PanTiltDrive-motion functions
     */

    virtual Result PanTiltDriveMotion(const unsigned char& value1, const unsigned char& value2, const unsigned char& value3, const unsigned char& value4, const unsigned char& value5) = 0;

    virtual Result PanTiltDriveUp(const unsigned char& value1, const unsigned char& value2, const unsigned char& value3) = 0;

    virtual Result PanTiltDriveDown(const unsigned char& value1, const unsigned char& value2, const unsigned char& value3) = 0;

    virtual Result PanTiltDriveLeft(const unsigned char& value1, const unsigned char& value2, const unsigned char& value3) = 0;

    virtual Result PanTiltDriveRight(const unsigned char& value1, const unsigned char& value2, const unsigned char& value3) = 0;

    virtual Result PanTiltDriveUpLeft(const unsigned char& value1, const unsigned char& value2, const unsigned char& value3) = 0;

    virtual Result PanTiltDriveUpRight(const unsigned char& value1, const unsigned char& value2, const unsigned char& value3) = 0;

    virtual Result PanTiltDriveDownLeft(const unsigned char& value1, const unsigned char& value2, const unsigned char& value3) = 0;

    virtual Result PanTiltDriveDownRight(const unsigned char& value1, const unsigned char& value2, const unsigned char& value3) = 0;

    virtual Result PanTiltDriveStop(const unsigned char& value1, const unsigned char& value2, const unsigned char& value3) = 0;
    /** 
     * PanTiltDrivePosition
     *
     * input  : camera-identifier,
     *          position-mode (relative or absolute)
     *          motion x-direction in coordinates
     *          morion y-direction in coordinates
     *          speed x-direction, speed y-direction
     * output : SUCCESS, if command send with success
     *          FAILURE, if an error occured
     * 
     * used for PanTiltDrive-position functions
     */

    virtual Result PanTiltDrivePosition(const unsigned char& value1, const unsigned char& value2, const int& value3, const int& value4, const unsigned char& value5, const unsigned char& value6) = 0;

    virtual Result PanTiltDriveAbsolute(const unsigned char& value1, const int& value2, const int& value3, const unsigned char& value4, const unsigned char& value5) = 0;

    virtual Result PanTiltDriveRelative(const unsigned char& value1, const int& value2, const int& value3, const unsigned char& value4, const unsigned char& value5) = 0;

    virtual Result PanTiltDriveAbsoluteDeg(const unsigned char& value1, const int& value2, const int& value3, const unsigned char& value4, const unsigned char& value5) = 0;

    virtual Result PanTiltDriveRelativeDeg(const unsigned char& value1, const int& value2, const int& value3, const unsigned char& value4, const unsigned char& value5) = 0;

    virtual Result PanTiltDriveHome(const unsigned char& value) = 0;

    virtual Result PanTiltDriveReset(const unsigned char& value) = 0;

    virtual Result PanTiltLimitSet(const unsigned char& value1, const unsigned char& value2, const int& value3, const int& value4) = 0;

    virtual Result PanTiltLimitClear(const unsigned char& value1, const unsigned char& value2) = 0;
    /**
     * Inquiry Command List 
     */
    /**
     * PowerInq
     *
     * input : The camera ID
     * ouput : The status of the Camera
     *          true = On
     *          false =Off(Standby)
     */

    virtual Result PowerInq(const unsigned char& value1, bool& value2) = 0;

    virtual Result AutoPowerOffInq(const unsigned char& value1, unsigned short& value2) = 0;

    virtual Result DZoomModeInq(const unsigned char& value1, bool& value2) = 0;

    virtual Result ZoomPosInq(const unsigned char& value1, unsigned short& value2) = 0;

    virtual Result FocusModeInq(const unsigned char& value1, bool& value2) = 0;

    virtual Result FocusPosInq(const unsigned char& value1, unsigned short& value2) = 0;

    virtual Result AFModeInq(const unsigned char& value1, bool& value2) = 0;

    virtual Result FocusNearLimitInq(const unsigned char& value1, unsigned short& value2) = 0;

    virtual Result WhiteBalanceModeInq(const unsigned char& value1, unsigned char& value2) = 0;

    virtual Result RGainInq(const unsigned char& value1, unsigned char& value2) = 0;

    virtual Result BGainInq(const unsigned char& value1, unsigned short& value2) = 0;

    virtual Result AutoExposureModeInq(const unsigned char& value1, unsigned char& value2) = 0;

    virtual Result SlowShutterModeInq(const unsigned char& value1, bool& value2) = 0;

    virtual Result ShutterPosInq(const unsigned char& value1, unsigned short& value2) = 0;

    virtual Result IrisPosInq(const unsigned char& value1, unsigned short& value2) = 0;

    virtual Result GainPosInq(const unsigned char& value1, unsigned short& value2) = 0;

    virtual Result BrightPosInq(const unsigned char& value1, unsigned short& value2) = 0;

    virtual Result ExposureCompensationModeInq(const unsigned char& value1, bool& value2) = 0;

    virtual Result ExposureCompensationPosInq(const unsigned char& value1, unsigned short& value2) = 0;

    virtual Result BackLightModeInq(const unsigned char& value1, bool& value2) = 0;

    virtual Result ApertureInq(const unsigned char& value1, unsigned short& value2) = 0;

    virtual Result WideModeInq(const unsigned char& value1, unsigned char& value2) = 0;

    virtual Result LR_ReverseModeInq(const unsigned char& value1, bool& value2) = 0;

    virtual Result FreezeModeInq(const unsigned char& value1, bool& value2) = 0;

    virtual Result PictureEffectModeInq(const unsigned char& value1, unsigned char& value2) = 0;

    virtual Result DigitalEffectModeInq(const unsigned char& value1, unsigned char& value2) = 0;

    virtual Result DigitalEffectLevelInq(const unsigned char& value1, unsigned char& value2) = 0;

    virtual Result MemoryInq(const unsigned char& value1, unsigned char& value2) = 0;

    virtual Result DatascreenInq(const unsigned char& value1, bool& value2) = 0;

    virtual Result PanTiltModeInq(const unsigned char& value1, unsigned short& value2) = 0;

    virtual Result PanTiltMaxSpeedInq(const unsigned char& value1, unsigned char& value2, unsigned char& value3) = 0;

    virtual Result PanTiltPosInq(const unsigned char& value1, int& value2, int& value3) = 0;

    virtual Result VideoSystemInq(const unsigned char& value1, unsigned char& value2) = 0;

    virtual Result DeviceTypeInq(const unsigned char& value1, unsigned short& value2, unsigned short& value3, unsigned short& value4, unsigned char& value5) = 0;
    /**
     * misc stuff
     */

    virtual void verbose(const string& value1, const string& value2) = 0;

    virtual bool islocked(const unsigned char& value) = 0;


  private:
    void registerEvents();
  };

}

#endif
