/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, 
 * USA.
 */

#include "iPAQProducer.hpp"

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/fb.h>
#include <sys/mman.h>
#include <sys/time.h>


#include "nmm/interfaces/video/display/IDisplayEvents.hpp"
#include "nmm/base/IReceiver.hpp"
#include "nmm/utils/utils.hpp"
#include "nmm/utils/xml/xml++.hpp"


namespace NMM {
  
  iPAQProducer::iPAQProducer() {

    /* Terminal */
    tty_handle = -2;
    tty_no = 0;
    tty_orig = 0;

    /* default" layout */
    keymap.clear();
    
    addKey( RECORD,            string("KEY_Record") );
    addKey( RECORD|RELEASED,   string("KEY_Record") );  
    
    addKey( CALENDAR,          string("KEY_Calendar") );
    addKey( CALENDAR|RELEASED, string("KEY_Calendar") ); 
      
    addKey( CONTACTS,          string("KEY_Contacts") ); 
    addKey( CONTACTS|RELEASED, string("KEY_Contacts") );
      
    addKey( MAIL,              string("KEY_Mail") );
    addKey( MAIL|RELEASED,     string("KEY_Mail") );
    
    addKey( START,             string("KEY_Start") );
    addKey( START|RELEASED,    string("KEY_Start") );
    
    addKey( LEFT,              string("KEY_Left") );
    addKey( LEFT|RELEASED,     string("KEY_Left") );

    addKey( RIGHT,             string("KEY_Right") );
    addKey( RIGHT|RELEASED,    string("KEY_Right") );
      
    addKey( UP,                string("KEY_Up") );
    addKey( UP|RELEASED,       string("KEY_Up") );
    
    addKey( DOWN,              string("KEY_Down") );
    addKey( DOWN|RELEASED,     string("KEY_Down") );
        
    addKey( ACTION,            string("KEY_Action") );
    addKey( ACTION|RELEASED,   string("KEY_Action") );

    keymaps[ string("default") ] = keymap;

    currentLayout = "default";
  }

  iPAQProducer::~iPAQProducer() {
    close_kb();
    close_vt();
  }


  bool iPAQProducer::open_vt(void) {
    const char* ttynames[] = {"/dev/tty0", "/dev/tty", NULL};
    const char* ttyfmts[] = {"/dev/tty%d", "/dev/tty%02x", "/dev/tty%x", "/dev/tty%02d", NULL};
    int cttyname = 0;
    char ttyname[11];
    int i = 0;

    vt_stat tty_vts;

    /* open current console */
    while ((tty_handle = open(ttynames[cttyname], O_WRONLY )) == -1) {
      GLOBAL_DEBUG_STREAM("WARNING: Virtual terminal open failed on \'" << ttynames[cttyname] << "\' ." << endl);
    
      if (ttynames[++cttyname] == NULL) {
        GLOBAL_ERROR_STREAM("ERROR: Failed to open virtual terminal." << endl);
        return false;
      } else {
        GLOBAL_DEBUG_STREAM("WARNING: Trying open virtual terminal \'" << ttynames[cttyname] << "\' " << endl);
      }
    }

    /* get number of free VT */
    if (ioctl(tty_handle, VT_OPENQRY, &tty_no) == -1) {
      GLOBAL_ERROR_STREAM("ERROR: No free ttys" << endl);
      return false;
    }

    /* close current console */
    if (close(tty_handle) == -1) {
      GLOBAL_ERROR_STREAM("ERROR: Failed to close \'" << ttynames[cttyname] << "\' ." << endl);
      return false;
    }

    /* open new VT */
    do {
      sprintf(ttyname, ttyfmts[i++], tty_no);
    } while ((ttyfmts[i] != NULL) && (tty_handle = open(ttyname, O_RDONLY)) == -1);

    if (tty_handle == -1) {
      GLOBAL_ERROR_STREAM("ERROR: Failed to open virtual terminal." << endl);
      return false;
    }

    if ((i = open("/dev/tty", O_RDWR)) >= 0) {
	ioctl(i, TIOCNOTTY, 0);
	close(i);
    }

    /* get current VT number (for switching back) */
    if (ioctl(tty_handle, VT_GETSTATE, &tty_vts) == -1) {
      GLOBAL_ERROR_STREAM("ERROR: Couldn't get virtual terminal state.\n" << endl);
      return false;
    }
    tty_orig = tty_vts.v_active;

    /* switch to new VT */
    if (ioctl(tty_handle, VT_ACTIVATE, tty_no) == -1) {
      GLOBAL_ERROR_STREAM("ERROR: Couldn't switch to virtual terminal \'" << tty_no << "\' " << endl);
      return false;
    }

    /* wait for new VT become active */
    if (ioctl(tty_handle, VT_WAITACTIVE, tty_no) == -1) {
      GLOBAL_ERROR_STREAM("ERROR: Virtual terminal " <<  tty_no << "didn't become active." << endl);
      return false;
    }

    /* get mode of new VT */
    if (ioctl(tty_handle, VT_GETMODE, &tty_vtm) == -1) {
      GLOBAL_ERROR_STREAM("ERROR: Couldn't get mode of virtual terminal " << tty_no << endl);
      return false;
    }

    tty_vtm.mode = VT_PROCESS;

    /* set new mode parameters */
    if (ioctl(tty_handle, VT_SETMODE, &tty_vtm) == -1) {
      GLOBAL_ERROR_STREAM("ERROR: Couldn't set mode of virtual terminal \'" << tty_no << "\' ." << endl);
      return false;
    }

    /* disable cursor */
    if (ioctl(tty_handle, KDSETMODE, KD_GRAPHICS) == -1) {
      GLOBAL_ERROR_STREAM("ERROR: Couldn't set keyboard graphics mode on virtual terminal \'" << tty_no << "\' " << endl);
      return false;
    }
    
    /* all fine */
    return true;   

  } /* open_vt */



  bool iPAQProducer::close_vt(void) {
    if (tty_handle > 0) {
      
      /* enable cursor */
      if (ioctl(tty_handle, KDSETMODE, KD_TEXT) == -1) {
        GLOBAL_ERROR_STREAM("ERROR: Couldn't set keyboard graphics mode on virtual terminal \'" << tty_no << "\' " << endl);
        return false;
      }

      /* restore mode parameters */
      tty_vtm.mode = VT_AUTO;
      if (ioctl(tty_handle, VT_SETMODE, &tty_vtm) == -1) {
        GLOBAL_ERROR_STREAM("ERROR: Couldn't set mode of virtual terminal \'" << tty_no << "\' " << endl);        
        return false;
      }

      /* switch back to original VT */
      if (ioctl(tty_handle, VT_ACTIVATE, tty_orig) == -1) {
        GLOBAL_ERROR_STREAM("ERROR: Couldn't switch to virtual terminal \'" << tty_no << "\' " << endl);
        return false;
      }

      /* wait for new VT become active */
      if (ioctl(tty_handle, VT_WAITACTIVE, tty_orig) == -1) {
        GLOBAL_ERROR_STREAM("ERROR: Virtual terminal \'" << tty_no << "\' didn't become active." << endl);
        return false;
      }

      /* deallocate VT */
      if (ioctl(tty_handle, VT_DISALLOCATE, tty_no) == -1) {
        GLOBAL_DEBUG_STREAM("WARNING: Couldn't deallocate virtual terminal \'" << tty_no << "\' " << endl);
      }

      /* close VT */
      if (close(tty_handle) == -1) {
        GLOBAL_ERROR_STREAM("ERROR: Failed to close virtual terminal." << endl);
        return false;
      } else {
        tty_handle = -1;
      }
    }
    
    /* all fine */
    return true;
        
  } /* close_vt */



  bool iPAQProducer::open_kb(void) {
    termios kb_newttysets;

    /* Build keyboard translation tables */
    // get_keymaps(tty_handle);

    /* Save old kbd state and tty settings */
    ioctl(tty_handle, KDGKBMODE, &kb_oldkbdmode);
    tcgetattr(tty_handle, &kb_oldttysets);

    /* Set new kbd mode and tty settings */
    ioctl(tty_handle, KDSKBMODE, K_MEDIUMRAW);
    kb_newttysets = kb_oldttysets;
    kb_newttysets.c_iflag = (IGNPAR | IGNBRK) & (~PARMRK) & (~ISTRIP);
    kb_newttysets.c_oflag = 0;
    kb_newttysets.c_cflag = CREAD | CS8;
    kb_newttysets.c_lflag = 0;
    kb_newttysets.c_cc[VTIME] = 0;
    kb_newttysets.c_cc[VMIN] = 1;
    cfsetispeed(&kb_newttysets, B9600);
    cfsetospeed(&kb_newttysets, B9600);
    tcsetattr(tty_handle, TCSANOW, &kb_newttysets);    

    return true;
  } /* open_kb */



  bool iPAQProducer::close_kb(void) {
    tcsetattr(tty_handle, TCSANOW, &kb_oldttysets);
    ioctl(tty_handle, KDSKBMODE, kb_oldkbdmode);     

    return true;
  } /* close_kb */



  void iPAQProducer::addKey(unsigned char key, string event) {
    MutexGuard mg( keymap_mutex );
    /* if actuel vector size < keycode, then resize the vector */
    if (key >= keymap.size()) {
      unsigned int tofill = key - keymap.size();
      for (unsigned int i=0; i<tofill; ++i) {
        keymap.push_back("x");
      }
      keymap.push_back(event);
    } else {
      keymap[key] = event;
    }  
  } /* addKey */



  string iPAQProducer::getKeyMapping(unsigned char key) {
    MutexGuard mg( keymap_mutex );
    if (key >= keymap.size()) {
      return "";
    }
    return keymap[key];
  } /* getKeyMapping */



  Result iPAQProducer::readConfig(string fname) {
    /* open XML document */
    XMLTree* xml = new XMLTree( fname );
    
    /* get pointer to XML root */
    XMLNode* rootnode = xml->root();
    if (rootnode == NULL) {
      GLOBAL_ERROR_STREAM("iPAQProducer: readConfig(" << fname << "): no XML root node found" << endl);
      return FAILURE;
    }
    
    if (rootnode->name() != "configuration") {
      GLOBAL_ERROR_STREAM("iPAQProducer: readConfig(" << fname << "): no <configuration> tag found" << endl);
      return FAILURE;
    }
    
    XMLNode* node1;
    
    /* parse the keyboard configuration */
    XMLNodeList klist, nlist;
    XMLNodeIterator iter, stop;
    
    klist = rootnode->children("keyboard");
    iter = klist.begin();
    if ( iter == klist.end() ) {
      GLOBAL_DEBUG_STREAM("iPAQProducer: readConfig(" << fname << "): no <keyboard> entry found in configuration" << endl);
      return FAILURE;
    }

    XMLProperty* kNodeprop;
    string layout = "";
    int i=0;

    /* go through all keyboard layouts */
    for (XMLNodeIterator ki=klist.begin(); ki != klist.end(); ++ki) {
      node1     = *ki;
      kNodeprop = node1->property( "layout" );
      if ( kNodeprop ) {
        layout = kNodeprop->value();
      } else {
        layout = string("Layout") + i;
        i++;
      }

      GLOBAL_DEBUG_STREAM("Found keyboard layout [" << layout << "]" << endl); 

      /* take curent layout */
      node1 = *ki; 
    
      /* Get a list with all <event> entries */
      nlist = node1->children( "event" );

      iter = nlist.begin();
      stop = nlist.end();
    
      /* clear all layout */
      keymap.clear();

      /* Process each <event> entry */
      while ( iter != stop ) {
        XMLProperty* prop;
        XMLNode* node = *iter;
        iter++;
      
        /* Try to find <event> entry with "key" attribute */
        string key;
        prop = node->property( "key" );
        if ( prop ) key = prop->value();
      
        /* now the cdata-string of <event> tag containing the event-string */
        string event = node->getData();
        GLOBAL_DEBUG_STREAM("  Found Key: " << key << " -> " << event << endl);

        if ( key == "RECORD" ) { 
          addKey( RECORD,            event );
          addKey( RECORD|RELEASED,   event ); 
        } 

        else if ( key == "CALENDAR" ) { 
          addKey( CALENDAR,          event );
          addKey( CALENDAR|RELEASED, event ); 
        }
      
        else if ( key == "CONTACTS" ) { 
          addKey( CONTACTS,          event ); 
          addKey( CONTACTS|RELEASED, event );
        }

        else if ( key == "MAIL" ) { 
          addKey( MAIL,              event );
          addKey( MAIL|RELEASED,     event ); 
        }

        else if ( key == "START" ) { 
          addKey( START,             event );
          addKey( START|RELEASED,    event ); 
        }

        else if ( key == "LEFT" ) { 
          addKey( LEFT,              event );
          addKey( LEFT|RELEASED,     event );
        }

        else if ( key == "RIGHT" ) { 
          addKey( RIGHT,             event );
          addKey( RIGHT|RELEASED,    event );
        }

        else if ( key == "UP" ) { 
          addKey( UP,                event );
          addKey( UP|RELEASED,       event );
        }

        else if ( key == "DOWN" ) {
          addKey( DOWN,              event );
          addKey( DOWN|RELEASED,     event );
        }
        
        else if ( key == "ACTION" ) {
          addKey( ACTION,            event );
          addKey( ACTION|RELEASED,   event );
        }

      } /* while iter != stop */


      /* save layout */
      GLOBAL_DEBUG_STREAM("Save keyboard layout [" << layout << "]" << endl); 
      keymaps[ layout ] = keymap;

    } /* for ki != klist.end() */

   
    return SUCCESS;   
  } /* read config */



  void iPAQProducer::start() {
    open_vt();
    open_kb();
    //    open_ts();
    
    setFileDescriptor(tty_handle);

    InputProducer::start();
  } /* start */



  void iPAQProducer::stop() {
    InputProducer::stop();
    //    close_ts();
    close_kb();
    close_vt();
  } /* stop */



  void iPAQProducer::setLayout(string layout) {
    if (!layout.empty()) {
      map< const string, vector<string> >::iterator p = keymaps.find( layout );
      if (p != keymaps.end()) {
        /* there exist a layout with the name <layout> */
        GLOBAL_DEBUG_STREAM("Switch to layout " << layout << endl);
        {
          MutexGuard mg( keymap_mutex );          
          keymap.clear();
          keymap = keymaps[layout];
          currentLayout = layout;
        }
      } else {
        GLOBAL_WARNING_STREAM("Keyboard layout [" << layout << "] not found" << endl); 
      }
    }
  } /* setLayout */



  string iPAQProducer::getLayout() {
    return currentLayout;
  } /* getLayout */



  void iPAQProducer::handleInput() {
    auto_ptr<Event> event;

    unsigned char tty_input = 0;
    string event_str = "";
    
    int r = read(tty_handle, &tty_input, sizeof(char));
    if (r != sizeof(char)) {

      GLOBAL_ERROR_STREAM("Keyboard error!" << endl);

    } else {

      /* get event for pressed|released key */
      event_str = getKeyMapping(tty_input); 

      /* Bit 7 is set == key released */
      if ((tty_input & 128) == 128) { 
        event.reset(IDisplayEvents::create_keyReleased( event_str ));
      } else {
        event.reset(IDisplayEvents::create_keyPressed( event_str ));
      }
      
      if (event.get()) { 
        sendEvent(*event);
      }
      
   } /* if */
    
  } /* handleInput */

} /* namespace NMM */

