/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *                     
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef TVGLIBXML2DATABASE_HPP
#define TVGLIBXML2DATABASE_HPP TVGLIBXML2DATABASE_HPP

#include "nmm/utils/DirReader.hpp"
#include "nmm/utils/xml/xml++.hpp"

#include "nmm/interfaces/misc/tvguide/ChannelDescription.hpp"
#include "nmm/interfaces/misc/tvguide/ProgramDescription.hpp"
#include "nmm/interfaces/misc/tvguide/ProgramFilter.hpp"

#include "nmm/interfaces/misc/tvguide/ITVGuide.hpp"

#include "nmm/utils/thread/Thread.hpp"
#include <string>
#include <list>

#include <time.h>


namespace NMM{
  class TVGLibXML2Database{

  private:
    list<ChannelDescription> channels; /** List with all available channels */
    string workDir; /** Directory, where the information is stored */
    list<string> categories; /** List with all available categories */
    
    int epsilon;
    
    DirReader directory;
    
    ThreadMutex tvmutex;
    ThreadMutex insmutex;
    
    XMLTree* xml;
    XMLNode* root;

    bool termUpdate;

    void transform(string &s){
      size_t sIt = string::npos;
      size_t sOld = 0;
      do{
        sIt = s.find_first_of(" ", sOld);
        if(sIt != string::npos) s.erase(sIt,1);
        sOld = sIt+1;
      }while(sIt != string::npos);

      for (unsigned int i = 0; i <= s.length(); i++)
        s[i] = toupper(s[i]);
    }
   
    ITVGuide_var itvguide; //to send events to TVGuide-Object when eg new channels arrive

  public:

    TVGLibXML2Database(ITVGuide* itvg);
    ~TVGLibXML2Database();

    /** Insert Channel into DB. Indicates, that information for this channel is available
     *@param id ID of the channel
     *@param name Name of the channel
     */
    void insertChannel(const string &id, const string &name);
    /** Set the directory, where information is stored
     *@param path Complete path of the directory
     */
    void setDirectory(const string &path);
    /** Get the directory, where the informations are stored
     *@return complete path of the directory
     */
    string getDirectory() throw (const NMM::Exception&) ;

    /** Get a list with all available categories
     *@return List of categories
     */
    list<string> getCategories() throw (const NMM::Exception&);
    
    /** Get a list with all available channels. 
     *@return list of channels
     */
    list<ChannelDescription> getChannelsList()throw (const NMM::Exception&);
    /** Look for name of channel with given ID
     *@param id ID of the channel
     *@return string name of channel, if the ID is known
     *@return empty-string if ID is unknown
     */
    string  getChannelById(const string &id);
    /** Look for ID of channel with given name
     *@param name name of the channel
     *@return string ID of channel, if the name is known
     *@return empty-string if name is unknown
     */
    string getChannelByName(const string &name);

    /** Insert an event to the database
     *@param pd ProgramDescription containing the information of the event
     */
    void insertEvent(const ProgramDescription &pd);
    /** look for a specific event in the time-inteval [start,end] on the channel with the given ID
     *@param start Start-Time of the interval. Time-representation: YYYYMMDDhhmm
     *@param end End-Time of the interval. Time-representation: YYYYMMDDhhmm
     *@param id ID of the Channel
     *@return ProgramDescription of the Event
     */
    ProgramDescription getEvent(const string &start, const string &end, const string &id);
    /** Get a list of all Events of a given channel on a specific day
     *@param day Representation: YYYYMMDD
     *@param id id of the channel
     *@return List of ProgramDescriptions containing the events found
     */
    list<ProgramDescription> getEventsOnDayOnChannel(const string &day, const string &chan);

    /** Search for Events matching the given filter
     *@param pf ProgramFilter to apply on the search
     *@return List with ProgramDescriptions of events found
     */
    list<ProgramDescription> search(const ProgramFilter &pf);

    /** Add an Extended Description to an event. if the event is not available in the database, it is added.
     *@param start Start-time of the event
     *@param end Stop-Time of the event
     *@param id ID of the channel
     *@param desc description to be added
     */
    void addExtDescription(const string& start, const string& end, const string& id, const string& desc);

    /** get the current event of the channel specified. Make sure, your hardware-clock is set up correctly
     *@param channel either name or id
     *@return ProgramDescription with event found
     */
    ProgramDescription getCurrentEventOnChannel(const string& channel);
    /** get the next event of the channel specified. Make sure, your hardware-clock is set up correctly
     *@param channel either name or id
     *@return ProgramDescription with event found
     */
    ProgramDescription getFollowingEventOnChannel(const string& channel);

    
  private:
    void readChannels(); //read channels from file
    void saveChannels(); //save channels to file

    void readCategories(); //read categories from file
    void saveCategories(); //save categories to file

    bool fileExists(string file); // test, if the file specified exists
    string transformSpecialCharsIn(string s); //transform special chars, eg umlaut-representation to 'real' umlauts
    string transformSpecialCharsOut(string s); //transform 'real' special chars to utf-8-representation

    int executeCommand(const string& command); //execute an update-command
    int readXMLTV(const string& file); //read the XML-File from executeCommand

    bool compareStrings(const string s1, const string s2); //test if 2 Strings are approximately equal

  };
}
#endif
