/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *                     
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef NMM_TVGUIDE_HPP
#define NMM_TVGUIDE_HPP NMM_TVGUIDE_HPP

#include "nmm/interfaces/misc/tvguide/ITVGuideImpl.hpp"
#include "nmm/interfaces/misc/tvguide/ITVGuideUpdateFinishedImpl.hpp"
#include "nmm/interfaces/misc/tvguide/ITVGuidePlugin.hpp"
#include "nmm/interfaces/misc/tvguide/ITVGuide.hpp"
#include "nmm/interfaces/misc/tvguide/ITVGuidePeerObjectImpl.hpp"

#include "nmm/misc/tvguide/TVGuidePluginManager.hpp"

#include <string>
#include <list>

#include <time.h>

#include "nmm/misc/tvguide/TVGLibXML2Database.hpp"
#include "nmm/misc/tvguide/TVGXercescDatabase.hpp"

#include "nmm/misc/tvguide/TVGuidePeerConnector.hpp"


namespace NMM{

  class TVGuide : public ITVGuideImpl,
                  public ITVGuideUpdateFinishedImpl,
                  public ITVGuidePeerSearchImpl
  {
    friend class CreateUsingNew<TVGuide>;

  private:

    list<TVGuidePlugin *> updatePlugins;

    //TVGLibXML2Database* database;
    TVGXercescDatabase* database;
    TVGuidePeerConnector* peerConnector;

  private:

    TVGuide();
    TVGuide(const TVGuide&);
    ~TVGuide();

  protected:    
    Result start();
    Result stop();
    Result init();
    Result deinit();

    
    /** Insert Channel into DB. Indicates, that information for this channel is available
     *@param id ID of the channel
     *@param name Name of the channel
     */
    Result insertChannel(const string &id, const string &name);
    /** Insert ChannelDescription into DB. Indicates, that information for this channel is available
     *@param cd ChanenlDescription
     */
    Result insertChannelDescription(const ChannelDescription &cd);
    /** Set the directory, where information is stored
     *@param path Complete path of the directory
     */
    void setDirectory(const string &path);
     /*@return complete path of the directory
     */
    string getDirectory() throw (const NMM::Exception&);

    /** Get a list with all available categories
     *@return List of categories
     */
    list<string> getCategories() throw (const NMM::Exception&);
    
    /** Get a list with all available channels. 
     *@return list of channels
     */
    list<ChannelDescription> getChannelsList()throw (const NMM::Exception&);
    /** Look for name of channel with given ID
     *@param id ID of the channel
     *@return string name of channel, if the ID is known
     *@return empty-string if ID is unknown
     */
    string  getChannelById(const string &id) ;
    /** Look for ID of channel with given name
     *@param name name of the channel
     *@return string ID of channel, if the name is known
     *@return empty-string if name is unknown
     */
    string getChannelByName(const string &name);

    /** Insert an event to the database
     *@param pd ProgramDescription containing the information of the event
     */
    Result insertEvent(const ProgramDescription &pd);
    /** look for a specific event in the time-inteval [start,end] on the channel with the given ID
     *@param start Start-Time of the interval. Time-representation: YYYYMMDDhhmm
     *@param end End-Time of the interval. Time-representation: YYYYMMDDhhmm
     *@param id ID of the Channel
     *@return ProgramDescription of the Event
     */
    ProgramDescription getEvent(const string &start, const string &end, const string &id);
    /** Get a list of all Events of a given channel on a specific day
     *@param day Representation: YYYYMMDD
     *@param id id of the channel
     *@return List of ProgramDescriptions containing the events found
     */
    list<ProgramDescription> getEventsOnDayOnChannel(const string &day, const string &chan);

    /** Search for Events matching the given filter
     *@param pf ProgramFilter to apply on the search
     *@param peer Interface to the application the it is notified of new events
     *@return List with ProgramDescriptions of events found
     */
    list<ProgramDescription> search(const ProgramFilter &pf, const ITVGuidePeerObject* peer = NULL);

    /** Add an Extended Description to an event. if the event is not available in the database, it is added.
     *@param start Start-time of the event
     *@param end Stop-Time of the event
     *@param id ID of the channel
     *@param desc description to be added
     */
    void addExtDescription(const string& start, const string& end, const string& id, const string& desc);

    /** get the current event of the channel specified. Make sure, your hardware-clock is set up correctly
     *@param channel either name or id
     *@return ProgramDescription with event found
     */
    ProgramDescription getCurrentEventOnChannel(const string& channel);
    /** get the next event of the channel specified. Make sure, your hardware-clock is set up correctly
     *@param channel either name or id
     *@return ProgramDescription with event found
     */
    ProgramDescription getFollowingEventOnChannel(const string& channel);

    /** get the interface specified in argument
     *@param type name of the interface 
     *@return Pointer to the grabber-interface
     */
    ITVGuidePlugin* getGrabberInterface(const string& type);

    /** add the address of a remote TVGuide. This TVGuide will be queried when distributing requests
     *@param host hostname of the remote TVGuide
     *@return port the remote TVGuide listens on
     */ 
    void addPeerHost(const string& host, const int& port);
    /** clear the list of remote TVGuide-Hosts
     */
    void clearPeerHosts();

    /** cleanup the database
     *@param days specify a number of days for wich data should be kept from the past
     */
    void cleanDatabase(const int &days);

    /* ITVGuideUpdateFinished-interface */
    Result partiallyFinished(){return SUCCESS;}
    Result finished(){return SUCCESS;}
    Result newChannel(){return SUCCESS;}
    Result newShow(const string& /*channelID*/){ return SUCCESS;};
    Result informationUpdate(){return SUCCESS;};

    /*ITVGuidePeerSearch-Interface */
    Result peerSearch(const ProgramFilter& filter, 
                      const list<Interface*>& history);

   Result peerChannelSearch(const list<Interface*>& history);

  };

  /* Typdef to access the TVGuide */
  typedef Singleton<TVGuide, CreateUsingNew> STVGuide;

}

#endif
