/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "TVTimerControl.hpp"

#include <list>

#include "nmm/base/ProxyObject.hpp"

namespace NMM{
  
  TVTimerControl::TVTimerControl()
    : itvguide(NULL)
  {
    //cerr << "Creating TVTimerControl-Object" << endl;
    scheduler = new TVTimerScheduler();
    recorder = new TVRecorderManager(this->getInterface<ITVTimer>());
  }
  
  TVTimerControl::~TVTimerControl(){
    if(scheduler){
      delete scheduler;
    }
    if (recorder){
      delete recorder;
    }
  }
   
  Result TVTimerControl::programTimer(const TVTimerEntry &entry) throw (const TimerAlreadyProgrammedException&){

    //MutexGuard mg(timerMutex);

    GLOBAL_DEBUG_STREAM("TVTimerControl::programTimer(TVTimerEntry)" << endl);
    //    TVTimerEntry newEntry = const_cast<TVTimerEntry &>(entry);
    TVTimerEntry newEntry = entry;

    //initiate a self-test of the timer, before scheduling
    if(newEntry.verify() == FAILURE){
      return FAILURE;
    }
    
    try{
      if(newEntry.getTimerInfo().empty()){
        if(itvguide){
          time_t now = time(NULL); 
          struct tm *abs_time = localtime(&now); //make sure, the timezone-vaiable is initialized
          char tmp[6];
          double d = -1*timezone/60/60*100;
          //If a TVGuide is connected, search for information 
          string c = newEntry.getChannelName();
          string channel = itvguide->getChannelByName(c);
          char start[25]; char endt[25];
          time_t s = newEntry.getStartTime();
          time_t e = newEntry.getEndTime();
          
          //respect Daylight Saving Time
          struct tm *tmp_tm = localtime(&s);
          if(tmp_tm->tm_isdst >=0){
            s += (abs_time->tm_isdst - tmp_tm->tm_isdst)*3600;              
            e += (abs_time->tm_isdst - tmp_tm->tm_isdst)*3600;
            d += tmp_tm->tm_isdst * 100;
          }
          
          //append timezone-string
          strftime(start, sizeof(start), "%Y%m%d%H%M", localtime(&s));
          sprintf(tmp, " %0+5.0f", d);
          string tz = string(tmp);
          
          strftime(endt, sizeof(endt), "%Y%m%d%H%M", localtime(&e));
          
          ProgramDescription pd = itvguide->getEvent(start+tz, endt+tz ,channel);
          pd.localizeTimes();
          newEntry.setTimerInfo(pd);
          
        }else{
          GLOBAL_DEBUG_STREAM("TVTimerControl: have no TVGuide" << endl);
        }
      }

      if(scheduler->schedule(newEntry) == SUCCESS){
        recorder->record(newEntry);

        try{
          ITVTimerEvents_var t(this->getCheckedInterface<ITVTimerEvents>());
          t->timerUpdate();
        }catch(const Exception &e){}
      }else{
        GLOBAL_ERROR_STREAM("ERROR scheduling the timer" << endl);
        return FAILURE;
      }
      
    }catch(const TimerAlreadyProgrammedException &e){
      throw TimerAlreadyProgrammedException(e.getTimerID());
    }catch(const TimerRecordedException &e){
      return SUCCESS;
    }catch(const Exception &e){
    }

    return SUCCESS;
  }
  
  Result TVTimerControl::changeTimer(const TVTimerEntry &newEntry) throw (const ActionNotAllowedException&){
    //MutexGuard mg(timerMutex);

    Result res = FAILURE;
    TVTimerEntry e = newEntry;
    TVTimerEntry oldEntry = STVTimerList::getInstance().getTimer(e.getID());

    try{
      res = deleteTimer(e.getID(), e.getUserName());
    }catch(const ActionNotAllowedException &e){
      throw ActionNotAllowedException("change Timer");
    }

    if((res = programTimer(newEntry)) == FAILURE){
      //readd the old timer on FAILURE. Adding suffices since the recorder is still registered
      STVTimerList::getInstance().addTimer(oldEntry);
    }
    
    try{
      this->getCheckedInterface<ITVTimerEvents>()->timerUpdate();
    }catch(const Exception &e){
    }

    return res;
  }

    
  Result TVTimerControl::deleteTimer(const int &i, const string &user) throw (const ActionNotAllowedException&){
    //MutexGuard mg(timerMutex);

    TVTimerEntry e;
    try{
      e = STVTimerList::getInstance().getTimer(i);
    }catch(const Exception &e){
    }

    if(e.getUserName() != user){
      throw ActionNotAllowedException("delete Timer");
    }

    if(e.getStatus() == TVTimerEntry::RECORDING){
      recorder->removeRecord(i);
    }

    Result res = STVTimerList::getInstance().removeTimer(i);

    //reprogram the conflicts
    list<int> conflicts= e.getConflictingTimers();
    for(list<int>::iterator i = conflicts.begin(); i != conflicts.end(); i++){
      try{
        TVTimerEntry e = STVTimerList::getInstance().getTimer(*i);

        //If the conflict-list is empty, there are no conflicts at the moment and we can try to program this timer
        //otherwise there are still conflicts with other timers. 
        if(e.getConflictingTimers().empty()){
          STVTimerList::getInstance().removeTimer(*i);
          programTimer(e);
        }

      }catch(const Exception &e){
        //cerr << e << endl;
      }
    }

    try{
      this->getCheckedInterface<ITVTimerEvents>()->timerUpdate();
    }catch(const Exception &e){
    }

    return res;
  }
    
  list<TVTimerEntry> TVTimerControl::getTimers(const string &day = "", const string &channel = ""){ 
    if(day.empty() && channel.empty()){
      return STVTimerList::getInstance().getTimers();
    }else if(channel.empty()){
      return STVTimerList::getInstance().getTimers(day);
    }else{
      return STVTimerList::getInstance().getTimers(day, channel);
    }
  }

  list<TVTimerEntry> TVTimerControl::getSleepingTimers(){
    return STVTimerList::getInstance().getSleepingTimers();
  }
  list<TVTimerEntry> TVTimerControl::getRecordingTimers(){
    return STVTimerList::getInstance().getRecordingTimers();
  }
  list<TVTimerEntry> TVTimerControl::getRecordedTimers(){
    return STVTimerList::getInstance().getRecordedTimers();
  }
  list<TVTimerEntry> TVTimerControl::getConflictTimers(){
    return STVTimerList::getInstance().getConflictTimers();
  }
  list<TVTimerEntry> TVTimerControl::getConflictsWithTimer(const TVTimerEntry &entry){
    return STVTimerList::getInstance().getConflictsWithTimer(entry);
  }

  Result TVTimerControl::loadTimer(const string &filename){ 
    return STVTimerList::getInstance().loadTimer(filename, this->getInterface<ITVTimer>());
  }

  Result TVTimerControl::saveTimer(const string &filename){
    return STVTimerList::getInstance().saveTimer(filename);
  }

  void TVTimerControl::setTVGuide(const ITVGuide *itvg){
    itvguide = dynamic_cast<ITVGuide *>(itvg->copy());
  }

  void TVTimerControl::setDirectory(const string &dir){
    if(recorder){
      recorder->setDirectory(dir);
    }
  }

  string TVTimerControl::getDirectory(){
    if(recorder){
      return recorder->getDirectory();
    }else{
      return string("");
    }
  }
 
  void TVTimerControl::setApplication(NMMApplication* app){
    if(recorder){
      recorder->setApplication(app);
    }

  }
  
  void TVTimerControl::setChannelConfig(TVChannelList *cc){
    if(scheduler){
      scheduler->setChannelConfig(cc);
    }
  }

  TVChannelList TVTimerControl::getChannelConfig(){
    return *(scheduler->getChannelConfig());
  }

  TVTimerEntry TVTimerControl::getTimer(const unsigned int &id){
    try{
      return STVTimerList::getInstance().getTimer(id);
    }catch(const Exception &e){
      throw NoTimerFoundException();
    }
  }

  Result TVTimerControl::tvcardUnavailable(const unsigned int &id){
    //cerr << "event tvcardUnavailable" << endl;
    TVTimerEntry entry = STVTimerList::getInstance().getTimer(id);
    STVTimerList::getInstance().removeTimer(id);

    //cerr << "reprogramming timer" << endl;
    if(scheduler->reSchedule(entry) == SUCCESS){
      //cerr << "reprogram successful" << endl;
      recorder->record(entry);
      
      try{
        this->getCheckedInterface<ITVTimerEvents>()->timerUpdate();
      }catch(const Exception &e){
      }

    }else{
      //cerr << "reprogramming failed, try shared" << endl;

      //readd the timer to the list. the recorder will alter the status
      STVTimerList::getInstance().addTimer(entry);

      //try to record the program with a shared node
      if(recorder->recordShared(entry) == SUCCESS){
        //cerr << "shared reprogram Successful" << endl;
      }else{
        //cerr << "shared reprogram failed" << endl;
      }
    }
    return SUCCESS;
    
  }
  
  Result TVTimerControl::timerUpdate() { 
    //cerr << "timerupdate" << endl; 
    return SUCCESS;
  }
  
  
}
