/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <fstream>

#include "TVTimerList.hpp"

#include "nmm/utils/xml/xml++.hpp"
#include "nmm/utils/utils.hpp"

#include "nmm/base/serialize/MagicNumberValueStream.hpp"
#include "nmm/base/serialize/StdStream.hpp"

namespace NMM{
  TVTimerList::TVTimerList() 
    : currentID(0),
      timerFilename("")
  {
  }

  TVTimerList::TVTimerList(const TVTimerList&){}

  TVTimerList::~TVTimerList(){
    saveTimer(timerFilename);
  }

  Result TVTimerList::addTimer(TVTimerEntry &entry){
    currentID++;
    entry.setID(currentID);
    TVTimerEntry toAdd = entry; //copy the entry to add it to the list
    timers.push_back(toAdd);
    timers.sort();
    saveTimer(timerFilename);
    return SUCCESS;
  }

  Result TVTimerList::changeTimer(TVTimerEntry &entry){
    for(list<TVTimerEntry>::iterator i = timers.begin(); i != timers.end(); i++){
      if(i->getID() == entry.getID()){
        timers.erase(i);
        TVTimerEntry toAdd = entry; //copy the entry to add it to the list
        timers.push_back(toAdd);
        timers.sort();
        saveTimer(timerFilename);
        return SUCCESS;
      }
    }
    return FAILURE;
  }

  list<TVTimerEntry> TVTimerList::getTimers(){
    return timers;
  }

  list<TVTimerEntry> TVTimerList::getTimers(string day){
    list<TVTimerEntry> returnList;
    for(list<TVTimerEntry>::iterator i = timers.begin(); i != timers.end(); i++){
      char start[25];
      time_t t = i->getStartTime();
      strftime(start, sizeof(start), "%Y%m%d", localtime(&t));
      //if time matches
      try{
        if(day.substr(0,8) == string(start)){
          returnList.push_back(*i);
        }
      }catch(...){}
    }
    return returnList;
  }

  list<TVTimerEntry> TVTimerList::getTimers(string day, string channel){
    list<TVTimerEntry> returnList;
    for(list<TVTimerEntry>::iterator i = timers.begin(); i != timers.end(); i++){
      char start[25];
      time_t t = i->getStartTime();
      strftime(start, sizeof(start), "%Y%m%d", localtime(&t));
      //if time matches
      try{
        if(day.substr(0,8) == string(start)){
          //if channel matches the channelname
          if(channel == i->getChannelName()){
            returnList.push_back(*i);
          }else if(channel == i->getTimerInfo().getChannelID()){
            //... or the channelid of timerinfo
            returnList.push_back(*i);
          }
        }
      }catch(...){}
    }
    return returnList;
  }

  list<TVTimerEntry> TVTimerList::getTimers(TVCardDescription d){
    list<TVTimerEntry> returnList;
    for(list<TVTimerEntry>::iterator i = timers.begin(); i != timers.end(); i++){
      try{
        if(i->getTVCard()== d){
          returnList.push_back(*i);
        }
      }catch(const Exception &e){}
    }
    return returnList;
  }

  TVTimerEntry TVTimerList::getTimer(unsigned int id)throw (NoTimerFoundException){
    for(list<TVTimerEntry>::iterator i = timers.begin(); i != timers.end(); i++){
      if(i->getID() == id){
        return *i;
      }
    }
    //if we don't find a timer with this ID, throw Exception
    throw NoTimerFoundException();
  }

  list<TVTimerEntry> TVTimerList::getSleepingTimers(){
    list<TVTimerEntry> returnList;
    for(list<TVTimerEntry>::iterator i = timers.begin(); i != timers.end(); i++){
      if(i->getStatus()== TVTimerEntry::SLEEPING){
        returnList.push_back(*i);
      }
    }
    return returnList;
  }

  list<TVTimerEntry> TVTimerList::getRecordingTimers(){
    list<TVTimerEntry> returnList;
    for(list<TVTimerEntry>::iterator i = timers.begin(); i != timers.end(); i++){
      if(i->getStatus() == TVTimerEntry::RECORDING){
        returnList.push_back(*i);
      }
    }
    return returnList;
  }

  list<TVTimerEntry> TVTimerList::getRecordedTimers(){
    list<TVTimerEntry> returnList;
    for(list<TVTimerEntry>::iterator i = timers.begin(); i != timers.end(); i++){
      if(i->getStatus() == TVTimerEntry::RECORDED || i->getStatus() == TVTimerEntry::RECORDED_WITH_ERROR){
        returnList.push_back(*i);
      }
    }
    return returnList;
  }

  list<TVTimerEntry> TVTimerList::getConflictTimers(){
    list<TVTimerEntry> returnList;
    for(list<TVTimerEntry>::iterator i = timers.begin(); i != timers.end(); i++){
      if(i->getStatus() == TVTimerEntry::CONFLICT){
        returnList.push_back(*i);
      }
    }
    return returnList;
  }

  list<TVTimerEntry> TVTimerList::getConflictsWithTimer(const TVTimerEntry &entry){
    list<TVTimerEntry> returnList;
    list<int> timerIDs = const_cast<TVTimerEntry&>(entry).getConflictingTimers();

    for(list<TVTimerEntry>::iterator i = timers.begin(); i != timers.end(); i++){
      for(list<int>::iterator j = timerIDs.begin(); j != timerIDs.end(); j++){
        if(i->getID() == (unsigned int)(*j)){
          returnList.push_back(*i);
        }
      }
    }
    return returnList;
  }


  Result TVTimerList::removeTimer(unsigned int id){
    Result res = FAILURE;
    for(list<TVTimerEntry>::iterator i = timers.begin(); i != timers.end(); i++){
      if(i->getID() == id){
        timers.erase(i);
        res = SUCCESS;
        break;
      }
    }
    for(list<TVTimerEntry>::iterator i = timers.begin(); i != timers.end(); i++){
      //update the conflicts of every timer
      if(i->hasConflictWith(id)){
        i->removeConflict(id);
      }
    }

    saveTimer(timerFilename);
    return res;
  }

  Result TVTimerList::loadTimer(string filename, ITVTimer *itimer){
    GLOBAL_DEBUG_STREAM("TVTimerList::loadTimer(" << filename << ")" << endl);

    timerFilename = filename;


    ifstream stdstream;
    stdstream.open(timerFilename.c_str());

    if(stdstream){
      try{
        
        StdIStream s(&stdstream);
        MagicNumberIValueStream vs;
        vs.setIPartnerStream(&s);
        vs.readData();
        
        while(!vs.isEmpty()){
          TVTimerEntry e;
          vs >> e;
          if(e.getEndTime() != 0)
            try{
              itimer->programTimer(e);
            }catch(const Exception &e){}
        }
        stdstream.close();

        return SUCCESS;

      }catch(const Exception &e){
        GLOBAL_ERROR_STREAM("TVTimerList: Error reading timerlist: e" << endl);
      }
      stdstream.close();
    }else{
      GLOBAL_ERROR_STREAM("TVTimerList: Error reading timerlist: Failed to open " << filename << endl);
    }
    
    return SUCCESS;
  }
  
  
  Result TVTimerList::saveTimer(string filename){
    GLOBAL_DEBUG_STREAM("TVTimerList::saveTimer(" << filename << ")" << endl);


    string saveFilename = filename;
    
    if(saveFilename.empty()){
      if(timerFilename.empty()){
        return FAILURE;
      }else{
        saveFilename = timerFilename;
      }
    }

    ofstream stdstream(saveFilename.c_str());
    if(!stdstream){
      GLOBAL_ERROR_STREAM("TVChannelList: Error writing channelList: Failed to open " << filename << endl);
      return FAILURE;
    }
    
    try{
      StdOStream s(&stdstream);
      MagicNumberOValueStream magicNumberStream;
      magicNumberStream.setOPartnerStream(&s);
      for(list<TVTimerEntry>::iterator i = timers.begin(); i != timers.end(); i++){
        magicNumberStream << *i;
      }

      magicNumberStream.writeData();
    }catch(const Exception &e){
      GLOBAL_ERROR_STREAM("TVChannelList: Error writing channelList: e" << endl);
      return FAILURE;
    }

    stdstream.close(); 
    return SUCCESS;
  }

  string TVTimerList::TimeToString(time_t ltime) {
    char temp[64];
    strftime(temp, sizeof(temp), "%Y%m%d%H%M",  localtime(&ltime));
    return( string(temp) );
  }


}
