/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        David Maass <dama@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "VorbisEncodeNode.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/registry/Plugin.hpp"

NMM_REGISTER_PLUGIN(NMM::VorbisEncodeNode, "VorbisEncodeNode")

namespace NMM {
   VorbisEncodeNode::VorbisEncodeNode(const char* name,
			  const StreamQueue::Mode queue_mode,
			  const int queue_size)
     : GenericProcessorNode(name,queue_mode,queue_size), 
       quality(0),
       setupdone(false),
       HeaderWritten(false),
       trackname(""),
       album("unknown"),
       year("unknown"),
       artist("unknown"),
       genre("unknown"),
       title("unknow")
   {}

  VorbisEncodeNode::~VorbisEncodeNode()
  {}

  Result VorbisEncodeNode::startTrack(const string& filename) {
    DEBUG_STREAM("VorbisEncodeNode: startTrack " << filename << endl);
    trackname = filename;
    if (!setupdone) setup();
    event = DATA;
    return SUCCESS;
  }

  Result VorbisEncodeNode::endTrack() {
    // set internal state to END and set working-flag to TRUE
    event = END;
    setWorkingFlag(true);
    return DELETE;
  }

  Result VorbisEncodeNode::finished() {
    return SUCCESS;
  }
   

  Result VorbisEncodeNode::TagInfo(const string& _title, const string& _album, const string& _artist, const string& _year, const string& _genre) {
    title = _title;
    album = _album;
    year = _year;
    genre = _genre;
    artist = _artist;
    return SUCCESS;
  }

  Result VorbisEncodeNode::doInit()
  {
    Format* myInputFormat
      = getOwnInputProperty()->addNewFormat("audio/raw");
    myInputFormat->addWildcard(Format::sample_rate_param);
    myInputFormat->addIntParamValue(Format::channels_param,1);
    myInputFormat->addIntParamValue(Format::channels_param,2);    
    myInputFormat->addIntParamValue(Format::bitpersample_param,8);    
    myInputFormat->addIntParamValue(Format::bitpersample_param,16);
    myInputFormat->addStringParamValue(Format::endian_param,"little");
    myInputFormat->addStringParamValue(Format::encoding_param,"linear");
    myInputFormat->addStringParamValue(Format::sign_param, "signed");
    getOwnInputProperty()->setDefaultFormat(myInputFormat);

    Format* myOutputFormat
      = getOwnOutputProperty()->addNewFormat(Format::audio_x_ogg_type);
    myOutputFormat -> addWildcard(Format::sample_rate_param);
    myOutputFormat -> addIntParamValue(Format::channels_param,1);
    myOutputFormat -> addIntParamValue(Format::channels_param,2);
    myOutputFormat -> addIntParamValue(Format::audio_codec_param, 0xFFFE);
    myOutputFormat -> addWildcard(Format::quality_param);
    getOwnOutputProperty()->setDefaultFormat(myOutputFormat);
    
    myInputFormat->setIOPartner(myOutputFormat);
    myOutputFormat->setIOPartner(myInputFormat);
    return SUCCESS;
  }

  /** 
   * doInitOutput() initialises the OutputFormat
   */
  Result VorbisEncodeNode::doInitOutput()
  {
    //Node must be connected now
    const Format* in_format = getInputFormat();
    if (in_format == NULL) {GLOBAL_ERROR_STREAM("Informat = NULL"<<endl); return FAILURE; }
    sample_rate = in_format->getIntValue(Format::sample_rate_param);
    channels = in_format->getIntValue(Format::channels_param);
    bitpersample = in_format->getIntValue(Format::bitpersample_param);

    
    //the only input format
    Format* myInputFormat
      = getOwnWorkingInputProperty()->addNewFormat("audio/raw");
    myInputFormat->addIntParamValue(Format::sample_rate_param,sample_rate);
    myInputFormat->addIntParamValue(Format::channels_param,channels);
    myInputFormat->addIntParamValue(Format::bitpersample_param,bitpersample);
    myInputFormat->addStringParamValue(Format::endian_param,"little");
    myInputFormat->addStringParamValue(Format::encoding_param,"linear");
    myInputFormat->addStringParamValue(Format::sign_param, "signed");
    getOwnWorkingInputProperty()->setDefaultFormat(myInputFormat);

    //setting spezified outputformat
    Format* myOutputFormat
      =getOwnWorkingOutputProperty()->addNewFormat(Format::audio_x_ogg_type);
    myOutputFormat -> addIntParamValue(Format::sample_rate_param,sample_rate);
    myOutputFormat -> addIntParamValue(Format::channels_param,channels);
    myOutputFormat -> addIntParamValue(Format::audio_codec_param, 0xFFFE);
    myOutputFormat -> addWildcard(Format::quality_param);
    getOwnWorkingOutputProperty()->setDefaultFormat(myOutputFormat);

    //connecting IOPartners
    myInputFormat->setIOPartner(myOutputFormat);
    myOutputFormat->setIOPartner(myInputFormat);

    return SUCCESS;
  }
  
  /**
   * doActivate() checks for right inputformat,
   * activates the VorbisEncoder with the given parameters for VBR Encoding
   * and inits some more vorbismethods
   */

  Result VorbisEncodeNode::setup() {
    DEBUG_STREAM("Quality is: "<<quality<<endl);
    vorbis_info_init(&VorbisInfo);
    int retval=vorbis_encode_init_vbr(&VorbisInfo,channels,sample_rate,quality);
    if (retval <0) {
      ERROR_STREAM("Vorbis Encoder Activation failed"<<endl);
      return FAILURE;
    }
       
    DEBUG_STREAM("adding a comment"<<endl);
    vorbis_comment_init(&VorbisComment);
    vorbis_comment_add(&VorbisComment,const_cast<char*>("encoded by MMBox Vorbis Encoder"));

    DEBUG_STREAM("setting up some more things"<<endl);
    vorbis_analysis_init(&VorbisDspState,&VorbisInfo);
    vorbis_block_init(&VorbisDspState,&VorbisBlock);
    srand(time(NULL));
    ogg_stream_init(&OggStreamState,rand());
    setupdone = true;
    return SUCCESS;
  }

  Result VorbisEncodeNode::doActivate()
  {
    try{
      const Format* in_format = getOutputFormat();
      quality = in_format->getFloatValue(Format::quality_param);
    }catch(const Exception &e){
      ERROR_STREAM("Error in getting quality. Setting it to 1" << endl);
      quality = 1;
    }

    return setup();
  }

  Result VorbisEncodeNode::doDeactivate()
  {
    return SUCCESS;
  }


  /**
   * processBuffer() encodes an incoming Audio-Buffer to an Vorbisaudio
   */

  Message* VorbisEncodeNode::processBuffer(Buffer *in_buffer)
  {
    if (getProducingFlag()) {
      //we had a bufferproblem and have some data in rescue_buffer to send - so we do
      setProducingFlag(false);
      return rescue_buffer;
    }

    int incoming_size=0;
    int working_size=incoming_size;
    int out_buffer_pos;
    char* data=NULL;
    float **buffer2=NULL;
    Buffer *return_buffer=NULL;

    switch(event) {
    case TEST:
      ERROR_STREAM("Don't call me" << endl);
      return NULL;
      break;
      
    case START:
      ERROR_STREAM("Don't call me" << endl);
      event = DATA;
      return NULL;
      break;

    case FINISHED:
      return NULL;
      break;
    
    case END:
      DEBUG_STREAM("VorbisEncoderNode encoded the audio and is cleaning up"<<
		   endl);
      setWorkingFlag(false);
      ogg_stream_clear(&OggStreamState);
      vorbis_block_clear(&VorbisBlock);
      vorbis_dsp_clear(&VorbisDspState);
      vorbis_comment_clear(&VorbisComment);
      DEBUG_STREAM("AudioStream got nominal Bitrate of: "<<VorbisInfo.bitrate_nominal<<endl);
      vorbis_info_clear(&VorbisInfo);
      HeaderWritten = false;
      setupdone=false;
      // david, please check!
      // return endTrack event (was delete before)
      return new CEvent(Event(ITrack::endTrack_event));
      break;

   case DATA:
      incoming_size = in_buffer->getUsed();
      //a normal oggpage has a size around 4k, so requesting a smaller buffer is useless and leads to problems
      //so we request a minimal buffer of 4500 to avoid these problems
      if (incoming_size<4500) {
        return_buffer = getNewBuffer(4500);
        working_size=4500;
      } else {
        return_buffer = getNewBuffer(incoming_size);
        working_size = incoming_size;
      }
      
      out_buffer_pos=0;
      data = in_buffer->getData();
      if (return_buffer) {
	if (!HeaderWritten) {
	  HeaderWritten = true;
	  DEBUG_STREAM("creating OggFile Header"<<endl);
          vorbis_comment_add_tag (&VorbisComment,  const_cast<char*>("TITLE"), const_cast<char*>(title.c_str()));
          vorbis_comment_add_tag (&VorbisComment,  const_cast<char*>("ALBUM"), const_cast<char*>(album.c_str()));
          vorbis_comment_add_tag (&VorbisComment,  const_cast<char*>("ARTIST"), const_cast<char*>(artist.c_str()));
          vorbis_comment_add_tag (&VorbisComment,  const_cast<char*>("DATE"), const_cast<char*>(year.c_str()));
          vorbis_comment_add_tag (&VorbisComment,  const_cast<char*>("GENRE"), const_cast<char*>(genre.c_str()));
	  ogg_packet header;
	  ogg_packet header_comm;
	  ogg_packet header_code;
	  vorbis_analysis_headerout(&VorbisDspState,
				    &VorbisComment,
				    &header,
				    &header_comm,
				    &header_code);

	  ogg_stream_packetin(&OggStreamState,&header);
	  ogg_stream_packetin(&OggStreamState,&header_comm);
	  ogg_stream_packetin(&OggStreamState,&header_code);
	  for (;;) {
	    int result = ogg_stream_flush(&OggStreamState,&OggPage);
	    if (result==0) 
	      break;
            memcpy(return_buffer->getData()+out_buffer_pos,OggPage.header,OggPage.header_len);
	    out_buffer_pos+=OggPage.header_len;
	    memcpy(return_buffer->getData()+out_buffer_pos,OggPage.body,OggPage.body_len);
	    out_buffer_pos+=OggPage.body_len;
	  }
	}
	
	/* data to encode */
	DEBUG_STREAM("Got Outbuffer and starting to encode"<<endl);

	/* expose the buffer to submit data */
	//buffer2=vorbis_analysis_buffer(&VorbisDspState,incoming_size/4);
	long realsamples = incoming_size/ ((bitpersample/8)*channels);
	buffer2=vorbis_analysis_buffer(&VorbisDspState, realsamples);

	if (bitpersample == 8) {
	        unsigned char *dat = (unsigned char *)data;
		for(int i=0;i<realsamples;i++) {
			for(int j=0;j<channels;j++) {
				buffer2[j][i]=((int)(dat[i*channels + j])-128)/128.0f;
			}
		}
	} else {
		for(int i=0;i<realsamples;i++) {
			for(int j=0;j<channels;j++) {
				buffer2[j][i]=((data[i*2*channels +2*j +1]<<8) |
						(data[i*2*channels+2*j] & 0xff))/32768.0f;
			}
		}
	}

	/* tell the library how much we actually submitted */
	vorbis_analysis_wrote(&VorbisDspState,realsamples);

	while(vorbis_analysis_blockout(&VorbisDspState,&VorbisBlock)==1){

	  /* analysis, assume we want to use bitrate management */
	     vorbis_analysis(&VorbisBlock,NULL);
	     vorbis_bitrate_addblock(&VorbisBlock);
	  while(vorbis_bitrate_flushpacket(&VorbisDspState,&OggPacket)){

	    /* weld the packet into the bitstream */
	    ogg_stream_packetin(&OggStreamState,&OggPacket);

	    /* write out pages (if any) */
	    for(;;){
	      int result=ogg_stream_pageout(&OggStreamState,&OggPage);
       	      if(result==0 || ogg_page_eos(&OggPage))break;
              if (OggPage.header_len+OggPage.body_len+out_buffer_pos>working_size) {
                //its possible to run out of buffer-space. to avoid segfaulting we have a rescue buffer which is used to get one more
                //oggpage and is sendet on next processbuffer (triggerd without in_buffer because producingflag==true
                GLOBAL_WARNING_STREAM("We get a buffer problem, if your oggfile is damaged inform us"<<endl);
                rescue_buffer = getNewBuffer(5000);
                memcpy(rescue_buffer->getData()+out_buffer_pos,OggPage.header,OggPage.header_len);
                memcpy(rescue_buffer->getData()+OggPage.header_len+out_buffer_pos,OggPage.body,OggPage.body_len);
                setProducingFlag(true);
                rescue_buffer->setUsed(OggPage.header_len+OggPage.body_len+out_buffer_pos);
                break;
              }
              memcpy(return_buffer->getData()+out_buffer_pos,OggPage.header,OggPage.header_len);
	      out_buffer_pos+=OggPage.header_len;
	      memcpy(return_buffer->getData()+out_buffer_pos,OggPage.body,OggPage.body_len);
	      out_buffer_pos+=OggPage.body_len;
	    }
	  } //while vorbis_bitrate_flushpacket;
	}	    

	return_buffer->setUsed(out_buffer_pos);
        DEBUG_STREAM("Got "<<incoming_size<<" and build "<<out_buffer_pos<<endl);
        in_buffer->release();
        if (out_buffer_pos==0) {
          //if we build nothing - return nothing
          return_buffer->release();
          return NULL;
        }
	return return_buffer;
      }
      else {
	DEBUG_STREAM("Buffer requested but not returned -> released." << endl);
	return NULL;
      }

      break;
    };//switch
    return NULL;
  };//processbuffer
};//namespace
