/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    Network-Multimedia Workgroup,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:	      Roger Dostert <djurig@graphics.cs.uni-sb.de 
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef NMM_FFMPEG_FFAVFORMAT_HPP
#define NMM_FFMPEG_FFAVFORMAT_HPP

#include "nmm/base/memorymgr/Buffer.hpp"

#include <list>

class Buffer;
class AVFormatContext;
class AVOutputFormat;
class AVStream;

namespace NMM {


	/** 
		* @typedef FFBufferList
		* @brief a type for lists holding NMM-Buffers
		*/
	typedef std::list<Buffer*> FFBufferList;

	/** 
		* @enum FFNeedBuffer
		* @brief enum-type that says what kind of buffer is required
		*/
	enum FFNeedBuffer { 
		NB_NONE    = 0, ///! no buffers needed
		NB_NEED_V  = 1, ///! video needed
		NB_NEED_A  = 2, ///! audio needed
		NB_NEED_AV = 3, ///! audio and video needed
		NB_ERROR   = 4  ///! wrong mux state, should never happen
	};
		

	/** this class encapsulates stuff related to the libavformat library */
	class FFAVFormat {
	public:
		/**
			* @brief constructor
			*/
		FFAVFormat(void);
		/**
			* @brief destructor
			*/
		~FFAVFormat(void);

		/**
			* @brief add an audio stream for muxing out
			* @param codecId ID of the FFMpeg codec to use
			* @param bitrate bitrate of the audio stream 
			* @param samplerate samplerate of the audio stream
			* @param channels number of audio channels
			* @return no return value
			*/
		void addAudioStream(int codecId,int bitrate,int samplerate,int channels);
		/**
			* @brief add an video stream for muxing out
			* @param codecId ID of the FFMpeg codec to use
			* @param bitrate bitrate of the video stream
			* @param width witdth of a video frame
			* @param height of a video frame
			* @param framerate framerate if the video
			* @return no return value
			*/
		void addVideoStream(int codecId,int bitrate,int width,int height,float framerate);

		/**
			* @brief write the header of the output stream
			* @return no return value
			*/
		void writeHeader(void);

		/**
			* @brief write the trailer of the output stream
			* @return no return value
		*/
		void writeTrailer(void);

		/**
			* @brief mux out as many buffers as possible
			* @param buffer the new data
			* @param video buffer is a video frame or not
			* @return what kind of buffers are needed
			*/
		int muxFrame(Buffer* buffer,bool video);

		/**
			* @brief read out muxed data
			* @param data buffer for the data
			* @param size size of data
			*/
		int readBack(char* data,const int size);

		/**
			* @brief deliver the amount of data in the output buffer
			* @return the value of io_buffersize
			*/
		int getIOBufferSize(void) { return io_buffersize; }

		void setOutputFormat(const char* short_name);

	private:
		AVFormatContext*          fmtctx;      ///! FFMpeg format context
		int                       streamCount; ///! total number of streams
		AVStream*                 a_stream;    ///! audio stream
		AVStream*                 v_stream;    ///! video stream

		unsigned char*    buffer;              ///! buffer used by io_writePacket
		unsigned char*    io_buffer;           ///! output buffer
		int               io_buffersize;       ///! size of the output buffer
		int               io_bufferused;       ///! used part in bytes of the output buffer

		FFBufferList      a_buffers;           ///! buffered audio data
		FFBufferList      v_buffers;           ///! buffered video data

		// some counters (for information)
		int               n_vbuffer;           ///! number of video buffers muxed out
		int               n_abuffer;           ///! number of audio buffers muxed out
		int               g_vbuffer;           ///! number of video buffers got delivered
		int               g_abuffer;           ///! number of audio buffers got delivered


		/**
			* @brief search for an specified output format in the list of supported formats of the libavformat library
			* @param short_name short name of the format
			* @param len length of the short_name string
			* @return if found,a copy of the format and NULL otherwise
			*/
		AVOutputFormat* findOutputFormat(const char* short_name,int len);

		/**
			* @brief write an audio buffer into the output stream
			* @return NB_NONE or NB_AUDIO
			*/
		int writeAudioFrame();
		/**
			* @brief write a video buffer into the output stream
			* @return NB_NONE or NB_VIDEO
			*/
		int writeVideoFrame();

	private:

		/**
			* @brief writes a packet into the output buffer
			* the static is required because the method is uses as a function pointer from libavformat
			*/
		static void io_writePacket(void* opaque,unsigned char* buf,int size);
	}; // class FFAVFormat

} // namespace NMM

#endif // NMM_FFMPEG_FFAVFORMAT_HPP
