/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    Network-Multimedia Workgroup,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Eric Peters <epeters@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "NetSourceNode.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/registry/Plugin.hpp"
#include "nmm/base/ConnectionManager.hpp"
#include "nmm/base/TSFactory.hpp"

NMM_REGISTER_PLUGIN(NMM::NetSourceNode, "NetSourceNode")

namespace NMM {

  
  NetSourceNode::NetSourceNode(const char* name, 
			       IQueued::Mode queue_mode /*= IQueued::MODE_SUSPEND*/,
			       unsigned int queue_size /*= 100*/) 
    : GenericSourceNode(name), 
      __transport_strategy(0),
      __current_message(0),
      __stream_queue((StreamQueue::Mode) queue_mode, queue_size, "BufferedSource"),
      __payload_type(0)
  {
    /* this object reads from the queue, so it is the parent of the queue */
    __stream_queue.setParentSignalReceiver(this);
  }

  NetSourceNode::~NetSourceNode() {
    if (__transport_strategy)
      __transport_strategy->scheduleDestruction();
  }

  void NetSourceNode::setRTPStrategy(const string& _address, const int& _rtp_port, const int& _rtcp_port) {
    TSAddress* address = new RTPAddress(UnicastIPAddress(_address), _rtp_port, _rtcp_port);
    setTransportStrategy(address);    
  }

  void NetSourceNode::setTransportStrategy(TransportStrategy* transport_strategy) {

    /* Begin> If an old strategy is available, release it. */
    if (__transport_strategy)
      __transport_strategy->scheduleDestruction();
    /* End< */

    /* stores the new strategy */
    __transport_strategy = transport_strategy;
    /* End< */

    /* the transport strategy writes to the queue, so it is the partner of the queue */
    __stream_queue.setPartnerSignalReceiver(__transport_strategy);
  }

  void NetSourceNode::setTransportStrategy(const TSAddress* address) throw(const BindingException&) {
  
    /* If the address is invalid throw an Exception */
    if (!address)
      throw BindingException("Invalid Address (0-ptr)");

    /* Create a TransportStrategy which can connect to the given address */
    TransportStrategy* transport_strategy = STSFactory::getInstance().createStrategy(address->getStrategyType());
    
    /* If it could not be created throw a BindingException */
    if (!transport_strategy)
      throw  BindingException("Can not create strategy with type "+ address->getStrategyType());

    /* The strategy connects to the given address */
    
    transport_strategy->setAddress(*address);
    
    setTransportStrategy(transport_strategy);
  }

  void NetSourceNode::setPayloadType(const int& type) {
    __payload_type = type;
  }

  int NetSourceNode::getPayloadType() const {
    return __payload_type;
  }

  Message* NetSourceNode::processBuffer(Buffer* /*buffer*/) {

    if (!__current_message) {
      setProducingFlag(false);
      return NULL;
    }

    __current_message_mutex.lock();
    Message* temp_message = __current_message;
    __current_message = 0;
    setProducingFlag(false);
    __transport_strategy->receiveSignal(FREE_QUEUE_SLOT);
    __current_message_mutex.unlock();
    
    
    return temp_message;
  }
  
  Result NetSourceNode::receiveMessage(Message* message, const IMessageSender* /*sender*/){
    
    if (!message) {
      return FAILURE;
    }

    if (__current_message) {
      ERROR_STREAM( "ERROR IN NetSourceNode __current_message != 0" << endl);
      return FAILURE;
    }

    /* Begin> Store the message pointer */
    __current_message_mutex.lock();
    __current_message = message;
    setProducingFlag(true);
    if (__current_message)
      __current_message_mutex.unlock();
    /* End< */

    /* Return blocked*/
    return BLOCKED;
  }

  Result NetSourceNode::doInit()  {
    
    return SUCCESS;
  }
  
  Result NetSourceNode::doInitOutput() {
    
    Format* strategy_format = 0;
    DEBUG_STREAM("__payload_type = " << __payload_type << endl);
    switch (__payload_type) {
    case 0:
      strategy_format = getOwnWorkingOutputProperty()->addNewFormat("audio/raw");
      strategy_format->addIntParamValue(Format::sample_rate_param,8000);
      strategy_format->addIntParamValue(Format::channels_param, 1);
      strategy_format->addIntParamValue(Format::bitpersample_param, 8);
      strategy_format->addStringParamValue(Format::endian_param, "little");
      strategy_format->addStringParamValue(Format::encoding_param, "mu_law");
      strategy_format->addStringParamValue(Format::sign_param, "signed");
      getOwnWorkingOutputProperty()->setDefaultFormat(strategy_format);
      break;
    case 8:
      strategy_format = getOwnWorkingOutputProperty()->addNewFormat("audio/raw");
      strategy_format->addIntParamValue(Format::sample_rate_param,8000);
      strategy_format->addIntParamValue(Format::channels_param, 1);
      strategy_format->addIntParamValue(Format::bitpersample_param, 8);
      strategy_format->addStringParamValue(Format::endian_param, "little");
      strategy_format->addStringParamValue(Format::encoding_param, "a_law");
      strategy_format->addStringParamValue(Format::sign_param, "signed");
      getOwnWorkingOutputProperty()->setDefaultFormat(strategy_format);
      break;
    case 10: 
      strategy_format = getOwnWorkingOutputProperty()->addNewFormat(Format::audio_raw_type);
      strategy_format->addIntParamValue(Format::sample_rate_param, 44100);
      strategy_format->addIntParamValue(Format::channels_param, 2);
      strategy_format->addIntParamValue(Format::bitpersample_param,  16);
      strategy_format->addStringParamValue(Format::endian_param, "little");
      strategy_format->addStringParamValue(Format::encoding_param, "linear");
      strategy_format->addStringParamValue(Format::sign_param, "signed");
      getOwnWorkingOutputProperty()->setDefaultFormat(strategy_format);
      break;
    case 11:
      strategy_format = getOwnWorkingOutputProperty()->addNewFormat(Format::audio_raw_type);
      strategy_format->addIntParamValue(Format::sample_rate_param, 44100);
      strategy_format->addIntParamValue(Format::channels_param, 1);
      strategy_format->addIntParamValue(Format::bitpersample_param,  16);
      strategy_format->addStringParamValue(Format::endian_param, "little");
      strategy_format->addStringParamValue(Format::encoding_param, "linear");
      strategy_format->addStringParamValue(Format::sign_param, "signed");
      getOwnWorkingOutputProperty()->setDefaultFormat(strategy_format);
      break;
    case 14:
      strategy_format = getOwnWorkingOutputProperty()->addNewFormat(Format::audio_mpeg_type);
      getOwnWorkingOutputProperty()->setDefaultFormat(strategy_format);
      break;
    case 26:
      strategy_format = getOwnWorkingOutputProperty()->addNewFormat(Format::video_jpeg_type);
      strategy_format->addWildcard(Format::x_resolution_param);
      strategy_format->addWildcard(Format::y_resolution_param);
      strategy_format->addWildcard(Format::ratio_x_param);
      strategy_format->addWildcard(Format::ratio_y_param);
      strategy_format->addWildcard(Format::framerate_param);
      getOwnWorkingOutputProperty()->setDefaultFormat(strategy_format);
      break;  
    case 32:
      strategy_format = getOwnWorkingOutputProperty()->addNewFormat(Format::video_mpeg_type);
      strategy_format->addIntParamValue(Format::version_param, 1);
      strategy_format->addIntParamValue(Format::version_param, 2);
      getOwnWorkingOutputProperty()->setDefaultFormat(strategy_format);
      break;
    case 34:
      strategy_format = getOwnWorkingOutputProperty()->addNewFormat(Format::video_h263plus_type);
      strategy_format->addWildcard(Format::x_resolution_param);
      strategy_format->addWildcard(Format::y_resolution_param);
      strategy_format->addWildcard(Format::ratio_x_param);
      strategy_format->addWildcard(Format::ratio_y_param);
      strategy_format->addWildcard(Format::framerate_param);
      strategy_format->addWildcard(Format::bitrate_param);
      getOwnWorkingOutputProperty()->setDefaultFormat(strategy_format);
      break;
    default:
      Format* out_format = getOwnWorkingOutputProperty()->addNewFormat("*/*");
      getOwnWorkingOutputProperty()->setDefaultFormat(out_format);
      ERROR_STREAM("Unknown payload type specified. Only types audio/raw (10/11), "
                   << "pcmu (0), audio/mpeg (14), video/jpeg (26), video/mpeg (32)"
                   << "and video/h263 (34)  are supported!"  
                   << endl << "Taking */* as default format." << endl);
      
      return FAILURE;
    }

    __transport_strategy->setMessageReceiver(this);
    __transport_strategy->setBufferManager(this);    
    __transport_strategy->configureForInputJack(*strategy_format);
    
    if(__payload_type==0 || __payload_type==8 || __payload_type==10|| __payload_type==11 ){
      string type = strategy_format->getFormatType();
      int channels = strategy_format->getIntValue(Format::channels_param);
      int bitpersample = strategy_format->getIntValue(Format::bitpersample_param);
      int sample_rate = strategy_format->getIntValue(Format::sample_rate_param);
      string encoding = strategy_format->getStringValue(Format::encoding_param);
      
      DEBUG_STREAM(endl << endl << "NetSourceNode OwnWorkingOutputProperty" << endl 
                   << "type = " << type << endl
                   <<  "encoding = " << encoding << endl 
                   << "channels = " << channels<< endl
                   << "bitpersample = " << bitpersample << endl
                   << "sample_rate = " << sample_rate << endl << endl);
    }
    
    return SUCCESS;

    /*
    const Format* strategy_format = getOwnWorkingOutputProperty()->getDefaultFormat();

    if (strategy_format) {

      __transport_strategy->setMessageReceiver(this);
      __transport_strategy->setBufferManager(this);
      __transport_strategy->configureForInputJack(*strategy_format);
    
      return SUCCESS;
    }
    return FAILURE;*/
  }
  
  Result NetSourceNode::doActivate() {
    setProducingFlag(true);
    return SUCCESS;
  }

} //end namespace
