/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


/**
 * This NMM node provides the possibility to control a 
 * video camera over the VISCA-protocol from Sony.
 * Should work up to 7 cameras
 *
 * This NMM node was developed in the context of 
 * the multimedia course in WS2001/02 at 
 * University of Saarbruecken/Germany
 */

#ifndef _VISCACAMERAREADNODE_HPP_
#define _VISCACAMERAREADNODE_HPP_ _VISCACAMERAREADNODE_HPP_

#include <string>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <unistd.h>
#include <fstream>
#include <iomanip>
#include <termios.h>
#include <fcntl.h>

#include "nmm/plugins/video/tv/ivtv/IVTVReadNode.hpp"
#include "nmm/interfaces/video/viscacamera/IVISCACameraImpl.hpp"

#define MAX_VISCA_DEVICES         8 /* do NOT change this!! */
#define MAX_SLOTS_PER_CAMERA      5
#define MAX_VISCA_COMMAND_LENGTH 15

namespace NMM {

  class VISCACameraReadNode : public IVTVReadNode,
			      public IVISCACameraImpl {
  public:

    typedef struct struct_camerapos {
      /* position in x direction */ 
      signed short int x;       /* [-1440;1440] */
      signed short int x_deg;   /* [-100;100] */
      
      /* position in y direction */ 
      signed short int y;       /* [-360;360] */
      signed short int y_deg;   /* [-25;25] */
      
      /* position of the zoom , 
         optical zoom : [    0;16384]
         digital zoom : [16384;28672] */
      signed short int z;  
    };

    typedef struct struct_visca_command {
      unsigned char command[MAX_VISCA_COMMAND_LENGTH];
      unsigned char length;
    };
    
    struct_camerapos camerapos;  

    /**
     * Constructor
     * 
     * input : string for the serial port device
     *         eg. "/dev/ttyS1"
     * ouput : none
     */
    VISCACameraReadNode(const char *name = "VISCACameraReadNode",
		   StreamQueue::Mode = StreamQueue::MODE_SUSPEND,
		   int = StreamQueue::DEFAULT_SIZE);

    /**
     * Destructor
     */
    ~VISCACameraReadNode();


    /**
     * init function
     *
     * Call this function to initialize all variable and 
     * constants needed for this library to do his job.
     *
     * input  : none
     * output : SUCCESS when success, else FAILURE
     */
    Result doInit();
    
    Result doDeinit();
    
    /**
     * init_serialdevice function
     * restore_serialdevice function
     *
     * This function sets the serial port paramater to 
     * 8 data bits, 1 start and 1 stop bit, no parity and
     * most significant byte first
     *
     * Please call this functions not directly, they are 
     * called automaticly
     *
     * input  : none
     * output : SUCCESS when success, else FAILURE
     */
    Result init_serialdevice();
    Result restore_serialdevice();
    
    

    /**
     * events functions 
     */

    Result setDevice(string&, string&);
    Result getDevice(string&, string&);
    Result setChannel(int&);

    /*******************************************************************
     * EVI-D100P functions 
     *******************************************************************/
    
    /**
     * IF_Clear 
     * 
     * Clears the command buffers in the camera and 
     * cancels the command currently beeing executed.
     *
     * this function will be called automaticly by the init
     * function 
     *
     * input  : the number of the camera [1;7,8]
     *          camera_id 8 means broadcast, all 
     *          cameras will clear their command
     *          buffers
     * output : SUCCESS when success, else FAILURE
     */
    Result IF_Clear( const unsigned char& );
    
    /**
     * AddressSet
     *
     * Scan all cameras connected to the VISCA-bus 
     * and dispatch one VISCA-address for each camera found.
     *
     * This function will be called automaticly by the 
     * init function
     *
     * input  : none
     * output : SUCCESS when success, else FAILURE
     */
    Result AddressSet();
    
    /**
     * getNumberOfCameras
     *
     * Returns the number of cameras connected to the VISCA-bus
     *
     * input  : none
     * output : number of cameras found
     */
    Result getNumberOfCameras(unsigned char&);
    
    
    /**
     * CommandCancel
     * 
     * Cancel the command currently executed by the camera
     * with the address <camera identificator> and the socket
     * number <socket id>.
     *
     * input  : camera identificator and socket id
     * output : SUCCESS, if the command is canceled
     *          FAILURE, if not
     */
    Result CommandCancel( const unsigned char&, const unsigned char& );
    
    
    /**
     * Power functions
     *
     * The following functions controlls the powerstate 
     * of the cameras. Each camera can be controlled seperatly,
     * selectecd by the <camera id>. The AutoPowerOff-function sets 
     * the timer after the camera goes in standby-mode, when no 
     * VISCA-cammand arrived 
     *
     * input  : camera id
     * output : SUCCESS if success, else FAILURE
     */
    Result PowerOn( const unsigned char& );
    Result PowerOff( const unsigned char& );
    Result AutoPowerOff( const unsigned char&, const int& );
    
    
    /**
     * Zoom functions
     *
     * The following functions controls the zoom of the camera. 
     * All that functions expect the <camera_id> as parameter.
     * Some of them allow additional speed settings 
     *
     * - The Range for the Zoom:
     *   optical zoom : [    0;16384]
     *   digital zoom : [16384;28672]
     * - ZoomDZoom(On|Off) turn the DigitalZoom On or Off.
     */
    Result ZoomStop( const unsigned char& );
    Result ZoomTeleStd( const unsigned char& );
    Result ZoomWideStd( const unsigned char& );
    Result ZoomTeleVar( const unsigned char&, const unsigned char&);
    Result ZoomWideVar( const unsigned char&, const unsigned char&);
    Result ZoomDirect( const unsigned char&, const int& ); 
    Result ZoomDigitalZoomOnOff(const unsigned char&, const bool&);
    
    
    /**
     * Focus functions
     * 
     * the Focus* functions are controling the settings of te
     * camera's focus. All functions expect the <camera_id> 
     * as first parameter, some an additional parameter
     */
    Result FocusStop(const unsigned char&);
    Result FocusFarStd(const unsigned char&);
    Result FocusNearStd(const unsigned char&);
    Result FocusFarVar(const unsigned char&,const unsigned char&);
    Result FocusNearVar(const unsigned char&,const unsigned char&);
    Result FocusDirect(const unsigned char&,const short unsigned int&);
    Result FocusAutoFocus(const unsigned char&); 
    Result FocusManualFocus(const unsigned char&);
    Result FocusAutoManual(const unsigned char&);
    Result FocusOnePushTrigger(const unsigned char&);
    Result FocusInfinity(const unsigned char&);
    Result FocusAFSensHigh(const unsigned char&);
    Result FocusAFSensLow(const unsigned char&);
    Result FocusNearLimit(const unsigned char&,const short unsigned int&);
    
    
    /**
     * WhiteBalance
     */
    Result WhiteBalance(const unsigned char&,const unsigned char&,const unsigned char&);
    Result WhiteBalanceAuto(const unsigned char&);
    Result WhiteBalanceIndoor(const unsigned char&);
    Result WhiteBalanceOutdoor(const unsigned char&);
    Result WhiteBalanceOnePush(const unsigned char&);
    Result WhiteBalanceAT(const unsigned char&);
    Result WhiteBalanceManual(const unsigned char&);
    Result WhiteBalanceOnePushTrigger(const unsigned char&);
    
    
    /**
     * RGain
     */
    Result RGainReset(const unsigned char&);
    Result RGainUp(const unsigned char&);
    Result RGainDown(const unsigned char&);
    Result RGainDirect(const unsigned char&,const short unsigned int&);
    
    
    
    /**
     * BGain
     */
    Result BGainReset(const unsigned char&);
    Result BGainUp(const unsigned char&);
    Result BGainDown(const unsigned char&);
    Result BGainDirect(const unsigned char&,const short unsigned int&);
    
    
    /**
     * Automatic Exposure 
     */
    Result AutomaticExposureFull(const unsigned char&);
    Result AutomaticExposureManual(const unsigned char&);
    Result AutomaticExposureShutterPriority(const unsigned char&);
    Result AutomaticExposureIrisPriority(const unsigned char&);
    Result AutomaticExposureGainPriority(const unsigned char&);
    Result AutomaticExposureBright(const unsigned char&);
    Result AutomaticExposureShutterAuto(const unsigned char&);
    Result AutomaticExposureIrisAuto(const unsigned char&);
    Result AutomaticExposureGainAuto(const unsigned char&);
    
    /**
     * Slow Shutter
     */
    Result SlowShutter(const unsigned char&,const bool&);
    
    /**
     * Shutter
     */
    Result ShutterReset(const unsigned char&);
    Result ShutterUp(const unsigned char&);
    Result ShutterDown(const unsigned char&);
    Result ShutterDirect(const unsigned char&,const short unsigned int&);
    

    /**
     * Iris
     */
    Result IrisReset(const unsigned char&);
    Result IrisUp(const unsigned char&);
    Result IrisDown(const unsigned char&);
    Result IrisDirect(const unsigned char&,const short unsigned int&);


    /**
     * Gain
     */
    Result GainReset(const unsigned char&);
    Result GainUp(const unsigned char&);
    Result GainDown(const unsigned char&);
    Result GainDirect(const unsigned char&,const short unsigned int&);

    /**
     * Bright
     */
    Result BrightReset(const unsigned char&);
    Result BrightUp(const unsigned char&);
    Result BrightDown(const unsigned char&);
    Result BrightDirect(const unsigned char&,const short unsigned int&);
    
    /**
     * Exposure Compensation
     */
    Result ExposureCompensationOnOff(const unsigned char&,const bool&);
    Result ExposureCompensationReset(const unsigned char&);
    Result ExposureCompensationUp(const unsigned char&);
    Result ExposureCompensationDown(const unsigned char&);
    Result ExposureCompensationDirect(const unsigned char&,const short unsigned int&);
    
    /** 
     * BackLight
     */
    Result BackLightOnOff(const unsigned char&,const bool&);
    
    
    /**
     * Aperture
     */
    Result ApertureReset(const unsigned char&);
    Result ApertureUp(const unsigned char&);
    Result ApertureDown(const unsigned char&);
    Result ApertureDirect(const unsigned char&,const short unsigned int&);
    
    
    /**
     * Wide Mode
     */
    Result WideModeOff(const unsigned char&);
    Result WideModeCinema(const unsigned char&);
    Result WideMode16to9Full(const unsigned char&);
    
    
    /**
     *  Mirror image on/off
     */
    Result LR_Reverse(const unsigned char&, const bool&);
    
    /**
     *  Freeze image on/off
     */
    Result Freeze(const unsigned char&, const bool&);
    
    
    /**
     * Picture Effects
     *
     * These functions allows to turn internal effects on or off. 
     * They expect as parameter the <camera_id>.
     */
    Result PictureEffectOff( const unsigned char& );
    Result PictureEffectPastel( const unsigned char& );
    Result PictureEffectNegArt( const unsigned char& );
    Result PictureEffectSepia( const unsigned char& );
    Result PictureEffectBW( const unsigned char& );
    Result PictureEffectSolarize( const unsigned char& );
    Result PictureEffectMosaic( const unsigned char& );
    Result PictureEffectSlim( const unsigned char& );
    Result PictureEffectStretch( const unsigned char& );
    

    /**
     * Digital Effects
     *
     * These functions allows to turn internal digital effects 
     * on or off. They expect as parameter the <camera_id>.
     *
     * Please don't call DigitalEffect( const unsigned char&, const unsigned char& ), its a parent
     * functions for all the other effect functions  
     */
    Result DigitalEffectOff( const unsigned char& );
    Result DigitalEffectStill( const unsigned char& );
    Result DigitalEffectFlash( const unsigned char& );
    Result DigitalEffectLumi( const unsigned char& );
    Result DigitalEffectTrail( const unsigned char& );
    Result DigitalEffectLevel( const unsigned char& , const unsigned char&, const unsigned char& );
    

    /**
     * Memory (user defined presets)
     */
    Result Memory(const unsigned char&,const unsigned char&,const unsigned char&);
    Result MemoryReset( const unsigned char&, const unsigned char& );
    Result MemorySet( const unsigned char&, const unsigned char& );
    Result MemoryRecall( const unsigned char&, const unsigned char& );
    
    
    /**
     * DataScreen (OSD)
     */
    Result DataScreenOnOff(const unsigned char&,const bool&);
    

    /**
     * IR (remote controller)
     */
    Result IRReceiveOnOff(const unsigned char&,const bool&);
    Result IRReceiveReturn(const unsigned char&,const bool&);
    

    /** 
     * PanTiltDriveMotion
     *
     * input  : camera-identifier,
     *          motion x-direction, motion y-direction, 
     *          speed x-direction, speed y-direction
     * output : SUCCESS, if command send with success
     *          FAILURE, if an error occured
     * 
     * used for PanTiltDrive-motion functions
     */
    Result PanTiltDriveMotion(const unsigned char&, const unsigned char&, const unsigned char&, const unsigned char&, const unsigned char&);
    
    Result PanTiltDriveUp( const unsigned char&, const unsigned char&, const unsigned char&);
    Result PanTiltDriveDown( const unsigned char&, const unsigned char&, const unsigned char&);
    Result PanTiltDriveLeft( const unsigned char&, const unsigned char&, const unsigned char&);
    Result PanTiltDriveRight( const unsigned char&, const unsigned char&, const unsigned char&);
    Result PanTiltDriveUpLeft( const unsigned char&, const unsigned char&, const unsigned char&);
    Result PanTiltDriveUpRight( const unsigned char&, const unsigned char&, const unsigned char&);
    Result PanTiltDriveDownLeft( const unsigned char&, const unsigned char&, const unsigned char&);
    Result PanTiltDriveDownRight( const unsigned char&, const unsigned char&, const unsigned char&);
    Result PanTiltDriveStop( const unsigned char&, const unsigned char&, const unsigned char&);
    
    /** 
     * PanTiltDrivePosition
     *
     * input  : camera-identifier,
     *          position-mode (relative or absolute)
     *          motion x-direction in coordinates
     *          morion y-direction in coordinates
     *          speed x-direction, speed y-direction
     * output : SUCCESS, if command send with success
     *          FAILURE, if an error occured
     * 
     * used for PanTiltDrive-position functions
     */
    Result PanTiltDrivePosition( const unsigned char&, const unsigned char&, const int&, const int&, const unsigned char&, const unsigned char& );
    Result PanTiltDriveAbsolute( const unsigned char&, const int&, const int&, const unsigned char&, const unsigned char&);
    Result PanTiltDriveRelative( const unsigned char&, const int&, const int&, const unsigned char&, const unsigned char&);
    
    Result PanTiltDriveAbsoluteDeg( const unsigned char&, const int&, const int&, const unsigned char&, const unsigned char&);
    Result PanTiltDriveRelativeDeg( const unsigned char&, const int&, const int&, const unsigned char&, const unsigned char&);
    
    Result PanTiltDriveHome( const unsigned char& );
    Result PanTiltDriveReset( const unsigned char& );
    Result PanTiltLimitSet( const unsigned char&, const unsigned char&, const int&, const int& );
    Result PanTiltLimitClear( const unsigned char&, const unsigned char& );
    
    
    /**
     * Inquiry Command List 
     */
    
    /**
     * PowerInq
     *
     * input : The camera ID
     * ouput : The status of the Camera
     *          true = On
     *          false =Off(Standby)
     */
    Result PowerInq( const unsigned char&,bool& );
    Result AutoPowerOffInq( const unsigned char&,short unsigned int& );
    Result DZoomModeInq(const unsigned char&, bool&);
    Result ZoomPosInq(const unsigned char&, short unsigned int&);
    Result FocusModeInq(const unsigned char&,bool&);
    Result FocusPosInq(const unsigned char&,short unsigned int&);
    Result AFModeInq(const unsigned char&,bool&);
    Result FocusNearLimitInq(const unsigned char&,short unsigned int&);
    Result WhiteBalanceModeInq(const unsigned char&,unsigned char&);
    //    Result RGainInq(const unsigned char&,short unsigned int&);
    Result RGainInq(const unsigned char&,unsigned char&);
    Result BGainInq(const unsigned char&,short unsigned int&);
    Result AutoExposureModeInq(const unsigned char&,unsigned char&);
    Result SlowShutterModeInq(const unsigned char&,bool&);
    Result ShutterPosInq(const unsigned char&,short unsigned int&);
    Result IrisPosInq(const unsigned char&,short unsigned int&);
    Result GainPosInq(const unsigned char&,short unsigned int&);
    Result BrightPosInq(const unsigned char&,short unsigned int&);
    Result ExposureCompensationModeInq(const unsigned char&,bool&);
    Result ExposureCompensationPosInq(const unsigned char&,short unsigned int&);
    Result BackLightModeInq(const unsigned char&,bool&);
    Result ApertureInq(const unsigned char&,short unsigned int&);
    Result WideModeInq(const unsigned char&,unsigned char&);
    Result LR_ReverseModeInq(const unsigned char&, bool&);
    Result FreezeModeInq(const unsigned char&, bool&);
    Result PictureEffectModeInq(const unsigned char&, unsigned char&);
    Result DigitalEffectModeInq(const unsigned char&, unsigned char&);
    Result DigitalEffectLevelInq(const unsigned char&, unsigned char&);
    Result MemoryInq(const unsigned char&,unsigned char&);
    Result DatascreenInq(const unsigned char&,bool&);
    Result PanTiltModeInq(const unsigned char&,short unsigned int&);
    Result PanTiltMaxSpeedInq(const unsigned char&,unsigned char&,unsigned char&);
    Result PanTiltPosInq(const unsigned char&, int&, int&);
    //    Result PanTiltPosInq( const unsigned char& , struct_camerapos* );
    Result VideoSystemInq(const unsigned char&,unsigned char&);
    Result DeviceTypeInq(const unsigned char&, short unsigned int&, short unsigned int&, short unsigned int&, unsigned char&);
    
    /**
     * misc stuff
     */
//     void verbose(const string&, struct_visca_command);
    void verbose(const string&, const string&);
    bool islocked(const unsigned char&);
    
  private:
    
    typedef struct struct_slots {
      bool active;
      unsigned char socket;
      struct_visca_command expected_answer;
      struct_visca_command comparison_mask;
      struct timeval expiredate;
    };
    
    typedef struct struct_camera_buffer {
      bool         locked;
      unsigned char         lockedslot;
      struct_slots slots[MAX_SLOTS_PER_CAMERA];
    };
    
    struct_visca_command visca_command;
    struct_visca_command visca_response1;
    struct_visca_command visca_response2;
    
    struct_camera_buffer slots[MAX_VISCA_DEVICES];
    
    
    /* stores port parameters */
    struct termios *old_termios, *new_termios;

    string serial_device;
    int spd; /* Serial Port Descriptor */
    
    int number_of_cameras;
    int number_of_tries;
    
    ThreadMutex mutex_command_send;

    /**
     * low level access functions to camera(s)
     */
    Result send_to_camera(struct_visca_command*);
    Result read_from_camera(struct_visca_command*);
    bool compare_commands( struct_visca_command*, struct_visca_command* );
    
    Result wait_for_ack(const unsigned char&,const unsigned char&);
    Result send_command_and_wait_for_ack(struct_visca_command *cmdsend, bool lockcam=false);
    Result send_inquiry_command(const unsigned char&, const unsigned char&, const unsigned char&, struct_visca_command*);
    
    
    /**
     * slot control
     */
    int check_slots_for_timeout(const unsigned char&);
    int find_free_slot(const unsigned char&);
    
    
    /**
     * misc stuff
     */
    void compose_command6(const unsigned char&,const unsigned char&,const unsigned char&,const unsigned char&,struct_visca_command*);
    int check_response(struct_visca_command *response, bool);
    Result process_response(struct_visca_command*);

  }; /* class */
  
}; /* namespace NMM */

#endif /* _CAMERAREADNODE_HPP_ */
