/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_VIDEOSCALERNODE_HPP_
#define _NMM_VIDEOSCALERNODE_HPP_ _NMM_VIDEOSCALERNODE_HPP_

#include "nmm/base/GenericProcessorNode.hpp"
#include "nmm/misc/scaler/VideoScaler.hpp"

#include "nmm/interfaces/video/IScalerImpl.hpp"

namespace NMM {
  
  /**
   * Video scaler node.
   * Scales YUV12 video frames of a given input format to the given output format.
   *
   * The scaler has two modes of operation. The keep_aspect_ratio event is used to switch between both modes.
   *
   * If (keep_aspect_ratio == false), the output resolution is set explicitly. Typically it is set by specifying
   * a format when connecting the nodes. It can also be set through the IScaler interface explicitely.
   * If the input resolution or aspect ratio changes, the output resolution is not affected.
   *
   * If (keep_aspect_ratio == true), the output resolution is calculated from the input resolution and aspect ratio.
   * The input resolution can be changed by sending a resolution_changed event. The aspect ratio is set by the
   * aspect_ratio event.
   * The output resolution is recalculated whenever the input resolution or the aspect ratio is changed.
   */
  class VideoScalerNode :
    public GenericProcessorNode,
    public IScalerImpl
  {  
  private:
    bool keep_aspect_ratio;
    bool send_format;
    int ratio_x;
    int ratio_y;
    string channelorder;
    string colorspace;
    string endian;
    string format;
    int bitperpixel;
    float framerate;
    Buffer* saved_buffer;
    
    VideoScaler scaler;
    
    void reaspect();
    void getAspectCorrection(int *w, int *h, int ratio_x, int ratio_y);
    
    /**
     * Setup output format.
     * Sets up the working input/output properties based on the given input format.
     * This is called from doInitOutput and in format changes.
     */
    Result setupOutputFormat(const Format* in_format);
    
  protected:
    virtual Message* processBuffer(Buffer* in_buffer);
    
    virtual Result doInit();
    virtual Result doDeinit();
    virtual Result doInitOutput();
    virtual Result doDeinitOutput();
    virtual Result doActivate();
    
    /**
     * Set auto aspect mode.
     * If the flag is set, the output resolution is calculated from both the input resolution and the aspect ratio
     * whenever the resolution or aspect ratio changes.
     *
     * If the flag is not set, the output resolution is set to match the input resolution whenever the input resolution
     * changes.
     *
     * You can also set this flag by sending a keep_aspect_ratio event.
     */
    void setAutoAspect(const bool& value);

    /**
     * Get auto aspect mode.
     */
    bool getAutoAspect() const;
    
    /**
     * Set output size.
     * Forces the output size to the given value.
     * In auto aspect mode, the output size changes if the input size or aspect ratio changes.
     */
    void setOutputSize(const int& w, const int& h);
    
    /**
     * Get output size.
     * Returns the current output size of the scaler
     */
    void getOutputSize(int& w, int& h) const;
    
    /**
     * Enable or disable filtering.
     * Enables bilinear filtering when the output resolution is greater than the input resolution.
     * Currently broken. Do not use.
     */
    void setFiltering(const bool& filtering);

    /**
     * Get filtering mode.
     */
    bool getFiltering() const;
    
    /**
     * Format change: do-method
     */
    Result doSetFormat(const Format& format);
    
  public:
    VideoScalerNode(const char *name = "VideoScalerNode",
		    const StreamQueue::Mode queue_mode = StreamQueue::MODE_SUSPEND, 
		    const int queue_size = 1 );
    ~VideoScalerNode();
  }; 
}

#endif // _NMM_VIDEOSCALERNODE_HPP_
