/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_DIVX4ENCODENODE_HPP_
#define _NMM_DIVX4ENCODENODE_HPP_ _NMM_DIVX4ENCODENODE_HPP_

#include <sys/time.h>
#include "nmm/base/GenericProcessorNode.hpp"

#include "nmm/interfaces/general/IEncodeImpl.hpp"
#include "nmm/interfaces/stats/IEncodeStatsImpl.hpp"
#include "nmm/interfaces/file/ITrackImpl.hpp"

extern "C" {
#include <encore2.h>

#include <inttypes.h>
#include <fcntl.h>
}


namespace NMM {

  /**
   * This class takes YUV-Video-frames and converts
   * them to divx4-frames by using the opendivx-library from
   * http://www.projectmayo.com or the DivX4-Library from
   * http://www.divx.com
   *
   * This Node has one Input Stream and one Output Stream.
   */
  class DivX4EncodeNode :
    public GenericProcessorNode,
    public IEncodeImpl,
    public ITrackImpl,
    public IEncodeStatsImpl
  {
  private:
    Result releaseEncoder(); ///< Release the Encoder-library
    
    // variables taken from ENC_PARAM struct from encore2.h
    ENC_PARAM* divx;         ///< structure from encore2.h used to deliver parameters at initialization
    int x_dim;		     ///< the x dimension of the frames to be encoded
    int y_dim;		     ///< the y dimension of the frames to be encoded
    float framerate;	     ///< the frame rate of the sequence to be encoded
    long bitrate;	     ///< the bitrate of the target encoded stream
    int rc_period;	     ///< the intended rate control averaging period
    int rc_reaction_period;  ///< the reaction period for rate control
    int rc_reaction_ratio;   ///< the ratio for down/up rate control
    int max_quantizer; 	     ///< the upper limit of the quantizer
    int min_quantizer; 	     ///< the lower limit of the quantizer
    int max_key_interval;    ///< the maximum interval between key frames
    int deinterlace;	     ///< fast deinterlace
    int quality;	     ///< the quality of compression ( 1 - fastest, 5 - best )
    void *handle;	     ///< the empty handle, which will be filled by encore
    
    ENC_FRAME encframe;      ///< structure from encore2.h used to encode frames
    size_t*    outframe;     ///< space for divx4-encoded frame(s)
    int length;              ///< length of the encoded bitstream
    int colorspace;          ///< the format of the image frame
    int quant;               ///< quantizer of this frame, only used in VBR modes
    int intra;               ///< force this frame to be intra/inter, only used in VBR 2-pass
    void* mvs;               ///< optional pointer to array of motion vectors
    
    int frame_count;         ///< count of frames that are encoded
    struct timeval tv;       ///< remember time to calculate frames per second
    struct timezone tz;      ///< timezone
    int startsec;            ///< the second we started
    int neededsecs;          ///< second since start
    float encfps;            ///< average frames per second
    
    int encoded_bytes;       ///< number of bytes that we encoded
    bool released;           ///< flag: is the encoder already released?
    
    const string compressor; ///< Name of Compressor, "DIVX" for us
    const int enc_buf_size;  ///< the default size for an encoded buffer
    
    ThreadMutex divx_mutex;  ///< Thread-Mutex
    
    Buffer* last_buffer;     ///< Buffer-Pointer to remember the Buffer if an Event is sent
    
    bool send_frame_count;   ///< Send frame count events
    int last_frame_count;    ///< Last frame count when a frame count event was sent
    
  protected:
    Result doActivate();     ///< The standard NMM doActivate()-Method
    Result doInit();         ///< The standard NMM doInit()-Method
    Result doDeactivate();   ///< The standard NMM doDeactivate()-Method
    Result doInitOutput();   ///< The standard NMM doInitOutput()-Method
    
    Message* processBuffer(Buffer*); ///< The standard NMM processBuffer method. Here, the raw video buffers are encoded
    
    /**
     * ITrack interface.
     * Instream event handlers for start and end of tracks.
     */
    virtual Result startTrack(const string& filename);
    virtual Result endTrack();
    virtual Result finished();
    
  public:

    DivX4EncodeNode(const char* name = "DivX4EncodeNode",
		    StreamQueue::Mode qmode = StreamQueue::MODE_SUSPEND,
		    int qsize = StreamQueue::DEFAULT_SIZE);
    ///< Constructor; takes the Nodename, the Mode of the Stream-Queue and the size of the Stream-Queue as arguments (all optional)
    ~DivX4EncodeNode(); ///< Destructor
    
    /**
     * IEncode interface.
     * Get and set encoding parameters.
     */
    virtual void setQuality(const float& quality);
    virtual void setBitrate(const int& bitrate);
    virtual float getQuality();
    virtual int getBitrate();
    
    /** IProgress interface. */
    virtual void sendFrameCount(const bool& value);
  };

}

#endif //_NMM_DIVX4ENCODENODE_HPP_
