/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:
 *                    Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_FFMPEGDECODENODE_HPP_
#define _NMM_FFMPEGDECODENODE_HPP_ _NMM_FFMPEGDECODENODE_HPP_

#include <sys/time.h>

#include "nmm/base/GenericProcessorNode.hpp"
#include "nmm/interfaces/file/ITrackImpl.hpp"
#include "nmm/interfaces/video/IVideoDelayImpl.hpp"
#include "nmm/interfaces/video/IKeyframeImpl.hpp"

class AVCodec;
class AVCodecContext;
class AVFrame;

namespace NMM {

  /**
   * This class takes encoded YUV-Video-frames and converts
   * them to decoded frames by using the FFMpeg-library from
   * http://ffmpeg.sourceforge.net/
   *
   * This Node has one Input Stream and one Output Stream.
   */
  class FFMpegDecodeNode :
    public GenericProcessorNode,
    public ITrackImpl,
    public IVideoDelayImpl,
    public IKeyframeImpl
  {
  private:
    class FFMpegHandle;

    FFMpegHandle* ffmpeg;    ///< class containing all avcodec relevant stuff
    int height;              ///< height of the frames
    int height_half;         ///< height DIV 2
    int width;               ///< width of the frames
    int width_half;          ///< width DIV 2
    float framerate;	     ///< the frame rate of the sequence to be encoded
    const int framerate_base; ///< constant to bring framerates to integer values
    int int_framerate;       ///< the framerate as integer (framerate * framerate_base);
    long bitrate;	     ///< the bitrate of the target encoded stream
    int gop_size;            ///< The size of one Group of Pictures

    int avframe_size;        ///< size of the Y-Plane of an image
    int out_size;            ///< the output size of an encoded frame

    int frame_count;         ///< count of frames that are encoded
    Timestamp timestamp;     ///< The timestamp to put in the next outgoing buffer.
    Timestamp in_buf_ts;

    struct timeval start_time; ///< start time of the stream, 0 in our case.
    struct timeval tv;       ///< remember time to calculate frames per second
    struct timezone tz;      ///< timezone
    int startsec;            ///< the second we started
    int neededsecs;          ///< second since start
    float encfps;            ///< average frames per second

    int got_picture;         ///< >0 if data could be decoded
    Buffer* out_buffer;      ///< pointer to outgoing buffer
    unsigned char* out_buffer_data1;  ///< pointer to outgoing data
    unsigned char* out_buffer_data2;  ///< pointer to outgoing data
    unsigned char* src1;     ///< pointer to avcodec data
    unsigned char* src2;     ///< pointer to avcodec data
    int out_buffer_size;     ///< precalculated outgoing buffer size
    int line;

    int vid_delay;           ///< video delay

    bool skip_until_keyframe;///< flag to indicat that all non-keyframes until the next kf should be dropped

    typedef std::map<std::string, int> CodecMap;
    typedef CodecMap::iterator CodecIterator;
    typedef CodecMap::const_iterator CodecConstIterator;

    CodecMap codecs;       ///< FourCC codes of all available codecs, mapped to codec IDs.
    std::string default_codec; ///< FourCC code of default codec.

  public:
    FFMpegDecodeNode(const char* name = "FFMpegDecodeNode",
		    StreamQueue::Mode qmode = StreamQueue::MODE_SUSPEND,
		    int qsize = StreamQueue::DEFAULT_SIZE);
    ~FFMpegDecodeNode();

  protected:
    /* GenericNode */
    virtual Result doInit();
    virtual Result doInitOutput();
    virtual Result doActivate();
    virtual Result doDeactivate();
    virtual Result doFlush();
    virtual Message* processBuffer(Buffer*);

    /* ITrackImpl */
    virtual Result startTrack(const string&);
    virtual Result endTrack();
    virtual Result finished();

    /* IVideoDelayImpl */
    virtual Result setVideoDelay(const int&);

    /* IKeyframeImpl */
    virtual Result setKeyframe();

  private:
    void detectAllCodecs() throw();
    void addCodec(const std::string& fourcc, int id);

    /**
     * struct FFMpegHandle: This is used to handle all ffmpeg relevant stuff
     */
   struct FFMpegHandle {
     AVCodec* ffAVCodec;
     AVCodecContext* ffavcc;
     AVFrame* avframe;
     FFMpegHandle();
   };

  };

}

#endif //_NMM_FFMPEGDECODENODE_HPP_
