/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:
 *                    Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "FFMpegEncodeNode.hpp"
#include "nmm/interfaces/video/IKeyframe.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/registry/Plugin.hpp"

#include <sys/time.h>

#include <ffmpeg/avcodec.h>

// BEGIN: patch from Ingo Hornberger
#ifndef FRAME_RATE_BASE
#define FRAME_RATE_BASE 10000
#endif

#ifndef UINT8
#define UINT8 uint8_t
#endif
// END: patch from Ingo Hornberger

NMM_REGISTER_PLUGIN(NMM::FFMpegEncodeNode, "FFMpegEncodeNode")

namespace NMM {

  /**
   * class FFMpegHandle: This is used to handle all ffmpeg relevant stuff
   * it avoids problems with including ffmpeg header in FFMpegEncodeNode header file
   */
  class FFMpegEncodeNode::FFMpegHandle {
  public:
    AVCodec* ffAVCodec;
    AVCodecContext* ffavcc;
    AVFrame avframe;
    enum PixelFormat pix_fmt;///< pixel format
    FFMpegHandle() : ffAVCodec(NULL), ffavcc(NULL), avframe(),
                     pix_fmt(PIX_FMT_YUV420P) {}
  };

  FFMpegEncodeNode::FFMpegEncodeNode(const char *name, StreamQueue::Mode qmode, int qsize) :
    GenericProcessorNode(name, qmode, qsize),
    ffmpeg(new FFMpegHandle()),
    height(0),
    width(0),
    framerate(0.0f),
    bitrate(12000),
    gop_size(250),
    qmin(2),
    qmax(31),
    max_qdiff(3),
    qblur(0.5f),
    qcompress(0.5f),
    me_method(ME_EPZS),
    avframe_size(0),
    out_size(0),
    compressor("DIVX"),
    enc_buf_size(0),
    frame_count(0),
    startsec(0),
    neededsecs(0),
    encfps(0.0f),
    last_buffer(NULL),
    send_frame_count(false),
    last_frame_count(0)
  { }

  FFMpegEncodeNode::~FFMpegEncodeNode() {
    /* release ffmpeg library */
    if (ffmpeg != NULL) {
      delete ffmpeg;
    }
  }

  Result FFMpegEncodeNode::startTrack(const string&) {
    DEBUG_STREAM("Received Event: startTrack");
    return SUCCESS;
  }

  Result FFMpegEncodeNode::endTrack() {
    DEBUG_STREAM("Received Event: endTrack");
    return SUCCESS;
  }

  Result FFMpegEncodeNode::finished() {
    DEBUG_STREAM("Received Event: finished");
    return SUCCESS;
  }

  void FFMpegEncodeNode::setQuality(const float&) {
    // quality not defined
  }

  void FFMpegEncodeNode::setBitrate(const int& bitrate) {
    MutexGuard ffmpeg_mg(ffmpeg_mutex);
    this->bitrate = bitrate;
    DEBUG_STREAM("Received Event: setBitrate; bitrate changed to " << bitrate << endl);
  }

  float FFMpegEncodeNode::getQuality() {
    // quality not defined
    return 1.0f;
  }

  int FFMpegEncodeNode::getBitrate() {
    return bitrate;
  }

  Result FFMpegEncodeNode::setFormat(const Format& format) {
    int res_x = format.getIntValue(Format::x_resolution_param);
    int res_y = format.getIntValue(Format::y_resolution_param);

    DEBUG_STREAM("Received format change; set size to to " << res_x << "x" << res_y << endl);
    setSize(res_x, res_y);
    return SUCCESS;
  }

  Result FFMpegEncodeNode::doInit() {
    DEBUG_STREAM("Got avcodec version: " << avcodec_version() << endl);
    DEBUG_STREAM("Got avcodec build: " << avcodec_build() << endl);

    /* initialize library */
    avcodec_init();

    /* register all the codecs */
    avcodec_register_all();

    /* detect all codecs */
    codecs.clear();
    
    __default_codec = "";
    detectAllCodecs();

    /* check if any codec was found */
    if(__default_codec.empty()) {
      ERROR_STREAM("No codecs with known FOURCC code available" << std::endl);
      return FAILURE;
    }

    /* general input format */
    Format* in_format = getOwnInputProperty()->addNewFormat(Format::video_raw_type);
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addStringParamValue(Format::channelorder_param,"yvu");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param,"yv12");
    in_format->addIntParamValue(Format::bitperpixel_param,12);
    in_format->addStringParamValue(Format::format_param, "planar");
    getOwnInputProperty()->setDefaultFormat(in_format);

    /* general output format */
    Format* out_format = getOwnOutputProperty()->addNewFormat(Format::video_mpeg4_type);
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addWildcard(Format::bitrate_param);
    for(CodecConstIterator i = codecs.begin(); i != codecs.end(); ++i) {
      out_format->addStringParamValue(Format::compressor_param, i->first);
    }

    Format* out_format2 = getOwnOutputProperty()->addNewFormat(Format::video_mpeg_type);
    out_format2->addWildcard(Format::x_resolution_param);
    out_format2->addWildcard(Format::y_resolution_param);
    out_format2->addWildcard(Format::version_param);
    out_format2->addWildcard(Format::ratio_x_param);
    out_format2->addWildcard(Format::ratio_y_param);
    out_format2->addWildcard(Format::framerate_param);
    out_format2->addWildcard(Format::bitrate_param);

    Format* out_format3 = getOwnOutputProperty()->addNewFormat(Format::video_jpeg_type);
    out_format3->addWildcard(Format::x_resolution_param);
    out_format3->addWildcard(Format::y_resolution_param);
    out_format3->addWildcard(Format::ratio_x_param);
    out_format3->addWildcard(Format::ratio_y_param);
    out_format3->addWildcard(Format::framerate_param);
    out_format3->addWildcard(Format::bitrate_param);

    Format* out_format4 = getOwnOutputProperty()->addNewFormat(Format::video_h263_type);
    out_format4->addWildcard(Format::x_resolution_param);
    out_format4->addWildcard(Format::y_resolution_param);
    out_format4->addWildcard(Format::ratio_x_param);
    out_format4->addWildcard(Format::ratio_y_param);
    out_format4->addWildcard(Format::framerate_param);
    out_format4->addWildcard(Format::bitrate_param);

    Format* out_format5 = getOwnOutputProperty()->addNewFormat(Format::video_h263plus_type);
    out_format5->addWildcard(Format::x_resolution_param);
    out_format5->addWildcard(Format::y_resolution_param);
    out_format5->addWildcard(Format::ratio_x_param);
    out_format5->addWildcard(Format::ratio_y_param);
    out_format5->addWildcard(Format::framerate_param);
    out_format5->addWildcard(Format::bitrate_param);
    
    /* default output format with default codec */
    Format* def_out_format = getOwnOutputProperty()->addNewFormat(Format::video_mpeg4_type);
    def_out_format->addWildcard(Format::x_resolution_param);
    def_out_format->addWildcard(Format::y_resolution_param);
    def_out_format->addWildcard(Format::ratio_x_param);
    def_out_format->addWildcard(Format::ratio_y_param);
    def_out_format->addWildcard(Format::framerate_param);
    def_out_format->addWildcard(Format::bitrate_param);
    def_out_format->addStringParamValue(Format::compressor_param, __default_codec);
    getOwnOutputProperty()->setDefaultFormat(def_out_format);

    return SUCCESS;
  }

  Result FFMpegEncodeNode::doInitOutput() {
    /* get input format */
    const Format* in_format = getInputFormat();
    if (in_format == NULL) {
      ERROR_STREAM("Missing input format" << std::endl);
      return FAILURE;
    }
    width     = in_format->getIntValue  (Format::x_resolution_param);
    height    = in_format->getIntValue  (Format::y_resolution_param);
    framerate = in_format->getFloatValue(Format::framerate_param);
    DEBUG_STREAM("Got parameters: width = " << width << ", height = " << height << ", framerate = " << framerate << endl);

    /* working input format */
    Format* i_format = new Format(*in_format);
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    getOwnWorkingInputProperty()->setDefaultFormat(i_format);

    
/* general working output format */
    Format* o_format = getOwnWorkingOutputProperty()->addNewFormat(Format::video_mpeg4_type);
    o_format->addIntParamValue(Format::x_resolution_param, width);
    o_format->addIntParamValue(Format::y_resolution_param, height);
    o_format->addIntParamValue(Format::ratio_x_param, in_format->getIntValue(Format::ratio_x_param));
    o_format->addIntParamValue(Format::ratio_y_param, in_format->getIntValue(Format::ratio_y_param));
    o_format->addFloatParamValue(Format::framerate_param, framerate);
    o_format->addWildcard(Format::bitrate_param);
    for(CodecConstIterator i = codecs.begin(); i != codecs.end(); ++i) {
      o_format->addStringParamValue(Format::compressor_param, i->first);
    }

    Format* o_format2 = getOwnWorkingOutputProperty()->addNewFormat(Format::video_mpeg_type);
    o_format2->addIntParamValue(Format::x_resolution_param, width);
    o_format2->addIntParamValue(Format::y_resolution_param, height);
    o_format2->addIntParamValue(Format::version_param,1);
    o_format2->addIntParamValue(Format::version_param,2);
    o_format2->addIntParamValue(Format::ratio_x_param, in_format->getIntValue(Format::ratio_x_param));
    o_format2->addIntParamValue(Format::ratio_y_param, in_format->getIntValue(Format::ratio_y_param));
    o_format2->addFloatParamValue(Format::framerate_param, framerate);
    o_format2->addWildcard(Format::bitrate_param);

    Format* o_format3 = getOwnWorkingOutputProperty()->addNewFormat(Format::video_jpeg_type);
    o_format3->addIntParamValue(Format::x_resolution_param, width);
    o_format3->addIntParamValue(Format::y_resolution_param, height);
    o_format3->addIntParamValue(Format::ratio_x_param, in_format->getIntValue(Format::ratio_x_param));
    o_format3->addIntParamValue(Format::ratio_y_param, in_format->getIntValue(Format::ratio_y_param));
    o_format3->addFloatParamValue(Format::framerate_param, framerate);
    o_format3->addWildcard(Format::bitrate_param);

    Format* o_format4 = getOwnWorkingOutputProperty()->addNewFormat(Format::video_h263_type);
    o_format4->addIntParamValue(Format::x_resolution_param, width);
    o_format4->addIntParamValue(Format::y_resolution_param, height);
    o_format4->addIntParamValue(Format::ratio_x_param, in_format->getIntValue(Format::ratio_x_param));
    o_format4->addIntParamValue(Format::ratio_y_param, in_format->getIntValue(Format::ratio_y_param));
    o_format4->addFloatParamValue(Format::framerate_param, framerate);
    o_format4->addWildcard(Format::bitrate_param);

    Format* o_format5 = getOwnWorkingOutputProperty()->addNewFormat(Format::video_h263plus_type);
    o_format5->addIntParamValue(Format::x_resolution_param, width);
    o_format5->addIntParamValue(Format::y_resolution_param, height);
    o_format5->addIntParamValue(Format::ratio_x_param, in_format->getIntValue(Format::ratio_x_param));
    o_format5->addIntParamValue(Format::ratio_y_param, in_format->getIntValue(Format::ratio_y_param));
    o_format5->addFloatParamValue(Format::framerate_param, framerate);
    o_format5->addWildcard(Format::bitrate_param);

    /* default working output format */
    Format* def_o_format = getOwnWorkingOutputProperty()->addNewFormat(Format::video_mpeg4_type);
    def_o_format->addIntParamValue(Format::x_resolution_param, width);
    def_o_format->addIntParamValue(Format::y_resolution_param, height);
    def_o_format->addIntParamValue(Format::ratio_x_param, in_format->getIntValue(Format::ratio_x_param));
    def_o_format->addIntParamValue(Format::ratio_y_param, in_format->getIntValue(Format::ratio_y_param));
    def_o_format->addFloatParamValue(Format::framerate_param, framerate);
    def_o_format->addWildcard(Format::bitrate_param);
    def_o_format->addStringParamValue(Format::compressor_param, __default_codec);
    getOwnWorkingOutputProperty()->setDefaultFormat(def_o_format);

    /* set I/O partners */
    def_o_format->setIOPartner(i_format);
    o_format->setIOPartner(i_format);
    o_format2->setIOPartner(i_format);
    i_format->setIOPartner(o_format);
    i_format->setIOPartner(o_format2);

    return SUCCESS;
  }

  Result FFMpegEncodeNode::doStart() {
    /* get and remember current time */
    gettimeofday(&tv, &tz);
    startsec = tv.tv_sec;
    return SUCCESS;
  }

  Result FFMpegEncodeNode::doActivate() {
    const Format* out_format = getOutputFormat();
    const Format* in_format  = getInputFormat();
    if (out_format == NULL) {
      ERROR_STREAM("Out format is 0" << endl);
      return FAILURE;
    }
    if (out_format->isParameterIncluded(Format::bitrate_param)) {
      bitrate = out_format->getIntValue(Format::bitrate_param);
    }


    if(out_format->getFormatType() == Format::video_mpeg_type) {
      unsigned int version = out_format->getIntValue(Format::version_param);
      if(version == 1) {
        compressor = "MPG1";
      } else if(version == 2) {
        compressor = "MPG2";
      } else {
        /* SHOULD NOT HAPPEN */
        ERROR_STREAM("Invalid MPEG video version" << endl);
        return FAILURE;
      }
    } else if(out_format->getFormatType() == Format::video_h263_type) {
      compressor = "H263";
    } else if(out_format->getFormatType() == Format::video_h263plus_type) {
      compressor = "H263P";
    } else if(out_format->getFormatType() == Format::video_jpeg_type) {
      compressor = "MJPG";
    } else if(out_format->getFormatType() == Format::video_mpeg4_type) {
      compressor = out_format->getStringValue(Format::compressor_param);
    } else {
      /* SHOULD NOT HAPPEN */
      ERROR_STREAM("Invalid input format type: " << in_format->getFormatType() << endl);
      return FAILURE;
    }

    frame_count = 0;
    neededsecs = 0;
    last_frame_count = 0;
    enc_buf_size = width * height * 3; // FFmpeg seems to dislike too small buffers
                    
    /* find the correct encoder */
    CodecConstIterator found = codecs.find(compressor);
    if(found == codecs.end()) {
      /* should not happen */
      ERROR_STREAM("FATAL: No codec listed for FourCC code " << compressor << std::endl);
    }

    /* find a codec with this ID */
    ffmpeg->ffAVCodec = avcodec_find_encoder(static_cast<CodecID>(found->second));
    if (!ffmpeg->ffAVCodec) { /* should not happen */
      ERROR_STREAM("FATAL: Unsupported codec" << endl);
      return FAILURE;
    }

    /* open the codec */
    return openCodec();
  }

  Result FFMpegEncodeNode::doDeactivate() {
    /* close codec */
    avcodec_close(ffmpeg->ffavcc);

    /* release ffmpeg context */
    if(ffmpeg->ffavcc) {
      free(ffmpeg->ffavcc);
      ffmpeg->ffavcc = NULL;
    }

    return SUCCESS;
  }

  Result FFMpegEncodeNode::setSize(int width, int height) {
    DEBUG_STREAM("set size: " << width << "x" << height << endl);
    if(this->width != width || this->height != height) {
      this->width = width;
      this->height = height;
      enc_buf_size = width * height * 3; // FFmpeg seems to dislike too small buffers

      return openCodec();
    } else {
      return SUCCESS;
    }
  }

  Result FFMpegEncodeNode::openCodec() {
    DEBUG_STREAM("Open codec" << endl);
    /* allocate ffmpeg context */
    ffmpeg->ffavcc = avcodec_alloc_context();

    /* init context */
    ffmpeg->ffavcc->width = width;
    ffmpeg->ffavcc->height = height;
    ffmpeg->ffavcc->frame_rate = static_cast<int>(framerate * FRAME_RATE_BASE);
    ffmpeg->ffavcc->frame_rate_base = FRAME_RATE_BASE;
    ffmpeg->ffavcc->bit_rate = bitrate;

    if (compressor == "DIVX") {
      // some default intializations grabbed from Avidemux/MPlayer
      ffmpeg->ffavcc->bit_rate_tolerance = 8000000;
      ffmpeg->ffavcc->pix_fmt = ffmpeg->pix_fmt;
      ffmpeg->ffavcc->gop_size = gop_size;
      ffmpeg->ffavcc->qmin = qmin;
      ffmpeg->ffavcc->qmax = qmax;
      ffmpeg->ffavcc->max_b_frames = 0;
      ffmpeg->ffavcc->mpeg_quant = 0;
      ffmpeg->ffavcc->me_method = me_method;
      ffmpeg->ffavcc->flags = 0;
      ffmpeg->ffavcc->max_qdiff = max_qdiff;
      ffmpeg->ffavcc->luma_elim_threshold = 0;
      ffmpeg->ffavcc->chroma_elim_threshold = 0;
      ffmpeg->ffavcc->lumi_masking = 0.0;;
      ffmpeg->ffavcc->dark_masking = 0.0;
      ffmpeg->ffavcc->qcompress = qcompress;
      ffmpeg->ffavcc->qblur = qblur;
      ffmpeg->ffavcc->b_quant_factor = 1.25;
      ffmpeg->ffavcc->rc_strategy = 2;
      ffmpeg->ffavcc->b_frame_strategy = 0;
      ffmpeg->ffavcc->b_quant_offset = 1.25;
      ffmpeg->ffavcc->rtp_payload_size = 0;
      ffmpeg->ffavcc->strict_std_compliance = 0;
      ffmpeg->ffavcc->i_quant_factor = 0.8;
      ffmpeg->ffavcc->i_quant_offset = 0.0;
      ffmpeg->ffavcc->rc_qsquish = 1.0;
      ffmpeg->ffavcc->rc_qmod_amp = 0;
      ffmpeg->ffavcc->rc_qmod_freq = 0;
      ffmpeg->ffavcc->rc_eq = const_cast<char *>("tex^qComp"); // rather evil but ought to work - let's blame it on the ffmpeg api
      ffmpeg->ffavcc->rc_max_rate = 000;
      ffmpeg->ffavcc->rc_min_rate = 000;
      ffmpeg->ffavcc->rc_buffer_size = 000;
      ffmpeg->ffavcc->rc_buffer_aggressivity = 1.0;
      ffmpeg->ffavcc->rc_initial_cplx = 0;
      ffmpeg->ffavcc->dct_algo = 0;
      ffmpeg->ffavcc->idct_algo = 0;
      ffmpeg->ffavcc->p_masking = 0.0;
    }

    
                { /* open the codec */
                        MutexGuard ffmpeg_mg(ffmpeg_mutex); // avcodec_open is not thread safe
                        if (avcodec_open(ffmpeg->ffavcc, ffmpeg->ffAVCodec) < 0) {
                                ERROR_STREAM("Could not open codec" << std::endl);
        
                                /* release ffmpeg context */
                                if(ffmpeg->ffavcc) {
                                        free(ffmpeg->ffavcc);
                                        ffmpeg->ffavcc = NULL;
                                }
        

                                return FAILURE;
                        }
                }

    ffmpeg->avframe.linesize[0] = ffmpeg->ffavcc->width;
    ffmpeg->avframe.linesize[1] = ffmpeg->ffavcc->width / 2;
    ffmpeg->avframe.linesize[2] = ffmpeg->ffavcc->width / 2;
    avframe_size = ffmpeg->ffavcc->width * ffmpeg->ffavcc->height;

    return SUCCESS;
  }

  Message* FFMpegEncodeNode::processBuffer(Buffer* in_buffer) {
    if (getWorkingFlag()) {
      DEBUG_STREAM("Send remembered buffer...");
      setWorkingFlag(false);
      return last_buffer;
    }

    DEBUG_STREAM("in_buffer->getUsed() = " << in_buffer->getUsed() << endl);

    Buffer* out_buffer;

    out_buffer = getNewBuffer(enc_buf_size, in_buffer);

    // encode the image
    ffmpeg->avframe.data[0] = reinterpret_cast<UINT8*>(in_buffer->getData());
    // have to swap U and V because libavcodec cannot take YV12 but (e.g.) YUV420P
    // as Pix-Format,
    ffmpeg->avframe.data[2] = ffmpeg->avframe.data[0] + avframe_size;
    ffmpeg->avframe.data[1] = ffmpeg->avframe.data[2] + avframe_size / 4;
    out_size = avcodec_encode_video(ffmpeg->ffavcc,
                                    reinterpret_cast<UINT8*>(out_buffer->getData()),
                                    enc_buf_size,
                                    &ffmpeg->avframe);
    out_buffer->setUsed(out_size);

    frame_count++;

    DEBUG_STREAM("Encoding results: "                          << endl <<
                 "Key_frame: "   << ffmpeg->ffavcc->coded_frame->key_frame << endl <<
                 "mv_bits: "     << ffmpeg->ffavcc->mv_bits                << endl <<
                 "header_bits: " << ffmpeg->ffavcc->header_bits            << endl <<
                 "i_tex_bits: "  << ffmpeg->ffavcc->i_tex_bits             << endl <<
                 "p_tex_bits: "  << ffmpeg->ffavcc->p_tex_bits             << endl <<
                 "i_count: "     << ffmpeg->ffavcc->i_count                << endl <<
                 "p_count: "     << ffmpeg->ffavcc->p_count                << endl <<
                 "skip_count: "  << ffmpeg->ffavcc->skip_count             << endl <<
                 "misc_bits: "   << ffmpeg->ffavcc->misc_bits              << endl <<
                 "frame_bits: "  << ffmpeg->ffavcc->frame_bits             << endl <<
                 "out_size:"     << out_size                               << endl);

    // the uncompressed Frame is not needed any more
    in_buffer->release();

    if (ffmpeg->ffavcc->coded_frame->key_frame) {
      // if we got a key frame, send an Event telling about it first
      setWorkingFlag(true);
      last_buffer = out_buffer;
      DEBUG_STREAM("Send Event key_frame...");
      return new CEvent(Event(IKeyframe::setKeyframe_event));
    }

    if (send_frame_count && frame_count - last_frame_count > 10) {
      setWorkingFlag(true);
      last_buffer = out_buffer;

      last_frame_count = frame_count;

      // check time and calculate encfps (encoding frames per second)
      gettimeofday(&tv, &tz);
      neededsecs = tv.tv_sec - startsec;
      //encoded_bytes += encframe.length;
      encfps = (float)frame_count/(float)neededsecs;

      return new CEvent( IEncodeStatsListener::create_setFrameCount((unsigned int) frame_count, encfps) );
    }

    return out_buffer;
  }

  void FFMpegEncodeNode::sendFrameCount(const bool& value) {
    send_frame_count = value;
  }

  void FFMpegEncodeNode::detectAllCodecs() {
    /* get all available codecs */
    AVCodec *p = first_avcodec;
    DEBUG_STREAM("Available Encoders:" << endl);
    while (p) {
      if (p->encode) {
        DEBUG_STREAM("  " << p->name << endl);
        switch(p->id) {
        case CODEC_ID_MSMPEG4V2:
          addCodec("MP42", p->id);
          break;
        case CODEC_ID_MSMPEG4V3:
          addCodec("MP43", p->id);
          break;
        case CODEC_ID_MPEG1VIDEO:
          addCodec("MPG1", p->id);
          addCodec("MPEG", p->id);
          break;
        case CODEC_ID_MPEG2VIDEO:
          addCodec("MPEG", p->id);
          addCodec("MPG2", p->id);
          break;
        case CODEC_ID_H263:
          addCodec("H263", p->id);
          break;
        case CODEC_ID_H263P:
          addCodec("H263P", p->id);
          break;
        case CODEC_ID_MJPEG:
          addCodec("MJPG", p->id);
          break;
        case CODEC_ID_MPEG4:
          addCodec("DIVX", p->id);
          addCodec("DX50", p->id);
          break;
        case CODEC_ID_WMV1:
          addCodec("WMV1", p->id);
          break;
        default:
          break;
        }
      }
      p = p->next;
    }
  }

  void FFMpegEncodeNode::addCodec(const std::string& fourcc, int id) {
    if(__default_codec.empty()) {
      /* if we have no default yet, use this one */
      __default_codec = fourcc;
    } else if(fourcc == "DIVX") {
      /* if DIVX is available, override default */
      __default_codec = fourcc;
    }

    /* add codec to map */
    codecs.insert(std::pair<std::string, int>(fourcc, id));

    DEBUG_STREAM("    " << fourcc << std::endl);
  }
}
