/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "YUVDeInterlaceNode.hpp"
#include "yv12_deinterlace.h"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/registry/Plugin.hpp"

NMM_REGISTER_PLUGIN(NMM::YUVDeInterlaceNode, "YUVDeInterlaceNode")

namespace NMM {

  YUVDeInterlaceNode::YUVDeInterlaceNode(const char *name,
					 const StreamQueue::Mode qmode,
					 const int qsize) :
    GenericFilterNode(name, qmode, qsize)
  {
    width = 0;
    height = 0;
    
    registerEvent("resolution_changed", new TEDObject2<YUVDeInterlaceNode,int,int>(this, &YUVDeInterlaceNode::resize));
  }

  YUVDeInterlaceNode::~YUVDeInterlaceNode() {}

  Result YUVDeInterlaceNode::resize(int& width_, int& height_) {
    width = width_;
    height = height_;
    return SUCCESS;
  }
  
  
  Result YUVDeInterlaceNode::doInit() {
    // register general input format
    Format* in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param); //
    in_format->addWildcard(Format::ratio_y_param);//
    in_format->addWildcard(Format::framerate_param);//
    in_format->addStringParamValue(Format::channelorder_param,"yvu");
    in_format->addStringParamValue(Format::colorspace_param,"yv12");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addIntParamValue(Format::bitperpixel_param, 12);
    in_format->addStringParamValue(Format::format_param, "planar");
    
    // register general output format
    
    Format* o_format = getOwnOutputProperty()->addNewFormat("video/raw");
    o_format->addWildcard(Format::x_resolution_param);
    o_format->addWildcard(Format::y_resolution_param);
    o_format->addWildcard(Format::ratio_x_param);
    o_format->addWildcard(Format::ratio_y_param);
    o_format->addWildcard(Format::framerate_param);
    o_format->addStringParamValue(Format::channelorder_param,"yvu");
    o_format->addStringParamValue(Format::endian_param, "little");
    o_format->addStringParamValue(Format::colorspace_param,"yv12");
    o_format->addIntParamValue(Format::bitperpixel_param, 12);
    o_format->addStringParamValue(Format::format_param, "planar");

    // set each format as default format
    getOwnOutputProperty()->setDefaultFormat(o_format);
    getOwnInputProperty()->setDefaultFormat(in_format);
    
    return SUCCESS;
  }
  
  Result YUVDeInterlaceNode::doDeinit() {
    getOwnInputProperty()->erase();
    getOwnOutputProperty()->erase();
    return SUCCESS;
  }
  
  Result YUVDeInterlaceNode::doInitOutput() {
    // get input format and make sure we have one
    const Format *in_f = getInputFormat();
    if (in_f == NULL) {
      DEBUG_STREAM("Couldn't get input Format!!!" << endl);
      return FAILURE;
    }
    
    return setupOutputFormat(in_f);
  }
  
  Result YUVDeInterlaceNode::setupOutputFormat(const Format* in_f) {
    // check for required parameters
    if (!in_f->isParameterIncluded(Format::x_resolution_param)) { return FAILURE; }
    if (!in_f->isParameterIncluded(Format::y_resolution_param)) { return FAILURE; }
    width  = in_f->getIntValue(Format::x_resolution_param);
    height = in_f->getIntValue(Format::y_resolution_param);
    
    DEBUG_STREAM("width = " << width << endl);
    DEBUG_STREAM("height = " << height << endl);
    
    if ((width <= 0) || (height <= 0)) {
      DEBUG_STREAM("height or width < 0 !!!" << endl);
      return FAILURE;
    }
    
    // set output format
    // this is equal to input format for all filter nodes
    Format* out_f = new Format(*in_f);
    getOwnWorkingOutputProperty()->addNewFormat(out_f);
    getOwnWorkingOutputProperty()->setDefaultFormat(out_f);
    
    return SUCCESS;
  }
  
  Message* YUVDeInterlaceNode::processBuffer(Buffer *in_buffer) {
    in_buffer = in_buffer->getWriteableInstance();
    
    yv12_deinterlace(in_buffer->getData(), width, height);
    
    return in_buffer;
  }
  
  Result YUVDeInterlaceNode::setFormat(const Format& format) {
    getWorkingInputProperty()->erase();
    getWorkingOutputProperty()->erase();
    
    return setupOutputFormat(&format);
  }
}
