/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_VIDEO_FILTER_OSDOVERLAYNODE_HPP_
#define _NMM_VIDEO_FILTER_OSDOVERLAYNODE_HPP_ _NMM_VIDEO_FILTER_OSDOVERLAYNODE_HPP_

#include "nmm/base/GenericFilterNode.hpp"
#include "nmm/base/EventDispatcher.hpp"

#include "nmm/interfaces/video/osd/IOSDOverlayImpl.hpp"

extern "C" {
#include "nmm/plugins/video/osd/font/font_load.h"
}

#include <png.h>

#define OSD_PB_START 0x10
#define OSD_PB_0 0x11
#define OSD_PB_END 0x12
#define OSD_PB_1 0x13

#define RGBA32 0
#define YVA12 1

namespace NMM {

  /**
   * This class takes an (RGBA32) video streams and puts an Onscreendisplay,
   * consisting of multiple texts or images,
   * onto the background. The output format is the one of the background stream.
   * 
   */
  class OSDOverlayNode : public GenericFilterNode,
                         public IOSDOverlayImpl {
    static unsigned short alphaMMXmul_const1[];
    static unsigned alphaMMXmul_0[];
    
    class OSDItem {
    public:
      size_t x_offset;
      size_t y_offset;
      size_t width;
      size_t height;
      size_t depth;
      bool enabled;
      
      OSDItem();
      
      virtual Result draw(char *data, int color, size_t width, size_t height) = 0;
      virtual unsigned char* getYUVData() = 0;
      virtual int getYUVSize() = 0;
    };
    
    class OSDImagePNG : public OSDItem {
    private:
      png_bytep *png_rows;               ///< stores the PNG image
      png_uint_32 png_width, png_height; ///< PNG image width and height
      unsigned char *yuv_buffer;         ///< PNG image in YUV format
      
      int image_width;        ///< width of the image in the input stream
      int image_height;       ///< height of the image in the input stream
      int color;              ///< colorspace of the image (YVA12 or RGBA32)
      int yuv_size;           ///< size of the YUV Image
      
      /**
       * Loads and prepares the PNG image.
       */
      Result preparePNGImage(const char *filename);
      
      /**
       * converts the PNG image to YV12-Colorspace
       */
      void convertPNGImageYUV();
      
      /**
       * converts a single pixel from rgba32 to yva12
       */
      void rgb2yuv(unsigned char, unsigned char, unsigned char, double*, double*, double*);
      
      /**
       * Draws the Image onto the background image for Colorspace YV12
       */
      void paintPNGImageAlphaYVA12(char *image, int x, int y);
      
    public:
      OSDImagePNG();
      virtual ~OSDImagePNG();
      Result load(const string url);
      virtual Result draw(char *data, int color, size_t width, size_t height);

      /**
       * Returns the YUV Version of the Image
       */
      virtual unsigned char* getYUVData();

      /**
       * Returns the size of the YUV Version of the Image
       */
      virtual int getYUVSize();
    };
    
    class OSDText : public OSDItem {
    private:
      char *current_data;
      
      int image_width;         
      int image_height;       
      int color;
      
      font_desc_t* osd_font;
      char* osd_text;
      int osd_progbar_type;
      int osd_progbar_value;   // 0..255
      int dxs;
      int dys;

      void draw_alpha(int x0,int y0, int w,int h, unsigned char* src, unsigned char *srca, int stride);
      void draw_text_progbar();
      void draw_text_osd();
      void draw_text();

    public:
      OSDText();
      virtual ~OSDText();
      virtual Result draw(char *data, int color, size_t width, size_t height);
      virtual unsigned char* getYUVData() { return 0; };
      /**
       * Returns the size of the YUV Version of the Image
       */
      virtual int getYUVSize() { return 0; };
      Result setText(const string text);
      Result setFont( font_desc_t* osd_font );
      Result setProgressBarValue(const int value ) ;
    };


  private:// Class OSDOverlayNode

    Buffer *lastBuffer;
    bool   paused;

    /**
     * sends a new Buffer containing all Changes since last update
     * ??? Still needed for filter ???
     * can be called via event "update"
     * private to avoid anyone from calling it directly
     */
    Result update();
    
  protected:
    
    typedef map<string, OSDItem *> items_t;
    items_t items; ///< all items to be overlayed

    font_desc_t* font; ///< font for overlayed text

    ThreadMutex osd_mutex;
  
    int image_width;        ///< width of the image in the input stream
    int image_height;       ///< height of the image in the input stream
    int color;              ///< colorspace: RGBA32 or YVA12

    Result doInit();
    Result doInitOutput();
    /**
     * The standard doActivate method from Node. Effects in OSDOverlayNode:
     * obtain and check the size of the incoming buffer image.
     */
    Result doActivate();

    /**
     * The standard processBuffer method from GenericProcessorNode.
     * Effects in OSDOverlayNode: display the items (text, images)
     * on the image in the buffer.
     */
    Buffer *processBuffer(Buffer *in_buffer);

    /**
     *  Handler to auto-clear the osd
     */
    static void SIGALRM_handler(int signum);
    static OSDOverlayNode* osd_node;

  public:

    OSDOverlayNode(const char *name = "OSDOverlayNode",
	     const StreamQueue::Mode queue_mode = StreamQueue::MODE_SUSPEND, 
	     const int queue_size = 1 );
    virtual ~OSDOverlayNode();
    
    Result addText(const string& id, const string& text, const size_t& x_offset, const size_t& y_offset);
    Result printText(const string& id, const string& text, const size_t& x_offset, const size_t& y_offset, const int& seconds);
    Result removeText(const string& id);
    Result addProgessBar(const string& id, const int& value, const int& x_offset, const int& y_offset);
    Result addImage(const string& id, const string& url, const int& x_offset, const int& y_offset);
    Result setItemOffset(const string& id, const int& x_offset, const int& y_offset);
    Result setItemText(const string& id, const string& text);
    Result setItemEnabled(const string& ,const bool&);
    Result disableItems();
    Result setFont( const string& font_name, const float& font_factor );
    Result doPause(const bool&);
    
    Result setFormat(const Format& format);
  };

}

#endif //_NMM_VIDEO_FILTER_OSDOVERLAYNODE_HPP_
