/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, 
 * USA.
 */


#include "RTJPEGDecodeNode.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/registry/Plugin.hpp"

NMM_REGISTER_PLUGIN(NMM::RTJPEGDecodeNode, "RTJPEGDecodeNode")

namespace NMM {

  RTJPEGDecodeNode::RTJPEGDecodeNode(const char* name,
				     const StreamQueue::Mode queue_mode,
				     const int queue_size) :
    GenericProcessorNode(name, queue_mode, queue_size )
  {
    // Setup benchmark- migrated the node
    benchmark.setNamedObject(this);
    benchmark.setEnabled(true);

    quality = 80;
    quality_size = sizeof(unsigned char);

    mode = IFrameEncodeMode::INTRA_FRAME;
    mode_size = sizeof(IFrameEncodeMode::Mode);

    rtjpeg_struct = new RTjpeg_struct();
    rtjpeg_tables = 0;

    rtjpeg_tables = new long unsigned int[128];
    rtjpeg_tables_size = 128 * sizeof(long unsigned int);

    first_time = true;

    inter_frame_buffer = 0; //
  }

  RTJPEGDecodeNode::~RTJPEGDecodeNode() {
    delete rtjpeg_struct;
    delete rtjpeg_tables;
    delete inter_frame_buffer;
  }
  
  Result RTJPEGDecodeNode::doInit() {
    NMM::Format *myInputFormat;
    myInputFormat=getOwnInputProperty()->addNewFormat("video/rtjpeg");
    myInputFormat->addWildcard(Format::x_resolution_param);
    myInputFormat->addWildcard(Format::y_resolution_param);
    myInputFormat->addWildcard(Format::ratio_x_param);
    myInputFormat->addWildcard(Format::ratio_y_param);
    myInputFormat->addStringParamValue(Format::colorspace_param,"yuv422");
    myInputFormat->addStringParamValue(Format::channelorder_param,"yuv");
    myInputFormat->addIntParamValue(Format::bitperpixel_param,16);
    myInputFormat->addStringParamValue(Format::endian_param, "little");
    myInputFormat->addWildcard(Format::framerate_param);
    
    NMM::Format *myOutputFormat;
    myOutputFormat=getOwnOutputProperty()->addNewFormat("video/raw");
    myOutputFormat->addWildcard(Format::x_resolution_param);
    myOutputFormat->addWildcard(Format::y_resolution_param);
    myOutputFormat->addWildcard(Format::ratio_x_param);
    myOutputFormat->addWildcard(Format::ratio_y_param);
    myOutputFormat->addStringParamValue(Format::colorspace_param,"yuv422");
//    myOutputFormat->addStringParamValue(Format::colorspace_param,"yuv420"); //according to the RTjpeg manual the format is YUV420P
    myOutputFormat->addStringParamValue(Format::channelorder_param,"yuv");
    myOutputFormat->addIntParamValue(Format::bitperpixel_param,16);
    myOutputFormat->addStringParamValue(Format::endian_param, "little");
    myOutputFormat->addWildcard(Format::framerate_param);
    myOutputFormat->addStringParamValue(Format::format_param, "planar");

    myInputFormat->setIOPartner(myOutputFormat);
    myOutputFormat->setIOPartner(myInputFormat);
    
    return SUCCESS;
  }
  
  Result RTJPEGDecodeNode::doInitOutput() {
    const Format *in_f = getInputFormat();
    
    if (in_f == NULL) {
      ERROR_STREAM( "Format is null" << endl);
      return FAILURE; // error, not without a format!
    }
    
    int x_resolution = in_f->getIntValue(Format::x_resolution_param);
    int y_resolution = in_f->getIntValue(Format::y_resolution_param);
    int ratio_x = in_f->getIntValue(Format::ratio_x_param);
    int ratio_y = in_f->getIntValue(Format::ratio_y_param);
    float framerate = in_f->getFloatValue(Format::framerate_param);
    
    Format* i_format = new Format( *in_f );
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    getOwnWorkingInputProperty()->setDefaultFormat(i_format);
    
    Format* o_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
    o_format->addIntParamValue(Format::x_resolution_param, x_resolution);
    o_format->addIntParamValue(Format::y_resolution_param, y_resolution);
    o_format->addIntParamValue(Format::ratio_x_param, ratio_x);
    o_format->addIntParamValue(Format::ratio_y_param, ratio_y);
    o_format->addStringParamValue(Format::colorspace_param,"yuv422");
    o_format->addStringParamValue(Format::channelorder_param,"yuv");
    o_format->addIntParamValue(Format::bitperpixel_param,16);
    o_format->addStringParamValue(Format::endian_param, "little");
    o_format->addFloatParamValue(Format::framerate_param, framerate);
    o_format->addStringParamValue(Format::format_param, "planar");
    getOwnWorkingOutputProperty()->setDefaultFormat(o_format);

    i_format->setIOPartner(o_format);
    o_format->setIOPartner(i_format);

    return SUCCESS;
  }
  
  Result RTJPEGDecodeNode::doActivate() {
    const Format *in_f = getInputFormat();
    
    if (in_f == NULL) {
      ERROR_STREAM( "Format is Null" << endl);
      return FAILURE; // error, not without a format!
    }

    image_width = in_f->getIntValue(Format::x_resolution_param);
    image_height = in_f->getIntValue(Format::y_resolution_param);

    if ((image_width <= 0) || (image_height <= 0)) {
      ERROR_STREAM( "image size is wrong" << endl);
      return FAILURE; // error!
    }
    
    if(mode == IFrameEncodeMode::INTER_FRAME) {
      inter_frame_buffer = new char[image_width * image_height * 2];
      memset(inter_frame_buffer, 0, image_width * image_height * 2);
    }

    return SUCCESS; 
  }

  Buffer* RTJPEGDecodeNode::processBuffer(Buffer *in_buffer) {
    if (in_buffer == 0)
      return 0;

    Buffer *out_buffer = getNewBuffer(image_width*image_height*2, in_buffer); 
 
    if (out_buffer) {

      // if rtjpeg_tables are not yet set OR quality has changed OR mode has changed
      // copy values from incoming buffer and init rtjpeg_struct
      if(first_time || quality != *(in_buffer->getData()) || mode != *(in_buffer->getData() + quality_size)) {
	DEBUG_STREAM( "got quality first time" << endl);
	memcpy(&quality     , in_buffer->getData()                           , quality_size);
	memcpy(&mode        , in_buffer->getData() + quality_size            , mode_size);
	memcpy(rtjpeg_tables, in_buffer->getData() + quality_size + mode_size, rtjpeg_tables_size);
	RTjpeg_init_decompress(rtjpeg_tables, image_width, image_height, rtjpeg_struct);
	if(mode == IFrameEncodeMode::INTER_FRAME) {
	  // if inter_frame_buffer is not yet allocated
	  if(inter_frame_buffer == 0) {
	    inter_frame_buffer = new char[image_width * image_height * 2];
	  }
	  // must always be set to zeros because of rtjpeg inter frame encoding
	  memset(inter_frame_buffer, 0, image_width * image_height * 2);
	  RTjpeg_init_mcompress(rtjpeg_struct);
	}
	first_time = false;
      }
      
      if(mode == IFrameEncodeMode::INTER_FRAME) {
	DEBUG_STREAM( "INTER_FRAME_COMPRESSION Dec" << endl);
	
	RTjpeg_decompressYUV422((__rtjpeg__s8 *) in_buffer->getData() + quality_size + mode_size + rtjpeg_tables_size, 
				(__rtjpeg__u8 *) inter_frame_buffer, rtjpeg_struct);

        memcpy(out_buffer->getData(), inter_frame_buffer, image_width*image_height*2);
      }
      else {
	RTjpeg_decompressYUV422((__rtjpeg__s8 *) in_buffer->getData() + quality_size + mode_size + rtjpeg_tables_size, 
				(__rtjpeg__u8 *) out_buffer->getData(), rtjpeg_struct);
      }

      out_buffer->setUsed(image_width*image_height*2);
      benchmark.benchmark(out_buffer->getUsed());

      in_buffer->release();
      return out_buffer; 
    }
    else {
      in_buffer->release();
      DEBUG_STREAM( "Could not allocate out Buffer!" << endl);
      return 0;
    }
  }
  
  void RTJPEGDecodeNode::setQuality(const float&) {
  }
  
  float RTJPEGDecodeNode::getQuality() {
    return (float) quality / 255.0f;
  }
  
  void RTJPEGDecodeNode::setBitrate(const int&) {
    return;
  }
  
  int RTJPEGDecodeNode::getBitrate() {
    return -1;
  }
  
  void RTJPEGDecodeNode::setFrameEncodeMode(const IFrameEncodeMode::Mode&) {
  }
  
  IFrameEncodeMode::Mode RTJPEGDecodeNode::getFrameEncodeMode() const {
    return mode;
  }
}

