/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:	       
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "SPUDecodeNode.hpp"
#include "nmm/interfaces/video/filter/IOverlay.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/registry/Plugin.hpp"
#include "nmm/base/memorymgr/Buffer.hpp"

NMM_REGISTER_PLUGIN(NMM::SPUDecodeNode, "SPUDecodeNode")

namespace NMM {
  
  SPUDecodeNode::SPUDecodeNode( const char * name, StreamQueue::Mode qmode, int qsize)
    : GenericProcessorNode(name, qmode, qsize) {
    
    decoder = new SPUDecoder();
    decoder->setMode( SPUDecoder::YV12A );
    
    menu_decoder = new SPUDecoder();
    menu_decoder->setMode( SPUDecoder::YV12A );
    
    /* create a default color palette */
    unsigned int YUV_palette[16] = { 0x00d19210, 0x00902236, 0x0051ef5a ,0x00902236,
				     0x00286def, 0x00eaef80, 0x00f01010, 0x00f020ef,
                                     0x00ff2010, 0x00ff80ef, 0x00f0ef20, 0x00f0ef20,
                                     0x00a910a5, 0x00bfef80, 0x00f05010, 0x00286def };
    
    for( unsigned int i=0; i<16; i++ )
      std_palette[i] = YUV_palette[i];
    
    /* set color palette */
    setPalette( std_palette );
  };
  
  SPUDecodeNode::~SPUDecodeNode() {
    delete decoder;
    delete menu_decoder;
  }
  
  void SPUDecodeNode::setPalette( unsigned int* pal ) {
    palette = pal;
  }

  Result SPUDecodeNode::setPaletteItem( const int& index,
                                        const int& color_y, const int& color_u, const int& color_v, const int& alpha ) {
    
    if ( index < 16 ) {
      
      // endian-safe conversion from color channels to integer
      union {
        u_int32_t color;
        unsigned char color_field[4];
      } color_value;
      
      color_value.color_field[0] = color_y;
      color_value.color_field[1] = color_u;
      color_value.color_field[2] = color_v;
      color_value.color_field[3] = alpha;
      
      palette[index] = color_value.color;
    }
    
    return SUCCESS;
  }
  
  Result SPUDecodeNode::initColorScheme( const int& scheme_nr, const int& selection, const int& activation ) {
    
    DEBUG_STREAM( "Initializing color scheme nr: " << scheme_nr << endl );
    menu_decoder->initColorScheme( scheme_nr, 0, selection );
    menu_decoder->initColorScheme( scheme_nr, 1, activation );
    
    return SUCCESS;
  }
  
  Result SPUDecodeNode::setDVDSelectedButton( const int& b_nr ) {
    MutexGuard mg_em(event_mutex);
    
    if  ( have_menu ) {
      selected_button  = b_nr;
      button_state     = 0;
      send_button_info = true;
      
      saved_working_flag = getWorkingFlag();
      setWorkingFlag( true );
    }
    
    return SUCCESS;
  }
  
  Result SPUDecodeNode::setDVDButtonCount( const int& ) {
    button_index = 0;
    scheme_index = 0;
    
    return SUCCESS;
  }
  
  Result SPUDecodeNode::setDVDButtonCoords( const int& x_start, const int& y_start, const int& x_end, const int& y_end ) {
    button[button_index].x_start = x_start;
    button[button_index].y_start = y_start;
    button[button_index].x_end   = x_end;
    button[button_index].y_end   = y_end;
    
    button_index++;
    
    return SUCCESS;
  }
  
  Result SPUDecodeNode::setDVDButtonScheme( const int& scheme ) {
    DEBUG_STREAM( "Setting button color scheme: " << scheme << endl );
    button[scheme_index].scheme = scheme;
    
    scheme_index++;
    
    return SUCCESS;
  }
  
  Result SPUDecodeNode::doActivate() {
    return SUCCESS;
  }
  
  Result SPUDecodeNode::doInit() {
    DEBUG_STREAM("Initializing spu decoder..." << endl);

    offset = 0;
    width  = 0;
    height = 0;
    memset( menu, 0, sizeof(menu) );
    send_menu = false;
    send_button_info = false;
    selected_button = 0;
    button_state = 0;
    send_lastbuffer = false;
    
    // init input property
    Format* i_format = getOwnInputProperty()->addNewFormat("video/spu");
    getOwnInputProperty()->setDefaultFormat(i_format);
    
    // init output property
    Format* o_format = getOwnOutputProperty()->addNewFormat("video/raw");
    o_format->addWildcard(Format::x_resolution_param);
    o_format->addWildcard(Format::y_resolution_param);
    o_format->addWildcard(Format::ratio_x_param);
    o_format->addWildcard(Format::ratio_y_param);
    o_format->addWildcard(Format::framerate_param);
    o_format->addStringParamValue(Format::colorspace_param,"yv12a");
    o_format->addStringParamValue(Format::channelorder_param,"yvua");
    o_format->addIntParamValue(Format::bitperpixel_param, 20);
    o_format->addStringParamValue(Format::endian_param, "little");
    o_format->addStringParamValue(Format::format_param, "planar");
    getOwnOutputProperty()->setDefaultFormat(o_format);
    
    o_format->setIOPartner(i_format);
    i_format->setIOPartner(o_format);
    
    setWorkingFlag(false);
    
    return SUCCESS;
  }
  
  Result SPUDecodeNode::doDeinit() {
    DEBUG_STREAM("exiting spu decoder..." << endl);
    
    return SUCCESS;
  }
  
  Result SPUDecodeNode::doInitOutput() {
    const Format* in_format = getInputFormat();
    if (in_format == NULL) { return FAILURE; }
    
    Format* i_format = new Format( *in_format );
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    getOwnWorkingInputProperty()->setDefaultFormat(i_format);
    
    Format* o_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
    o_format->addIntParamValue(Format::x_resolution_param, 0);
    o_format->addIntParamValue(Format::y_resolution_param, 0);
    o_format->addIntParamValue(Format::ratio_x_param, 1);
    o_format->addIntParamValue(Format::ratio_y_param, 1);
    o_format->addFloatParamValue(Format::framerate_param, 0.0f);
    o_format->addStringParamValue(Format::colorspace_param,"yv12a");
    o_format->addStringParamValue(Format::channelorder_param,"yvua");
    o_format->addIntParamValue(Format::bitperpixel_param, 20);
    o_format->addStringParamValue(Format::endian_param, "little");
    o_format->addStringParamValue(Format::format_param, "planar");
    getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
    
    o_format->setIOPartner(i_format);
    i_format->setIOPartner(o_format);
    
    return SUCCESS;
  }
  
  Message* SPUDecodeNode::processBuffer( Buffer *in_buffer ) {
    MutexGuard mg_em(event_mutex);
    
    if ( send_button_info ) {
      send_button_info = false;
      
      menu_decoder->decodeCS( menu, 0x10000 );
      if ( menu_decoder->isCSOk() ) {
        CEvent* event = new CEvent();
        Event event1( IOverlay::setForeground_event, 4 );
        event1.setValue( 0, new TInValue<int>(menu_decoder->getXPosition()) );
        event1.setValue( 1, new TInValue<int>(menu_decoder->getYPosition()) );
        event1.setValue( 2, new TInValue<int>(menu_decoder->getWidth()) );
        event1.setValue( 3, new TInValue<int>(menu_decoder->getHeight()) );
        event->insertEvent( 0, event1 );
        
        Event event2( IOverlay::setForegroundSegment_event, 4 );
        event2.setValue( 0, new TInValue<int>(button[selected_button].x_start) );
        event2.setValue( 1, new TInValue<int>(button[selected_button].y_start) );
        event2.setValue( 2, new TInValue<int>(button[selected_button].x_end - button[selected_button].x_start));
        event2.setValue( 3, new TInValue<int>(button[selected_button].y_end - button[selected_button].y_start));
        event->insertEvent( 1, event2 );
        
        send_menu = true;
        return( event );
      }
      
      setWorkingFlag( saved_working_flag );
      return( 0 );
    }
    
    if ( send_menu ) {
      DEBUG_STREAM( "Processbuffer send_menu" << endl );
      send_menu = false;
      
      unsigned int dim = menu_decoder->getWidth() * menu_decoder->getHeight();
      unsigned int isize = (dim << 1) + ((dim+1) >> 1); // fw: don't use float ops on byte sizes. was (unsigned int) (dim*2.5);
      Buffer* menu_out = getNewBuffer( isize );
      menu_out->setUsed( isize );
      menu_decoder->setColorScheme( button[selected_button].scheme, button_state );
      menu_decoder->decodeBitmap( menu, palette, (unsigned char*) menu_out->getData() );
      
      setWorkingFlag( saved_working_flag );
      return( menu_out );
    }
    
    if ( send_lastbuffer ) {
      setWorkingFlag( w_flag );
      send_lastbuffer = false;
      
      DEBUG_STREAM( "sending picture (last buffer)" << endl );
      return( out );
    }
    
    if ( !getWorkingFlag() ) {
       old_in_buffer = in_buffer;

       data = (unsigned char*) in_buffer->getData();
       size = in_buffer->getUsed();
       memcpy( temp+offset, data, size );
       size += offset;
    }
    
    int ret = decoder->decodeCS( temp, size );
    if ( ret < 0 ) {
      offset = size;
      old_in_buffer->release();
      setWorkingFlag( false );
      return( 0 );
    }
    
    if ( decoder->isCSOk() ) {
      if ( decoder->isMenu() ) {
        memcpy( menu, temp, sizeof( menu ) );
        have_menu = true;
      }

      // XXX: check if user would have subtitle
      //if ( decoder->isMenu() ) {
      unsigned int dim = decoder->getWidth() * decoder->getHeight();
      unsigned int isize = (dim << 1) + ((dim+1) >> 1); // fw: don't use float ops on byte sizes. was (unsigned int) (dim*2.5);
      out = getNewBuffer( isize );
      out->setUsed( isize );
      unsigned char* image = (unsigned char*) out->getData();

      decoder->decodeBitmap( temp, palette, image );
      //} else out = NULL;
    }
    
    if ( ret == 0 ) {
      offset = 0;
      old_in_buffer->release();
      setWorkingFlag( false );
    } else
    if ( ret > 0 ) {
      offset = 0;
      memcpy( &temp, data+(size-ret), ret );
      data = temp;
      size = ret;
      setWorkingFlag( true );
    };

    if ( (width != decoder->getWidth()) || (height != decoder->getHeight()) ||
         (x_offset != decoder->getXPosition()) || (y_offset != decoder->getYPosition()) ) {
      width    = decoder->getWidth();
      height   = decoder->getHeight();
      x_offset = decoder->getXPosition();
      y_offset = decoder->getYPosition();
      send_lastbuffer = true;
      w_flag = getWorkingFlag();
      setWorkingFlag( true );
      
      Event event( IOverlay::setForeground_event, 4 );
      event.setValue( 0, new TInValue<int>(x_offset) );
      event.setValue( 1, new TInValue<int>(y_offset) );
      event.setValue( 2, new TInValue<int>(width) );
      event.setValue( 3, new TInValue<int>(height) );
      return new CEvent( event );
    }
    
    DEBUG_STREAM( "sending picture" << endl );
    return( out );
  }
}
