/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_AVERAGE_HPP_
#define _NMM_AVERAGE_HPP_

#include <vector>

namespace NMM {
  
  /**
   *  @brief Average computation template
   *
   *  This class computes an average and variance over a sequence of values of type @c T.
   */
  template<typename T>
  class TAverage {
  public:
    /**
     *  @brief Default constructor
     *
     *  Set @a n to 10 and reset all counters. The acverage is undefined.
     */
    TAverage();
    
    /**
     *  @brief Set averaging interval @a n
     *  @param n The new value for @a n
     */
    void setAverageInterval(unsigned int n);
    
    /**
     *  @brief Put a value into the accumulator
     *  @param value Value to add
     *
     *  This method adds @c value to the accumulator and increments the averaging count. If the counter
     *  reaches the averaging interval @a n, an average is computed over the last @a n values. If this method
     *  is called for the first time, the average is set to @c value. The average is always defined when this
     *  method returns.
     */
    void put(const T& value);
    
    /**
     *  @brief Get the current average value
     *  @return The average, if available
     *
     *  The method may only be called if the average is available, i.e. put has been called at least once.
     */
    const T& get() const;

    /**
     *  @brief Get the current variance
     *  @return The variance, if available
     *
     *  The method may only be called if the variance is available, i.e. put has been called at least once.
     */    
    const T& getVariance() const;

    /**
     *  @brief Get the maximum variance
     *  @return The maximum variance, if available
     *
     *  The method may only be called if the maximum variance is available, i.e. put has been called at least once.
     */    
    const T& getMaxVariance() const;
    
  private:
    bool __initial;             ///< Average not set yet
    unsigned int __count;       ///< Number of values added to accumulator
    unsigned int __n;           ///< Number of values over which to compute average
    T __average;                ///< Current average value
    T __variance;               ///< Variance
    T __max_variance;           ///< Maximum variance
    std::vector<T> __samples;   ///< Samples
  };

  template<typename T>
  inline TAverage<T>::TAverage()
    : __initial(true),
      __count(0),
      __n(10),
      __average(T(0)),
      __samples(10)
  {}
  
  template<typename T>
  inline void TAverage<T>::setAverageInterval(unsigned int n) {
    __n = n;
    __samples.resize(n);
  }
  
  template<typename T>
  inline void TAverage<T>::put(const T& value) {
    __count += 1;
    if(__initial) {
      /* set first value as initial average value */
      __average = value;
      __variance = 0;
      __initial = false;
    }
    if(__count >= __n) {
      /* compute average and reset accumulator and counter */
      T acc(0);
      for(unsigned int i=0; i<__count; ++i) {
        acc = acc + __samples[i];
      }
      __average = acc;
      __average /= __count;
      T var(0);
      T max_var(0);
      for(unsigned int i=0; i<__count; ++i) {
        T diff = __samples[i] - __average;
        if(diff < T(0)) {
          diff = -diff;
        }
        var = var + diff;
        max_var = max(var, max_var);
      }
      __variance = var;
      __max_variance = max_var;
      __variance /= __count;
      __count = 0;
    }
    __samples[__count] = value;
  }
  
  template<typename T>
  inline const T& TAverage<T>::get() const {
    return __average;
  }
  
  template<typename T>
  inline const T& TAverage<T>::getVariance() const {
    return __variance;
  }

  template<typename T>
  inline const T& TAverage<T>::getMaxVariance() const {
    return __max_variance;
  }
  
}

#endif
