/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *                    Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef NMM_INETADDR_HPP
#define NMM_INETADDR_HPP NMM_INETADDR_HPP

#include "nmm/base/serialize/ValueStream.hpp"
#include "nmm/base/serialize/ComplexValue.hpp"

#include <sys/socket.h>
#include <netinet/ip.h>

namespace NMM {

  /** A C++ Wrapper for an internet address.
   * This class avoids some standard error with network addresses. So the IP-address and the port are
   * automatically converted from host byte-order to network byte-order.
   */
  class INetAddr : public SerializedValue {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Default Constructors & Destructor
     * Initialize the IP-Address and port to 0
     */
    INetAddr();

    /** Constructor.
     * @param addr set the IP-Address as long to addr. The IP-Address is automatically converted from byte-odrder to host-order if need.
     * @param port set the address-port to port. The port is automatically converted from byte-odrder to host-order if need.
     */
    INetAddr(unsigned long addr, unsigned int port);

    /** Constructor.
     * @param addr set the IP-Address from a valid hostname to addr(the char* can be a valid IP-Address too). The IP-Address is automatically converted from byte-odrder to host-order if need.
     * @param port set the address-port to port. The port is automatically converted from byte-odrder to host-order if need.
     */
    INetAddr(const char* addr, unsigned int port);
    
    /** Copy-Constructor.
     * @param addr Creates a copy from a given INetAddr.
     */
    INetAddr(const INetAddr& addr);
    
    /** Destroys an INetAddr.
     */
    ~INetAddr();
    //@}
    
    /** @name Get Methods
     */
    //@{
    /** virtual Copy-Constructor
     */
    virtual INetAddr*    copy()         const { return new INetAddr(*this); }

    /** This method returns the current port.
     * @return Returns the port in host-order. You must not convert it again!
     */
    inline unsigned int  getPort()      const { return ntohs(_address.sin_port); }

    /** This method returns the internal stored sockaddr. This method is for compatibility reason to use this class with C-Code
     * that uses the C-structs.
     * @return Returns internal C-Struct sockaddr.
     */
    inline sockaddr*     getAddress()   const { return reinterpret_cast<sockaddr*> (&_address); }

    /** This method returns the size of the internal sock_address.
     * This method is for compatibility reason to use this class with C-Code that uses the C-structs.
     * @return Returns the size of the sock_address
     */
    inline size_t        getSize()      const { return sizeof (_address); }

    /** This method returns the current IP-Address in host byte-order.
     * @return Returns the current IP-Address in host byte-order
     */
    inline unsigned long getIpAddress() const { return ntohl(_address.sin_addr.s_addr); }

    /** This method returns the transport protocol which can be eithe "TCP" or "UDP"
     * @return Returns the current transport protocol
     */
    inline const string& getTransportProtocol() const { return _trans_prot.getRValue(); }
    
    /** Convert address to human readable string of the form ip_address:port
     */
    string toString() const;
    //@}

    /** @name Set Methods
     */
    //@{
    /** Sets the port for the INetAddr. This method converts the port from host byte-order to network byte-order.
     * @param port Expects the port in host byte-order
     */
    inline void          setPort(unsigned int port)   { _address.sin_port = htons(port); _port = port; }

    /** Sets the transport protocol which can be either "tcp" or "udp"
     * @param protocol The Prototcol for the INetAddr.
     */
    inline void          setTransportProtocol(const string& protocol) { _trans_prot = protocol; }

    /** Sets a hostname to INetAddr. The hostname automatically converted to an IP-Address.
     * @param addr The new hostname for this INetAddr.
     */
    void          setAddress(const char* addr)  ;

    /** Sets the IP-Address for the INetAddr. This method converts the IP-Address from host byte-order to network byte-order.
     * @param addr The given IP-Address from host byte-order to network byte-order.
     */
    void          setAddress(unsigned long addr);

    /** Sets the an C-struct addr. All internal parameters are copied from the addr-struct.
     * @param addr The valid C-struct sockaddr_in.
     */
    void          setAddress(struct sockaddr_in* addr);
    //@}
    
    friend std::ostream& operator<<(std::ostream& ostr, INetAddr& addr) {
      return ostr << addr.toString();
    }
    
  protected:
    /** Prepare all members for serialization
     */
    void updateAttributes();
  private:
    /** Register all members by the base class SerializedValue.
     */
    void registerAttributes();
  private:
    mutable sockaddr_in     _address;           ///< The internal C-Struct soccaddr_in.
    TValue<string>          _trans_prot;        ///< Stores the Transport-protocol, which can be "TCP" or "UDP"
    TValue<unsigned int>    _port;              ///< Stores the port of the INetAddr
    TValue<unsigned long >  _ip_address;        ///< Stores the IP-Address of the INetAddr
  };

}

#endif
