/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */



#ifndef NMM_NETREACTOR_HPP
#define NMM_NETREACTOR_HPP

#include "nmm/utils/thread/ThreadedObject.hpp"
#include "nmm/utils/thread/WorkerThread.hpp"
#include "nmm/utils/Singleton.hpp"
#include "nmm/utils/NamedObject.hpp"

#include <sys/poll.h>
#include <vector>
#include <map>

namespace NMM {

class NetReactor;

/** \class NetHandler
  * This is the abstract base class for all NetHandler. The NetReactor calls the receiveEvent method
  * with the events occured on the socket.
  */
class NetHandler {
public:
  enum NetResult { VOID, REMOVE };
/* Destructor */
  virtual ~NetHandler(){};
  
/** @name Handler methods
  */
//@{
/** This method is called by the NetReactor and must be implement by all sub classes
  * \param socket The socket on which an event occurs.
  * \param event The events which have been registered.
  * \param revent The events which have been occured on the socket
  */
  virtual NetResult receiveEvent(int socket, short events, short revents) = 0;

/** This method returns true, if the NetHandler requires an own thread. 
  * \return true it the NetHandler requires an own thread, otherwise false
  */
  virtual bool requireThread() const = 0;
//@}
};

/** \class NetHandler
  * This class dispatches incoming events from the NetReactor and maps them to a method that
  * can be implmeneted by a sub class.
  */
class DispatchedNetHandler : public NetHandler {
public:
/** @name Handler methods
  */
//@{
/** This method dispatches the events and maps them to one of the following method.
  * \param socket The socket on which an event occurs.
  * \param event The events which were occured on the socket.
  */
  virtual NetResult receiveEvent(int socket, short event, short revents);

/** This method is calle if the event POLLIN, POLLRDNORM, POLLRDBANd, POLLPRI occurs, which means
  * that normal or priority Data can be read.
  * \param event The corresponding socket.
  */
  virtual NetResult receiveData(int)        { return VOID; };

/** This method is called if the event POLLOUT, POLLWRNORM or POLLWRBAND occurs, which means that
  * normal or priority Data can be send.
  * \param event The corresponding socket.
  */
  virtual NetResult writeData(int)          { return VOID; };

/** This method is called if the event POLLHUP occurs.
  * \param event The corresponding socket.
  */
  virtual NetResult closeConnection(int)    { return VOID; };

/** This method is called if the event POLLINVAL occurs.
  * \param event The corresponding socket.
  */
  virtual NetResult invalidSocket(int)      { return VOID; };

/** This method is called if the event POLLERR occurs.
  * \param event The corresponding socket.
  */
  virtual NetResult errorOccured(int)       { return VOID; };

/** This method returns true, if the NetHandler requires an own thread.
  * \return true it the NetHandler requires an own thread, otherwise false
  */
  virtual bool requireThread()     const    { return false; }
//@}
protected:
  NetResult handleEvents(int socket, short events);
};

class ThreadedNetHandler : public DispatchedNetHandler {
public:
/** @name Constructor
  */
//@{
  ThreadedNetHandler();
//@}

  virtual NetResult receiveEvent(int socket, short event, short revents);
  virtual bool requireThread()     const ;
protected:
  void disableNotification();
private:
  NetReactor&       __net_reactor;                // address to the global NetReactor
  bool              __notification;        
  ThreadMutex       __event_mutex;
  ThreadCondition   __event_condition;
};

/** \class NetReactor
  * The NetReactor is used to wait for incoming network events and to notify registered NetHandler
  * for a specific network action like a closed connection or incomming data. The events are defined
  * in an enum to allow a unified use on different systems. The NetReactor uses the poll-method to
  * wait for changes on registered sockets and uses therfore a own thread.
  *
  * This class is designed as a singleton.
  */
class NetReactor : public NamedObject {
/** The CreateUsingNew is friend to call the constructor */
friend class CreateUsingNew<NetReactor>;
public:
/** Same values as poll events. includes also the values defined in asm/poll.h
  */
  enum PollEvent {NR_POLLIN  = POLLIN,   NR_POLLPRI = POLLPRI, NR_POLLOUT  = POLLOUT,
                  NR_POLLERR  = POLLERR, NR_POLLHUP = POLLHUP, NR_POLLNVAL = POLLNVAL,
                  NR_POLLRDNORM = 0x0040, NR_POLLRDBAND = 0x0080, NR_POLLWRNORM = 0x0100,
                  NR_POLLWRBAND = 0x0100, NR_POLLMSG    = 0x0400};
/** @name Registration methods
  */
//@{
/** This method adds a new socket and a corresponding NetHandler wich is informed about incoming events.
  * Every NetHandler is informed about errors on the given socket.
  * \param socket The new socket.
  * \param handler The NetHandler which is informed about incoming events.
  * \param events The interesting events for the sockets. Several events can combined with or-operator(operator|()).
  */
  void registerSocket(int socket, NetHandler* handler, short events = NR_POLLIN|NR_POLLERR);
  void internalRegisterSocket(int socket, NetHandler* handler, short events = NR_POLLIN|NR_POLLERR);

/** This method changes the events a NetHandler is listening too.
  * \param socket The socket to change events
  * \param events The new events for the socket. Several events can combined with or-operator(operator|()).
  */
  void changeEvents(int socket, short events);

/** This method removes a registered socket and its corresponding NetHandler. The NetHandler is not destroyed.
  * \param socket The socket that is removed.
  */
  void unregisterSocket(int socket);
  void internalUnregisterSocket(int socket);

/** This method sets the timeout used by the poll()-method in milliseconds. The default value is 100 ms.
  * \param timeout The new timeout for the poll()-method
  */
  void setTimeout(int timeout);
//@}

  void requestLeaderThread();

protected:
/** The run method is used to call poll on registred sockets
  */
  virtual void run();

private:
/** @name Constructor & Destructor
  */
//@{
/** Default Constructor.
  */
  NetReactor() ;

/** Disabled Copy Constructor
  */
  NetReactor(const NetReactor& );

/** Destructor
  */
  ~NetReactor();
//@}

/** This methos checks the registered sockets. If this mehtod returns true, the thread can not be reused.
  * by the NetReactor, because another thread was used as leader.
  */
  bool checkSockets();

/** Returns an error as string for an errn occured.
  * \param error The errno-value
  */
  const string& getError(int error);

/** Test, if a specific socket is registered
  */
  bool isSocketRegisterd(int socket);
private:

  void internalLock();
  void internalUnlock();
  //private class
  class InternalNetHandler : public NetHandler {
  public:
    InternalNetHandler(NetReactor&);
    virtual NetResult receiveEvent(int socket, short event, short revents);
    /** This method returns true, if the NetHandler requires an own thread.
      * \return true it the NetHandler requires an own thread, otherwise false
      */
    virtual bool requireThread() const { return false; }
  private:
    int         __receive_signal;
    char        __internal_char;
    NetReactor& __reactor;
  };

  friend class InternalNetHandler;

private: //member
  vector<pollfd>            __socket_array;         ///< stores all registered sockets in a pollfd-struct
  map<int, NetHandler*>     __handler;              ///< stores all registered sockets with the corresponding NetHandler
  ThreadMutex               __handler_mutex;        ///< Mutex for the internal handler
  ThreadMutex               __handler_cond_mutex;   ///< Mutex for the internal handler
  int                       __timeout;              ///< Timeout used for the poll() method
  vector<int>               __pipe_socket;          ///< sockets for the pipe
  char                      __internal_char;        ///< Dummy byte which is written in the pipe to interrupt poll
  InternalNetHandler        __internal_handler;     ///< Internal handler to interrupt the poll method
  TTask<NetReactor>         __reactor_task;
  WorkerThread*             __leader_thread;

  static bool               __is_running;           ///<  running flag for the thread
  
  static const string EBADF_error;                  ///< static error string for errno EBADF
  static const string ENOMEM_error;                 ///< static error string for errno ENOMEM
  static const string EFAULT_error;                 ///< static error string for errno EFAULT
  static const string EINTR_error;                  ///< static error string for errno EINTR
  static const string unknown_error;                ///< static error string for an unknown error
};

/* Typdef to access the NetReactor */
typedef Singleton<NetReactor, CreateUsingNew> SNetReactor;


}//end namespace

#endif
