/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "SignalReactor.hpp"

#include <signal.h>
#include <iostream>


namespace NMM{
void generic_handler(int i) {
  SignalReactor::local_instance->pushSignal( SignalReactor::local_instance->toPosixSignal(i) );
}

void generic_exit(int sig ) {
  SignalReactor::PosixSignal p_sig = SignalReactor::local_instance->toPosixSignal(sig);
  GLOBAL_WARNING_STREAM("Signal      : " << SignalReactor::name_array[p_sig] << endl);
  GLOBAL_WARNING_STREAM("Description : " << SignalReactor::description_array[p_sig] << endl);
  cerr << "Signal      : " << SignalReactor::name_array[p_sig] << endl;
  cerr << "Description : " << SignalReactor::description_array[p_sig] << endl;
  cerr << "EXIT with " << sig << " !!!!!!!!!!!!!!!!!! " << endl;
  exit(sig);
}

SignalReactor::SignalReactor()
  : NamedObject("SignalReactor"), handler_array(SR_SIGTTOU+1), converter_array(SIGTTOU+1) {
  condition = new ThreadCondition(mutex);
  local_instance = this;

  for (unsigned int i = 0; i< converter_array.size(); ++i)
    converter_array[i] = static_cast<PosixSignal>(0);

  for (unsigned int i = 0; i<= SR_SIGTTOU; ++i) {
    converter_array[signal_array[i]] = static_cast<PosixSignal>(i);
    handler_array[i] = new SigHandlerStruct();
    handler_array[i]->default_handler = signal(signal_array[i], generic_handler);
    if (handler_array[i]->default_handler == 0) {
      handler_array[i]->default_handler = generic_exit; //set default handle
    }
  }
  is_running = true;
  activateThread();
}

void SignalReactor::setAction(PosixSignal signal, SignalAction action) {
  handler_array[signal]->default_action = action;
}

SignalReactor::~SignalReactor() {
  DEBUG_STREAM(">Destroy SignalReactor" << endl);
  for (unsigned int i = 0; i < handler_array.size(); ++i)
    delete handler_array[i];
  is_running = false;
  condition->notify();//notify
  DEBUG_STREAM("<Destroy SignalReactor" << endl);
}

void SignalReactor::pushSignal(PosixSignal signal) {
  queue_mutex.lock();
  signal_queue.push(signal);
  condition->notify();
  queue_mutex.unlock();
}

void SignalReactor::run() {
  while (is_running) {
    condition->wait();
    while(!signal_queue.empty()) {
      processSignal(popSignal());
    }
  }
}

SignalReactor::PosixSignal SignalReactor::popSignal() {
  queue_mutex.lock();
  PosixSignal p = signal_queue.front();
  signal_queue.pop();
  queue_mutex.unlock();
  return p;
}

void SignalReactor::processSignal(PosixSignal signal) {
  for (list<SignalHandler*>::iterator it = handler_array[signal]->object_list.begin();
       it != handler_array[signal]->object_list.end(); ++it)
  {
    (*it)->receiveSignal(signal);
  }
  for (list< void(*)(int) >::iterator it = handler_array[signal]->method_list.begin();
       it != handler_array[signal]->method_list.end(); ++it)
  {
    (*it)(signal_array[signal]);
  }

  if ((handler_array[signal]->default_action == SIG_DEFAULT) && handler_array[signal]->default_handler) {
    handler_array[signal]->default_handler(signal_array[signal]);
  } else {
    WARNING_STREAM("Ignore default handling for signal " << name_array[signal]<< endl);
  }

}

void SignalReactor::unregisterSignal(SignalHandler* sig_handler) {
  for (unsigned int i = 0; i <= handler_array.size(); ++i)
    unregisterSignal(static_cast<PosixSignal>(i), sig_handler);
}

void SignalReactor::unregisterSignal(signal_handler sig_handler) {
  for (unsigned int i = 0; i <= handler_array.size(); ++i)
    unregisterSignal(static_cast<PosixSignal>(i), sig_handler);
}


void SignalReactor::disableSignal(PosixSignal sig, SignalAction action) {
  switch (action) {
    case SIG_DEFAULT : signal(signal_array[sig], SIG_DFL); break;
    case SIG_IGNORE  : signal(signal_array[sig], SIG_IGN); break;
  }

  for (list<SignalHandler*>::iterator it = handler_array[sig]->object_list.begin();
       it != handler_array[sig]->object_list.begin(); ++it)
  {
    (*it)->disableSignal(sig, action);
  }
}

void SignalReactor::enableSignal(PosixSignal sig, SignalAction action) {
  signal_handler old_handler = signal(signal_array[sig], generic_handler);
  if (generic_handler != old_handler) {
    for (list<SignalHandler*>::iterator it = handler_array[sig]->object_list.begin();
         it != handler_array[sig]->object_list.begin(); ++it)
    {
      (*it)->enableSignal(sig, action);
    }
  }
}

const char* SignalReactor::description_array[] = {
 "Hangup detected on controlling terminal or death of controlling process",
 "Interrupt from keyboard",
 "Quit from keyboard",
 "Illegal Instruction",
 "Abort signal from abort(3)",
 "Floating point exception",
 "Kill signal",
 "Invalid memory reference",
 "Broken pipe: write to pipe with no readers",
 "Timer signal from alarm(2)",
 "Termination signal",
 "User-defined signal 1",
 "User-defined signal 2",
 "Child stopped or terminated",
 "Continue if stopped",
 "Stop process",
 "Stop typed at tty",
 "tty input for background process",
 "tty output for background process"
};

const char* SignalReactor::name_array[] = {
  "SIGHUP", "SIGINT", "SIGQUIT", "SIGILL",
  "SIGABRT", "SIGFPE","SIGKILL", "SIGSEGV",
  "SIGPIPE", "SIGALRM", "SIGTERM", "SIGUSR1",
  "SIGUSR2", "SIGCHLD", "SIGCONT", "SIGSTOP",
  "SIGTSTP", "SIGTTIN", "SIGTTOU"
};

const int SignalReactor::signal_array[] = {
  SIGHUP, SIGINT, SIGQUIT, SIGILL,
  SIGABRT, SIGFPE, SIGKILL, SIGSEGV,
  SIGPIPE, SIGALRM, SIGTERM, SIGUSR1,
  SIGUSR2, SIGCHLD, SIGCONT, SIGSTOP,
  SIGTSTP, SIGTTIN, SIGTTOU
};

SignalReactor* SignalReactor::local_instance = 0;

}//end Namespace NMM
