/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *                    Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "SockStream.hpp"

#include <errno.h>       // for errnos
#include <netdb.h>       // for gethostbyname, getprotobyname
#include <unistd.h>      // for close
#include <signal.h>      // for signals
#include <sys/time.h>

#include <iostream>

namespace NMM {

  /****************** class SockStream ******************/
  
  SockStream::SockStream(bool owner)
    :  __socket(-1),
       __own_socket(owner)
  {
    // XXX: fw: not necessary to create a socket here !! (SockConnector will do this)
    struct protoent* protocol = getprotobyname("tcp");
    
    __socket = ::socket(PF_INET, SOCK_STREAM, protocol->p_proto);
    if ( __socket == -1 ) {
      switch (errno){
        case EACCES : GLOBAL_ERROR_STREAM( " SockStream(): Permission to create a socket of the specified type and/or protocol is denied. " << endl); break;

        case EAFNOSUPPORT: GLOBAL_ERROR_STREAM( " SockStream(): The implementation does not support the specified address family." << endl); break;

        case EINVAL : GLOBAL_ERROR_STREAM( " SockStream(): Unknown protocol, or protocol family not available." << endl); break;

        case EMFILE : GLOBAL_ERROR_STREAM( " SockStream(): Process file table overflow." << endl); break;

        case ENFILE : GLOBAL_ERROR_STREAM( " SockStream(): The system limit on the total number of open files has been reached." << endl); break;

        case ENOBUFS : GLOBAL_ERROR_STREAM( " SockStream(): Insufficient memory is available.  The socket cannot be created until sufficient resources are freed." << endl); break;
	
	case ENOMEM : GLOBAL_ERROR_STREAM( " SockStream(): Insufficient memory is available.  The socket cannot be created until sufficient resources are freed." << endl); break;

        case EPROTONOSUPPORT : GLOBAL_ERROR_STREAM( " SockStream(): The protocol type or the specified protocol is not supported within this domain." << endl); break;   
      }
    }
  }
  
  SockStream::SockStream(int socket, bool owner)
    : __socket(socket),
      __own_socket(owner)
  {}
  
  SockStream::~SockStream() {
    /* if this object is the owner, close the socket */
    if (__own_socket && __socket >= 0) {
      close();
    }
  }
  
  void SockStream::setSocket(int socket, bool owner) {
    /* If this object is the owner, close the old socket */
    if (__own_socket && __socket >= 0) {
      close();
    } 
    /* set the new socket */
    __socket = socket;
    
    /* set the owner flag */
    __own_socket = owner;
  }
  
  ssize_t SockStream::recv (void* buffer,size_t len, int flags)
    throw (const TransmissionException&)
  {
    ssize_t size = ::recv(__socket, buffer, len, flags);
    if (size < 0) {
      createReadTransmissionException(errno);
    }
    return size;
  }
  
  ssize_t SockStream::send (const void* buffer,size_t len, int flags)
    throw (const TransmissionException&)
  {
    ssize_t size = ::send(__socket, buffer, len, flags);
    if (size < 0) {
      createWriteTransmissionException(errno);
    }
    return size;
  }
  
  void SockStream::close() {
    /* close the connection with c-shutdown method */
    if ( __socket ) {  
       ::shutdown(__socket, SHUT_RDWR);
       
      if ( ::close(__socket ) == -1) {
        GLOBAL_ERROR_STREAM( "SockeStream::close() : Closing socket " << __socket << " failed" << endl);    
      } 
      __socket = -1;
    }
  }
  
  void SockStream::readStream(char* data, size_t length, size_t max_packet_size /* = 0 */ ) 
    throw (const TransmissionException&, const ConnectionClosed&)
  {
    if (max_packet_size == 0) {
      max_packet_size = length;
    }
    
    int received = 0;
    
    /* While not all data are received call recv() */
    while (length > 0)  {
      /* receive the */
      if ( (received = recv(data, length, MSG_NOSIGNAL)) == 0) {
        /* the connection was closed */
        throw ConnectionClosed(EPIPE);
      }
      
      length -= received;
      data   += received;
    }
  }
  
  void SockStream::writeStream (const char* data, size_t length, size_t max_packet_size)
    throw (const TransmissionException&, const ConnectionClosed&)
  {
    if (max_packet_size == 0) {
      max_packet_size = length;
    }
    
    int transmit = 0;
    
    /* Wile not all data are written call send */
    while (length > 0)  {    
      if (length < max_packet_size)
        // calls SockStream::send and disables all signals
        transmit = send(data, length, MSG_NOSIGNAL); 
      else
        // calls SockStream::send and disables all signals
        transmit = send(data, max_packet_size, MSG_NOSIGNAL); 
      
      /* If return value is 0 the connection is closed */
      if (transmit == 0) {
        /* Throw a ConnectionClosed exception */
        // XXX: fw: I don't think this will ever happen (see send(2))
        throw ConnectionClosed(EPIPE);
      }
      
      /* go to the memory section which was not written */
      data   += transmit;
      /* remove the written bytes */
      length -=  transmit;
    }
  }
  
  ssize_t SockStream::read (char* buffer, size_t len)
    throw (const TransmissionException&)
  {
    int retval = 0;
    signal(SIGPIPE, SIG_IGN);
    retval = ::read (__socket, buffer, len);
    signal(SIGPIPE, SIG_DFL);
    
    if (retval < 0) {
      createReadTransmissionException(errno);
    }
    return retval;
  }
  
  ssize_t SockStream::write(const char* buffer, size_t len)
    throw (const TransmissionException&)
  {
    signal(SIGPIPE, SIG_IGN);
    ssize_t retval = 0;
    retval = ::write(__socket, buffer, len);
    signal(SIGPIPE, SIG_DFL);
    
    if (retval < 0) {
      createWriteTransmissionException(errno);
    }
    return retval;
  }
  
  INetAddr SockStream::getLocalAddress() const {
    struct sockaddr_in addr;
    socklen_t len_addr = sizeof(addr);
    
    if (getsockname(__socket, reinterpret_cast<struct sockaddr*>(&addr), &len_addr) == 0) {
      INetAddr address;
      address.setAddress(&addr);
      return address;
    } else {
      GLOBAL_ERROR_STREAM("Failed to get stream socket address" << endl);
      return INetAddr();
    }
  }
  
  const INetAddr& SockStream::getPeerAddress() const {
    return __peer_addr;
  }
  
  void SockStream::setPeerAddress(const INetAddr& addr) {
    __peer_addr = addr;
  }
  
}
