/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */



#include "Thread.hpp"

#include <sys/time.h>

#ifdef PROFILING
#include "gprof_pthreads.h"
#endif

namespace NMM {


ThreadMutex::ThreadMutex(MutexKind mk) {
  mutex_kind   = mk;
  lock_counter = 0;
  current_thread = 0;
  pthread_mutex_init(&mutex, 0);
}

ThreadMutex::~ThreadMutex() {
  pthread_mutex_destroy(&mutex);
}

Result ThreadMutex::tryLock() const {
  switch(mutex_kind) {
    case FAST:
      if (pthread_mutex_trylock(&mutex) == 0)  {
        current_thread = pthread_self();
        lock_counter = 1;
        return SUCCESS;
      }
      return FAILURE;

    case RECURSIVE :
      if ( lock_counter > 0 ) {
        if ( pthread_equal(current_thread, pthread_self() ) != 0 ) {
          lock_counter++;
          return SUCCESS;
        } else
          return FAILURE;
      } else {
          if (pthread_mutex_trylock(&mutex) == 0)  {
            current_thread = pthread_self();
            lock_counter = 1;
            return SUCCESS;
          }
      }
      break;
  }
  return FAILURE;
}

void ThreadMutex::lock() const {
  switch (mutex_kind) {
    case FAST :
      pthread_mutex_lock(&mutex);
      break;
    case RECURSIVE :
      if (lock_counter > 0)	{
        if ( pthread_equal(current_thread, pthread_self() ) != 0) {
          lock_counter++;
        } else {
          pthread_mutex_lock(&mutex);
          current_thread = pthread_self();
          lock_counter = 1;
        }
      } else	{
        pthread_mutex_lock(&mutex);
        current_thread = pthread_self();
        lock_counter = 1;		  	
      }
     break;
   }
}		


void ThreadMutex::unlock() const {
  switch (mutex_kind)  {
    case FAST:
      pthread_mutex_unlock(&mutex);
      break;
    case RECURSIVE:
      lock_counter--;
      if (lock_counter <= 0)	
        pthread_mutex_unlock(&mutex);
      break;
  }
}


ThreadCondition::ThreadCondition(const ThreadMutex& mutex) : mutex(mutex) {
  if(mutex.mutex_kind != ThreadMutex::FAST) {
    GLOBAL_ERROR_STREAM("Cannot use RECURSICE mutex with ThreadCondition" << std::endl);
    throw Exception("Cannot use RECURSICE mutex with ThreadCondition");
  }

  pthread_cond_init(&cond, 0);
}


ThreadCondition::~ThreadCondition() {
  pthread_cond_destroy(&cond);
}

Result ThreadCondition::timedWait(long sec, long nsec) const {
  struct timespec timeout;
  timeout.tv_sec  = sec;
  timeout.tv_nsec = nsec;
  int result = pthread_cond_timedwait(&cond, &(mutex.mutex), &timeout);
  if(result != 0) {
    return TIME_OUT;
  } else {
    return SUCCESS;
  }
}

Result ThreadCondition::relativeTimedWait(long sec, long nsec) const {
  struct timeval current_time;
  ::gettimeofday(&current_time, 0);

  struct timespec timeout;
  timeout.tv_sec  = current_time.tv_sec + sec;
  timeout.tv_nsec = current_time.tv_usec * 1000 + nsec;

  int result = pthread_cond_timedwait(&cond, &(mutex.mutex), &timeout);

  if(result != 0) {
    return TIME_OUT;
  } else {
    return SUCCESS;
  }
}

void ThreadCondition::wait() const {
  pthread_cond_wait(&cond, &(mutex.mutex));
}

Thread::Thread(){
  fifo_scheduling_enabled = rr_scheduling_enabled = false;
}

Thread::~Thread() {
}

void Thread::start(  ThrFunction function, void *arg, pthread_attr_t* attr ) throw (const Exception&) {
#ifdef PROFILING
  if ( gprof_pthread_create(&thread, attr, function, arg) != 0) {
    GLOBAL_DEBUG_STREAM2("pthread_create failed!" << endl);
    throw (Exception("not enough system resources to create a process for the new thread"));
  }
  else {
    GLOBAL_DEBUG_STREAM2("gprof_pthread_create successful!" << endl);
  }
#else
  if ((fifo_scheduling_enabled) || (rr_scheduling_enabled)) {
    createDefaultAttribute();
    attr = &custom_sched_attr;
  }
  if ( pthread_create(&thread, attr, function, arg) != 0) {
    GLOBAL_DEBUG_STREAM2("pthread_create failed!" << endl);
    throw (Exception("not enough system resources to create a process for the new thread"));
  }
  //getSchedPriotity();
#endif
}


Result Thread::join() {
  if (pthread_join(thread, 0) == 0)
    return SUCCESS;
  return FAILURE;
}

void Thread::kill()
{
  pthread_kill(thread, 9);
}

void Thread::enableCustomScheduling(Scheduling_Policy policy) {
  switch (policy) {
    case FIFO:
      fifo_scheduling_enabled = true;
      rr_scheduling_enabled   = false;
      break;
    case ROUND_ROBIN:
      fifo_scheduling_enabled = false;
      rr_scheduling_enabled   = true;
      break;
    case OTHER:
      fifo_scheduling_enabled = false;
      rr_scheduling_enabled   = false;
      break;
  }
}

void Thread::createDefaultAttribute() {
  GLOBAL_DEBUG_STREAM2("creating default pthread attribute" << endl);

  pthread_attr_init(&custom_sched_attr);
  pthread_attr_setinheritsched(&custom_sched_attr, PTHREAD_EXPLICIT_SCHED);
  if (rr_scheduling_enabled) {
      pthread_attr_setschedpolicy(&custom_sched_attr, SCHED_RR);
      rr_param.sched_priority = sched_get_priority_max(SCHED_RR);
      pthread_attr_setschedparam(&custom_sched_attr, &rr_param);
      GLOBAL_DEBUG_STREAM2("default pthread attribute (EXPLICIT ROUND ROBIN SCHEDULING) created" << endl);
  }
  if (fifo_scheduling_enabled) {
      pthread_attr_setschedpolicy(&custom_sched_attr, SCHED_FIFO);
      fifo_param.sched_priority = sched_get_priority_max(SCHED_FIFO);
      pthread_attr_setschedparam(&custom_sched_attr, &fifo_param);
      GLOBAL_DEBUG_STREAM2("default pthread attribute (EXPLICIT FIFO SCHEDULING) created" << endl);
  }
}

void Thread::setMinPriority() {
  // get the lowest possible priority
  int sched_priority;
  if (rr_scheduling_enabled) {
      sched_priority = sched_get_priority_min(SCHED_RR);
      rr_param.sched_priority = sched_priority;
      pthread_setschedparam(thread, SCHED_RR, &rr_param);
      getSchedPriotity();
//      GLOBAL_DEBUG_STREAM2("setting pthread priority to minimum ("<<  sched_priority << ")" << endl);
  }
  if (fifo_scheduling_enabled) {
      sched_priority = sched_get_priority_min(SCHED_FIFO);
      fifo_param.sched_priority = sched_priority;
      pthread_setschedparam(thread, SCHED_FIFO, &fifo_param);
      getSchedPriotity();
//      GLOBAL_DEBUG_STREAM2("setting pthread priority to minimum ("<<  sched_priority << ")" << endl);
  }
}
void Thread::setMidPriority() {
  // compute middle priority
  int sched_priority;
  if (rr_scheduling_enabled) {
      sched_priority = (sched_get_priority_min(SCHED_RR) + sched_get_priority_max(SCHED_RR)) / 2;
      rr_param.sched_priority = sched_priority;
      pthread_setschedparam(thread, SCHED_RR, &rr_param);
      getSchedPriotity();
//      GLOBAL_DEBUG_STREAM2("setting pthread middle priority ("<<  sched_priority << ")" << endl);
  }
  if (fifo_scheduling_enabled) {
      sched_priority = (sched_get_priority_min(SCHED_FIFO) + sched_get_priority_max(SCHED_FIFO)) / 2;
      fifo_param.sched_priority = sched_priority;
      pthread_setschedparam(thread, SCHED_FIFO, &fifo_param);
      getSchedPriotity();
//      GLOBAL_DEBUG_STREAM2("setting pthread middle priority ("<<  sched_priority << ")" << endl);
  }
}
void Thread::setMaxPriority() {
  // get the highest possible priority
  int sched_priority;
  if (rr_scheduling_enabled) {
      sched_priority = sched_get_priority_max(SCHED_RR);
      rr_param.sched_priority = sched_priority;
      pthread_setschedparam(thread, SCHED_RR, &rr_param);
      getSchedPriotity();
//      GLOBAL_DEBUG_STREAM2("setting pthread priority to maximum ("<<  sched_priority << ")" << endl);
  }
  if (fifo_scheduling_enabled) {
      sched_priority = sched_get_priority_max(SCHED_FIFO);
      fifo_param.sched_priority = sched_priority;
      pthread_setschedparam(thread, SCHED_FIFO, &fifo_param);
      getSchedPriotity();
//      GLOBAL_DEBUG_STREAM2("setting pthread priority to maximum ("<<  sched_priority << ")" << endl);
  }
}

int Thread::getSchedPriotity() {
  int returned_policy;
  struct sched_param parameters;
  int return_value = pthread_getschedparam(thread, &returned_policy, &parameters);
  if (return_value != 0) {
    GLOBAL_ERROR_STREAM2("pthread getsched failed!" << endl);
  }
  switch (returned_policy) {
    case SCHED_RR:
      GLOBAL_DEBUG_STREAM2( "pthread priority : " << parameters.sched_priority << " (round-robin scheduling)" << endl);
      break;
    case SCHED_FIFO:
      GLOBAL_DEBUG_STREAM2( "pthread priority : " << parameters.sched_priority << " (FIFO scheduling)" << endl);
      break;
    case SCHED_OTHER:
      GLOBAL_DEBUG_STREAM2( "pthread priority : " << parameters.sched_priority << " (other scheduling)" << endl);
      break;
  }
  return parameters.sched_priority;
}

}

#ifdef MutexTest
NMM::ThreadMutex mutex(NMM::ThreadMutex::RECURSIVE);
NMM::ThreadMutex wait_mutex;
NMM::ThreadCondition cond(&wait_mutex);
timeval tv;
bool winner = 0;
void moreLock(int v)
{
  for (int i = 0; i< 500; i++) {
    mutex.lock();
    cout << i <<".st lock from Thread " << v << endl;
  }
  for (int i = 0; i< 500; i++) {
    mutex.unlock();
    cout << i <<".st unlock from Thread " << v << endl;
  }
}

static void* race(void* value) {
  int *i = (int*)value;
	
  cout << "Thread " << (*i) << " lock" << endl;
  mutex.lock();
  if (winner)
    winner = false;
  else
    winner = true;
  	
  cout <<"Winner is Thread"<< (*i) <<endl;
  tv.tv_sec = 5;
  tv.tv_usec = 0;
  select(0, 0, NULL, NULL, &tv);	
  moreLock(*i);
  mutex.unlock();
  return 0;
}


int main(int argc, char*argv[])
{
  NMM::Thread t1, t2, t3, t4, t5;
  int i = 1;
  int j = 2;
  int k = 3;
  int l = 4;
  int m = 5;
  t2.start(race, (void*)&j);
  t3.start(race, (void*)&k);
  t4.start(race, (void*)&l);
  t1.start(race, (void*) &i);	
  t5.start(race, (void*)&m);
  t1.join();	
  t2.join();
  t3.join();
  t4.join();
  t5.join();
  return 0;
}
#endif
